if(typeof VTL === "undefined") VTL = {};

VTL.CameraControl = class {
    canvas;
    camera;
    lookAtVec;
    lon;
    lat;
    constructor(camera, canvas){
        this.camera = camera;
        this.canvas = canvas;
        this.FOV_MAX = 135;
        this.FOV_MIN = 45;
        this.FOV_INITIAL = 90;
        this.lookAtVec = new THREE.Vector3(-1, 0, 0);
        this.lookAtVec.applyQuaternion(this.camera.quaternion);

        this.initLookAtVec = {x:this.lookAtVec.x, y:this.lookAtVec.y, z:this.lookAtVec.z};

        this.init();
    }
}

VTL.CameraControl.prototype.resetPosition = function(){
    this.lon = 180;
    this.lat = 0;
    this.lookAtVec.set(this.initLookAtVec.x, this.initLookAtVec.y, this.initLookAtVec.z);
    this.camera.lookAt(this.lookAtVec);
    this.setFov(this.FOV_INITIAL);
    
}

VTL.CameraControl.prototype.init = function(){
    let s = this;
    let downX = 0, downY = 0, downLon = 0, downLat = 0, downDist = 0, downFov = 0;
    let deltaX = 0; deltaY = 0;
    let oneFinderTouched = false;

    // mouse events
    this.canvas.addEventListener("mousedown", onCanvasMouseDown, false);
    document.addEventListener("wheel", onDocumentMouseWheel, {passive: false});

    // touch events
    s.canvas.addEventListener("touchstart", onCanvasTouchStart, {passive: false});
    document.addEventListener("touchmove", onDocumentTouchMove, {passive: false});
    document.addEventListener("touchend", onDocumentTouchEnd, {passive: true});
    

    // mouse events
    function onCanvasMouseDown(e) {
        if(e.button != 0) return;
        e.preventDefault();
        deltaX = 0;
        deltaY = 0;
        downX = e.clientX;
        downY = e.clientY;
        downLat = s.lat;
        downLon = s.lon;

        document.addEventListener("mousemove", onDocumentMouveMove, false);
        document.addEventListener("mouseup", onDocumentMouseUp, false);
    }

    function onDocumentMouveMove(e) {
        deltaX = e.clientX - downX;
        deltaY = e.clientY - downY;
        rotateCamera(deltaX, deltaY, downLon, downLat);
        s.canvas.dispatchEvent(new MouseEvent('modeldrag', e));
    }

    function onDocumentMouseUp(e) {
        document.removeEventListener("mousemove", onDocumentMouveMove, false);
        document.removeEventListener("mouseup", onDocumentMouseUp, false);

        if(deltaX == 0 && deltaY == 0){
            s.canvas.dispatchEvent(new MouseEvent('pointclick', e));
        }
        
    }

    function onDocumentMouseWheel(e) {
        if (e.target != s.canvas) return;
        s.setFov(s.camera.fov + e.deltaY * 0.05);
    }

    // touch events
    let trackingTouchIds = [];
    function onCanvasTouchStart(e) {
        e.preventDefault();
        
        // add id to track
        for(let i=0;i<e.changedTouches.length;i++){
            trackingTouchIds.push(e.changedTouches[i].identifier);
        }

        oneFinderTouched = false;
        switch(trackingTouchIds.length){
            case 1:
                downX = e.changedTouches[0].pageX;
                downY = e.changedTouches[0].pageY;
                downLon = s.lon;
                downLat = s.lat;
                deltaX = 0;
                deltaY = 0;

                oneFinderTouched = e.touches.length == 1;
                break;

            case 2:
                let touch0 = e.touches[findTouchIndexById(e.touches, trackingTouchIds[0])];
                let touch1 = e.touches[findTouchIndexById(e.touches, trackingTouchIds[1])];
                downFov = s.camera.fov;
                downDist = Math.hypot(touch0.pageX - touch1.pageX, touch0.pageY - touch1.pageY);
                break;

            default:
                break;
        }
    }

    function onDocumentTouchMove(e) {
        //e.preventDefault();

        let trackingTouches = [];
        let trackingInfos = [];
        
        for(let i=0;i<trackingTouchIds.length;i++)
        {
            let idx = findTouchIndexById(e.touches, trackingTouchIds[i]);
            if(idx != -1){
                trackingTouches.push(e.touches[idx]);
                trackingInfos.push(trackingTouches[i]);
            }
        }

        switch(trackingTouches.length){
            case 1:
                deltaX = trackingTouches[0].pageX - downX;
                deltaY =  trackingTouches[0].pageY - downY;
                rotateCamera(deltaX, deltaY, downLon, downLat);
                break;
            case 2:
                let dist = Math.hypot(trackingTouches[0].pageX - trackingTouches[1].pageX, trackingTouches[0].pageY - trackingTouches[1].pageY);
                s.setFov(Math.floor(downFov + ((downDist - dist)  / 5)));
                break;
            case 0:
            default:
                break;
        }
        
        s.canvas.dispatchEvent(new MouseEvent('dragmove', e));
    }

    function onDocumentTouchEnd(e) {
        e.preventDefault();
        // clean up touchIds
        for(let i=trackingTouchIds.length-1;i >= 0;i--){
            if(findTouchIndexById(e.touches, trackingTouchIds[i]) == -1)
            {
                trackingTouchIds.splice(i, 1);
            }
        }

        
        if(trackingTouchIds.length == 1){
            let touch = e.touches[findTouchIndexById(e.touches, trackingTouchIds[0])];
            downX = touch.pageX;
            downY = touch.pageY;
            downLon = s.lon;
            downLat = s.lat;
        }

        if(e.touches.length === 0){
            if (oneFinderTouched){
                if(deltaX == 0 && deltaY == 0){
                    s.canvas.dispatchEvent(new TouchEvent('pointtouch', e));
                }
                oneFinderTouched = false;
            }
        }
    }

    const movementSpeed = 0.3;
    // 
    function rotateCamera(deltaX, deltaY, downLon, downLat){

        s.lon = downLon + deltaX * (-movementSpeed);
        s.lat = downLat + deltaY * (-movementSpeed);

        s.lat = Math.max(-85, Math.min(85, s.lat));
        let phi = THREE.Math.degToRad(90 - s.lat);
        let theta = THREE.Math.degToRad(s.lon);
        s.lookAtVec.x = 100 * Math.sin(phi) * Math.cos(theta);
        s.lookAtVec.y = 100 * -Math.cos(phi);
        s.lookAtVec.z = 100 * Math.sin(phi) * Math.sin(theta);
        
        s.camera.lookAt(s.lookAtVec);
    }

    function findTouchIndexById(touches, id){
        for(let i=0;i<touches.length;i++){
            if(touches[i].identifier == id){
                return i;
            }
        }
        return -1;
    }

    VTL.CameraControl.prototype.setFov = function(fov){
        
        this.camera.fov = Math.max(Math.min(fov, this.FOV_MAX), this.FOV_MIN);
        this.camera.updateProjectionMatrix();
    }
}
