if(typeof VTL === "undefined") VTL = {};
VTL.RealapsData = class{
    version = "1.02";
    description = "supported L*a*b*";
    src
    size;
    projection;
    ra = {
        moniMaxLum:0,
        compRatio:0
    };
    images = []; // {name, data, width, height}
    constructor(size, projection, ra, images)
    {
        this.size = size;
        this.projection = projection;
        this.ra = ra;
        this.images = images;
    }
}

VTL.RealapsDataLoader = class{
}

VTL.RealapsDataLoader.prototype.loadFromJpegFile = async function (fileJpeg){
    let self = this;
    
    return await new Promise((resolve, reject) => {
        const fr = new FileReader(fileJpeg);
        fr.onload = function(){
            try{
                resolve(self.loadFromJpegArrayBuffer(fr.result));
            }
            catch(e){
                reject(e);
            }
        }
        fr.onerror = function(e){
            reject(e);
        }
        fr.readAsArrayBuffer(fileJpeg);
    });
}

VTL.RealapsDataLoader.prototype.loadFromJpegUrl = async function(jpegUrl){
    let self = this;

    return await new Promise((resolve, reject) => {
        let xhr = new XMLHttpRequest();

        xhr.open( "GET", jpegUrl, true );
        xhr.responseType = "arraybuffer";

        xhr.onload = function( e ) {
            try{
                resolve(self.loadFromJpegArrayBuffer(this.response));
            }
            catch(err){
                reject(err);
            }
        };

        xhr.send(null);
    });
}

VTL.RealapsDataLoader.prototype.loadFromJpegArrayBuffer = function(arrayBuffer) {
    const soi = "FFD8";
    const eoi = "FFD9";
    const sos = "FFDA";
    const app1 = "FFE1";
    const app11 = "FFEB";
    const sof0 = "FFC0";
    const sof2 = "FFC2";

    const vtl0 = "VTL0"; 

    let reader = new VTL.JpegReader(arrayBuffer);
    let info = {};
    let isEquirectangular = false;
    let hasVTLSegment = false;
    let images = [];

    let imgIdx = 0;
    let totalByteCount = 0;
    let marker = reader.readMarker();
    let data = null;
    let keyword = "";
    let imageBytes
    if(marker !== soi){
        throw "file is not jpeg";
    }
    
    let bytes = null;
    while(true){
        marker = reader.readMarker();
        data = reader.readData();
        
        if(marker === app1){
            let dataText = byte2String(data);
            let NULIndex = dataText.indexOf("\u0000");
            keyword = dataText.slice(0, NULIndex);
            switch(keyword){
                case "Exif": continue;
                case "http://ns.adobe.com/xap/1.0/":
                    dataText = dataText.slice(NULIndex+1, dataText.length);
                    let endIndex = dataText.length;
                    if(dataText.slice(0, 9) == "<?xpacket"){
                        const regex = /<\?xpacket end=['"][wr]['"]\?>/;
                        endIndex = dataText.search(regex) + 19;
                    }
                    
                    xmlString = dataText.slice(0, endIndex);
                    isEquirectangular = isEquirectangularProjectionType(string2Xml(xmlString));
                    break;
            }
        }
        else if(marker === sof0 || marker === sof2){
            if(!info.size){
                info.size = {
                    height: (data[1] << 8) + data[2],
                    width: (data[3] << 8) + data[4]};
            }
        }
        else if(marker === app11){
            if(!hasVTLSegment){
                let itr = 0;
                if(data.length <= vtl0.length || byte2String(data.subarray(itr,(itr+=vtl0.length))) != vtl0){
                    throw new "file has not vtl-data segment";
                }
                hasVTLSegment = true;
                
                let xmlLength = (data[itr++] << 8 | data[itr++]) - 2; // except length bytes
                let xmlString = new TextDecoder().decode(data.subarray(itr, itr+xmlLength));
                itr += xmlLength;
                let xml = string2Xml(xmlString);
                info = VTL.ImageInfo.fromXML(xml);
                bytes = data.slice(itr);
            }
            else{
                bytes = data;
            }
            
            while(bytes.length != 0){
                if(totalByteCount == 0){
                    imageBytes = new Uint8Array(info.decodeParams[imgIdx].size);
                }

                if(totalByteCount + bytes.length < imageBytes.length){
                    imageBytes.set(bytes, totalByteCount);
                    totalByteCount += bytes.length;
                    break;
                }
                else{
                    let len = imageBytes.length - totalByteCount;
                    imageBytes.set(bytes.subarray(0, len), totalByteCount);
                    
                    let floats = Uint8ArrayToFloat32Array(imageBytes, info.size.width, info.size.height, info.decodeParams[imgIdx]);
                    images.push({type: info.decodeParams[imgIdx].type, name:info.decodeParams[imgIdx].name, value:floats});
                    bytes = bytes.slice(len);
                    
                    // reset
                    imgIdx++;
                    totalByteCount = 0;
                }
            }
        } 
        else if (marker === sos || marker === eoi){
            break;
        }
    }

    return this.createData(isEquirectangular, info.size, info.projection, info.ra, info.color, images);

    function Uint8ArrayToFloat32Array(uint8array, width, height, decodeParam){
        if(decodeParam.isZipped){
            let inflate = new Zlib.RawInflate(uint8array);
            uint8array = inflate.decompress();
        }
        
        return decodeBytes(uint8array, width, height, decodeParam);
    }

    // returns float32 array
    function decodeBytes(bytes, width, height, decodeParam){
        //this.BitUnit が8バイト以下は考慮していない
        let float0xx = new VTL.Float0xx(decodeParam.fractBit, decodeParam.expoBit);
        let values = new Float32Array(height*width);
        for (let y = 0; y < height; y++)
        {
            let yIdx = y * width;
            for (let x = 0; x < width; x++)
            {
                let idx = yIdx + x;
                let num = decodeParam.bitUnit * idx;
                let val = 0;
                for (let i = 0; i < decodeParam.bitUnit; i++)
                {
                    val += (getBit(i + num, bytes) << (decodeParam.bitUnit - i - 1));
                }
                
                let norm = float0xx.getValue(val);
                
                values[idx] = norm * decodeParam.parameter + decodeParam.offset;
            }
        }

        return values;

        function getBit(num, data)
        {
            let n = (num / 8)^0;
            let x = num % 8;
            
            if (n >= data.Length){
                return 0;
            }
            else{
                return (data[n] >> (7 - x)) & 1;
            }
        }
    }

    function byte2String(bytes){
        let str = "";
        for(var i=0;i<bytes.length;i++){
            str += String.fromCharCode(bytes[i]);
        }
        return str;
    }

    function string2Xml(xmlString){
        let parser = new DOMParser();
        return parser.parseFromString(xmlString, "text/xml");
    }

    function isEquirectangularProjectionType(xml){
        let root = xml.documentElement;
        let descs = root.getElementsByTagName("rdf:Description");
        if(descs.length == 0) return false;

        let targetName = "GPano:ProjectionType";
        let desc = descs[0];
        let projectionType = desc.getAttribute(targetName);
        
        if(projectionType === null){
            let ePros = desc.getElementsByTagName("GPano:ProjectionType");
            if(ePros.length == 0) return false;
            projectionType = ePros[0].innerHTML;
        }

        return (projectionType === "equirectangular");
    }
}


VTL.RealapsDataLoader.prototype.getImageFromName = function(name){
    for(let i=0;i<this.images.length;i++){
        if(this.images[i].name == name){
            return this.images[i];
        }
    }
    return null;
}

VTL.RealapsDataLoader.prototype.createData = function(isEquirectangular, size, projection, ra, color, images){
    return {
        isEquirectangular: isEquirectangular,
        size: size,
        projection: projection,
        ra: ra,
        color: color,
        images: images
    };
}

VTL.ImageInfo = class {
    constructor(size, projection, ra, color, decodeParams)
    {
        this.size = size;
        this.projection = projection;
        this.ra = ra;
        this.color = color;
        this.decodeParams = decodeParams;
    }
}

VTL.ImageInfo.fromXML = function(xml){
    let size, projection, ra, color, normalizationParams = [];

    let root = xml.documentElement;
    // size
    let sizeNodes = root.getElementsByTagName("Size");
    if(sizeNodes.length != 0){
        let sizeNode = sizeNodes[0];
        size = {
            width: parseInt(sizeNode.getAttribute("Width")),
            height: parseInt(sizeNode.getAttribute("Height")),
            vaov: parseFloat(sizeNode.getAttribute("VerticalViewAngle")),
            haov: parseFloat(sizeNode.getAttribute("HorizontalViewAngle")),
        };
    }

    // projection
    let projectionNodes = root.getElementsByTagName("Projection");
    if(projectionNodes.length != 0){
        projection = {
            type: projectionNodes[0].getAttribute("Type"),
        };
    }

    // ra 
    let raNodes = root.getElementsByTagName("RA");
    if(raNodes.lenght != 0){
        ra = {
            moniLum: parseFloat(raNodes[0].getAttribute("MoniLuminance")),
            compRatio: parseFloat(raNodes[0].getAttribute("CompRatio")),
        };
    }

    let colorNodes = root.getElementsByTagName("Color");
    if(colorNodes.length != 0){
        let colorNode = colorNodes[0];
        color = {
            munsellBackRef: colorNode.getAttribute("MunsellBackRef"),
            backReflectance: parseFloat(colorNode.getAttribute("BackReflectance"))
        }

        let normalizationNodes = colorNode.getElementsByTagName("Munsell");
        for(let i=0;i<normalizationNodes.length;i++){
            let node = normalizationNodes[i];
            let param = {
                hue: node.getAttribute("HueName"),
                angle: parseFloat(node.getAttribute("Angle_ToNorm")),
                chroma: parseFloat(node.getAttribute("Chroma_ToNorm")),
                lStar: parseFloat(node.getAttribute("L_star_ToNorm"))
            };
            normalizationParams.push(param);
        }
        color.normalizationParams = normalizationParams;
    }

    // decode params
    let imageNodes = root.getElementsByTagName("Images")[0].children;
    let decodeParams = new Array(imageNodes.length);
    for(var i=0;i<imageNodes.length;i++){
        let imageNode = imageNodes[i];
        let decodeParam = {
            type: imageNode.getAttribute("Type"),
            name: imageNode.getAttribute("Name"),
            unitType: imageNode.getAttribute("UnitType"),
            bitUnit: parseInt(imageNode.getAttribute("BitUnit")),
            expoBit: parseInt(imageNode.getAttribute("ExpoBit")),
            fractBit: parseInt(imageNode.getAttribute("FractBit")),
            signBit: parseInt(imageNode.getAttribute("SignBit")),
            bias: parseFloat(imageNode.getAttribute("Baias")),
            parameter: parseFloat(imageNode.getAttribute("Parameter")),
            offset: parseFloat(imageNode.getAttribute("Offset")),
        };
        
        decodeParam.isZipped = imageNode.getAttribute("Zip") == "True";
        let byteLength = imageNode.getAttribute("Size");
        decodeParam.size = byteLength ? parseFloat(byteLength) : Math.ceil((decodeParam.bitUnit * size.width * size.height)/8);
        decodeParams[i] = decodeParam;
    }
    return new VTL.ImageInfo(size, projection, ra, color, decodeParams);
}

VTL.JpegReader = class{
    idx = 0;
    arr;
    constructor(arrayBuffer){
        this.arr = new Uint8Array(arrayBuffer);
    }
};

VTL.JpegReader.prototype.readData = function(){
    let length = this.readWord() - 2; /*exclude length bytes*/
    return this.arr.subarray(this.idx, (this.idx+=length));
};

VTL.JpegReader.prototype.readLength = function(){

};

VTL.JpegReader.prototype.readWord = function(){
    return this.arr[this.idx++] << 8 | this.arr[this.idx++];
};

VTL.JpegReader.prototype.readMarker = function(){
    return (this.arr[this.idx++].toString(16) + this.arr[this.idx++].toString(16)).toUpperCase();
};

VTL.Float0xx = class {
    fractBit;
    expoBit;
    kpara;
    MIN_VALUE = 0.0;
    MAX_VALUE;
    bias;
    constructor(fractBit, expoBit){
        this.fractBit = fractBit;
        this.expoBit = expoBit;
        this.kpara = getKpara(fractBit);
        let factor = getFactor(fractBit, expoBit, this.kpara);
        this.MIN_VALUE = factor.min;
        this.MAX_VALUE = factor.max;
        this.bias = factor.bias;

        function getKpara(fractBit){
            let kpara = new Array(fractBit);
            for (var i = 0; i < fractBit; i++)
            {
                kpara[fractBit - i - 1] = Math.pow(2.0, -(i + 1));
            }
            return kpara;
        }

        function getFactor(fractBit, expoBit, kpara){
            let sisu = (1 << expoBit) - 1;
            let kasu = 0.0;
            for (let i = 0; i < fractBit; i++)
            {
                kasu += kpara[i];
            }
        
            let pkasu = 1.0 + kasu;
            let bias = 0.0;
            let ab = 1.0;
            let fval = Number.MAX_VALUE;
        
            for (let i = 0; i < 1000; i++)
            {
                fval = Math.pow(2.0, sisu - bias) * pkasu;
        
                if (fval == 1.0)
                {
                    break;
                }
        
                if (fval > 1.0)
                {
                    bias += ab;
                }
                else
                {
                    bias -= ab;
                    ab = ab / 10.0;
                }
            }
        
            if (fval < 1.0)
            {
                bias -= ab;
            }
        
            return {bias:bias, min:0.0, max:Math.pow(2.0, sisu - bias) * pkasu};
        }
    }
};

VTL.Float0xx.prototype.getValue = function(data){
    let sisu = (data >> this.fractBit) & ((1 << this.expoBit) - 1);
    let kasu = data & ((1 << this.fractBit) - 1);
    let value = 0.0;
	
	

    if (kasu <= 0 && sisu <= 0)
    {
        return value;
    }

    if (kasu >= Math.pow(2.0, this.fractBit)^0 || sisu >= Math.pow(2.0, this.expoBit)^0)
    {
        return this.MAX_VALUE;
    }

    let k = 0.0;
	
    for (let i = 0; i < this.fractBit; i++)
    {
		
        k += ((kasu >> i) & 1) == 1 ? this.kpara[i] : 0.0;
    }

    if (sisu == 0)
    {
        value = Math.pow(2, 1.0 - this.bias) * k;
    }
    else
    {
        value = Math.pow(2, sisu - this.bias) * (1.0 + k);
    }
    return value;
};