if(typeof VTL === "undefined") VTL = {};

VTL.ImageType = {LM: "lm", BR: "br", EBR: "ebr", UGR: "ugr", RA: "ra", IL: "il", LAB: "lab"};
VTL.DataType = {LM: "lm", BR: "br", UGR: "ugr", RA: "ra", LStar: "L*", AStar: "a*", BStar:"b*", NLStar: "nL*", NAStar: "na*", NBStar:"nb*", ChromaX: "chromaX", ChromaY:"chromaY"};

VTL.RealapsDataViewer = class {
	constructor(wrapperNode, rdvParam, searchParams) {
		const self = this;
        this.scalingSettings = {};
		this.viewSettings = (function(){
			let vSettings = {};
			vSettings[VTL.ImageType.LM] = {division:5, MAX: 100000, MIN: 0.01, MIN_DIFF: 0.0001};
			vSettings[VTL.ImageType.IL] = {division: 5, MAX: 100000, MIN: 0.01, MIN_DIFF: 0.0001, REF_MAX:1.0, REF_MIN: 0.0};
			vSettings[VTL.ImageType.BR] = {division: 13, MAX: 20, MIN: 0, MIN_DIFF: 0.0001};
			vSettings[VTL.ImageType.UGR] = {division:5, MAX: 5, MIN: 0, MIN_DIFF: 0.0001};
			vSettings[VTL.ImageType.LAB] = {CHROMA_MIN: 0, CHROMA_MAX: 2.0, BRIGHTNESS_MIN:-2.0, BRIGHTNESS_MAX:2.0};
			return vSettings;
		})();
		this.viewParams = {};
		this.resetViewParams();

		this.imageTypePriorities = [
			{imageType: VTL.ImageType.UGR, showRA: true},
			{imageType: VTL.ImageType.BR, showRA: true},
			{imageType: VTL.ImageType.EBR, showRA: true},
			{imageType: VTL.ImageType.LAB, showRA: true},
			{imageType: VTL.ImageType.LM, showRA: true},
			{imageType: VTL.ImageType.IL, showRA: true}
		];
		if(searchParams.image) this.imageTypePriorities.unshift({imageType: searchParams.image, showRA: false});
		
		this.currentImageType = null;

		this.state = {
			needsUpdateMarkerPosition: false,
			showRA: true,
			displayInfo: true,
			displayRemarks: true,
			displayLegend: true,
			useMarkdown: false
		};
		if(rdvParam.remarks && rdvParam.remarks.useMarkdown){
			this.state.useMarkdown = rdvParam.remarks.useMarkdown;
		}

		this.options = (function (rdvOptions){
			let options = {};
			let keys = Object.keys(rdvOptions);
			for(let i=0;i<keys.length;i++){
				let key = keys[i];
				options[key] = rdvOptions[key];
			}
	
			return options;
		})(rdvParam.options);
		
        this.fileInfo = {
			name: "",
			isValid: false,
			isEquirectangular: false,
			isVTLFormated: false,
			size: {width: 0, height: 0},
			raConfig: {backLumi: 0, compRatio: 0.5},
			colorConfig: {backMunsell: "N3", backRefrectance: 0.5},
		};
		this.datas = {};
		this.materials = {};
		this.markers = [];

		this.dragOverCounter = 0;
		this.DEG1 = 180.0/Math.PI;

		window.addEventListener("dragover", function (e) { e.preventDefault(); }, true);

		this.elems = this.createElems(wrapperNode);

		this.viewer = new VTL.OmniViewer(this.elems.stage);
		this.viewer.onRequestAnimation = function(){
			if(self.state.needsUpdateMarkerPosition){
				self.updateMarkersPosition(self.viewer.control.lookAtVec, self.markers);
				self.state.needsUpdateMarkerPosition = false;
			}
		}
		this.viewer.render();

		// legendpart
		this.colorbar = new VTL.Colorbar();
		this.colorbar.elems.min.addEventListener("change", function(e){self.onScaleMinChange(e);}, false);
		this.colorbar.elems.max.addEventListener("change", function(e){self.onScaleMaxChange(e);}, false);
		this.huechart = new VTL.HueChart();
		this.huechart.setChromaLim(this.viewParams[VTL.ImageType.LAB].chromaLim);
		this.huechart.setBrightnessLim(this.viewParams[VTL.ImageType.LAB].brightnessLim);
		this.huechart.elems.chroma.addEventListener("change", function(e){self.onHueChromaLimChange(e);}, false);
		this.huechart.elems.brightness.addEventListener("change", function(e){self.onHueBrightnessLimChange(e);}, false);
		this.elems.legendPart.appendChild(this.colorbar.elems.self);
		this.elems.legendPart.appendChild(this.huechart.elems.self);

		let frag = document.createDocumentFragment();
		frag.appendChild(this.elems.logo);
		frag.appendChild(this.elems.modelView);
		frag.appendChild(this.elems.loadingView);
		frag.appendChild(this.elems.messagebox);

		if(this.options.acceptUserFile) {
			this.elems.modelView.addEventListener("dragover", function (e) { self.onModelViewDragOver(e) }, true);
			frag.appendChild(this.elems.dropView);
		}
		wrapperNode.appendChild(frag);

		if(rdvParam.source){
			this.load(rdvParam.source, rdvParam.remarks.value);
		}
		else if(this.options.acceptUserFile) {
			this._show(this.elems.dropView);	
		}
	}
}

VTL.RealapsDataViewer.prototype.startRendering = function (){
	this.viewer.render();
}

VTL.RealapsDataViewer.prototype.resetViewParams = function (){
    this.viewParams[VTL.ImageType.LM] = {min:1, max: 10000.0};
    this.viewParams[VTL.ImageType.IL] = {min: 1, max:300.0, ref: 0.2};
    this.viewParams[VTL.ImageType.BR] = {min: 1, max:13.0};
    this.viewParams[VTL.ImageType.UGR] = {min: 0, max:1};
    this.viewParams[VTL.ImageType.LAB] = {brightnessLim: -0.75, chromaLim: 0.4};
}
VTL.RealapsDataViewer.prototype.createElems = function (wrapper){
	const self = this;

	let elems = {self: wrapper};
	addLogoTo(elems);
	addLoadingViewTo(elems);
	addDropViewTo(elems);
	addModelViewTo(elems, wrapper, this.options.pointable);
	addMessageBoxTo(elems);

	return elems;

	function addLogoTo(elems){
        let logo = document.createElement("div");
        logo.className = "logo";
        elems.logo = logo;
    }

    function addLoadingViewTo(elems){
        let view = document.createElement("div");
        view.className = "loadingview";

		let loader = document.createElement("div");
		loader.className = "loader";

		view.appendChild(loader);

		elems.loadingView = view;
    }

    function addDropViewTo(elems){
		this.dragOverCounter = 0;

		let dropView = document.createElement("div");
		dropView.className = "dropview";
		dropView.addEventListener("dragenter", function (e) { self.onDropViewDragEnter(e); }, true);
		dropView.addEventListener("dragleave", function (e) { self.onDropViewDragLeave(e);}, false);

		let dropZone = document.createElement("div");
		dropZone.className = "dropzone";
		dropZone.addEventListener("drop", async function (e) { self.onDropZoneDrop(e);}, false);
		dropZone.addEventListener("click", async function (e) { elems.fileLoader.click();}, false);
		
		let text = document.createElement("p");
		text.innerHTML = "ここにREALAPS-Jpegをドロップ<br />またはクリックしてREALAPS-Jpegを選択";
		
		dropZone.appendChild(text);
		dropView.appendChild(dropZone);

		elems.dropView = dropView;
    }

	function addMessageBoxTo(elems){
		let messagebox = document.createElement("div");
		messagebox.className = "messagebox";
		elems.messagebox = messagebox;
	}

	function addModelViewTo(elems, wrapper, pointable){
		let modelView = document.createElement("div");
		modelView.className = "modelview";

		const rect = wrapper.getBoundingClientRect();
		let stage = document.createElement("canvas");
		stage.width = rect.width;
		stage.height = rect.height;
		stage.addEventListener("modeldrag", function(e){ self.state.needsUpdateMarkerPosition = true;}, false);
		stage.addEventListener("wheel", function(e){self.state.needsUpdateMarkerPosition = true;}, false);
		if(pointable){
			stage.addEventListener("pointclick", function(e){self.onCanvasPointClick(e);}, false);
			stage.addEventListener("pointtouch", function(e){self.onCanvasPointTouch(e)}, false);
		}
		

		createImageSwitcher(elems);
		

		addSubButtonsTo(elems, self.options);
		addLegentPartTo(elems);
		addRemarksPartTo(elems);

		modelView.appendChild(stage);
		modelView.appendChild(elems.imageSwitcher);
		modelView.appendChild(elems.subButtons);
		modelView.appendChild(elems.legendPart);
		modelView.appendChild(elems.remarksPart);

		elems.modelView = modelView;
		elems.stage = stage;
	
		function createImageSwitcher(elems){
			const imageSwitcher = document.createElement("div");
			imageSwitcher.className = "imageswitcher";

			let div = document.createElement("div");
			const title = document.createElement("p");
			title.innerText = "分布図";

			let items = [
				{type: VTL.ImageType.UGR, name:"グレア"},
				{type: VTL.ImageType.BR, name:"明るさ"},
				{type: VTL.ImageType.EBR, name:"明るさ検討"},
				{type: VTL.ImageType.LAB, name:"色彩"},
				{type: VTL.ImageType.LM, name:"輝度"},
				{type: VTL.ImageType.IL, name:"照度(反射率)"}];

			let ul = document.createElement("ul");
			for(var i=0;i<items.length;i++){
				let li = document.createElement("li");
				let label = document.createElement("label");
				let input = document.createElement("input");
				input.type = "radio";
				input.value = items[i].type;
				input.name = "datatype";
				input.addEventListener("change", function(e){self.onImageTypeChanged(e);}, true);


				label.appendChild(input);
				label.appendChild(document.createTextNode(items[i].name));
				if(items[i].type == VTL.ImageType.IL){
					let text = document.createElement("input");
					text.className = "ilref";
					text.type = "text";
					text.value = 1.0;
					text.placeholder = "1.00";
					text.addEventListener("change", function(e){self.onIlRefChange(e);}, false);
					elems.ilref = text;
					label.appendChild(text);
				}
				li.appendChild(label);
				ul.appendChild(li);
				elems[items[i].type] = input;
			}

			div.appendChild(title);
			div.appendChild(ul);
			imageSwitcher.appendChild(div);

			div = document.createElement("div");
			items = [{type: VTL.ImageType.RA, name:"リアルアピアランス"}];

			ul = document.createElement("ul");
			for(var i=0;i<items.length;i++){
				let li = document.createElement("li");
				let label = document.createElement("label");
				let input = document.createElement("input");
				input.type = "checkbox";
				input.value = items[i].type;
				input.addEventListener("change", function(e){self.onRAChecked(e);}, false);

				label.appendChild(input);
				label.appendChild(document.createTextNode(items[i].name));
				li.appendChild(label);
				ul.appendChild(li);
				elems[items[i].type] = input;
			}
			div.appendChild(ul);
			imageSwitcher.appendChild(div);

			elems.imageSwitcher = imageSwitcher;
		}
		function addLegentPartTo(elems){
			let legendPart = document.createElement("div");
			legendPart.className ="legendpart part";
			
			let title = document.createElement("h4");
			title.innerText = "カラー分布";
			let switcher = document.createElement("button");
			switcher.className = "switcher";
			switcher.addEventListener("click", onSwitchClick, false);

			legendPart.appendChild(title);
			legendPart.appendChild(switcher);

			elems.legendPartTitle = title;
			elems.legendPart = legendPart;
		}
		function addRemarksPartTo(elems){
			let remarksPart = document.createElement("div");
			remarksPart.className = "remarkspart part";

			let title = document.createElement("h4");
			title.innerText = "備考";

			let remarks = document.createElement("div");
			remarks.className ="remarks";

			let switcher = document.createElement("button");
			switcher.className = "switcher";
			switcher.addEventListener("click", onSwitchClick, false);

			remarksPart.appendChild(title);
			remarksPart.appendChild(remarks);
			remarksPart.appendChild(switcher);

			elems.remarksPart = remarksPart;
			elems.remarksPartTotle = title;
			elems.remarks = remarks;
		}
		function onSwitchClick(e){
			e.target.parentNode.classList.toggle("hide");
		}
		function addSubButtonsTo(elems, options){
			const subButtons = document.createElement("div");
			subButtons.className = "subbuttons";
			
			let fileloader;
			if(options.acceptUserFile){
				fileloader = document.createElement("div");
				fileloader.className ="fileloader";
				fileloader.title = "REALAPS-Jpegを読み込む";

				let input = document.createElement("input");
				input.type = "file";
				input.accept = "image/jpeg";
				input.multiple = false;
				input.addEventListener("change", function(e){self.onFileLoaderChange(e);}, false)

				fileloader.addEventListener("click", function(e){input.click();}, false);

				fileloader.appendChild(input);
				subButtons.appendChild(fileloader);
			}
			
			let resetCameraPosition = document.createElement("div");
			resetCameraPosition.className ="resetcameraposition";
			resetCameraPosition.title = "カメラの位置を原点に戻す";
			resetCameraPosition.addEventListener("click", function(){ self.resetCameraPosition();}, false);
			subButtons.appendChild(resetCameraPosition);

			let clearMarkers;
			if(options.pointable){
				clearMarkers = document.createElement("div");
				clearMarkers.className ="clearmarkers";
				clearMarkers.title = "すべてのマーカーを除く";
				clearMarkers.addEventListener("click", function(){self.clearMarkers();}, false);
				subButtons.appendChild(clearMarkers);
			}

			elems.subButtons = subButtons;
			elems.fileLoader = fileloader;
			elems.resetCameraPosition = resetCameraPosition;
			elems.clearMarkers = clearMarkers;

		}
	}
}
VTL.RealapsDataViewer.prototype.reloadView = function (){
	let material;
	const imageType = this.currentImageType;

	if(this.state.showRA){
		material = this.materials[VTL.ImageType.RA];
		this._hide(this.elems.legendPart);
	}
	else{
		material = this.materials[imageType];
		let legendPart = this.elems.legendPart;

		switch(imageType){
			case VTL.ImageType.LAB:
				legendPart.classList.remove("bar");
				legendPart.classList.add("chart");
				this.setLegendPartTitle("色相図");
				break;
			default:
				legendPart.classList.add("bar");
				legendPart.classList.remove("chart");
				let title, unit;
				switch(imageType){
					case VTL.ImageType.LM:
						title = "輝度", unit = "[cd/m<sup>2</sup>]";
						break;
					case VTL.ImageType.IL:
						title = "照度", unit = "[ lx ]";
						break;
					case VTL.ImageType.BR:
					case VTL.ImageType.EBR:
						title = "NB値", unit ="[ NB ]";
						break;
					case VTL.ImageType.UGR:
						title = "UGR_n", unit = "[ - ]";
						break;
				}
				this.setLegendPartTitle(title);
				this.setColorbarUnit(unit);
				this.colorbar.displayMajor(imageType);
				break;
				
		}
		this._show(this.elems.legendPart);
	}
	this.viewer.setMaterial(material);
}
VTL.RealapsDataViewer.prototype.updateAllMarkersValue = function (imageType){
	for(let i=0;i<this.markers.length;i++){
		this.markers[i].updateLabel(imageType);
	}
}
VTL.RealapsDataViewer.prototype.reloadAllMarkersILValue = function (update){
	for(let i=0;i<this.markers.length;i++){
		let marker = this.markers[i];
		this.setLabelILValueToMarker(marker);
		if(update){
			marker.updateLabel(VTL.ImageType.IL);
		}
	}
}
VTL.RealapsDataViewer.prototype.setupMojorInfos = function (){
	const keys = Object.keys(this.materials);
	for(let i=0;i<keys.length;i++){
		let key = keys[i];
		let viewParam = this.viewParams[key];
		let viewSetting = this.viewSettings[key];

		let majorInfos = this.getMajorInfo(key, viewParam, viewSetting);
		if(majorInfos){
			this.colorbar.loadMajorInfo(key, majorInfos);
		}
	}
}
VTL.RealapsDataViewer.prototype.getMajorInfo = function (imageType, viewParam, viewSetting){
	let majorInfo;
	switch(imageType){
		case VTL.ImageType.LM:
		case VTL.ImageType.IL:
			majorInfo = VTL.ColorbarMajorInfo.createLogInfo(viewParam.min, viewParam.max, viewSetting.division);
			break;
		case VTL.ImageType.BR:
		case VTL.ImageType.UGR:
			majorInfo = VTL.ColorbarMajorInfo.createLinearInfo(viewParam.min, viewParam.max, viewSetting.division);
			break;
		case VTL.ImageType.EBR:
			majorInfo = VTL.ColorbarMajorInfo.createEbrInfo();
			break;
		default:
			break;
	}
	return majorInfo;
}
VTL.RealapsDataViewer.prototype.updateILMaterialRef = function (){
	this.materials[VTL.ImageType.IL].uniforms.uRefRatio.value = this.viewParams[VTL.ImageType.IL].ref;
}
VTL.RealapsDataViewer.prototype.updateMaterialRange = function (imageType, min, max){
	this.materials[imageType].uniforms.uMin.value = min;
	this.materials[imageType].uniforms.uMax.value = max;
}
VTL.RealapsDataViewer.prototype.updateLabMaterial = function (){
	let material = this.materials[VTL.ImageType.LAB];
	let viewParam = this.viewParams[VTL.ImageType.LAB];
	material.uniforms.uChromaLim.value = viewParam.chromaLim;
	material.uniforms.uBrightnessLim.value = viewParam.brightnessLim;
}
VTL.RealapsDataViewer.prototype.setRemarks = function (text){
	if(this.state.useMarkdown){
		this.elems.remarks.innerHTML = marked(text);
	} else {
		this.elems.remarks.innerText = text;
	}
}
VTL.RealapsDataViewer.prototype.setLegendPartTitle = function (text){
	this.elems.legendPartTitle.innerHTML = text;
}
VTL.RealapsDataViewer.prototype.setColorbarUnit = function (text){
	this.colorbar.elems.unit.innerHTML = text;
}
VTL.RealapsDataViewer.prototype.hidePart = function (elem){
	elem.classList.add("hide");
}
VTL.RealapsDataViewer.prototype.showPart = function (elem){
	elem.classList.remove("hide");
}
VTL.RealapsDataViewer.prototype.load = async function(source, remarks){
	this._show(this.elems.loadingView, true);
	this._hide(this.elems.dropView);
	try{
		let loader = new VTL.RealapsDataLoader();
		let rdata;
		let url = "";
		let filename = "";
		if(source instanceof File){
			const isJpeg = source.type == "image/jpeg";
			if(!isJpeg){
				throw "Jpeg形式ではないファイルが読み込まれました。";
			}

			rdata = await loader.loadFromJpegFile(source);
			url = await getDataUrl(source);
			filename = source.name;
		}else if(typeof(source) == "string"){
			rdata = await loader.loadFromJpegUrl(source);
			url = source;
		}else{
			throw "unknown source";
		}

		if(!rdata.isEquirectangular){
			throw "360度画像(正距円筒図法)ではない画像ファイルが読み込まれました。";
		}
		if(rdata.images.length == 0 && !this.options.acceptNoDataFile){
			throw "REALAPS-dataの無いJpegファイルが読み込まれました。REALAPS-Omniで作成したREALAPS-Jpegを選択してください。";
		}
		if(hasImage(rdata.images, "L_star") && (!rdata.color.normalizationParams || rdata.color.normalizationParams.length == 0)){
			throw "ファイルに不備があります。「最新版のREALAPS-Omni-Color」から出力したファイルをご利用ください。";
		}

		this._loadParameters(rdata, url, filename, remarks);
	}
	catch(e){
		let message = e;
		switch(message){
			case "file is not jpeg":
				message = "Jpeg形式ではないファイルが読み込まれました。";
				break;

		}
		this.showMessage(message);
		if(this.options.acceptUserFile && !this.fileInfo.isValid){
			this._show(this.elems.dropView);
		}
	}


	this._fadeout(this.elems.loadingView);

	async function getDataUrl(file){
		return await new Promise((resolve, reject)=>{
			let fr = new FileReader();
			fr.onload = function(){resolve(fr.result);}
			fr.onerror = function(e){reject(e);}
			fr.readAsDataURL(file);
		});
	}

	function hasImage(images, type){
		for(let i=0;i<images.length;i++){
			if(images[i].type === type){
				return true;
			}
		}
		return false;
	}
}
VTL.RealapsDataViewer.prototype._loadParameters = async function(rdata, imgUrl, filename, remarks) {// imgUrl may data-url
	this.fileInfo.isValid = true;
	this.fileInfo.size.width = rdata.size.width;
	this.fileInfo.size.height = rdata.size.height;
	this.fileInfo.isEquirectangular = rdata.isEquirectangular;
	this.fileInfo.isVTLFormated = rdata.images.length != 0;
	this.fileInfo.colorConfig.backMunsell = rdata.color ? rdata.color.munsellBackRef : "None";

	this.clearMarkers();
	this.datas = getDatas(rdata);
	if(this.datas[VTL.DataType.LStar]){
		appendNormalizedLab(this.datas, rdata.color.normalizationParams);
	}
	let imageTypeObj = getInitialImageObj(this.datas, this.imageTypePriorities);
	this.state.showRA = imageTypeObj.showRA;
	this.currentImageType = imageTypeObj.imageType;

	this.viewer.setMaterial(null);
	disposeMaterials(this.materials);
	this.materials = createMaterials(imgUrl, this.datas, this.fileInfo.size, this.viewParams);
	//this.viewer.setMaterial(this.materials[VTL.ImageType.RA]);

	this.setupMojorInfos();

	this.loadCompleted(filename, remarks);
	
	function getInitialImageObj(datas, imageTypePriorities){
		for(let i=0;i<imageTypePriorities.length;i++){
			let priorityObj = imageTypePriorities[i];
			switch(priorityObj.imageType){
				case VTL.ImageType.UGR: 
					if(datas[VTL.DataType.UGR]) return priorityObj;
					break;
				case VTL.ImageType.BR:
				case VTL.ImageType.EBR:
					if(datas[VTL.DataType.BR]) return priorityObj;
					break;
				case VTL.ImageType.LAB:
					if(datas[VTL.DataType.LStar]) return priorityObj;
					break;
				case VTL.ImageType.LM:
				case VTL.ImageType.IL:
					if(datas[VTL.DataType.LM]) return priorityObj;
					break;
			}
		}
		return {imageType: VTL.ImageType.RA, showRA: true};
	}
	function getDatas(realapsData){
		let datas = {};
		for(var i=0;i<realapsData.images.length;i++){
			let image = realapsData.images[i];
			
			switch(image.type){
				case "Luminance":
					datas[VTL.DataType.LM] = image.value;
					break;
				case "NB":
					datas[VTL.DataType.BR] = image.value;
					break;
				case "Glare":
					datas[VTL.DataType.UGR] = image.value;
					break;
				case "L_star":
					datas[VTL.DataType.LStar] = image.value;
					break;
				case "a_star":
					datas[VTL.DataType.AStar] = image.value;
					break;
				case "b_star":
					datas[VTL.DataType.BStar] = image.value;
					break;
				case "Chroma_x":
					datas[VTL.DataType.ChromaX] = image.value;
					break;
				case "Chroma_y":
					datas[VTL.DataType.ChromaY] = image.value;
					break;
			}
		}
		return datas;
	}

	function appendNormalizedLab(datas, nParams){
		let lStars = datas[VTL.DataType.LStar]; 
		let aStars = datas[VTL.DataType.AStar];
		let bStars = datas[VTL.DataType.BStar];
		let nLStars = new Float32Array(lStars.length);
		let nAStars = new Float32Array(aStars.length);
		let nBStars = new Float32Array(bStars.length);
		for(let i=0;i<lStars.length;i++){
			let l = lStars[i];
			let a = aStars[i];
			let b = bStars[i];
			let nLab = VTL.Munsell.getNormalizedValue(l, a, b, nParams);
			nLStars[i] = nLab.lStar;
			nAStars[i] = nLab.aStar;
			nBStars[i] = nLab.bStar;
		}
		datas[VTL.DataType.NLStar] = nLStars;
		datas[VTL.DataType.NAStar] = nAStars;
		datas[VTL.DataType.NBStar] = nBStars;
	}
	function disposeMaterials(materials){
		let keys = Object.keys(materials);
		for(let i=0;i<keys.length;i++){
			materials[keys[i]].dispose();
		}
	}
	function createMaterials(source, datas, size, viewParams){		
		let materials = {};	
		let raTex = VTL.OmniViewer.createTexture(source);
		materials[VTL.ImageType.RA] = VTL.OmniViewer.createMaterial(raTex);
		
		let tex;
		let viewParam;
		let keys = Object.keys(datas);
		for(let i=0;i<keys.length;i++){
			let key = keys[i];
			switch(key){
				case VTL.DataType.UGR:
					tex = VTL.OmniViewer.createDataTexture(datas[key], size.width, size.height);
					viewParam = viewParams[VTL.ImageType.UGR];
					materials[VTL.ImageType.UGR] = VTL.OmniViewer.createUGRMaterial(raTex, tex, viewParam.min, viewParam.max);
					break;
				case VTL.DataType.LM:
					tex = VTL.OmniViewer.createDataTexture(datas[key], size.width, size.height);
					viewParam = viewParams[VTL.ImageType.LM];
					materials[VTL.ImageType.LM] = VTL.OmniViewer.createLMMaterial(tex, viewParam.min, viewParam.max);
					viewParam = viewParams[VTL.ImageType.IL];
					materials[VTL.ImageType.IL] = VTL.OmniViewer.createILMaterial(tex, viewParam.min, viewParam.max, viewParam.ref);
					break;
				case VTL.DataType.BR:
					tex = VTL.OmniViewer.createDataTexture(datas[key], size.width, size.height);
					viewParam = viewParams[VTL.ImageType.BR];
					materials[VTL.ImageType.BR] = VTL.OmniViewer.createBRMaterial(tex, viewParam.min, viewParam.max);
					materials[VTL.ImageType.EBR] = VTL.OmniViewer.createEBRMaterial(tex);
					break;
				case VTL.DataType.LStar:
					viewParam = viewParams[VTL.ImageType.LAB];
					let l = datas[VTL.DataType.NLStar];
					let a = datas[VTL.DataType.NAStar];
					let b = datas[VTL.DataType.NBStar];
					let lTex = VTL.OmniViewer.createDataTexture(l, size.width, size.height);
					let aTex = VTL.OmniViewer.createDataTexture(a, size.width, size.height);
					let bTex = VTL.OmniViewer.createDataTexture(b, size.width, size.height);
					materials[VTL.ImageType.LAB] = VTL.OmniViewer.createLabMaterial(raTex, lTex, aTex, bTex, viewParam.chromaLim, viewParam.brightnessLim);
					break;
			}
		}

		return materials;
	}
	
}
VTL.RealapsDataViewer.prototype.loadCompleted = function(filename, remarks){
	this.viewer.resetPosition();

	// set default values
	if(this.fileInfo.isVTLFormated){
		// ra 
		this.elems.ra.checked = this.state.showRA;

		// image switchers
		const keys = Object.keys(VTL.ImageType);
		for(let i=0;i<keys.length;i++){
			let imageType = VTL.ImageType[keys[i]];
			let elem = this.elems[imageType];
			if(this.materials[imageType]) {
				this._show(elem.parentNode.parentNode);
			} else {
				this._hide(elem.parentNode.parentNode);
			}

			if(this.currentImageType == imageType){
				elem.checked = true;
			}

			if(imageType == VTL.ImageType.IL){
				this.elems.ilref.value = this.viewParams[imageType].ref;
			}
		}

		// legend
		this.huechart.setBackMunsellName(this.fileInfo.colorConfig.backMunsell);
	}
	this.reloadView();

	// remarks
	if(remarks === "") {
		this._hide(this.elems.remarksPart);
	}
	else{
		this.setRemarks(remarks);
		this._show(this.elems.remarksPart);
	}

	if(this.fileInfo.isVTLFormated) this.elems.modelView.classList.remove("common");
	else this.elems.modelView.classList.add("common");

	this._show(this.elems.modelView);

	// document title
	if(filename !== "") document.title = filename;
}

// markers
VTL.RealapsDataViewer.prototype.loadValuesToMarker = function(marker){
	let index = marker.imagePoint.y * this.fileInfo.size.width + marker.imagePoint.x;
	let value;
	let values = {};
	let chromaCoords;
	let keys = Object.keys(this.datas);
	for(let i=0;i<keys.length;i++){
		let key = keys[i];
		let data = this.datas[key];
		switch(key){
			case VTL.DataType.LM:
				value = data[index];
				values[VTL.ImageType.LM] = value.toFixed(4);
				values[VTL.ImageType.IL] = (value * Math.PI / this.viewParams[VTL.ImageType.IL].ref).toFixed(4);
				break;
			case VTL.DataType.BR:
				value = data[index];
				values[VTL.ImageType.BR] = values[VTL.ImageType.EBR] = value.toFixed(4);
				break;
			case VTL.DataType.UGR:
				value = data[index];
				values[VTL.ImageType.UGR] = value.toFixed(4);
				break;
			case VTL.DataType.LStar:
				let m = VTL.Munsell.getClosestValue(this.datas[VTL.DataType.LStar][index], this.datas[VTL.DataType.AStar][index], this.datas[VTL.DataType.BStar][index]);
				
				const lLimit = 150.0;
				let l = this.datas[VTL.DataType.LStar][index];
				values[VTL.ImageType.LAB] = 
				"L*： " + ((lLimit <= l) ? "-" : (l.toFixed(2) + " (" + this.datas[VTL.DataType.NLStar][index].toFixed(2)+ ")")) + "<br />" + 
				"a*： " + this.datas[VTL.DataType.AStar][index].toFixed(2) + " (" + this.datas[VTL.DataType.NAStar][index].toFixed(2)+ ")<br />" + 
				"b*： " + this.datas[VTL.DataType.BStar][index].toFixed(2) + " (" + this.datas[VTL.DataType.NBStar][index].toFixed(2)+ ")<br />" + 
				"M： " + m.name;
				break;
			case VTL.DataType.ChromaX:
				let x = this.datas[VTL.DataType.ChromaX][index];
				let y = this.datas[VTL.DataType.ChromaY][index];
				chromaCoords = "<br/>("+x.toFixed(3)+", "+y.toFixed(3)+")";
		}
	}
	
	marker.values = values;
	marker.chromaCoords = chromaCoords;
}
VTL.RealapsDataViewer.prototype.setLabelILValueToMarker = function(marker){
	let index = marker.imagePoint.y * this.fileInfo.size.width + marker.imagePoint.x;
	marker.values[VTL.ImageType.IL] = (this.datas[VTL.DataType.LM][index] * Math.PI / this.viewParams[VTL.ImageType.IL].ref).toFixed(4);
}
VTL.RealapsDataViewer.prototype.clearMarkers = function(){
	for(let i=0;i<this.markers.length;i++){
		let marker = this.markers[i];
		marker.removeEventHandler();
		this.elems.modelView.removeChild(marker.elems.self);
	}
	this.markers = [];
}
VTL.RealapsDataViewer.prototype.resetCameraPosition = function(){
	this.viewer.resetPosition();
	this.state.needsUpdateMarkerPosition = true;
}
VTL.RealapsDataViewer.prototype.addMarker = function(clientX, clientY){
	const self = this;
	let marker = createMarker(clientX, clientY);
	this.loadValuesToMarker(marker);
	marker.updateLabel(this.currentImageType);

	let wp = marker.worldPoint;
	let pos = this.viewer.worldPointToClientLocation(wp.x, wp.y, wp.z);
	marker.updatePosition(pos.x, pos.y);
	
	this.markers.push(marker);
	this.elems.modelView.appendChild(marker.elems.self);

	function createMarker(offsetX, offsetY){
		let info = self.viewer.locationToSphereInfo(offsetX, offsetY);
		let worldPoint = info.point;
		let imagePoint = {
			x:Math.floor(info.uv.x * self.fileInfo.size.width),
			y:Math.floor(self.fileInfo.size.height - info.uv.y * self.fileInfo.size.height)};
		
		let marker = new VTL.Marker(worldPoint, imagePoint);
		marker.addEventHandler(onMarkerClick);
		return marker;

		function onMarkerClick(e){
			e.preventDefault();
			e.stopPropagation();
			for(let i=0;i<self.markers.length;i++){
				let marker = self.markers[i];
				if(marker.elems.finder == e.target){
					marker.removeEventHandler();
					self.elems.modelView.removeChild(marker.elems.self);
					self.markers.splice(i, 1);
					break;
				}		
			}
		}
	}
}
VTL.RealapsDataViewer.prototype.updateMarkersPosition = function(lookAt, markers){
	if(this.markers.length != 0){
		let sqrtq = Math.sqrt(lookAt.x*lookAt.x + lookAt.y*lookAt.y + lookAt.z*lookAt.z);
		for(let i=0;i<markers.length;i++){
			let marker = this.markers[i];
			let wPoint = marker.worldPoint;
			let sqrtp = Math.sqrt(wPoint.x*wPoint.x + wPoint.y*wPoint.y + wPoint.z*wPoint.z);
			let dot = (wPoint.x*lookAt.x + wPoint.y*lookAt.y + wPoint.z*lookAt.z);
			let angle = Math.acos(dot/(sqrtp*sqrtq))* this.DEG1;
			marker.setVisible(angle < 90.0)
			let pos = this.viewer.worldPointToClientLocation(wPoint.x, wPoint.y, wPoint.z);
			marker.updatePosition(pos.x, pos.y);
		}
	}
}

// utilities
VTL.RealapsDataViewer.prototype._show = function(elem, force){
	elem.style.display = "block";
	if(force) {
		elem.style.display = "none";
		elem.offsetHeight;
		elem.style.display = "block"; //getBoundingClientRect(); // force rendering
	}
}
VTL.RealapsDataViewer.prototype._hide = function(elem){
	elem.style.display = "none";
}
VTL.RealapsDataViewer.prototype._wait = function(time){
	
}
VTL.RealapsDataViewer.prototype._fadeout = function(elem){
	elem.classList.add("fadeout");
	setTimeout(function(){
		elem.style.display = "none";
		elem.classList.remove("fadeout");
	}, 700);
}
let messageCount = 0;
VTL.RealapsDataViewer.prototype.showMessage = function(text){
	messageCount++;
	let messagebox = this.elems.messagebox;
	this._show(messagebox);
	messagebox.innerHTML = text;

	setTimeout(fadeOutElem, 4000);

	let self = this;
	function fadeOutElem(){
		messageCount--;
		if(messageCount == 0){
			self._fadeout(messagebox);
		}
	}
}
VTL.RealapsDataViewer.prototype.setSize = function(width, height){
	this.viewer.updateSize(width, height);

	this.state.needsUpdateMarkerPosition = true;
}

VTL.RealapsDataViewer.prototype.onModelViewDragOver = function (e) {
	e.preventDefault();
	if(e.dataTransfer.items[0].kind == "file"){
		this.dragOverCounter = 0;
		this._show(this.elems.dropView);
	}
}
VTL.RealapsDataViewer.prototype.onDropViewDragEnter = function (e) {
	e.preventDefault();
	this.dragOverCounter++;	
}
VTL.RealapsDataViewer.prototype.onDropViewDragLeave = function (e) {
	e.preventDefault();
	this.dragOverCounter--;
	if (this.dragOverCounter == 0) {
		this._hide(this.elems.dropView);
	}
}
VTL.RealapsDataViewer.prototype.onDropZoneDrop = async function (e) {
	e.preventDefault();
	this.dragOverCounter = 0;
	this._hide(this.elems.dropView);
	this.load(e.dataTransfer.files[0], "");
}
VTL.RealapsDataViewer.prototype.onImageTypeChanged = function(e){
	this.currentImageType = e.target.value;
	
	this.updateAllMarkersValue(this.currentImageType);
	this.reloadView();
}
VTL.RealapsDataViewer.prototype.onRAChecked = function(e){
	this.state.showRA = e.target.checked;
	this.reloadView();
}
VTL.RealapsDataViewer.prototype.onIlRefChange = function(e){
	if(!e.isTrusted) return;

	const viewParam = this.viewParams[VTL.ImageType.IL];
	let f = parseFloat(e.target.value);
	if(isFinite(f) && !isNaN(f)){
		const setting = this.viewSettings[VTL.ImageType.IL];
		viewParam.ref = Math.max(Math.min(setting.REF_MAX, f), setting.REF_MIN);
		
	}
	this.updateILMaterialRef();
	this.reloadAllMarkersILValue(this.currentImageType === VTL.ImageType.IL);
	this.reloadView();

	e.target.value = viewParam.ref.toFixed(2);
}
VTL.RealapsDataViewer.prototype.onCanvasPointClick = function(e){
	e.preventDefault();
	if(this.fileInfo.isVTLFormated){
		this.addMarker(e.offsetX, e.offsetY);
	}
}
VTL.RealapsDataViewer.prototype.onCanvasPointTouch = function(e){
	e.preventDefault();	
	if(this.fileInfo.isVTLFormated){
		this.addMarker(e.changedTouches[0].pageX, e.changedTouches[0].pageY);
	}
}
VTL.RealapsDataViewer.prototype.onMarkerClick = function(e){
	e.preventDefault();
	e.stopPropagation();
	for(let i=0;i<this.markers.length;i++){
		if(this.markers[i].elems.finder == e.target){
			this.removeMarkerFrom(this.elems.wrapper, this.markers[i]);
			this.markers.splice(i, 1);
			break;
		}		
	}
}
VTL.RealapsDataViewer.prototype.onScaleMaxChange = function(e){
	const imageType = this.currentImageType;
	const vParam = this.viewParams[imageType];
	const setting = this.viewSettings[imageType];
	if(setting == null) return;

	let f = parseFloat(e.target.value);
	if(isFinite(f) && !isNaN(f)){
		if(f <= vParam.min){
			f = vParam.min + setting.MIN_DIFF;
		}
		else if(setting.MAX_LIM < f){
			f = setting.MAX_LIM;
		}
		vParam.max = f;
	}

	this.colorbar.loadMajorInfo(imageType, this.getMajorInfo(imageType, vParam, setting), true);
	this.updateMaterialRange(imageType, vParam.min, vParam.max);
	this.colorbar.elems.max.value = vParam.max;
}
VTL.RealapsDataViewer.prototype.onScaleMinChange = function(e){	
	const imageType = this.currentImageType;
	const vParam = this.viewParams[imageType];
	const setting = this.viewSettings[imageType];
	if(setting == null) return;

	let f = parseFloat(e.target.value);
	if(isFinite(f) && !isNaN(f)){
		if(setting.max <= f){
			f = vParam.max - setting.MIN_DIFF;
		}
		else if(f < setting.MIN_LIM){
			f = setting.MIN_LIM;
		}
		vParam.min = f;
	}

	this.colorbar.loadMajorInfo(imageType, this.getMajorInfo(imageType, vParam, setting), true);
	this.updateMaterialRange(imageType, vParam.min, vParam.max);
	this.colorbar.elems.min.value = vParam.min;
}
VTL.RealapsDataViewer.prototype.onFileLoaderChange = function(e){
	this.load(e.target.files[0], "");
	e.target.value = ""; // clear for next time
}
VTL.RealapsDataViewer.prototype.onHueChromaLimChange = function(e){
	let vParam = this.viewParams[VTL.ImageType.LAB];
	let vSetting = this.viewSettings[VTL.ImageType.LAB];
	let f = parseFloat(e.target.value);
	if(isFinite(f) && !isNaN(f)){
		if(f < vSetting.CHROMA_MIN) f = vSetting.CHROMA_MIN;
		else if (vSetting.CHROMA_MAX < f)  f = vSetting.CHROMA_MAX;
		vParam.chromaLim = f;

		this.updateLabMaterial();
	}

	e.target.value = vParam.chromaLim;
}
VTL.RealapsDataViewer.prototype.onHueBrightnessLimChange = function(e){
	let vParam = this.viewParams[VTL.ImageType.LAB];
	let vSetting = this.viewSettings[VTL.ImageType.LAB];
	let f = parseFloat(e.target.value);
	if(!isNaN(f) && isFinite(f)){
		if(f < vSetting.BRIGHTNESS_MIN) f = vSetting.BRIGHTNESS_MIN;
		else if (vSetting.BRIGHTNESS_MAX < f)  f = vSetting.BRIGHTNESS_MAX;
		vParam.brightnessLim = f;

		this.updateLabMaterial();
	}

	e.target.value = vParam.brightnessLim;
}

VTL.Marker = class {	
	constructor(worldPoint, imagePoint){
		this.worldPoint = worldPoint;
		this.imagePoint = imagePoint;
		this.isOutScreen = false;
		this.labelValues = {};
		this.elems = createElement();
		
		function createElement(){
			let marker = document.createElement("div");
			marker.className = "marker";
			let label = document.createElement("p");
			label.className = "label";
			let finder = document.createElement("div");
			finder.className = "finder";

			marker.appendChild(label);
			marker.appendChild(finder);

			return {self: marker, finder: finder, label, label};
		}
	}
}

VTL.Marker.prototype.updatePosition = function(x, y){
	this.elems.self.style.top = y + "px";
	this.elems.self.style.left = x + "px";
}

VTL.Marker.prototype.setVisible = function(tf){
	if(this._visible != tf){
		this._visible = tf;
		this.elems.self.style.display = tf ? "block" : "none";
	}
}

VTL.Marker.prototype.setValue = function(dataType, valueStr){
	this.values[dataType] = valueStr;
}
VTL.Marker.prototype.updateLabel = function(imageType){
	const hasType = this.values[imageType];
	if(hasType){
		let html = this.values[imageType];
		
		switch(imageType){
			case VTL.ImageType.LAB:
				this.elems.label.classList.add("lab");
				break;
			default:
				this.elems.label.classList.remove("lab");
				break;
		}

		if(this.chromaCoords){
			if(imageType == VTL.ImageType.LM ){
				this.elems.label.classList.add("chroma");
				html += this.chromaCoords;
			}else {
				this.elems.label.classList.remove("chroma");
			}
		}
		this.elems.label.innerHTML = html;
	}

	this.elems.label.style.display = hasType ? "block" : "none";
}

VTL.Marker.prototype.addEventHandler = function(handler){
	this.handler = handler;
	this.elems.finder.addEventListener("click", handler, false);
	this.elems.finder.addEventListener("touchend", handler, false);
}
VTL.Marker.prototype.removeEventHandler = function(){
	if(!this.handler) return;
	let handler = this.handler;
	this.elems.finder.removeEventListener("click", handler);
	this.elems.finder.removeEventListener("touchend", handler);
	this.handler = null;
}



VTL.Colorbar = class {
	constructor(){
		this.barHeight = 180;
		this.majors = {};
		this.elems = this.createElements();
	}
}

VTL.Colorbar.prototype.createElements = function(){
	let colorbar = document.createElement("div");
	colorbar.className = "colorbar";

	let barWrapper = document.createElement("div");
	barWrapper.className ="barwrapper"

	let barImage = document.createElement("div");
	barImage.className ="barimage";

	let majors = document.createElement("div");
	majors.className ="majors";

	let max = document.createElement("input");
	max.type ="text";
	max.className ="max";

	let min = document.createElement("input");
	min.type ="text";
	min.className ="min";

	let unit = document.createElement("p");
	unit.className ="unit";


	barWrapper.appendChild(barImage);
	barWrapper.appendChild(max);
	barWrapper.appendChild(min);
	barWrapper.appendChild(majors);
	barWrapper.appendChild(unit);
	colorbar.appendChild(barWrapper);

	return {
		self: colorbar,
		min: min,
		max: max,
		majors: majors,
		unit: unit
	};
}

VTL.Colorbar.prototype.loadMajorInfo = function(imageType, majorInfo, display){
	this.majors[imageType] = majorInfo;
	if(display){
		this.displayMajor(imageType);
	}
}

VTL.Colorbar.prototype.displayMajor = function(imageType){
	let majorInfo = this.majors[imageType];
	if(!majorInfo) return;

	const isEbr = majorInfo.style === "ebr";
	if(isEbr) this.elems.self.classList.add("ebr");
	else this.elems.self.classList.remove("ebr");
	
	let majors = majorInfo.majors;


	let begin = 0; end = majors.length;
	if(!isEbr){
		this.elems.min.value = majors[0].value;
		this.elems.max.value = majors[majors.length-1].value;

		begin++;
		end--;
	}
	let frag = document.createDocumentFragment();
	for(let i=begin;i<end;i++){
		let tick = majors[i];
		let tickNode = document.createElement("span");
		tickNode.className = "tick";
		tickNode.style.top = this.barHeight - (this.barHeight * tick.position) - 8 + "px"; // 8 is half elem size
		tickNode.innerText = tick.value;

		frag.appendChild(tickNode);
	}

	this.elems.majors.innerHTML = "";
	this.elems.majors.appendChild(frag);
} 

VTL.ColorbarMajorInfo = {
	createLinearInfo: function(min, max, division){
		let range = max - min;
		let sampler = 10;
		let stepSample = range / sampler;
		let stepSampleStrs = Math.abs(stepSample).toString().split('.');
		let factor = (stepSampleStrs[0] != '0') ? Math.pow(10, -(stepSampleStrs[0].length-1)) :Math.pow(10, stepSampleStrs[1].length) ;
		let step = range/(division-1);
		let majors = new Array(division);
		majors[0] = {position: 0.0, value: min};
		for(let i=1;i < division-1;i++){
			let seed = (i * step) + min;
			let val = Math.round((seed) * factor)/factor;
			val = seed;
			val = Math.round(val * 10)/10;
			let pos = (val - min) / range;
			
			majors[i] = {position: pos, value:val};
		}
		majors[division-1] = {position: 1.0, value:max};

		return this._createObject("normal", majors);
	},
	createLogInfo: function(min, max, division){
		let lmin = Math.log10(min);
		let lmax = Math.log10(max);
		let range = lmax - lmin;
		let factor;
		
		let step = range/(division-1);
		let val = 0, pos = 0;
		let majors = new Array(division);
		majors[0] = {position: 0.0, value: min};
		for(let i=1;i < division-1;i++){
			let seed = (i * step) + lmin;
			factor = Math.pow(10, Math.floor(seed));
			val = Math.round(Math.pow(10, seed) / factor) * factor;			
			pos = ((val != 0 ? Math.log10(val) : 0) - lmin) / range;
			
			majors[i] = {position: pos, value:val};
		}
		majors[division-1] = {position: 1.0, value:max};
		return this._createObject("normal", majors);
	},
	createEbrInfo: function(){
		const majors = [
			{position:0.000, value:1.0},
			{position:0.250, value:4.0},
			{position:0.458, value:6.5},
			{position:0.542, value:7.5},
			{position:0.667, value:9.0},
			{position:1.000, value:13.0},
		];
		return this._createObject("ebr", majors);
	},

	_createObject: function(style, majors){
		return {
			style: style,
			majors: majors
		};
	}
}

VTL.HueChart = class {
	constructor(){
		this.elems = this.createElements();
	}
}

VTL.HueChart.prototype.createElements = function(){
	let area = document.createElement("div");
	area.className ="huechart";

	let chartWrapper = document.createElement("div");
	chartWrapper.className ="chartwrapper";

	let chart = document.createElement("div");
	chart.className ="chart";

	let a = document.createElement("p");
	a.className ="xtitle";
	a.innerText = "a*";

	let b = document.createElement("p");
	b.className ="ytitle";
	b.innerText = "b*";
	
	let bgMunsell = document.createElement("p");
	bgMunsell.className ="bgmunsell";
	
	let distancegroup = document.createElement("div");
	distancegroup.className = "distancegroup";
	//let dimage = document.createElement("div");
	//dimage.className ="dimage";
	let chroma = document.createElement("input");
	chroma.type = "text";
	chroma.className ="chroma";
	let brightness = document.createElement("input");
	brightness.type = "text";
	brightness.className ="brightness";

	//distancegroup.appendChild(dimage);
	distancegroup.appendChild(document.createTextNode("正規化彩度 ≧"));
	distancegroup.appendChild(chroma);
	distancegroup.appendChild(document.createElement("br"));
	distancegroup.appendChild(document.createTextNode("正規化明度 ≧"));
	distancegroup.appendChild(brightness);

	chartWrapper.appendChild(chart);
	chartWrapper.appendChild(a);
	chartWrapper.appendChild(b);

	area.appendChild(chartWrapper);
	area.appendChild(bgMunsell);
	area.appendChild(distancegroup);

	return {
		self:area, 
		bgMunsell: bgMunsell, 
		chroma: chroma,
		brightness: brightness
	};
}

VTL.HueChart.prototype.setBackMunsellName = function(munsellName){
	this.elems.bgMunsell.innerHTML = "背景：" + munsellName;
}
VTL.HueChart.prototype.setChromaLim = function(chroma){
	this.elems.chroma.value = chroma;
}
VTL.HueChart.prototype.setBrightnessLim = function(brightness){
	this.elems.brightness.value = brightness;
}