<?php

/*
Plugin Name: Realaps Data Viewer
Plugin URI: http://www.vtl.co.jp
Description: RealapsData埋め込みパノラマ画像を表示する。
Author: Visual Technology Laboratory
Version: 1.07
Author URI: http://www.vtl.co.jp
*/

new RealapsDataViewer();

class RealapsDataViewer
{
    private $name = 'Realaps Data Viewer';
    private $post_type = 'realaps-data-viewer';

    // custom fields
    const CUSTOM_FIELDS = array(
        "IMAGE_URL"=>"image-url",
        "REMARKS"=>"remarks",
        "USE_MARKDOWN"=>"use-markdown",
        "ACCEPT_USER_FILE"=>"accept-user-file",
        "ACCEPT_NO_DATA_FILE"=>"accept-no-data-file",
        "POINTABLE"=>"pointable",
        "SET_INITIAL_IMAGE_TYPE"=>"set-initial-image-type",
        "INITIAL_IMAGE_TYPE"=>"initial-image-type",
        "IL_REF"=>"il-ref",
    );

    // paths
    private $root_dir;
    private $root_url;

    public function __construct() {
        $this->root_dir = plugin_dir_path(__FILE__);
        $this->root_url = plugins_url('/', __FILE__);
        
        add_action('init', array($this, 'init'));
        register_activation_hook(__FILE__, array($this, 'activation_hook'));
		register_deactivation_hook(__FILE__, array($this, 'deactivation_hook'));
        
        add_filter('single_template', array($this, 'single_template'));
        add_filter('post_type_link', array($this, 'post_type_link'), 1000, 2);
        add_filter('preview_post_link', array($this, 'preview_post_link'), 1000, 2);
        add_action('wp_enqueue_scripts', array($this, 'wp_dequeue_styles'), 1000);
        add_action('wp_enqueue_scripts', array($this, 'wp_enqueue_scripts'), 1001);
        
        add_action('admin_enqueue_scripts', array($this, 'admin_enqueue_scripts'), 1000);
        add_action('add_meta_boxes_'.$this->post_type, array($this, 'add_meta_boxes_post_type'));
        add_action('save_post_'.$this->post_type, array($this, 'save_post_post_type'));
        add_action('before_delete_post', array($this, 'before_delete_post'));
    }

    public function init(){
        register_post_type(
            $this->post_type, array(
            'public'=> true,
            'label'=> __($this->name, $this->post_type),
            'labels'=> array(
            'all_items'=> __('イメージ一覧', $this->post_type),
            ),
            'menu_icon'=> 'dashicons-images-alt2',
            'exclude_from_search'=> true,
            'supports'=> array(
                'title', 
                'thumbnail'
            ))
        );
    }
	
	public function deactivation_hook(){
		unregister_post_type($this->post_type);
		flush_rewrite_rules();
	}
	
	public function activation_hook(){
		$this->init();
		flush_rewrite_rules();
	}

    public function before_delete_post($post_ID){
        $post_type = get_post_type($post_ID);
        if($post_type === $this->post_type){
            foreach(array_keys(self::CUSTOM_FIELDS) as $key){
                delete_post_meta($post_ID, $key);
            }
        }
    }

    public function wp_enqueue_scripts(){
        global $post;
        $post_type = $this->post_type;
        $scripts_url = $this->get_scripts_url();
        if(is_singular($post_type)){
            wp_enqueue_style($post_type.'-style', $this->get_css_url().'style.css');
            wp_enqueue_style($post_type.'-realapsdataviewer', $this->get_css_url().'realapsdataviewer.css');

            //wp_enqueue_script($post_type.'-markdown', 'https://cdn.jsdelivr.net/npm/marked/marked.min.js');
			wp_enqueue_script($post_type.'-markdown', 'https://cdnjs.cloudflare.com/ajax/libs/marked/2.1.3/marked.min.js');
            //wp_enqueue_script($post_type.'-three', $scripts_url.'three.min.js');
			wp_enqueue_script($post_type.'-three', 'https://cdnjs.cloudflare.com/ajax/libs/three.js/r128/three.min.js');
            wp_enqueue_script($post_type.'-rawinflate', $scripts_url.'rawinflate.min.js');
            wp_enqueue_script($post_type.'-cameracontrol', $scripts_url.'cameracontrol.js');
            wp_enqueue_script($post_type.'-omniviewer', $scripts_url.'omniviewer.js');
            wp_enqueue_script($post_type.'-munsell', $scripts_url.'munsell.js');
            wp_enqueue_script($post_type.'-realapsdataloader', $scripts_url.'realapsdataloader.js');
            wp_enqueue_script($post_type.'-realapsdataviewer', $scripts_url.'realapsdataviewer.js');
            wp_enqueue_script($post_type.'-main', $scripts_url.'main.js');

            wp_add_inline_script($this->post_type.'-main', $this->create_js_object($post->ID));
        }
    }

    private function create_js_object($post_ID){
        $values = array(
            "source"=> get_post_meta($post_ID, self::CUSTOM_FIELDS["IMAGE_URL"], true), 
            "remarks"=> array(
                'value'=> get_post_meta($post_ID, self::CUSTOM_FIELDS["REMARKS"], true), 
                "useMarkdown"=> (bool)get_post_meta($post_ID, self::CUSTOM_FIELDS["USE_MARKDOWN"], true),
            ),
            'options'=>array(
                "acceptUserFile"=> (bool)get_post_meta($post_ID, self::CUSTOM_FIELDS["ACCEPT_USER_FILE"], true),
                "acceptNoDataFile"=> (bool)get_post_meta($post_ID, self::CUSTOM_FIELDS["ACCEPT_NO_DATA_FILE"], true),
                "pointable"=> (bool)get_post_meta($post_ID, self::CUSTOM_FIELDS["POINTABLE"], true)
            )
        );
        return 'const rdvparam = ' . json_encode($values);
    }

    // dequeue parent thema styles
    public function wp_dequeue_styles() {
        global $post;
        
        if(is_singular($this->post_type)){
            global $wp_styles;
            //$thema_name = get_template();
            foreach( $wp_styles->queue as $handle ) {
                if(strpos($wp_styles->registered[$handle]->src, '/wp-includes/css/') === false)
                {
                    wp_dequeue_style($handle);
                }
				//$formated_handle = strtolower(str_replace("_", "-", $handle));
                //if (strpos($formated_handle, $thema_name) !== false) {
                //    wp_dequeue_style($handle);
                //}
            }
        }
    }


    public function post_type_link($post_link, $post){
        return $this->rewrite_link($post_link, $post);
    }
    public function preview_post_link($post_link, $post){
        return $this->rewrite_link($post_link, $post);
    }

    private function rewrite_link($link, $post){
        if($post->post_type == $this->post_type){
            if(get_post_meta($post->ID, self::CUSTOM_FIELDS["SET_INITIAL_IMAGE_TYPE"], true)){
                return add_query_arg(
                    'image',
                    get_post_meta($post->ID, self::CUSTOM_FIELDS["INITIAL_IMAGE_TYPE"], true),
                    $link
                );
            }
        }
        return $link;
    }

    public function add_meta_boxes_post_type(){
        add_meta_box(
            $this->post_type.'-details',
            __('詳細', $this->post_type),
            array($this, 'admin_view'),
            $this->post_type,
            'normal',
            'core'
        );
    }

    public function admin_view($post, $box){
        // get_post_meta returns string value.
        // if value is not set, returns empty string.
        // when it is bool value, returns "1" or "0".

        $thumbnail_url = get_post_meta($post->ID, self::CUSTOM_FIELDS["IMAGE_URL"], true);
        $remarks = get_post_meta($post->ID, self::CUSTOM_FIELDS["REMARKS"], true);
        $use_markdown = get_post_meta($post->ID, self::CUSTOM_FIELDS["USE_MARKDOWN"], true) === "1";
        $pointable = get_post_meta($post->ID, self::CUSTOM_FIELDS["POINTABLE"], true) !== "0";
        $accpet_user_file = get_post_meta($post->ID, self::CUSTOM_FIELDS["ACCEPT_USER_FILE"], true) !== "0";
        $accpet_no_data_file = get_post_meta($post->ID, self::CUSTOM_FIELDS["ACCEPT_NO_DATA_FILE"], true) !== "0";
        $set_initial_image_type = get_post_meta($post->ID, self::CUSTOM_FIELDS["SET_INITIAL_IMAGE_TYPE"], true) === "1";
        $initial_image_type = get_post_meta($post->ID, self::CUSTOM_FIELDS["INITIAL_IMAGE_TYPE"], true);

        $initial_image_types = array("lm"=>"輝度", "il"=>"照度", "br"=>"明るさ", "ebr"=>"明るさ検討", "ugr"=>"グレア", "lab"=>"色彩");
?>
<div class="image-block block">
    <h4>画像</h4>
    <p>表示するREALAPSデータをもったJPEGファイルを指定します。</p>
    <div id="inputgroup">
        <input name="<?php echo self::CUSTOM_FIELDS["IMAGE_URL"]; ?>" type="text" autocomplete="off" value="<?php echo $thumbnail_url; ?>"/>
        <button id="image-selector">ファイル選択</button>
    </div>
    <div class="thumbnail-wrapper">
        <img class="thumbnail" src="<?php echo $thumbnail_url;?>"/>
    </div>
</div>
<div class="remarks-block block">
    <h4>備考</h4>
    <p>データに関する情報をここに記述できます。記述がないときはビューワーの備考欄は表示されません。</p>
    <label><input type="checkbox" name="<?php echo self::CUSTOM_FIELDS["USE_MARKDOWN"]; ?>" <?php echo $use_markdown ? "checked" : ""; ?>/>マークダウン記法を使用する</label>
    <textarea name="<?php echo self::CUSTOM_FIELDS["REMARKS"]; ?>"><?php echo $remarks; ?></textarea>
</div>
<div class="option-block">
    <h4>オプション</h4>
    <ul>
        <li><label><input type="checkbox" name="<?php echo self::CUSTOM_FIELDS["POINTABLE"]; ?>" <?php echo $pointable ? "checked" : ""; ?>/>クリックで値を確認できるようにする</label></li>
        <li><label><input type="checkbox" name="<?php echo self::CUSTOM_FIELDS["ACCEPT_USER_FILE"]; ?>" <?php echo $accpet_user_file ? "checked" : ""; ?>/>ファイルのドラッグアンドドロップやファイル選択によってデータを表示できるようにする</label></li>
        <li><label><input type="checkbox" name="<?php echo self::CUSTOM_FIELDS["ACCEPT_NO_DATA_FILE"]; ?>" <?php echo $accpet_no_data_file ? "checked" : ""; ?>/>REALAPSデータを持たない360度画像ファイルも表示できるようにする</label></li>
        <li><label><input type="checkbox" name="<?php echo self::CUSTOM_FIELDS["SET_INITIAL_IMAGE_TYPE"]; ?>" <?php echo $set_initial_image_type ? "checked" : ""; ?>/>はじめに表示する画像データを指定する</label>
            <select name="<?php echo self::CUSTOM_FIELDS["INITIAL_IMAGE_TYPE"]; ?>" value="<?php echo $initial_image_type; ?>">
            <?php 
            foreach($initial_image_types as $key => $value) { ?>
                <option value="<?php echo $key; ?>" <?php echo $initial_image_type == $key ? "selected" :""; ?>><?php echo $value; ?></option>
            <?php
            }
            ?>
            </select>
        </li>
    </ul>
</div>
<?php
    }

    public function admin_enqueue_scripts($hook_suffix){
        global $post_type;

        if($post_type == $this->post_type && ($hook_suffix == 'post-new.php' || $hook_suffix == 'post.php')) {
            wp_enqueue_media(); // necessary to use media-library(javascript)

            wp_enqueue_style($this->post_type.'-edit', $this->get_css_url().'edit.css');
            wp_enqueue_script($this->post_type.'-edit', $this->get_scripts_url().'edit.js');
        }
    }

    public function save_post_post_type($post_ID){
        // for string values
        $keys = array(
            self::CUSTOM_FIELDS["IMAGE_URL"], 
            self::CUSTOM_FIELDS["REMARKS"], 
            self::CUSTOM_FIELDS["INITIAL_IMAGE_TYPE"]);
        foreach($keys as $key){
            if(isset($_POST[$key])){
                update_post_meta($post_ID, $key, $_POST[$key]);
            }
        }

        // for boolean values
        $keys = array(
            self::CUSTOM_FIELDS["SET_INITIAL_IMAGE_TYPE"],
            self::CUSTOM_FIELDS["ACCEPT_USER_FILE"],
            self::CUSTOM_FIELDS["ACCEPT_NO_DATA_FILE"], 
            self::CUSTOM_FIELDS["POINTABLE"],
            self::CUSTOM_FIELDS["USE_MARKDOWN"],
        );
        
        foreach($keys as $key){
            update_post_meta($post_ID, $key, isset($_POST[$key]));
        }
    }

    public function single_template($single){
        global $post;
        if($post->post_type===$this->post_type) {
            $single = $this->get_assets_dir().'post.php';
        }
        return $single;
    }

    private function get_assets_dir(){
        return $this->root_dir . 'assets/';    
    }

    private function get_assets_url(){
        return $this->root_url . 'assets/';    
    }

    private function get_images_url(){
        return $this->get_assets_url() . 'images/';
    }

    private function get_css_url(){
        return $this->get_assets_url() . 'css/';
    }
    private function get_scripts_url(){
        return $this->get_assets_url() . 'scripts/';
    }

}