if(typeof VTL === "undefined") VTL = {};
VTL.Munsell = {
    getClosestValue: function(l, a, b){
        let closest = Infinity;
        let closestMunsell = undefined;
        for(let i=0;i<this.values.length;i++){
            let munsell = this.values[i];
            let distance = Math.sqrt((munsell.L_star - l) * (munsell.L_star - l) + (munsell.a_star - a) * (munsell.a_star - a)  + (munsell.b_star - b) * (munsell.b_star - b));
            if(distance < closest) {
                closest = distance;
                closestMunsell = munsell;
            }
        }
        if(20 < closest) return { "name": "-","L_star": 0, "a_star": 0, "b_star": 0};
        
        return closestMunsell; 
    },
    getNormalizedValue(l, a, b, nParams)
    {
        let rotAngle = 180.0 * Math.atan2(b, a) / Math.PI;
        //Hueデータの代表回転角度が、求めた回転角より大きい側と小さい側で最も近いHueデータを求める
        //N0～N10は除外する
        let istart = 1;
        let sa;
        let saMinCW = new Array(2);
        let saMinCCW = new Array(2);
        let saCW = new Array(2);
        let saCCW = new Array(2);
        //反時計回りに引き算して正になるもののの一番小さいものを求める
        saMinCCW[0] = saMinCCW[1] = Number.MAX_VALUE;
        saMinCW[0] = saMinCW[1] = Number.MAX_VALUE;
        for (let i = istart; i < nParams.length; i++)
        {
            //反時計回りに引き算して正になるもののの一番小さいものを求める
            sa = nParams[i].angle - rotAngle;
            if (sa >= 0 && saMinCCW[0] > sa)
            {
                saMinCCW[0] = sa;
                saCCW[0] = i;
            }
            //時計回りに引き算して正になるもののの一番小さいものを求める
            sa = rotAngle - nParams[i].angle;
            if (sa >= 0 && saMinCW[0] > sa)
            {
                saMinCW[0] = sa;
                saCW[0] = i;
            }
            //-180と180をまたぐところが問題なので、0～360にしてもう一度やってみる
            {
                //反時計回りに引き算して正になるもののの一番小さいものを求める
                sa = _0to360(nParams[i].angle) - _0to360(rotAngle);
                if (sa >= 0 && saMinCCW[1] > sa)
                {
                    saMinCCW[1] = sa;
                    saCCW[1] = i;
                }
                //時計回りに引き算して正になるもののの一番小さいものを求める
                sa = _0to360(rotAngle) - _0to360(nParams[i].angle);
                if (sa >= 0 && saMinCW[1] > sa)
                {
                    saMinCW[1] = sa;
                    saCW[1] = i;
                }
            }
        }
        let staIndex = saMinCW[0] < saMinCW[1] ? saCW[0] : saCW[1];
        
        let endIndex = saMinCCW[0] < saMinCCW[1] ? saCCW[0] : saCCW[1];
        if (endIndex - staIndex > 1)
        {
            throw new Error("Error: " + staIndex.ToString() + " ～ " + endIndex.ToString());
        }
        let ratio = 0.5;
        if (nParams[staIndex].angle < nParams[endIndex].angle)
        {
            ratio = (rotAngle - nParams[staIndex].angle) / (nParams[endIndex].angle - nParams[staIndex].angle);
        }
        else
        {
            ratio = (_0to360(rotAngle) - _0to360(nParams[staIndex].angle)) / (_0to360(nParams[endIndex].angle) - _0to360(nParams[staIndex].angle));
        }
        let chroma = (nParams[endIndex].chroma - nParams[staIndex].chroma) * ratio + nParams[staIndex].chroma;
        let lStar = (nParams[endIndex].lStar - nParams[staIndex].lStar) * ratio + nParams[staIndex].lStar;
        norm_a_star = a / chroma;
        norm_b_star = b / chroma;
        //param.NormChroma = param.Chroma / chroma;
        if (l >= lStar)
        {
            norm_l_star = (l - lStar) / (100.0 - lStar);
        }
        else
        {
            norm_l_star = (lStar - l) / (0.0 - lStar);
        }

        return {lStar: norm_l_star, aStar: norm_a_star, bStar: norm_b_star};

        function _0to360(angle)
        {
            let tmp = angle;
            while (true)
            {
                if (tmp >= 360.0)
                {
                    tmp -= 360.0;
                }
                else if (tmp < 0)
                {
                    tmp += 360.0;
                }
                else
                {
                    break;
                }
            }
            return tmp;
        }
    },
    values: [
{ "name": "N10","L_star": 100, "a_star": 0, "b_star": 0},
{ "name": "N9.5","L_star": 95.0555748001624, "a_star": 5.55111512312578E-14, "b_star": 0},
{ "name": "N9","L_star": 90.1787402752534, "a_star": 0, "b_star": 0},
{ "name": "N8.5","L_star": 85.3420891480087, "a_star": 0, "b_star": 0},
{ "name": "N8","L_star": 80.5270589174699, "a_star": 0, "b_star": 0},
{ "name": "N7.5","L_star": 75.7056880951435, "a_star": 5.55111512312578E-14, "b_star": -2.22044604925031E-14},
{ "name": "N7","L_star": 70.8571284675481, "a_star": 5.55111512312578E-14, "b_star": 0},
{ "name": "N6.5","L_star": 65.9738265861701, "a_star": 0, "b_star": 0},
{ "name": "N6","L_star": 61.0454854659928, "a_star": 0, "b_star": 0},
{ "name": "N5.5","L_star": 56.0577450332982, "a_star": 5.55111512312578E-14, "b_star": 0},
{ "name": "N5","L_star": 51.0016082780202, "a_star": 0, "b_star": 0},
{ "name": "N4.5","L_star": 45.8933391934866, "a_star": 0, "b_star": 0},
{ "name": "N4","L_star": 40.7352896595015, "a_star": 0, "b_star": 1.11022302462516E-14},
{ "name": "N3.5","L_star": 35.553305272152, "a_star": 2.77555756156289E-14, "b_star": -1.11022302462516E-14},
{ "name": "N3","L_star": 30.3782397967149, "a_star": 0, "b_star": -1.11022302462516E-14},
{ "name": "N2.5","L_star": 25.2538488392805, "a_star": 2.77555756156289E-14, "b_star": 0},
{ "name": "N2","L_star": 20.2351149553776, "a_star": 0, "b_star": 0},
{ "name": "N1.5","L_star": 15.3343144444061, "a_star": 0, "b_star": 0},
{ "name": "N1","L_star": 10.4090021335651, "a_star": 0, "b_star": 0},
{ "name": "2.5R9/6","L_star": 89.8268304826796, "a_star": 26.9449250928079, "b_star": 10.3412634902248},
{ "name": "2.5R9/4","L_star": 90.0539870712774, "a_star": 16.3093654362047, "b_star": 6.3216371206511},
{ "name": "2.5R9/3","L_star": 90.1140900455589, "a_star": 10.7128390451883, "b_star": 4.1811589020035},
{ "name": "2.5R9/2","L_star": 90.1371884550925, "a_star": 5.2574071488814, "b_star": 2.03411890587959},
{ "name": "2.5R9/1","L_star": 90.1556599466689, "a_star": 1.89805357082584, "b_star": 0.772048385450552},
{ "name": "2.5R8/10","L_star": 79.908855158823, "a_star": 44.5854427742164, "b_star": 16.2880126630493},
{ "name": "2.5R8/8","L_star": 80.2133297573375, "a_star": 34.4725550283241, "b_star": 12.7942726677866},
{ "name": "2.5R8/6","L_star": 80.3928535877543, "a_star": 24.4790741650918, "b_star": 9.23255900647197},
{ "name": "2.5R8/4","L_star": 80.4320354035271, "a_star": 15.4501340672125, "b_star": 5.94763717009721},
{ "name": "2.5R8/3","L_star": 80.4600029327644, "a_star": 10.4983952289986, "b_star": 4.04363653181481},
{ "name": "2.5R8/2","L_star": 80.4823652850047, "a_star": 5.94476857010956, "b_star": 2.27709959456441},
{ "name": "2.5R8/1","L_star": 80.4991302489017, "a_star": 2.25209224698963, "b_star": 0.889234862503407},
{ "name": "2.5R7/16","L_star": 70.1969359540931, "a_star": 70.1724614872202, "b_star": 25.8982447448487},
{ "name": "2.5R7/14","L_star": 70.322801700102, "a_star": 61.1517174210787, "b_star": 21.8546913563405},
{ "name": "2.5R7/12","L_star": 70.524816740343, "a_star": 51.6450224703026, "b_star": 18.7787507656976},
{ "name": "2.5R7/10","L_star": 70.6358706940783, "a_star": 41.7178318255328, "b_star": 15.318078440566},
{ "name": "2.5R7/8","L_star": 70.6843673931195, "a_star": 33.2568551570115, "b_star": 12.2400389490426},
{ "name": "2.5R7/6","L_star": 70.7258928449287, "a_star": 24.362229621202, "b_star": 9.00831401382662},
{ "name": "2.5R7/4","L_star": 70.7742890005338, "a_star": 15.6942362209064, "b_star": 5.71557197342942},
{ "name": "2.5R7/3","L_star": 70.7950136941808, "a_star": 11.4144659074782, "b_star": 4.18091351118175},
{ "name": "2.5R7/2","L_star": 70.8088246600745, "a_star": 7.23758609805492, "b_star": 2.70911058404191},
{ "name": "2.5R7/1","L_star": 70.8295328728496, "a_star": 3.40283008000347, "b_star": 1.33753388194082},
{ "name": "2.5R6/18","L_star": 60.6667259617269, "a_star": 77.1215919741456, "b_star": 26.2093069126355},
{ "name": "2.5R6/16","L_star": 60.6932726141122, "a_star": 68.6795550935825, "b_star": 22.9965092994177},
{ "name": "2.5R6/14","L_star": 60.7286395882807, "a_star": 59.0062999219482, "b_star": 20.2178463680589},
{ "name": "2.5R6/12","L_star": 60.7728025078901, "a_star": 50.9251875029935, "b_star": 17.7363240581454},
{ "name": "2.5R6/10","L_star": 60.8169146771618, "a_star": 42.1541700095602, "b_star": 14.8947461585401},
{ "name": "2.5R6/8","L_star": 60.8697824947294, "a_star": 33.371041793552, "b_star": 11.8758636172922},
{ "name": "2.5R6/6","L_star": 60.9049873462251, "a_star": 25.4395026935619, "b_star": 9.04791509738474},
{ "name": "2.5R6/4","L_star": 60.9489481351208, "a_star": 16.5100434835887, "b_star": 5.69279287081734},
{ "name": "2.5R6/3","L_star": 60.9753005166793, "a_star": 12.092807827642, "b_star": 4.16809710883652},
{ "name": "2.5R6/2","L_star": 60.9928587518952, "a_star": 7.81241911799613, "b_star": 2.69602022767064},
{ "name": "2.5R6/1","L_star": 61.0104089824442, "a_star": 3.74434156759251, "b_star": 1.373618763133},
{ "name": "2.5R5/20","L_star": 50.7456524262333, "a_star": 87.6343779094315, "b_star": 26.3977475371562},
{ "name": "2.5R5/18","L_star": 50.6755045527692, "a_star": 78.7180186955532, "b_star": 23.6836405471377},
{ "name": "2.5R5/16","L_star": 50.6169350351417, "a_star": 69.9175675505038, "b_star": 21.3428601063606},
{ "name": "2.5R5/14","L_star": 50.6755045527692, "a_star": 61.0802827681015, "b_star": 19.1114102141495},
{ "name": "2.5R5/12","L_star": 50.7105969300653, "a_star": 53.5547719709137, "b_star": 17.0033715817142},
{ "name": "2.5R5/10","L_star": 50.7690023152908, "a_star": 44.31473929712, "b_star": 14.3133317985768},
{ "name": "2.5R5/8","L_star": 50.8156531617798, "a_star": 35.5001757422417, "b_star": 11.6598979574617},
{ "name": "2.5R5/6","L_star": 50.8622389556151, "a_star": 26.6541116388931, "b_star": 8.70497347297094},
{ "name": "2.5R5/4","L_star": 50.8971357463743, "a_star": 17.5858847989351, "b_star": 5.638487876234},
{ "name": "2.5R5/3","L_star": 50.9203800622637, "a_star": 12.9099076312196, "b_star": 4.09973478262542},
{ "name": "2.5R5/2","L_star": 50.9436082418668, "a_star": 8.35287467822105, "b_star": 2.60042348156364},
{ "name": "2.5R5/1","L_star": 50.966820313159, "a_star": 4.00405147095395, "b_star": 1.23415941502425},
{ "name": "2.5R4/18","L_star": 40.4428345189212, "a_star": 80.2461968236263, "b_star": 21.3966909090999},
{ "name": "2.5R4/16","L_star": 40.4917875655919, "a_star": 71.2329202296744, "b_star": 19.2930507280336},
{ "name": "2.5R4/14","L_star": 40.3937964103769, "a_star": 63.1393046429698, "b_star": 17.3582705124407},
{ "name": "2.5R4/12","L_star": 40.4428345189212, "a_star": 54.0698357997671, "b_star": 15.292295656635},
{ "name": "2.5R4/10","L_star": 40.4917875655919, "a_star": 45.4864971850352, "b_star": 13.0995164433187},
{ "name": "2.5R4/8","L_star": 40.5406559183046, "a_star": 34.9258226010457, "b_star": 12.3952686215144},
{ "name": "2.5R4/6","L_star": 40.5894399424351, "a_star": 28.0550835876821, "b_star": 8.33846930971447},
{ "name": "2.5R4/4","L_star": 40.6219159552695, "a_star": 19.1034768584799, "b_star": 5.73277328787235},
{ "name": "2.5R4/3","L_star": 40.6543547569545, "a_star": 14.5698070463368, "b_star": 4.35520739672569},
{ "name": "2.5R4/2","L_star": 40.6705602368878, "a_star": 9.90048882254527, "b_star": 2.96303643771763},
{ "name": "2.5R4/1","L_star": 40.686756453899, "a_star": 5.08026756619642, "b_star": 1.50068227120886},
{ "name": "2.5R3/16","L_star": 29.7902728718213, "a_star": 74.2810734139486, "b_star": 16.4777201489327},
{ "name": "2.5R3/14","L_star": 30.0321751958295, "a_star": 65.9147707246397, "b_star": 15.2588463594213},
{ "name": "2.5R3/12","L_star": 30.1277389127189, "a_star": 57.9005149266903, "b_star": 13.9165929024555},
{ "name": "2.5R3/10","L_star": 30.108168380496, "a_star": 48.7248804601997, "b_star": 12.1271451801837},
{ "name": "2.5R3/8","L_star": 30.1595056550719, "a_star": 39.4292527401068, "b_star": 10.2285850650974},
{ "name": "2.5R3/6","L_star": 30.2107289912647, "a_star": 29.7767825763274, "b_star": 7.88012675916103},
{ "name": "2.5R3/4","L_star": 30.2569762620997, "a_star": 20.9656998198777, "b_star": 5.62075958845689},
{ "name": "2.5R3/3","L_star": 30.2788506236177, "a_star": 15.433909573709, "b_star": 4.97351058058286},
{ "name": "2.5R3/2","L_star": 30.3031312422504, "a_star": 11.2544596709823, "b_star": 3.05385567568072},
{ "name": "2.5R3/1","L_star": 30.3273864228217, "a_star": 5.93008785973809, "b_star": 1.59488919944403},
{ "name": "2.5R2/14","L_star": 19.5031772121629, "a_star": 64.7121086988978, "b_star": 2.04576342669827},
{ "name": "2.5R2/12","L_star": 19.8019844306396, "a_star": 55.6650369310409, "b_star": 2.52506426413315},
{ "name": "2.5R2/10","L_star": 19.987741340579, "a_star": 46.6705649535333, "b_star": 2.97245437853191},
{ "name": "2.5R2/8","L_star": 20.0238615468325, "a_star": 37.7828320407082, "b_star": 2.83164824262246},
{ "name": "2.5R2/6","L_star": 20.0759076516696, "a_star": 28.3138779256128, "b_star": 2.53700015262669},
{ "name": "2.5R2/4","L_star": 20.115841166286, "a_star": 19.8977872685911, "b_star": 2.01739118378502},
{ "name": "2.5R2/3","L_star": 20.135774850191, "a_star": 15.5894265699165, "b_star": 1.69406952638759},
{ "name": "2.5R2/2","L_star": 20.1556865661722, "a_star": 10.9993081400376, "b_star": 1.25555401346962},
{ "name": "2.5R2/1","L_star": 20.1755763746542, "a_star": 4.92678100808017, "b_star": 1.45353348449128},
{ "name": "2.5R1/10","L_star": 9.94585687546375, "a_star": 45.3247869666011, "b_star": -5.13738022754118},
{ "name": "2.5R1/8","L_star": 10.1528804424542, "a_star": 37.7652195679653, "b_star": -3.53779935313766},
{ "name": "2.5R1/6","L_star": 10.2287302884101, "a_star": 29.8322972623231, "b_star": -2.29581782551824},
{ "name": "2.5R1/4","L_star": 10.2589477463405, "a_star": 21.6455327575549, "b_star": -1.35538100002903},
{ "name": "2.5R1/3","L_star": 10.281565283574, "a_star": 17.645299120054, "b_star": -0.942642889895601},
{ "name": "2.5R1/2","L_star": 10.2966220415417, "a_star": 13.1063325796451, "b_star": -0.605652788496897},
{ "name": "2.5R1/1","L_star": 10.3191749016881, "a_star": 7.52846555756828, "b_star": -0.334570640269499},
{ "name": "5R9/6","L_star": 89.8407660675763, "a_star": 26.287471653658, "b_star": 14.3516603470993},
{ "name": "5R9/4","L_star": 90.0493609431551, "a_star": 16.4199192054656, "b_star": 8.85901556642039},
{ "name": "5R9/3","L_star": 90.1094691567687, "a_star": 10.7767093679032, "b_star": 5.93465639614745},
{ "name": "5R9/2","L_star": 90.132569577483, "a_star": 5.79212837658499, "b_star": 3.23536479270923},
{ "name": "5R9/1","L_star": 90.1556599466689, "a_star": 1.91258574271208, "b_star": 1.19422034368124},
{ "name": "5R8/10","L_star": 79.8975411496159, "a_star": 43.918036013134, "b_star": 22.6419284257586},
{ "name": "5R8/8","L_star": 80.1852185742504, "a_star": 34.2671924530985, "b_star": 17.6423592295837},
{ "name": "5R8/6","L_star": 80.3816529307179, "a_star": 24.1660763943196, "b_star": 12.8021317205187},
{ "name": "5R8/4","L_star": 80.4264399506073, "a_star": 15.5417975424853, "b_star": 8.25848432896934},
{ "name": "5R8/3","L_star": 80.4544107243379, "a_star": 10.6342622284528, "b_star": 5.57544725278969},
{ "name": "5R8/2","L_star": 80.4767756687889, "a_star": 5.99287944857874, "b_star": 3.06712657195245},
{ "name": "5R8/1","L_star": 80.4991302489017, "a_star": 2.21726501561159, "b_star": 1.16172034236446},
{ "name": "5R7/14","L_star": 70.301849562944, "a_star": 60.0375387375722, "b_star": 30.5918710431778},
{ "name": "5R7/12","L_star": 70.5039623609524, "a_star": 50.9597820112432, "b_star": 25.9084119453827},
{ "name": "5R7/10","L_star": 70.6289381605852, "a_star": 41.3353159472725, "b_star": 21.195621178088},
{ "name": "5R7/8","L_star": 70.6774426153785, "a_star": 32.9593501842113, "b_star": 16.8227475258338},
{ "name": "5R7/6","L_star": 70.7189746977356, "a_star": 24.1684443334062, "b_star": 12.3561617048199},
{ "name": "5R7/4","L_star": 70.7673785689525, "a_star": 15.6681872764644, "b_star": 7.95416428946258},
{ "name": "5R7/3","L_star": 70.7881065627101, "a_star": 11.4371807958759, "b_star": 5.71380165181008},
{ "name": "5R7/2","L_star": 70.8088246600745, "a_star": 7.34714249805468, "b_star": 3.58379117633771},
{ "name": "5R7/1","L_star": 70.8295328728496, "a_star": 3.48732813751385, "b_star": 1.66937003337371},
{ "name": "5R6/18","L_star": 60.6667259617269, "a_star": 75.265141455748, "b_star": 39.2295238470106},
{ "name": "5R6/16","L_star": 60.675576888377, "a_star": 66.9959279303363, "b_star": 34.077655101922},
{ "name": "5R6/14","L_star": 60.7198009014627, "a_star": 58.2326790531798, "b_star": 29.4269141877728},
{ "name": "5R6/12","L_star": 60.7639739886647, "a_star": 50.2388409968351, "b_star": 25.3172164520674},
{ "name": "5R6/10","L_star": 60.8080962963385, "a_star": 41.7214546732553, "b_star": 21.0537438074046},
{ "name": "5R6/8","L_star": 60.852167970168, "a_star": 33.0313388895663, "b_star": 16.5440855779527},
{ "name": "5R6/6","L_star": 60.8961891551695, "a_star": 25.1275191541624, "b_star": 12.5174115526462},
{ "name": "5R6/4","L_star": 60.940159995696, "a_star": 16.4898996767213, "b_star": 8.00301146376821},
{ "name": "5R6/3","L_star": 60.9665183944677, "a_star": 12.0807123853891, "b_star": 5.77258576986015},
{ "name": "5R6/2","L_star": 60.9928587518952, "a_star": 7.83053809675444, "b_star": 3.64390564890116},
{ "name": "5R6/1","L_star": 61.0104089824442, "a_star": 3.74903863724085, "b_star": 1.68179981424699},
{ "name": "5R5/20","L_star": 50.7456524262333, "a_star": 84.4140747030641, "b_star": 42.6769611558806},
{ "name": "5R5/18","L_star": 50.6755045527692, "a_star": 77.209816412767, "b_star": 38.8029614874748},
{ "name": "5R5/16","L_star": 50.6052087657505, "a_star": 68.6604516597436, "b_star": 34.1138401098752},
{ "name": "5R5/14","L_star": 50.6637988810833, "a_star": 59.9795451511969, "b_star": 29.8029379681316},
{ "name": "5R5/12","L_star": 50.6989035736593, "a_star": 52.476262489321, "b_star": 25.9748151994198},
{ "name": "5R5/10","L_star": 50.7573294125503, "a_star": 43.6475911458442, "b_star": 21.6419577768026},
{ "name": "5R5/8","L_star": 50.8039965577119, "a_star": 34.9347835363699, "b_star": 17.1224455803729},
{ "name": "5R5/6","L_star": 50.8505985934447, "a_star": 26.3214365464769, "b_star": 12.6867745346043},
{ "name": "5R5/4","L_star": 50.8971357463743, "a_star": 17.5733920815705, "b_star": 8.21725273198859},
{ "name": "5R5/3","L_star": 50.9203800622637, "a_star": 12.9021612092903, "b_star": 5.9121088814361},
{ "name": "5R5/2","L_star": 50.9436082418668, "a_star": 8.26132182820943, "b_star": 3.70130388047674},
{ "name": "5R5/1","L_star": 50.966820313159, "a_star": 3.88914274198232, "b_star": 1.72725664215478},
{ "name": "5R4/18","L_star": 40.4917875655919, "a_star": 77.971255357602, "b_star": 37.62559784987},
{ "name": "5R4/16","L_star": 40.4917875655919, "a_star": 69.6344651837345, "b_star": 33.647960374825},
{ "name": "5R4/14","L_star": 40.3774314075978, "a_star": 61.451039967594, "b_star": 29.4942826981035},
{ "name": "5R4/12","L_star": 40.4264979523013, "a_star": 52.561552807031, "b_star": 25.2481222434782},
{ "name": "5R4/10","L_star": 40.4754793119243, "a_star": 44.4572197745986, "b_star": 21.1701298115852},
{ "name": "5R4/8","L_star": 40.5243758552341, "a_star": 36.3402081570786, "b_star": 17.1963641189821},
{ "name": "5R4/6","L_star": 40.5731879484504, "a_star": 27.5498426687842, "b_star": 12.8958054518358},
{ "name": "5R4/4","L_star": 40.6219159552695, "a_star": 19.0145852187319, "b_star": 8.74993215390143},
{ "name": "5R4/3","L_star": 40.6381400008437, "a_star": 14.4868599617937, "b_star": 6.56087260306445},
{ "name": "5R4/2","L_star": 40.6705602368878, "a_star": 9.77148769234712, "b_star": 4.3434545783272},
{ "name": "5R4/1","L_star": 40.686756453899, "a_star": 4.97403739999785, "b_star": 2.14934775279113},
{ "name": "5R3/16","L_star": 29.7927541914789, "a_star": 73.0143423865586, "b_star": 28.3659066278757},
{ "name": "5R3/14","L_star": 30.0395403439865, "a_star": 64.6943838703573, "b_star": 25.7194395212977},
{ "name": "5R3/12","L_star": 30.1301840618839, "a_star": 56.7690063614576, "b_star": 23.1478294201912},
{ "name": "5R3/10","L_star": 30.0983768812887, "a_star": 47.6163132685528, "b_star": 20.0516748304184},
{ "name": "5R3/8","L_star": 30.1521787492893, "a_star": 38.3172426515735, "b_star": 16.3855862763187},
{ "name": "5R3/6","L_star": 30.2034183236864, "a_star": 29.0773174007357, "b_star": 12.3225724280351},
{ "name": "5R3/4","L_star": 30.2496802089405, "a_star": 20.6852994887553, "b_star": 8.56662233773965},
{ "name": "5R3/3","L_star": 30.2739914414501, "a_star": 16.0162702937356, "b_star": 6.5263048280297},
{ "name": "5R3/2","L_star": 30.3007043262, "a_star": 11.0254372215855, "b_star": 4.41816160372791},
{ "name": "5R3/1","L_star": 30.3249620475793, "a_star": 5.74660288063431, "b_star": 2.23933916767149},
{ "name": "5R2/14","L_star": 19.4494345468919, "a_star": 64.9840319928495, "b_star": 11.0243650016276},
{ "name": "5R2/12","L_star": 19.7613465019228, "a_star": 55.5279929468625, "b_star": 10.7564688905919},
{ "name": "5R2/10","L_star": 19.9756851449229, "a_star": 46.8360774029047, "b_star": 10.025519518051},
{ "name": "5R2/8","L_star": 20.0118295280528, "a_star": 37.7819504278203, "b_star": 8.85033423230829},
{ "name": "5R2/6","L_star": 20.063910341955, "a_star": 28.1272262892908, "b_star": 6.78142789967741},
{ "name": "5R2/4","L_star": 20.1078615281194, "a_star": 19.7117015630783, "b_star": 4.8439130692783},
{ "name": "5R2/3","L_star": 20.1278040161726, "a_star": 15.3955621668898, "b_star": 3.82688698409196},
{ "name": "5R2/2","L_star": 20.1517059775056, "a_star": 10.8424098783517, "b_star": 2.70662214175993},
{ "name": "5R2/1","L_star": 20.1755763746542, "a_star": 5.85320412008539, "b_star": 1.42085468523663},
{ "name": "5R1/10","L_star": 9.89940045222461, "a_star": 45.5237379827566, "b_star": 0.762608741596915},
{ "name": "5R1/8","L_star": 10.1147899713589, "a_star": 38.0834713584489, "b_star": 1.49367419972353},
{ "name": "5R1/6","L_star": 10.2211650355277, "a_star": 30.020877332999, "b_star": 1.77509725424883},
{ "name": "5R1/4","L_star": 10.2513999023643, "a_star": 21.7073755466259, "b_star": 1.56040407871101},
{ "name": "5R1/3","L_star": 10.2740304307288, "a_star": 17.6187085462634, "b_star": 1.41769728464304},
{ "name": "5R1/2","L_star": 10.2966220415417, "a_star": 13.0440940841753, "b_star": 1.12501029033447},
{ "name": "5R1/1","L_star": 10.3191749016881, "a_star": 7.5182439904929, "b_star": 0.609931999743346},
{ "name": "7.5R9/6","L_star": 89.8593411409776, "a_star": 25.100652065316, "b_star": 19.2805813511503},
{ "name": "7.5R9/4","L_star": 90.0308523933759, "a_star": 16.3240384254527, "b_star": 11.870057607333},
{ "name": "7.5R9/3","L_star": 90.1002261716037, "a_star": 10.9262849062928, "b_star": 7.93994763966432},
{ "name": "7.5R9/2","L_star": 90.1279502978124, "a_star": 5.89852519665929, "b_star": 4.38627886492764},
{ "name": "7.5R9/1","L_star": 90.1510426764292, "a_star": 1.97506030668332, "b_star": 1.63347396933653},
{ "name": "7.5R8/10","L_star": 79.8749051187621, "a_star": 42.3603793674457, "b_star": 30.3618360143411},
{ "name": "7.5R8/8","L_star": 80.1852185742504, "a_star": 33.2164000958732, "b_star": 23.8064642207745},
{ "name": "7.5R8/6","L_star": 80.3760516258334, "a_star": 23.8010620710501, "b_star": 17.1061827649374},
{ "name": "7.5R8/4","L_star": 80.420843848224, "a_star": 15.411486375175, "b_star": 10.885730186388},
{ "name": "7.5R8/3","L_star": 80.448817867389, "a_star": 10.5105500924914, "b_star": 7.44957592568793},
{ "name": "7.5R8/2","L_star": 80.4767756687889, "a_star": 5.9161964882361, "b_star": 4.24448282945153},
{ "name": "7.5R8/1","L_star": 80.4935425748736, "a_star": 2.14784771086013, "b_star": 1.70631712792204},
{ "name": "7.5R7/16","L_star": 70.1899326280991, "a_star": 65.2302623324311, "b_star": 50.4516583090154},
{ "name": "7.5R7/14","L_star": 70.3088347389812, "a_star": 57.2780214421259, "b_star": 42.2496459324057},
{ "name": "7.5R7/12","L_star": 70.5317659666793, "a_star": 48.833145459562, "b_star": 35.3400908045844},
{ "name": "7.5R7/10","L_star": 70.622004517354, "a_star": 39.9150795806199, "b_star": 28.6058397894248},
{ "name": "7.5R7/8","L_star": 70.6635897395448, "a_star": 32.2128976194788, "b_star": 22.7384819541155},
{ "name": "7.5R7/6","L_star": 70.7120554465534, "a_star": 23.4219020922098, "b_star": 16.3277305534635},
{ "name": "7.5R7/4","L_star": 70.7604670364579, "a_star": 15.5119672910047, "b_star": 10.5548259091828},
{ "name": "7.5R7/3","L_star": 70.7811983316404, "a_star": 11.3726308916725, "b_star": 7.61601079688092},
{ "name": "7.5R7/2","L_star": 70.8088246600745, "a_star": 7.33886737399442, "b_star": 4.8526036952113},
{ "name": "7.5R7/1","L_star": 70.8295328728496, "a_star": 3.46170917014976, "b_star": 2.3053226741075},
{ "name": "7.5R6/18","L_star": 60.675576888377, "a_star": 71.4583420703946, "b_star": 55.9467898982289},
{ "name": "7.5R6/16","L_star": 60.6667259617269, "a_star": 63.8740921911544, "b_star": 48.3630616871474},
{ "name": "7.5R6/14","L_star": 60.7109601776115, "a_star": 56.0656475780506, "b_star": 41.335900038514},
{ "name": "7.5R6/12","L_star": 60.7551434382602, "a_star": 48.2995061783426, "b_star": 34.7022877695583},
{ "name": "7.5R6/10","L_star": 60.7992758901635, "a_star": 40.6005077619592, "b_star": 28.5046628065061},
{ "name": "7.5R6/8","L_star": 60.852167970168, "a_star": 32.1346619978782, "b_star": 22.074149346431},
{ "name": "7.5R6/6","L_star": 60.8961891551695, "a_star": 24.1582888620018, "b_star": 16.1591695945123},
{ "name": "7.5R6/4","L_star": 60.940159995696, "a_star": 16.1242442801017, "b_star": 10.663661267002},
{ "name": "7.5R6/3","L_star": 60.957734267663, "a_star": 11.9483145133185, "b_star": 7.85178547854137},
{ "name": "7.5R6/2","L_star": 60.984080635441, "a_star": 7.83973399552923, "b_star": 5.05898546100356},
{ "name": "7.5R6/1","L_star": 61.0104089824442, "a_star": 3.79497198681561, "b_star": 2.46477784152932},
{ "name": "7.5R5/20","L_star": 50.7456524262333, "a_star": 79.7120227631106, "b_star": 61.0739799416475},
{ "name": "7.5R5/18","L_star": 50.68720611576, "a_star": 73.098761528453, "b_star": 54.5146212648313},
{ "name": "7.5R5/16","L_star": 50.6052087657505, "a_star": 65.7995813135181, "b_star": 48.0175504436277},
{ "name": "7.5R5/14","L_star": 50.6520890970938, "a_star": 57.8901168050734, "b_star": 41.2037049858845},
{ "name": "7.5R5/12","L_star": 50.6989035736593, "a_star": 50.5062768758383, "b_star": 35.0574417816157},
{ "name": "7.5R5/10","L_star": 50.7456524262333, "a_star": 42.1796238869509, "b_star": 28.8462223262908},
{ "name": "7.5R5/8","L_star": 50.7923358843108, "a_star": 33.905762282902, "b_star": 22.7913138654128},
{ "name": "7.5R5/6","L_star": 50.8389541761052, "a_star": 25.5010704314196, "b_star": 16.5254278175125},
{ "name": "7.5R5/4","L_star": 50.8855075285586, "a_star": 17.1309506401024, "b_star": 10.6434928237081},
{ "name": "7.5R5/3","L_star": 50.9203800622637, "a_star": 12.6423689745374, "b_star": 7.68998566517962},
{ "name": "7.5R5/2","L_star": 50.9436082418668, "a_star": 8.14003256824569, "b_star": 4.89255345769963},
{ "name": "7.5R5/1","L_star": 50.966820313159, "a_star": 3.86510740384449, "b_star": 2.3201950984443},
{ "name": "7.5R4/20","L_star": 40.3937964103769, "a_star": 81.8058460937141, "b_star": 59.6309322469613},
{ "name": "7.5R4/18","L_star": 40.5731879484504, "a_star": 73.9232652573451, "b_star": 52.5453079908717},
{ "name": "7.5R4/16","L_star": 40.5080864088407, "a_star": 66.3140627633011, "b_star": 46.1338270184226},
{ "name": "7.5R4/14","L_star": 40.3937964103769, "a_star": 58.6947139869721, "b_star": 40.1805589736269},
{ "name": "7.5R4/12","L_star": 40.4264979523013, "a_star": 50.7141529649056, "b_star": 33.8353354064087},
{ "name": "7.5R4/10","L_star": 40.4754793119243, "a_star": 43.0216660253403, "b_star": 28.0275980534779},
{ "name": "7.5R4/8","L_star": 40.5243758552341, "a_star": 35.329295846744, "b_star": 22.4573598592393},
{ "name": "7.5R4/6","L_star": 40.5731879484504, "a_star": 26.9256849090768, "b_star": 16.5111045315237},
{ "name": "7.5R4/4","L_star": 40.6219159552695, "a_star": 18.664952176897, "b_star": 11.0034177263092},
{ "name": "7.5R4/3","L_star": 40.6381400008437, "a_star": 14.189200584268, "b_star": 8.14237827803735},
{ "name": "7.5R4/2","L_star": 40.6705602368878, "a_star": 9.57207908299188, "b_star": 5.30505927251829},
{ "name": "7.5R4/1","L_star": 40.686756453899, "a_star": 4.8458925365435, "b_star": 2.587821534418},
{ "name": "7.5R3/16","L_star": 29.8670690803586, "a_star": 70.3200168297587, "b_star": 43.207089983392},
{ "name": "7.5R3/14","L_star": 30.0934795714802, "a_star": 62.1415536533418, "b_star": 38.1322229926691},
{ "name": "7.5R3/12","L_star": 30.1521787492893, "a_star": 54.4405154599694, "b_star": 33.4694620185477},
{ "name": "7.5R3/10","L_star": 30.0983768812887, "a_star": 45.4077027720649, "b_star": 27.9539784176776},
{ "name": "7.5R3/8","L_star": 30.1497359303174, "a_star": 36.7384236558252, "b_star": 22.0676455606045},
{ "name": "7.5R3/6","L_star": 30.2009809203097, "a_star": 28.0781407364523, "b_star": 16.34749416642},
{ "name": "7.5R3/4","L_star": 30.247247679603, "a_star": 19.8108975259088, "b_star": 11.2099069211931},
{ "name": "7.5R3/3","L_star": 30.2715614676263, "a_star": 15.3976503784275, "b_star": 8.47395137305937},
{ "name": "7.5R3/2","L_star": 30.2982771557024, "a_star": 10.6764844880498, "b_star": 5.67589192646761},
{ "name": "7.5R3/1","L_star": 30.3249620475793, "a_star": 5.60758317960111, "b_star": 2.85156162152137},
{ "name": "7.5R2/14","L_star": 19.4494345468919, "a_star": 63.4226273262759, "b_star": 22.9316714203784},
{ "name": "7.5R2/12","L_star": 19.7613465019228, "a_star": 54.4390215719248, "b_star": 20.8348701361109},
{ "name": "7.5R2/10","L_star": 19.9716646168222, "a_star": 45.6814219662514, "b_star": 18.2671399776877},
{ "name": "7.5R2/8","L_star": 20.0078170679051, "a_star": 36.4060922504552, "b_star": 14.9570925441293},
{ "name": "7.5R2/6","L_star": 20.0599094643603, "a_star": 27.4356353365219, "b_star": 11.2382362495368},
{ "name": "7.5R2/4","L_star": 20.1038703861031, "a_star": 19.2787439347146, "b_star": 7.68734447375666},
{ "name": "7.5R2/3","L_star": 20.1278040161726, "a_star": 15.0292985802104, "b_star": 5.83768585350525},
{ "name": "7.5R2/2","L_star": 20.1477245120574, "a_star": 10.5333488842282, "b_star": 3.94584283049459},
{ "name": "7.5R2/1","L_star": 20.1755763746542, "a_star": 5.65035276770265, "b_star": 1.9835273751273},
{ "name": "7.5R1/10","L_star": 9.82940173128609, "a_star": 45.5085074671118, "b_star": 6.76699626752676},
{ "name": "7.5R1/8","L_star": 10.0842373959684, "a_star": 38.1711150077986, "b_star": 6.60882211991931},
{ "name": "7.5R1/6","L_star": 10.2135954147157, "a_star": 30.0458736600966, "b_star": 5.86970784856497},
{ "name": "7.5R1/4","L_star": 10.2438477155545, "a_star": 21.6252359394662, "b_star": 4.44782392614798},
{ "name": "7.5R1/3","L_star": 10.2664912537208, "a_star": 17.4108768651343, "b_star": 3.6571506767446},
{ "name": "7.5R1/2","L_star": 10.2890958184513, "a_star": 12.8022293292412, "b_star": 2.64661475901323},
{ "name": "7.5R1/1","L_star": 10.3191749016881, "a_star": 7.34407862352861, "b_star": 1.38332630981615},
{ "name": "10R9/6","L_star": 89.8871915345649, "a_star": 23.6063942166624, "b_star": 24.0552647395766},
{ "name": "10R9/4","L_star": 90.0447344113918, "a_star": 15.4350022550417, "b_star": 15.0391255522134},
{ "name": "10R9/3","L_star": 90.0956040750537, "a_star": 10.4820892265985, "b_star": 10.0853887593656},
{ "name": "10R9/2","L_star": 90.1279502978124, "a_star": 5.81850461827815, "b_star": 5.57292405198409},
{ "name": "10R9/1","L_star": 90.1510426764292, "a_star": 2.03553413307839, "b_star": 2.02002822128657},
{ "name": "10R8/10","L_star": 79.9314751726709, "a_star": 38.5273670238396, "b_star": 38.2228005607318},
{ "name": "10R8/8","L_star": 80.2695028958329, "a_star": 30.3702771264062, "b_star": 30.269301759},
{ "name": "10R8/6","L_star": 80.3704496697858, "a_star": 22.3288983306313, "b_star": 21.9506458033738},
{ "name": "10R8/4","L_star": 80.4152470961886, "a_star": 14.7399866568508, "b_star": 14.0383435526695},
{ "name": "10R8/3","L_star": 80.4432243617297, "a_star": 10.2209206752238, "b_star": 9.55078409118575},
{ "name": "10R8/2","L_star": 80.4711854048021, "a_star": 5.88250637569754, "b_star": 5.43716319871499},
{ "name": "10R8/1","L_star": 80.4935425748736, "a_star": 2.24694166055917, "b_star": 2.1217020971096},
{ "name": "10R7/16","L_star": 70.2529216390413, "a_star": 58.8973310192533, "b_star": 66.471798275509},
{ "name": "10R7/14","L_star": 70.3716505069624, "a_star": 52.1525653476421, "b_star": 55.5740432320642},
{ "name": "10R7/12","L_star": 70.5595517166717, "a_star": 44.4053309005901, "b_star": 46.1216391008816},
{ "name": "10R7/10","L_star": 70.622004517354, "a_star": 36.8358438947433, "b_star": 36.7591962176474},
{ "name": "10R7/8","L_star": 70.6635897395448, "a_star": 30.0256028870554, "b_star": 29.2211661442227},
{ "name": "10R7/6","L_star": 70.7051350909417, "a_star": 22.2801714002321, "b_star": 21.2382987847868},
{ "name": "10R7/4","L_star": 70.7535544026117, "a_star": 14.6861691483839, "b_star": 13.6873251981491},
{ "name": "10R7/3","L_star": 70.7811983316404, "a_star": 10.8607501161763, "b_star": 9.82466708412899},
{ "name": "10R7/2","L_star": 70.8019197264899, "a_star": 7.0296218420895, "b_star": 6.1341636020992},
{ "name": "10R7/1","L_star": 70.8226312328179, "a_star": 3.39552622121492, "b_star": 2.83982141708772},
{ "name": "10R6/18","L_star": 60.6490179749643, "a_star": 63.3583671777159, "b_star": 79.5866940787614},
{ "name": "10R6/16","L_star": 60.675576888377, "a_star": 57.4095754766445, "b_star": 65.8408546274208},
{ "name": "10R6/14","L_star": 60.7109601776115, "a_star": 51.4765613159205, "b_star": 55.7247469135154},
{ "name": "10R6/12","L_star": 60.7551434382602, "a_star": 44.6851848796584, "b_star": 46.1218872632357},
{ "name": "10R6/10","L_star": 60.7992758901635, "a_star": 37.5835160335413, "b_star": 37.3030581811054},
{ "name": "10R6/8","L_star": 60.8433576791392, "a_star": 30.069646292607, "b_star": 28.6989474767129},
{ "name": "10R6/6","L_star": 60.887388950337, "a_star": 22.6641419800058, "b_star": 21.1381164306439},
{ "name": "10R6/4","L_star": 60.9313698482419, "a_star": 15.4429089728224, "b_star": 13.970352157975},
{ "name": "10R6/3","L_star": 60.957734267663, "a_star": 11.6158891108775, "b_star": 10.2091797433442},
{ "name": "10R6/2","L_star": 60.984080635441, "a_star": 7.71639818769748, "b_star": 6.53927796036289},
{ "name": "10R6/1","L_star": 61.0016348671828, "a_star": 3.7700382926093, "b_star": 3.10629502466528},
{ "name": "10R5/18","L_star": 50.8039965577119, "a_star": 63.634019116414, "b_star": 87.1764403096531},
{ "name": "10R5/16","L_star": 50.6286571777442, "a_star": 57.6127359162317, "b_star": 67.0823359830711},
{ "name": "10R5/14","L_star": 50.6637988810833, "a_star": 52.0580207050107, "b_star": 56.3168465596701},
{ "name": "10R5/12","L_star": 50.6989035736593, "a_star": 46.19682798956, "b_star": 47.5810156065161},
{ "name": "10R5/10","L_star": 50.7456524262333, "a_star": 39.1988174336689, "b_star": 38.3267706154004},
{ "name": "10R5/8","L_star": 50.7923358843108, "a_star": 31.739941404373, "b_star": 29.6562793739245},
{ "name": "10R5/6","L_star": 50.8389541761052, "a_star": 23.946598216371, "b_star": 21.6135231714284},
{ "name": "10R5/4","L_star": 50.8855075285586, "a_star": 16.015422312444, "b_star": 13.8558724368026},
{ "name": "10R5/3","L_star": 50.9087599231063, "a_star": 11.9361463729181, "b_star": 10.0502499572072},
{ "name": "10R5/2","L_star": 50.9319961673508, "a_star": 7.83953555097316, "b_star": 6.42151910347848},
{ "name": "10R5/1","L_star": 50.9552162893063, "a_star": 3.84197125248292, "b_star": 3.04655839743309},
{ "name": "10R4/16","L_star": 40.6219159552695, "a_star": 58.6975131790803, "b_star": 74.0780908490975},
{ "name": "10R4/14","L_star": 40.4754793119243, "a_star": 53.3423957275981, "b_star": 57.8779849351377},
{ "name": "10R4/12","L_star": 40.4428345189212, "a_star": 46.6706070191517, "b_star": 46.0225470439551},
{ "name": "10R4/10","L_star": 40.4754793119243, "a_star": 40.1007202657051, "b_star": 36.85845165472},
{ "name": "10R4/8","L_star": 40.5243758552341, "a_star": 32.8439685021443, "b_star": 28.8721511921821},
{ "name": "10R4/6","L_star": 40.5731879484504, "a_star": 25.2122967538719, "b_star": 21.2066339898284},
{ "name": "10R4/4","L_star": 40.6056826069156, "a_star": 17.7026147060902, "b_star": 14.0984532328773},
{ "name": "10R4/3","L_star": 40.6381400008437, "a_star": 13.5579675144154, "b_star": 10.4408759389549},
{ "name": "10R4/2","L_star": 40.6543547569545, "a_star": 9.17882826117286, "b_star": 6.80711207452141},
{ "name": "10R4/1","L_star": 40.686756453899, "a_star": 4.64318829036423, "b_star": 3.31129836033528},
{ "name": "10R3/14","L_star": 30.3079843112769, "a_star": 58.6151270923281, "b_star": 61.3571658104849},
{ "name": "10R3/12","L_star": 30.1814724282108, "a_star": 50.7574256244607, "b_star": 45.2251803244587},
{ "name": "10R3/10","L_star": 30.1057208955728, "a_star": 42.5752048618885, "b_star": 35.2459677445307},
{ "name": "10R3/8","L_star": 30.1521787492893, "a_star": 34.7606391777191, "b_star": 27.5613827724172},
{ "name": "10R3/6","L_star": 30.2009809203097, "a_star": 26.6060806996439, "b_star": 20.2605054208094},
{ "name": "10R3/4","L_star": 30.247247679603, "a_star": 18.6262368537912, "b_star": 13.6110733249879},
{ "name": "10R3/3","L_star": 30.2715614676263, "a_star": 14.4255315748986, "b_star": 10.337022857474},
{ "name": "10R3/2","L_star": 30.2982771557024, "a_star": 10.0240129067758, "b_star": 7.01576008962376},
{ "name": "10R3/1","L_star": 30.3249620475793, "a_star": 5.2879254109941, "b_star": 3.57969964336275},
{ "name": "10R2/14","L_star": 19.4783931362464, "a_star": 61.4568286976433, "b_star": 41.707778291248},
{ "name": "10R2/12","L_star": 19.8627689012203, "a_star": 52.518900187995, "b_star": 33.3440792278841},
{ "name": "10R2/10","L_star": 20.0118295280528, "a_star": 43.549169888098, "b_star": 27.4035489238105},
{ "name": "10R2/8","L_star": 20.0038037133168, "a_star": 35.0055309955812, "b_star": 21.8485456840082},
{ "name": "10R2/6","L_star": 20.0559076987671, "a_star": 26.2824514579245, "b_star": 16.1250753709964},
{ "name": "10R2/4","L_star": 20.0998783614812, "a_star": 18.4104647738659, "b_star": 10.834239460441},
{ "name": "10R2/3","L_star": 20.1238172798771, "a_star": 14.2350379533607, "b_star": 8.17209019068337},
{ "name": "10R2/2","L_star": 20.1477245120574, "a_star": 9.84373335404179, "b_star": 5.46625734227318},
{ "name": "10R2/1","L_star": 20.1716001626663, "a_star": 5.20806072407673, "b_star": 2.72846924472814},
{ "name": "10R1/10","L_star": 9.7668605622588, "a_star": 45.436232409722, "b_star": 14.840016825198},
{ "name": "10R1/8","L_star": 10.068934231766, "a_star": 37.733921671076, "b_star": 13.1167427628085},
{ "name": "10R1/6","L_star": 10.2287302884101, "a_star": 29.4436590748471, "b_star": 10.8142016831449},
{ "name": "10R1/4","L_star": 10.2362911796586, "a_star": 21.5552104346266, "b_star": 7.85730232337638},
{ "name": "10R1/3","L_star": 10.2664912537208, "a_star": 17.1508692695889, "b_star": 6.17691324109458},
{ "name": "10R1/2","L_star": 10.2890958184513, "a_star": 12.34547181768, "b_star": 4.28538922168524},
{ "name": "10R1/1","L_star": 10.3191749016881, "a_star": 6.89997583975593, "b_star": 2.18120449305861},
{ "name": "2.5YR9/6","L_star": 89.9150272854619, "a_star": 20.5278068413886, "b_star": 28.8053748562508},
{ "name": "2.5YR9/4","L_star": 90.0678630346794, "a_star": 13.7474617592251, "b_star": 18.3334185154782},
{ "name": "2.5YR9/3","L_star": 90.0956040750537, "a_star": 9.68513358502249, "b_star": 12.7799938883491},
{ "name": "2.5YR9/2","L_star": 90.1279502978124, "a_star": 5.70264858852931, "b_star": 7.53817100246543},
{ "name": "2.5YR9/1","L_star": 90.1510426764292, "a_star": 2.26128758900301, "b_star": 3.14446702516018},
{ "name": "2.5YR8/12","L_star": 79.8296009631123, "a_star": 40.1600102589922, "b_star": 56.8560411825307},
{ "name": "2.5YR8/10","L_star": 80.0331334703953, "a_star": 33.444068865605, "b_star": 45.3258048534787},
{ "name": "2.5YR8/8","L_star": 80.3143942536544, "a_star": 26.8596525668223, "b_star": 36.0102740266602},
{ "name": "2.5YR8/6","L_star": 80.3648470623858, "a_star": 19.9219597879579, "b_star": 26.1793342965627},
{ "name": "2.5YR8/4","L_star": 80.4152470961886, "a_star": 13.2626955104981, "b_star": 17.321062221105},
{ "name": "2.5YR8/3","L_star": 80.4432243617297, "a_star": 9.44577982204764, "b_star": 12.1345356761394},
{ "name": "2.5YR8/2","L_star": 80.4711854048021, "a_star": 5.64562337629482, "b_star": 7.17700256226876},
{ "name": "2.5YR8/1","L_star": 80.4991302489017, "a_star": 2.31683774324837, "b_star": 3.00144332745478},
{ "name": "2.5YR7/20","L_star": 70.2389320318477, "a_star": 55.5077210859809, "b_star": 103.572339350039},
{ "name": "2.5YR7/18","L_star": 70.2948632559789, "a_star": 52.8250286169188, "b_star": 90.4146192117144},
{ "name": "2.5YR7/16","L_star": 70.3995393208176, "a_star": 49.2034241198344, "b_star": 78.5059455767861},
{ "name": "2.5YR7/14","L_star": 70.524816740343, "a_star": 44.5117327361419, "b_star": 66.9956121982918},
{ "name": "2.5YR7/12","L_star": 70.5664953675691, "a_star": 39.022734130966, "b_star": 54.8804336611277},
{ "name": "2.5YR7/10","L_star": 70.6150697639405, "a_star": 32.6745091523964, "b_star": 43.6739158068526},
{ "name": "2.5YR7/8","L_star": 70.6566616405674, "a_star": 26.6443470951971, "b_star": 34.7436271100897},
{ "name": "2.5YR7/6","L_star": 70.7051350909417, "a_star": 20.2645390351617, "b_star": 25.7839626896169},
{ "name": "2.5YR7/4","L_star": 70.7535544026117, "a_star": 13.3060709768707, "b_star": 16.5949397032025},
{ "name": "2.5YR7/3","L_star": 70.7811983316404, "a_star": 9.92689860443979, "b_star": 12.1568129641621},
{ "name": "2.5YR7/2","L_star": 70.8019197264899, "a_star": 6.53967152493856, "b_star": 7.86468359844619},
{ "name": "2.5YR7/1","L_star": 70.8295328728496, "a_star": 3.18753410717598, "b_star": 3.83657864750404},
{ "name": "2.5YR6/18","L_star": 60.6667259617269, "a_star": 50.5467397865709, "b_star": 95.0918951653995},
{ "name": "2.5YR6/16","L_star": 60.6932726141122, "a_star": 47.2773223916805, "b_star": 78.3211055434746},
{ "name": "2.5YR6/14","L_star": 60.7198009014627, "a_star": 43.7622003823075, "b_star": 66.3993145851775},
{ "name": "2.5YR6/12","L_star": 60.7463108555081, "a_star": 39.1192041595793, "b_star": 55.1442482813773},
{ "name": "2.5YR6/10","L_star": 60.7904534574736, "a_star": 33.5546075944639, "b_star": 44.4215320886018},
{ "name": "2.5YR6/8","L_star": 60.8345453674015, "a_star": 27.1836312816061, "b_star": 34.446218811836},
{ "name": "2.5YR6/6","L_star": 60.8785867305749, "a_star": 20.7299512593784, "b_star": 25.6040190605878},
{ "name": "2.5YR6/4","L_star": 60.9313698482419, "a_star": 13.7832905460387, "b_star": 16.6675216350014},
{ "name": "2.5YR6/3","L_star": 60.957734267663, "a_star": 10.4433523672115, "b_star": 12.4685973743677},
{ "name": "2.5YR6/2","L_star": 60.984080635441, "a_star": 7.10240476308338, "b_star": 8.32545855205564},
{ "name": "2.5YR6/1","L_star": 61.0104089824442, "a_star": 3.68544264711723, "b_star": 4.18454034448545},
{ "name": "2.5YR5/16","L_star": 50.6755045527692, "a_star": 46.2288037208115, "b_star": 85.2515145960085},
{ "name": "2.5YR5/14","L_star": 50.6755045527692, "a_star": 42.8283258308526, "b_star": 67.6745405210969},
{ "name": "2.5YR5/12","L_star": 50.7105969300653, "a_star": 39.0170920802497, "b_star": 56.0869091875477},
{ "name": "2.5YR5/10","L_star": 50.7456524262333, "a_star": 34.3635635594328, "b_star": 45.6958945584603},
{ "name": "2.5YR5/8","L_star": 50.7806711380232, "a_star": 28.3693258190473, "b_star": 35.5952897233317},
{ "name": "2.5YR5/6","L_star": 50.8389541761052, "a_star": 21.5085686250842, "b_star": 25.7918549195685},
{ "name": "2.5YR5/4","L_star": 50.8855075285586, "a_star": 14.4385986258789, "b_star": 16.551676660568},
{ "name": "2.5YR5/3","L_star": 50.9087599231063, "a_star": 10.8843707524877, "b_star": 12.2738966707879},
{ "name": "2.5YR5/2","L_star": 50.9319961673508, "a_star": 7.31845429244599, "b_star": 8.16293616292649},
{ "name": "2.5YR5/1","L_star": 50.966820313159, "a_star": 3.70711409522578, "b_star": 4.0976982363589},
{ "name": "2.5YR4/12","L_star": 40.4428345189212, "a_star": 38.627502746348, "b_star": 58.6584807827653},
{ "name": "2.5YR4/10","L_star": 40.4754793119243, "a_star": 34.190699903515, "b_star": 45.6888766201136},
{ "name": "2.5YR4/8","L_star": 40.5243758552341, "a_star": 28.8335505442172, "b_star": 35.1221790604656},
{ "name": "2.5YR4/6","L_star": 40.5569266115533, "a_star": 22.4601579923425, "b_star": 25.7395596046551},
{ "name": "2.5YR4/4","L_star": 40.6056826069156, "a_star": 15.9421111107999, "b_star": 17.2051621248172},
{ "name": "2.5YR4/3","L_star": 40.6381400008437, "a_star": 12.2618532000376, "b_star": 12.9003083026919},
{ "name": "2.5YR4/2","L_star": 40.6705602368878, "a_star": 8.34418797515915, "b_star": 8.58500867071587},
{ "name": "2.5YR4/1","L_star": 40.686756453899, "a_star": 4.24717910436845, "b_star": 4.29890290715656},
{ "name": "2.5YR3/10","L_star": 30.1155092767163, "a_star": 35.4354057509386, "b_star": 49.3213476821968},
{ "name": "2.5YR3/8","L_star": 30.1546213096934, "a_star": 29.7102611702427, "b_star": 35.2803384591749},
{ "name": "2.5YR3/6","L_star": 30.1985432597279, "a_star": 23.501845315576, "b_star": 25.475830788412},
{ "name": "2.5YR3/4","L_star": 30.2496802089405, "a_star": 16.4053977521441, "b_star": 16.5366638291687},
{ "name": "2.5YR3/3","L_star": 30.2739914414501, "a_star": 12.791491771132, "b_star": 12.4632122452215},
{ "name": "2.5YR3/2","L_star": 30.3007043262, "a_star": 8.97117056176725, "b_star": 8.42948156801323},
{ "name": "2.5YR3/1","L_star": 30.3298105443489, "a_star": 4.79648409766967, "b_star": 4.30468951567808},
{ "name": "2.5YR2/8","L_star": 19.9917582779202, "a_star": 32.0129513050578, "b_star": 33.2285078066218},
{ "name": "2.5YR2/6","L_star": 20.0479015016129, "a_star": 24.1519479908584, "b_star": 21.7752306629134},
{ "name": "2.5YR2/4","L_star": 20.0998783614812, "a_star": 16.9444123293981, "b_star": 14.0868339167889},
{ "name": "2.5YR2/3","L_star": 20.1278040161726, "a_star": 13.0262277960734, "b_star": 10.4154087702309},
{ "name": "2.5YR2/2","L_star": 20.1517059775056, "a_star": 8.93382466352405, "b_star": 6.84163961778433},
{ "name": "2.5YR2/1","L_star": 20.1795517127494, "a_star": 4.62662706806474, "b_star": 3.37090757720911},
{ "name": "2.5YR1/8","L_star": 10.1452712245553, "a_star": 36.097431307098, "b_star": 27.1328222456263},
{ "name": "2.5YR1/6","L_star": 10.1908602815276, "a_star": 27.8787316705431, "b_star": 19.5911358227166},
{ "name": "2.5YR1/4","L_star": 10.2362911796586, "a_star": 20.6700415795686, "b_star": 13.6648404579246},
{ "name": "2.5YR1/3","L_star": 10.2664912537208, "a_star": 16.1800909864302, "b_star": 10.1944407590902},
{ "name": "2.5YR1/2","L_star": 10.2966220415417, "a_star": 11.2062753659946, "b_star": 6.71835076085325},
{ "name": "2.5YR1/1","L_star": 10.3341886955329, "a_star": 5.8898130363108, "b_star": 3.28949871079693},
{ "name": "5YR9/6","L_star": 89.9567534988206, "a_star": 16.6503703398065, "b_star": 32.811799263023},
{ "name": "5YR9/4","L_star": 90.0771116603561, "a_star": 11.002721996672, "b_star": 21.8152198197922},
{ "name": "5YR9/3","L_star": 90.1048478654795, "a_star": 7.90022455305606, "b_star": 15.844296083073},
{ "name": "5YR9/2","L_star": 90.132569577483, "a_star": 4.86960303542411, "b_star": 9.90386911424543},
{ "name": "5YR9/1","L_star": 90.1556599466689, "a_star": 2.07187253373525, "b_star": 4.53404299030453},
{ "name": "5YR8/14","L_star": 79.8522583941405, "a_star": 36.3910943549481, "b_star": 76.2344175822756},
{ "name": "5YR8/12","L_star": 80.0049166463375, "a_star": 32.2107784687997, "b_star": 63.681646631264},
{ "name": "5YR8/10","L_star": 80.2919538068371, "a_star": 26.5518541462911, "b_star": 52.1419700436971},
{ "name": "5YR8/8","L_star": 80.3312177289753, "a_star": 21.9412031848868, "b_star": 41.9368678781022},
{ "name": "5YR8/6","L_star": 80.3760516258334, "a_star": 16.300757993974, "b_star": 30.3859154252177},
{ "name": "5YR8/4","L_star": 80.420843848224, "a_star": 10.8831100160522, "b_star": 20.2476927795717},
{ "name": "5YR8/3","L_star": 80.448817867389, "a_star": 7.93712908159705, "b_star": 14.8291725099926},
{ "name": "5YR8/2","L_star": 80.4767756687889, "a_star": 5.09103301559383, "b_star": 9.46096404150893},
{ "name": "5YR8/1","L_star": 80.4991302489017, "a_star": 2.31858447188593, "b_star": 4.47286064353594},
{ "name": "5YR7/20","L_star": 70.3646754887211, "a_star": 42.5898621818557, "b_star": 117.012156002825},
{ "name": "5YR7/18","L_star": 70.5178663975755, "a_star": 40.4124922318558, "b_star": 100.216673108853},
{ "name": "5YR7/16","L_star": 70.5317659666793, "a_star": 38.3993547886389, "b_star": 87.9153767324016},
{ "name": "5YR7/14","L_star": 70.5595517166717, "a_star": 35.474294915678, "b_star": 75.5519273822399},
{ "name": "5YR7/12","L_star": 70.5873196395698, "a_star": 31.6316381386857, "b_star": 63.1360880934602},
{ "name": "5YR7/10","L_star": 70.6289381605852, "a_star": 26.7612767021658, "b_star": 51.2016701986898},
{ "name": "5YR7/8","L_star": 70.670516731001, "a_star": 21.8721434573614, "b_star": 40.2351808821401},
{ "name": "5YR7/6","L_star": 70.7120554465534, "a_star": 16.7930200775391, "b_star": 30.2004378674625},
{ "name": "5YR7/4","L_star": 70.7604670364579, "a_star": 11.2098266467662, "b_star": 19.7270215097895},
{ "name": "5YR7/3","L_star": 70.7811983316404, "a_star": 8.48515405984968, "b_star": 14.648092490289},
{ "name": "5YR7/2","L_star": 70.8088246600745, "a_star": 5.69523162830249, "b_star": 9.74838178395658},
{ "name": "5YR7/1","L_star": 70.8295328728496, "a_star": 2.90666618021646, "b_star": 4.8461507516546},
{ "name": "5YR6/18","L_star": 60.5425978200924, "a_star": 40.3802422623528, "b_star": 112.455734720423},
{ "name": "5YR6/16","L_star": 60.7286395882807, "a_star": 37.4192389297287, "b_star": 89.2231606631691},
{ "name": "5YR6/14","L_star": 60.7463108555081, "a_star": 34.9087682597483, "b_star": 75.0599526621325},
{ "name": "5YR6/12","L_star": 60.7728025078901, "a_star": 31.6359853482732, "b_star": 63.1619471219835},
{ "name": "5YR6/10","L_star": 60.8080962963385, "a_star": 27.6350645484801, "b_star": 51.8095600481996},
{ "name": "5YR6/8","L_star": 60.8433576791392, "a_star": 22.8021040406585, "b_star": 40.862691590647},
{ "name": "5YR6/6","L_star": 60.887388950337, "a_star": 17.3814441709139, "b_star": 30.2801629473529},
{ "name": "5YR6/4","L_star": 60.940159995696, "a_star": 11.6627088517435, "b_star": 19.5957791142464},
{ "name": "5YR6/3","L_star": 60.9665183944677, "a_star": 8.90504849395374, "b_star": 14.6977384832671},
{ "name": "5YR6/2","L_star": 60.984080635441, "a_star": 6.08708390270496, "b_star": 9.8928000652831},
{ "name": "5YR6/1","L_star": 61.0104089824442, "a_star": 3.13046615122303, "b_star": 5.04198879564679},
{ "name": "5YR5/14","L_star": 50.7105969300653, "a_star": 34.4408793371704, "b_star": 78.0117589871615},
{ "name": "5YR5/12","L_star": 50.7339713527665, "a_star": 31.5591801241791, "b_star": 63.1878263223698},
{ "name": "5YR5/10","L_star": 50.7573294125503, "a_star": 28.1476909688523, "b_star": 52.0398088702818},
{ "name": "5YR5/8","L_star": 50.8039965577119, "a_star": 23.6735536543853, "b_star": 41.4758597515177},
{ "name": "5YR5/6","L_star": 50.8389541761052, "a_star": 18.2189639826199, "b_star": 30.506795942141},
{ "name": "5YR5/4","L_star": 50.8971357463743, "a_star": 12.1775165784617, "b_star": 19.7409894457817},
{ "name": "5YR5/3","L_star": 50.9203800622637, "a_star": 9.25055920602635, "b_star": 14.6597023678329},
{ "name": "5YR5/2","L_star": 50.9436082418668, "a_star": 6.28916925449857, "b_star": 9.7106014246787},
{ "name": "5YR5/1","L_star": 50.966820313159, "a_star": 3.25227768453873, "b_star": 4.85523634593505},
{ "name": "5YR4/12","L_star": 40.4264979523013, "a_star": 31.7379718216437, "b_star": 70.1142212524183},
{ "name": "5YR4/10","L_star": 40.5080864088407, "a_star": 27.9030581856314, "b_star": 52.2100632732354},
{ "name": "5YR4/8","L_star": 40.5406559183046, "a_star": 23.7651107361636, "b_star": 40.3400827799713},
{ "name": "5YR4/6","L_star": 40.5731879484504, "a_star": 18.9083577358785, "b_star": 30.1020098320978},
{ "name": "5YR4/4","L_star": 40.6219159552695, "a_star": 13.5893480477481, "b_star": 20.4257099018954},
{ "name": "5YR4/3","L_star": 40.6381400008437, "a_star": 10.4712639873892, "b_star": 15.3526000191179},
{ "name": "5YR4/2","L_star": 40.6705602368878, "a_star": 7.14703233864136, "b_star": 10.2375860219907},
{ "name": "5YR4/1","L_star": 40.7029434212134, "a_star": 3.61732802923678, "b_star": 5.13110593482402},
{ "name": "5YR3/8","L_star": 30.1741524926899, "a_star": 24.2190627999722, "b_star": 41.6518558973611},
{ "name": "5YR3/6","L_star": 30.2131653664994, "a_star": 19.4833641988631, "b_star": 29.645217116239},
{ "name": "5YR3/4","L_star": 30.2594077684258, "a_star": 13.7012862474032, "b_star": 19.1294049603555},
{ "name": "5YR3/3","L_star": 30.2837087855916, "a_star": 10.692842751034, "b_star": 14.4658346571702},
{ "name": "5YR3/2","L_star": 30.3079843112769, "a_star": 7.50733607101625, "b_star": 9.84409286919645},
{ "name": "5YR3/1","L_star": 30.3370813873033, "a_star": 4.01928962510229, "b_star": 5.09342537235329},
{ "name": "5YR2/6","L_star": 20.0599094643603, "a_star": 20.3671984155287, "b_star": 29.6103284581544},
{ "name": "5YR2/4","L_star": 20.111851788018, "a_star": 14.2553371092853, "b_star": 17.7692245626962},
{ "name": "5YR2/3","L_star": 20.135774850191, "a_star": 10.9855496239709, "b_star": 12.8929465414763},
{ "name": "5YR2/2","L_star": 20.1636451150901, "a_star": 7.54525001262366, "b_star": 8.37187464140167},
{ "name": "5YR2/1","L_star": 20.1914724884762, "a_star": 3.91810106398741, "b_star": 4.08168447402322},
{ "name": "5YR1/4","L_star": 10.2362911796586, "a_star": 18.1379610708475, "b_star": 22.5770277015119},
{ "name": "5YR1/3","L_star": 10.281565283574, "a_star": 13.902088875193, "b_star": 15.0718705613954},
{ "name": "5YR1/2","L_star": 10.3116615770145, "a_star": 9.50618819448623, "b_star": 9.55454077572602},
{ "name": "5YR1/1","L_star": 10.3491853893754, "a_star": 4.74459704153164, "b_star": 4.32996096669897},
{ "name": "7.5YR9/8","L_star": 89.9243026184882, "a_star": 16.1843259351702, "b_star": 48.7190563524098},
{ "name": "7.5YR9/6","L_star": 90.0215956948791, "a_star": 12.2589812182849, "b_star": 36.2050606653573},
{ "name": "7.5YR9/4","L_star": 90.0956040750537, "a_star": 8.1474011023267, "b_star": 24.6344338954002},
{ "name": "7.5YR9/3","L_star": 90.1187105319379, "a_star": 5.9522342508852, "b_star": 18.4260049134874},
{ "name": "7.5YR9/2","L_star": 90.1418069307284, "a_star": 3.69825763795106, "b_star": 12.1707687084434},
{ "name": "7.5YR9/1","L_star": 90.1648932823636, "a_star": 1.65620554477858, "b_star": 5.99232390488216},
{ "name": "7.5YR8/20","L_star": 79.7955947143509, "a_star": 32.9389082960194, "b_star": 122.120231266017},
{ "name": "7.5YR8/18","L_star": 79.8635830939616, "a_star": 31.819741804417, "b_star": 108.881097807279},
{ "name": "7.5YR8/16","L_star": 31.0435334088221, "a_star": 14.5589460152661, "b_star": 46.9275325278843},
{ "name": "7.5YR8/14","L_star": 80.2133297573375, "a_star": 26.7439260009711, "b_star": 82.5201531305231},
{ "name": "7.5YR8/12","L_star": 80.2919538068371, "a_star": 23.7312969925542, "b_star": 69.9942036134036},
{ "name": "7.5YR8/10","L_star": 80.3256105568015, "a_star": 20.3209905455936, "b_star": 57.6754416060293},
{ "name": "7.5YR8/8","L_star": 80.3592438034441, "a_star": 16.7890524876853, "b_star": 46.2019071226217},
{ "name": "7.5YR8/6","L_star": 80.3984529402843, "a_star": 12.6036497378607, "b_star": 33.976345599458},
{ "name": "7.5YR8/4","L_star": 80.4432243617297, "a_star": 8.16695087591357, "b_star": 22.7622794561701},
{ "name": "7.5YR8/3","L_star": 80.4600029327644, "a_star": 6.04187955351504, "b_star": 17.0783781107254},
{ "name": "7.5YR8/2","L_star": 80.4823652850047, "a_star": 3.93639864588391, "b_star": 11.3457328961503},
{ "name": "7.5YR8/1","L_star": 80.5047172759086, "a_star": 1.93334545412049, "b_star": 5.65553903168525},
{ "name": "7.5YR7/18","L_star": 70.524816740343, "a_star": 29.8502792991238, "b_star": 109.271149313685},
{ "name": "7.5YR7/16","L_star": 70.5942588381602, "a_star": 28.351708491556, "b_star": 95.5822104231155},
{ "name": "7.5YR7/14","L_star": 70.6081338999003, "a_star": 26.6077134178002, "b_star": 82.7551672404962},
{ "name": "7.5YR7/12","L_star": 70.6289381605852, "a_star": 24.0844992563585, "b_star": 69.8892903271093},
{ "name": "7.5YR7/10","L_star": 70.6635897395448, "a_star": 20.6640519311901, "b_star": 57.0055337998505},
{ "name": "7.5YR7/8","L_star": 70.6982136304596, "a_star": 17.0368236180246, "b_star": 45.1823603570803},
{ "name": "7.5YR7/6","L_star": 70.732809888573, "a_star": 13.2019074323808, "b_star": 33.9125928533534},
{ "name": "7.5YR7/4","L_star": 70.7742890005338, "a_star": 8.99808492498172, "b_star": 22.3904073561793},
{ "name": "7.5YR7/3","L_star": 70.7950136941808, "a_star": 6.77592486510753, "b_star": 16.8574953339213},
{ "name": "7.5YR7/2","L_star": 70.8157284953717, "a_star": 4.50552101356161, "b_star": 11.3001236246159},
{ "name": "7.5YR7/1","L_star": 70.8433328624133, "a_star": 2.21266804082687, "b_star": 5.73820072938529},
{ "name": "7.5YR6/16","L_star": 60.613577434494, "a_star": 28.4774862403518, "b_star": 102.302600936583},
{ "name": "7.5YR6/14","L_star": 60.7992758901635, "a_star": 26.1724071442528, "b_star": 82.4311662119953},
{ "name": "7.5YR6/12","L_star": 60.8169146771618, "a_star": 24.2452548626168, "b_star": 69.938974112294},
{ "name": "7.5YR6/10","L_star": 60.8433576791392, "a_star": 21.4478382521511, "b_star": 57.8633057338898},
{ "name": "7.5YR6/8","L_star": 60.8697824947294, "a_star": 18.0405065072419, "b_star": 45.5786255109816},
{ "name": "7.5YR6/6","L_star": 60.9137835246555, "a_star": 13.8654350806345, "b_star": 33.7566660481006},
{ "name": "7.5YR6/4","L_star": 60.9489481351208, "a_star": 9.48837523565138, "b_star": 22.0754201828626},
{ "name": "7.5YR6/3","L_star": 60.9753005166793, "a_star": 7.20284325120879, "b_star": 16.6592513542079},
{ "name": "7.5YR6/2","L_star": 60.9928587518952, "a_star": 4.9035724979225, "b_star": 11.2564065044158},
{ "name": "7.5YR6/1","L_star": 61.0191810988183, "a_star": 2.4806264317534, "b_star": 5.7816885950835},
{ "name": "7.5YR5/14","L_star": 50.5817438320609, "a_star": 26.3968280257283, "b_star": 90.6659480392029},
{ "name": "7.5YR5/12","L_star": 50.7806711380232, "a_star": 23.8877878365304, "b_star": 70.6366538570306},
{ "name": "7.5YR5/10","L_star": 50.7923358843108, "a_star": 21.6747492312016, "b_star": 57.6257246672299},
{ "name": "7.5YR5/8","L_star": 50.8273057000631, "a_star": 18.8018827300134, "b_star": 46.1302534356958},
{ "name": "7.5YR5/6","L_star": 50.8622389556151, "a_star": 14.8465828390821, "b_star": 34.2504305805891},
{ "name": "7.5YR5/4","L_star": 50.9087599231063, "a_star": 10.0239885022542, "b_star": 22.2297129460629},
{ "name": "7.5YR5/3","L_star": 50.9319961673508, "a_star": 7.61020086854153, "b_star": 16.5109553323335},
{ "name": "7.5YR5/2","L_star": 50.9552162893063, "a_star": 5.12226213454869, "b_star": 10.9076670600353},
{ "name": "7.5YR5/1","L_star": 50.9784203169096, "a_star": 2.59903445955351, "b_star": 5.43169424779373},
{ "name": "7.5YR4/10","L_star": 40.5406559183046, "a_star": 21.3325032910262, "b_star": 59.0292020551401},
{ "name": "7.5YR4/8","L_star": 40.5569266115533, "a_star": 18.6029536214085, "b_star": 45.0564563203334},
{ "name": "7.5YR4/6","L_star": 40.5894399424351, "a_star": 15.1703530885583, "b_star": 33.7513441538966},
{ "name": "7.5YR4/4","L_star": 40.6381400008437, "a_star": 11.0169132736146, "b_star": 23.1468673762197},
{ "name": "7.5YR4/3","L_star": 40.6543547569545, "a_star": 8.56313266442113, "b_star": 17.3556627271648},
{ "name": "7.5YR4/2","L_star": 40.686756453899, "a_star": 5.85948764932601, "b_star": 11.5009966218511},
{ "name": "7.5YR4/1","L_star": 40.7029434212134, "a_star": 2.99390748471293, "b_star": 5.7026631511213},
{ "name": "7.5YR3/8","L_star": 30.2082923590962, "a_star": 18.5530449506734, "b_star": 48.078610688249},
{ "name": "7.5YR3/6","L_star": 30.2375150023693, "a_star": 15.141406658533, "b_star": 33.1159800460547},
{ "name": "7.5YR3/4","L_star": 30.2764211600916, "a_star": 10.8862059898585, "b_star": 21.5567606080102},
{ "name": "7.5YR3/3","L_star": 30.2982771557024, "a_star": 8.57639180141448, "b_star": 16.2380666254136},
{ "name": "7.5YR3/2","L_star": 30.3201125356833, "a_star": 6.04591371690427, "b_star": 10.955635366012},
{ "name": "7.5YR3/1","L_star": 30.3467722965866, "a_star": 3.23718925024616, "b_star": 5.58910260591562},
{ "name": "7.5YR2/6","L_star": 20.063910341955, "a_star": 15.9895222257637, "b_star": 38.9019084801686},
{ "name": "7.5YR2/4","L_star": 20.1317898727821, "a_star": 11.0912446903098, "b_star": 20.9672743391016},
{ "name": "7.5YR2/3","L_star": 20.1556865661722, "a_star": 8.54288199767636, "b_star": 15.1008676107295},
{ "name": "7.5YR2/2","L_star": 20.1795517127494, "a_star": 5.8698523343117, "b_star": 9.8094963991683},
{ "name": "7.5YR2/1","L_star": 20.2033854164229, "a_star": 3.07353617063966, "b_star": 4.83369373514626},
{ "name": "7.5YR1/2","L_star": 10.3341886955329, "a_star": 7.76626193579856, "b_star": 11.6782182206723},
{ "name": "7.5YR1/1","L_star": 10.3716484739028, "a_star": 3.63821676640089, "b_star": 4.74258766761381},
{ "name": "10YR9/8","L_star": 90.0169667394202, "a_star": 9.88472643922345, "b_star": 52.6239138650816},
{ "name": "10YR9/6","L_star": 90.0956040750537, "a_star": 6.99105251110022, "b_star": 39.8004085748674},
{ "name": "10YR9/4","L_star": 90.1279502978124, "a_star": 4.42250088128043, "b_star": 27.4083925873434},
{ "name": "10YR9/3","L_star": 90.1418069307284, "a_star": 3.07320270867978, "b_star": 20.9268578929713},
{ "name": "10YR9/2","L_star": 90.1602768152846, "a_star": 1.84065782944437, "b_star": 14.1877614521264},
{ "name": "10YR9/1","L_star": 90.1741250122608, "a_star": 0.734007603633624, "b_star": 7.25289487995759},
{ "name": "10YR8/20","L_star": 79.9201664993169, "a_star": 22.6542775074762, "b_star": 130.760874949874},
{ "name": "10YR8/18","L_star": 80.2470415084344, "a_star": 20.4734985494427, "b_star": 114.417615391432},
{ "name": "10YR8/16","L_star": 80.3200027317579, "a_star": 19.0191539381274, "b_star": 101.244924791288},
{ "name": "10YR8/14","L_star": 80.3368242484694, "a_star": 17.4869431463086, "b_star": 88.4908590504772},
{ "name": "10YR8/12","L_star": 80.3592438034441, "a_star": 15.4988740088433, "b_star": 75.4271999033266},
{ "name": "10YR8/10","L_star": 80.3816529307179, "a_star": 13.0868861417251, "b_star": 62.6049918648893},
{ "name": "10YR8/8","L_star": 80.4096496943127, "a_star": 10.507287221141, "b_star": 50.5707783613157},
{ "name": "10YR8/6","L_star": 80.4376302071718, "a_star": 7.62967711140339, "b_star": 37.8413353695237},
{ "name": "10YR8/4","L_star": 80.4711854048021, "a_star": 4.74162839558379, "b_star": 25.5193932899412},
{ "name": "10YR8/3","L_star": 80.4879542536371, "a_star": 3.42356561873924, "b_star": 19.4539365574955},
{ "name": "10YR8/2","L_star": 80.4991302489017, "a_star": 2.16595681976528, "b_star": 13.2437486684057},
{ "name": "10YR8/1","L_star": 80.5158893896084, "a_star": 1.01027301322698, "b_star": 6.8178991206453},
{ "name": "10YR7/18","L_star": 70.1478887518532, "a_star": 21.8129298726101, "b_star": 123.64947659076},
{ "name": "10YR7/16","L_star": 70.6635897395448, "a_star": 18.9549311450984, "b_star": 101.887964179078},
{ "name": "10YR7/14","L_star": 70.6774426153785, "a_star": 17.9103467747973, "b_star": 88.8557487771618},
{ "name": "10YR7/12","L_star": 70.691291064666, "a_star": 16.0971934603023, "b_star": 75.7226900812647},
{ "name": "10YR7/10","L_star": 70.7120554465534, "a_star": 13.8502530985469, "b_star": 62.3502218835277},
{ "name": "10YR7/8","L_star": 70.7397258291086, "a_star": 11.1714394094941, "b_star": 49.9850818084031},
{ "name": "10YR7/6","L_star": 70.7673785689525, "a_star": 8.32037199789831, "b_star": 37.8092960661142},
{ "name": "10YR7/4","L_star": 70.8019197264899, "a_star": 5.42679783216538, "b_star": 25.4461059330106},
{ "name": "10YR7/3","L_star": 70.8157284953717, "a_star": 4.04683038050646, "b_star": 19.2654113103351},
{ "name": "10YR7/2","L_star": 70.8364334159028, "a_star": 2.59018584157383, "b_star": 13.0033931087259},
{ "name": "10YR7/1","L_star": 70.8502312128167, "a_star": 1.22760824346796, "b_star": 6.63934293351205},
{ "name": "10YR6/14","L_star": 60.8609762416457, "a_star": 17.7760031409704, "b_star": 89.1126859295769},
{ "name": "10YR6/12","L_star": 60.8697824947294, "a_star": 16.4851134684311, "b_star": 75.3992364607604},
{ "name": "10YR6/10","L_star": 60.8961891551695, "a_star": 14.5585318374573, "b_star": 62.9356985276491},
{ "name": "10YR6/8","L_star": 60.9137835246555, "a_star": 12.1048438912548, "b_star": 50.3847272753937},
{ "name": "10YR6/6","L_star": 60.9489481351208, "a_star": 9.29014109643694, "b_star": 37.7589691862641},
{ "name": "10YR6/4","L_star": 60.984080635441, "a_star": 5.92987118783767, "b_star": 24.9583342795577},
{ "name": "10YR6/3","L_star": 60.9928587518952, "a_star": 4.43034473078363, "b_star": 18.943012515989},
{ "name": "10YR6/2","L_star": 61.0104089824442, "a_star": 2.9911818807985, "b_star": 12.8654511191163},
{ "name": "10YR6/1","L_star": 61.0279512174433, "a_star": 1.51068595953213, "b_star": 6.61822632180551},
{ "name": "10YR5/12","L_star": 50.8389541761052, "a_star": 16.1952209564784, "b_star": 76.6673977522447},
{ "name": "10YR5/10","L_star": 50.8505985934447, "a_star": 14.8919190480058, "b_star": 62.7603539398951},
{ "name": "10YR5/8","L_star": 50.8738752661451, "a_star": 12.8935002230373, "b_star": 50.5274776706926},
{ "name": "10YR5/6","L_star": 50.8971357463743, "a_star": 10.160511542249, "b_star": 38.1939953593116},
{ "name": "10YR5/4","L_star": 50.9319961673508, "a_star": 6.7561340570389, "b_star": 25.0596774442919},
{ "name": "10YR5/3","L_star": 50.9436082418668, "a_star": 5.03235636312849, "b_star": 18.7300301450707},
{ "name": "10YR5/2","L_star": 50.966820313159, "a_star": 3.30332429736652, "b_star": 12.4724300666186},
{ "name": "10YR5/1","L_star": 50.9900163040383, "a_star": 1.61598490702963, "b_star": 6.26937836144856},
{ "name": "10YR4/10","L_star": 40.5894399424351, "a_star": 14.8958176703485, "b_star": 65.5523557844363},
{ "name": "10YR4/8","L_star": 40.6056826069156, "a_star": 12.9996040258205, "b_star": 49.1692064457933},
{ "name": "10YR4/6","L_star": 40.6219159552695, "a_star": 10.562794844474, "b_star": 37.11254480413},
{ "name": "10YR4/4","L_star": 40.6543547569545, "a_star": 7.53341782825548, "b_star": 25.5570497960344},
{ "name": "10YR4/3","L_star": 40.6705602368878, "a_star": 5.72265319078336, "b_star": 19.3761345662245},
{ "name": "10YR4/2","L_star": 40.7029434212134, "a_star": 3.77114550416216, "b_star": 13.0605328685051},
{ "name": "10YR4/1","L_star": 40.7191211520259, "a_star": 1.78790384661653, "b_star": 6.62346972636352},
{ "name": "10YR3/8","L_star": 30.2082923590962, "a_star": 13.4066593092203, "b_star": 55.6833909436325},
{ "name": "10YR3/6","L_star": 30.2739914414501, "a_star": 10.6711506908951, "b_star": 36.4575190294416},
{ "name": "10YR3/4","L_star": 30.3055579039205, "a_star": 7.5068805669011, "b_star": 23.6203335797238},
{ "name": "10YR3/3","L_star": 30.3225374185553, "a_star": 5.79706362183691, "b_star": 17.7605462667841},
{ "name": "10YR3/2","L_star": 30.3419273485797, "a_star": 3.99069575687727, "b_star": 11.9852866445642},
{ "name": "10YR3/1","L_star": 30.3613010662833, "a_star": 2.09739171790474, "b_star": 6.11375083733979},
{ "name": "10YR2/4","L_star": 20.1596662785396, "a_star": 8.08876076022696, "b_star": 23.6658210768382},
{ "name": "10YR2/3","L_star": 20.1795517127494, "a_star": 6.10628196877705, "b_star": 16.7811802743358},
{ "name": "10YR2/2","L_star": 20.199415311619, "a_star": 4.11908497124536, "b_star": 10.864669590738},
{ "name": "10YR2/1","L_star": 20.2192571349912, "a_star": 2.09741974986782, "b_star": 5.36948908548506},
{ "name": "10YR1/2","L_star": 10.3641650318228, "a_star": 5.67390164309957, "b_star": 13.7289128808273},
{ "name": "10YR1/1","L_star": 10.3940733558595, "a_star": 2.54127697210803, "b_star": 5.02192741899275},
{ "name": "2.5Y9/12","L_star": 90.0030774471429, "a_star": 6.92527809980936, "b_star": 81.4092493931836},
{ "name": "2.5Y9/10","L_star": 90.0909815757417, "a_star": 5.03707333864017, "b_star": 68.072458810097},
{ "name": "2.5Y9/8","L_star": 90.1094691567687, "a_star": 3.52568515203483, "b_star": 55.4328797567846},
{ "name": "2.5Y9/6","L_star": 90.132569577483, "a_star": 1.94865287446222, "b_star": 42.3560250132227},
{ "name": "2.5Y9/4","L_star": 90.1510426764292, "a_star": 0.568330608918421, "b_star": 29.3416874102498},
{ "name": "2.5Y9/3","L_star": 90.1602768152846, "a_star": 0.112007744319575, "b_star": 22.6524482255176},
{ "name": "2.5Y9/2","L_star": 90.1695093479933, "a_star": -0.13351494525865, "b_star": 15.6074720477739},
{ "name": "2.5Y9/1","L_star": 90.1833551370584, "a_star": -0.238522691273724, "b_star": 8.12362831279416},
{ "name": "2.5Y8/20","L_star": 79.886224470121, "a_star": 13.2781027602053, "b_star": 144.918278557759},
{ "name": "2.5Y8/18","L_star": 80.050055608802, "a_star": 11.8118002228247, "b_star": 119.801648021608},
{ "name": "2.5Y8/16","L_star": 80.2919538068371, "a_star": 10.0995566296389, "b_star": 106.688057999388},
{ "name": "2.5Y8/14","L_star": 80.3984529402843, "a_star": 8.8388795472229, "b_star": 93.520923119369},
{ "name": "2.5Y8/12","L_star": 80.4152470961886, "a_star": 7.55381487576401, "b_star": 80.1922639639072},
{ "name": "2.5Y8/10","L_star": 80.4320354035271, "a_star": 6.09354593429995, "b_star": 66.6376266143657},
{ "name": "2.5Y8/8","L_star": 80.448817867389, "a_star": 4.37094304984015, "b_star": 53.6636479746499},
{ "name": "2.5Y8/6","L_star": 80.4711854048021, "a_star": 2.65971820127209, "b_star": 40.8276405673905},
{ "name": "2.5Y8/4","L_star": 80.4935425748736, "a_star": 1.1772096233314, "b_star": 27.5901224637753},
{ "name": "2.5Y8/3","L_star": 80.5047172759086, "a_star": 0.549721543154025, "b_star": 21.2758032244438},
{ "name": "2.5Y8/2","L_star": 80.5158893896084, "a_star": 0.103482795678567, "b_star": 14.8016941813226},
{ "name": "2.5Y8/1","L_star": 80.5214744766754, "a_star": -0.155190676284001, "b_star": 7.83011763536561},
{ "name": "2.5Y7/16","L_star": 70.3158187845481, "a_star": 11.3877221746762, "b_star": 108.119933331397},
{ "name": "2.5Y7/14","L_star": 70.732809888573, "a_star": 9.2380397183518, "b_star": 93.5384887152488},
{ "name": "2.5Y7/12","L_star": 70.746640666975, "a_star": 8.23480929653986, "b_star": 80.3254796550023},
{ "name": "2.5Y7/10","L_star": 70.7673785689525, "a_star": 6.80401437702738, "b_star": 67.0481299786057},
{ "name": "2.5Y7/8","L_star": 70.7811983316404, "a_star": 5.25867888969533, "b_star": 53.487354840812},
{ "name": "2.5Y7/6","L_star": 70.8019197264899, "a_star": 3.78343481959398, "b_star": 40.5050216098549},
{ "name": "2.5Y7/4","L_star": 70.8226312328179, "a_star": 2.12117845567422, "b_star": 27.3799684584009},
{ "name": "2.5Y7/3","L_star": 70.8364334159028, "a_star": 1.20715626839385, "b_star": 20.9520628183432},
{ "name": "2.5Y7/2","L_star": 70.8502312128167, "a_star": 0.385101060137227, "b_star": 14.3801532427869},
{ "name": "2.5Y7/1","L_star": 70.8640246270425, "a_star": -0.0466176992960921, "b_star": 7.46257170220093},
{ "name": "2.5Y6/14","L_star": 60.569230598101, "a_star": 10.631258586873, "b_star": 95.5473936294097},
{ "name": "2.5Y6/12","L_star": 60.9313698482419, "a_star": 8.67901814071681, "b_star": 79.442368345627},
{ "name": "2.5Y6/10","L_star": 60.940159995696, "a_star": 7.60985904474676, "b_star": 67.2771877681331},
{ "name": "2.5Y6/8","L_star": 60.957734267663, "a_star": 6.16712517399226, "b_star": 54.2235841953053},
{ "name": "2.5Y6/6","L_star": 60.984080635441, "a_star": 4.20922505249532, "b_star": 40.881410547423},
{ "name": "2.5Y6/4","L_star": 61.0016348671828, "a_star": 2.34749619702068, "b_star": 27.0756508033037},
{ "name": "2.5Y6/3","L_star": 61.0191810988183, "a_star": 1.4923693736742, "b_star": 20.653538231256},
{ "name": "2.5Y6/2","L_star": 61.0279512174433, "a_star": 0.799904474590385, "b_star": 14.1289394297869},
{ "name": "2.5Y6/1","L_star": 61.0367193394561, "a_star": 0.252644421262604, "b_star": 7.33271428233733},
{ "name": "2.5Y5/12","L_star": 50.6169350351417, "a_star": 9.9885318682309, "b_star": 83.3798809614409},
{ "name": "2.5Y5/10","L_star": 50.8971357463743, "a_star": 8.07913493367052, "b_star": 66.1083695387132},
{ "name": "2.5Y5/8","L_star": 50.9087599231063, "a_star": 6.81511741273888, "b_star": 53.8712648316071},
{ "name": "2.5Y5/6","L_star": 50.9319961673508, "a_star": 5.152087442378, "b_star": 41.2716582699076},
{ "name": "2.5Y5/4","L_star": 50.9552162893063, "a_star": 3.17960916276744, "b_star": 27.1308361404712},
{ "name": "2.5Y5/3","L_star": 50.966820313159, "a_star": 2.27768516080484, "b_star": 20.2753766583493},
{ "name": "2.5Y5/2","L_star": 50.9784203169096, "a_star": 1.41004550919999, "b_star": 13.534961912673},
{ "name": "2.5Y5/1","L_star": 50.9900163040383, "a_star": 0.635569880884257, "b_star": 6.80177031523443},
{ "name": "2.5Y4/10","L_star": 40.4591616342428, "a_star": 9.33650423742544, "b_star": 72.9841862243165},
{ "name": "2.5Y4/8","L_star": 40.6543547569545, "a_star": 7.1852701253978, "b_star": 52.8329540857321},
{ "name": "2.5Y4/6","L_star": 40.6705602368878, "a_star": 5.59613982129159, "b_star": 39.7003174810108},
{ "name": "2.5Y4/4","L_star": 40.686756453899, "a_star": 3.7950873984216, "b_star": 27.3217593359945},
{ "name": "2.5Y4/3","L_star": 40.7029434212134, "a_star": 2.72826701589507, "b_star": 20.7161903053081},
{ "name": "2.5Y4/2","L_star": 40.7191211520259, "a_star": 1.6850435195174, "b_star": 13.9424067053495},
{ "name": "2.5Y4/1","L_star": 40.7191211520259, "a_star": 0.681596904727994, "b_star": 7.0662037366741},
{ "name": "2.5Y3/6","L_star": 30.3079843112769, "a_star": 6.10766002594432, "b_star": 39.2148759578547},
{ "name": "2.5Y3/4","L_star": 30.3322344122274, "a_star": 3.95240311014516, "b_star": 25.2565044545111},
{ "name": "2.5Y3/3","L_star": 30.3443499492087, "a_star": 2.89093138306781, "b_star": 18.8947776818438},
{ "name": "2.5Y3/2","L_star": 30.3564591549096, "a_star": 1.87125569046073, "b_star": 12.6550750139127},
{ "name": "2.5Y3/1","L_star": 30.3709818560962, "a_star": 0.868559164226373, "b_star": 6.40666660544785},
{ "name": "2.5Y2/4","L_star": 20.187499769181, "a_star": 4.62557335539757, "b_star": 26.3090265868022},
{ "name": "2.5Y2/3","L_star": 20.199415311619, "a_star": 3.29711506172453, "b_star": 18.2401609436527},
{ "name": "2.5Y2/2","L_star": 20.2152905094929, "a_star": 2.05113866464923, "b_star": 11.6658726975254},
{ "name": "2.5Y2/1","L_star": 20.2271877805556, "a_star": 0.939021744594065, "b_star": 5.72377709699051},
{ "name": "2.5Y1/2","L_star": 10.3866026299139, "a_star": 2.99623390352195, "b_star": 14.97740341763},
{ "name": "2.5Y1/1","L_star": 10.4015398550854, "a_star": 1.2826521018091, "b_star": 5.04145634416108},
{ "name": "5Y9/20","L_star": 90.0678630346794, "a_star": -2.05723501009825, "b_star": 138.251250364027},
{ "name": "5Y9/18","L_star": 90.0863586735799, "a_star": -2.48374204395996, "b_star": 125.847141706693},
{ "name": "5Y9/16","L_star": 90.1140900455589, "a_star": -2.87985907128285, "b_star": 112.952423638053},
{ "name": "5Y9/14","L_star": 90.1418069307284, "a_star": -3.46298605254924, "b_star": 99.2724659188483},
{ "name": "5Y9/12","L_star": 90.1464250044782, "a_star": -3.79964851399045, "b_star": 85.3457916791674},
{ "name": "5Y9/10","L_star": 90.1556599466689, "a_star": -4.01019049174683, "b_star": 71.3672270161669},
{ "name": "5Y9/8","L_star": 90.1602768152846, "a_star": -4.10827879523951, "b_star": 58.2536223422105},
{ "name": "5Y9/6","L_star": 90.1695093479933, "a_star": -3.93357722371851, "b_star": 44.5871711727541},
{ "name": "5Y9/4","L_star": 90.1787402752534, "a_star": -3.44171419711486, "b_star": 30.7151425488158},
{ "name": "5Y9/3","L_star": 90.1833551370584, "a_star": -2.8713460763406, "b_star": 23.7565352682347},
{ "name": "5Y9/2","L_star": 90.1833551370584, "a_star": -2.06400516613098, "b_star": 16.4575680061885},
{ "name": "5Y9/1","L_star": 90.1879695977629, "a_star": -1.1711736070576, "b_star": 8.69410376785851},
{ "name": "5Y8/18","L_star": 79.9992712907649, "a_star": 0.693750066330656, "b_star": 124.672519063093},
{ "name": "5Y8/16","L_star": 80.1514634500276, "a_star": -0.0547129283394732, "b_star": 111.237265602698},
{ "name": "5Y8/14","L_star": 80.3984529402843, "a_star": -1.30793571207666, "b_star": 97.9472924969483},
{ "name": "5Y8/12","L_star": 80.4823652850047, "a_star": -2.06366823450693, "b_star": 84.5873031513307},
{ "name": "5Y8/10","L_star": 80.4879542536371, "a_star": -2.50630058943357, "b_star": 70.4215586137399},
{ "name": "5Y8/8","L_star": 80.4991302489017, "a_star": -2.932689690332, "b_star": 56.7371234896342},
{ "name": "5Y8/6","L_star": 80.5103036560818, "a_star": -3.11214250145947, "b_star": 43.0361143868354},
{ "name": "5Y8/4","L_star": 80.5214744766754, "a_star": -2.91161994274464, "b_star": 29.2361640265714},
{ "name": "5Y8/3","L_star": 80.5214744766754, "a_star": -2.44340554063582, "b_star": 22.570526329171},
{ "name": "5Y8/2","L_star": 80.5270589174699, "a_star": -1.80053964187482, "b_star": 15.6483714604372},
{ "name": "5Y8/1","L_star": 80.5270589174699, "a_star": -0.923579138013331, "b_star": 8.23760398439366},
{ "name": "5Y7/16","L_star": 70.1969359540931, "a_star": 2.05442750869261, "b_star": 112.632521603618},
{ "name": "5Y7/14","L_star": 70.5109149379283, "a_star": 0.890351101348552, "b_star": 96.3469894258793},
{ "name": "5Y7/12","L_star": 70.8088246600745, "a_star": -0.432433267755605, "b_star": 83.8038333766441},
{ "name": "5Y7/10","L_star": 70.8157284953717, "a_star": -0.926182004673481, "b_star": 70.6880775038147},
{ "name": "5Y7/8","L_star": 70.8295328728496, "a_star": -1.54602194361392, "b_star": 56.1996050612487},
{ "name": "5Y7/6","L_star": 70.8433328624133, "a_star": -2.07773916034837, "b_star": 42.785162394766},
{ "name": "5Y7/4","L_star": 70.8502312128167, "a_star": -2.15737494773738, "b_star": 29.0037866811429},
{ "name": "5Y7/3","L_star": 70.8571284675481, "a_star": -1.9013952053954, "b_star": 22.1201963201071},
{ "name": "5Y7/2","L_star": 70.8640246270425, "a_star": -1.42864298398659, "b_star": 15.0501741846581},
{ "name": "5Y7/1","L_star": 70.8640246270425, "a_star": -0.81419102789998, "b_star": 7.75605040867675},
{ "name": "5Y6/14","L_star": 60.3913274696579, "a_star": 2.87398374075953, "b_star": 102.960693315908},
{ "name": "5Y6/12","L_star": 60.8433576791392, "a_star": 1.15158701647805, "b_star": 82.1318927592106},
{ "name": "5Y6/10","L_star": 60.9928587518952, "a_star": 0.234193124043547, "b_star": 70.1714366275127},
{ "name": "5Y6/8","L_star": 61.0016348671828, "a_star": -0.387492425977187, "b_star": 57.0456276817679},
{ "name": "5Y6/6","L_star": 61.0191810988183, "a_star": -0.972444637273695, "b_star": 43.0362353562532},
{ "name": "5Y6/4","L_star": 61.0279512174433, "a_star": -1.55533446340506, "b_star": 28.5343498529681},
{ "name": "5Y6/3","L_star": 61.0367193394561, "a_star": -1.52514346594934, "b_star": 21.6899577238257},
{ "name": "5Y6/2","L_star": 61.0454854659928, "a_star": -1.22607957357468, "b_star": 14.7948382220509},
{ "name": "5Y6/1","L_star": 61.0454854659928, "a_star": -0.749874918960614, "b_star": 7.6523747156418},
{ "name": "5Y5/12","L_star": 50.4877182964346, "a_star": 3.43453678893069, "b_star": 90.811025953087},
{ "name": "5Y5/10","L_star": 50.9436082418668, "a_star": 1.40444987846744, "b_star": 69.2360116414111},
{ "name": "5Y5/8","L_star": 50.9552162893063, "a_star": 0.896686654286383, "b_star": 56.1895168866838},
{ "name": "5Y5/6","L_star": 50.966820313159, "a_star": 0.142406589684008, "b_star": 43.2494137521556},
{ "name": "5Y5/4","L_star": 50.9784203169096, "a_star": -0.688841504890347, "b_star": 28.6066920928326},
{ "name": "5Y5/3","L_star": 50.9900163040383, "a_star": -0.88207352752423, "b_star": 21.3505186357938},
{ "name": "5Y5/2","L_star": 51.0016082780202, "a_star": -0.894794788121156, "b_star": 14.18998869894},
{ "name": "5Y5/1","L_star": 51.0016082780202, "a_star": -0.658382567133908, "b_star": 7.10688510669288},
{ "name": "5Y4/8","L_star": 40.686756453899, "a_star": 1.74335088460986, "b_star": 55.2220156045058},
{ "name": "5Y4/6","L_star": 40.7029434212134, "a_star": 0.899483372321319, "b_star": 41.6140405006449},
{ "name": "5Y4/4","L_star": 40.7191211520259, "a_star": 0.097221080490123, "b_star": 28.5227404863854},
{ "name": "5Y4/3","L_star": 40.7191211520259, "a_star": -0.232689809802039, "b_star": 21.4985889953235},
{ "name": "5Y4/2","L_star": 40.7191211520259, "a_star": -0.384864620098374, "b_star": 14.3462339108925},
{ "name": "5Y4/1","L_star": 40.7352896595015, "a_star": -0.406378892440218, "b_star": 7.17979927883079},
{ "name": "5Y3/6","L_star": 30.3395044946333, "a_star": 1.74777287585737, "b_star": 40.9999983123851},
{ "name": "5Y3/4","L_star": 30.3540378198749, "a_star": 0.642934359757485, "b_star": 26.1139913143423},
{ "name": "5Y3/3","L_star": 30.3613010662833, "a_star": 0.223742492252793, "b_star": 19.4019230498382},
{ "name": "5Y3/2","L_star": 30.3709818560962, "a_star": -0.0892481252520172, "b_star": 12.8608571330669},
{ "name": "5Y3/1","L_star": 30.3782397967149, "a_star": -0.190861611820997, "b_star": 6.42346351960038},
{ "name": "5Y2/4","L_star": 20.2073546506852, "a_star": 1.44496009611814, "b_star": 27.9659882175132},
{ "name": "5Y2/3","L_star": 20.2192571349912, "a_star": 0.815515275011613, "b_star": 18.8542825531857},
{ "name": "5Y2/2","L_star": 20.2271877805556, "a_star": 0.323061708544153, "b_star": 11.8196080754976},
{ "name": "5Y2/1","L_star": 20.2351149553776, "a_star": 0.029614415006235, "b_star": 5.69505170472464},
{ "name": "5Y1/2","L_star": 10.4015398550854, "a_star": 0.956774375665215, "b_star": 14.8950325804442},
{ "name": "5Y1/1","L_star": 10.4090021335651, "a_star": 0.409777432970923, "b_star": 4.73375375347919},
{ "name": "7.5Y9/18","L_star": 90.1741250122608, "a_star": -10.3925029251436, "b_star": 128.68220286677},
{ "name": "7.5Y9/16","L_star": 90.1787402752534, "a_star": -10.4481291825382, "b_star": 114.367701036021},
{ "name": "7.5Y9/14","L_star": 90.1879695977629, "a_star": -10.5967745834689, "b_star": 101.212999079798},
{ "name": "7.5Y9/12","L_star": 90.192583657454, "a_star": -10.3336742993994, "b_star": 86.9936579854693},
{ "name": "7.5Y9/10","L_star": 90.192583657454, "a_star": -9.79117018251308, "b_star": 72.6976116490583},
{ "name": "7.5Y9/8","L_star": 90.1971973162189, "a_star": -9.02943490973657, "b_star": 59.2745198068087},
{ "name": "7.5Y9/6","L_star": 90.1971973162189, "a_star": -7.75950383176166, "b_star": 45.2741063688651},
{ "name": "7.5Y9/4","L_star": 90.1971973162189, "a_star": -6.04300013078746, "b_star": 31.2214175145763},
{ "name": "7.5Y9/3","L_star": 90.1971973162189, "a_star": -4.97868855170758, "b_star": 24.3028511699366},
{ "name": "7.5Y9/2","L_star": 90.1971973162189, "a_star": -3.69061654000119, "b_star": 17.0122997744034},
{ "name": "7.5Y9/1","L_star": 90.192583657454, "a_star": -1.8913299651539, "b_star": 8.88073372129585},
{ "name": "7.5Y8/18","L_star": 80.0162053661366, "a_star": -7.46852117258351, "b_star": 127.151252810523},
{ "name": "7.5Y8/16","L_star": 80.1402064735798, "a_star": -7.86018070511574, "b_star": 113.070437162932},
{ "name": "7.5Y8/14","L_star": 80.3424301154735, "a_star": -8.62360263867379, "b_star": 99.6140923258384},
{ "name": "7.5Y8/12","L_star": 80.5270589174699, "a_star": -9.11338582230614, "b_star": 86.1305544638246},
{ "name": "7.5Y8/10","L_star": 80.5326427121788, "a_star": -8.83113996437518, "b_star": 71.7006160104971},
{ "name": "7.5Y8/8","L_star": 80.5326427121788, "a_star": -8.06676330606743, "b_star": 57.8965038863905},
{ "name": "7.5Y8/6","L_star": 80.5382258609889, "a_star": -6.88573547239985, "b_star": 43.7506436698663},
{ "name": "7.5Y8/4","L_star": 80.5382258609889, "a_star": -5.29196250027558, "b_star": 29.7710479112624},
{ "name": "7.5Y8/3","L_star": 80.5382258609889, "a_star": -4.34932421883211, "b_star": 23.0650057408467},
{ "name": "7.5Y8/2","L_star": 80.5382258609889, "a_star": -3.27636377771517, "b_star": 16.1028333331734},
{ "name": "7.5Y8/1","L_star": 80.5382258609889, "a_star": -1.9416017801217, "b_star": 8.54677658622645},
{ "name": "7.5Y7/16","L_star": 70.1689158185295, "a_star": -5.75160251032408, "b_star": 116.518652785491},
{ "name": "7.5Y7/14","L_star": 70.4204441218562, "a_star": -6.49065735850607, "b_star": 98.5086725732155},
{ "name": "7.5Y7/12","L_star": 70.7673785689525, "a_star": -7.47594064461982, "b_star": 85.2227615637825},
{ "name": "7.5Y7/10","L_star": 70.8640246270425, "a_star": -7.55852032819343, "b_star": 72.1886971703302},
{ "name": "7.5Y7/8","L_star": 70.8640246270425, "a_star": -7.08666020520793, "b_star": 57.3589449812127},
{ "name": "7.5Y7/6","L_star": 70.8709196917347, "a_star": -6.1320183926431, "b_star": 43.3582156007412},
{ "name": "7.5Y7/4","L_star": 70.8709196917347, "a_star": -4.88592206099631, "b_star": 29.4136801549126},
{ "name": "7.5Y7/3","L_star": 70.8709196917347, "a_star": -3.98041890094708, "b_star": 22.378503091583},
{ "name": "7.5Y7/2","L_star": 70.8709196917347, "a_star": -2.94025872080056, "b_star": 15.206473536627},
{ "name": "7.5Y7/1","L_star": 70.8709196917347, "a_star": -1.63084468486091, "b_star": 7.77753395534253},
{ "name": "7.5Y6/14","L_star": 60.2931265197337, "a_star": -4.25609548934092, "b_star": 109.692903555119},
{ "name": "7.5Y6/12","L_star": 60.7198009014627, "a_star": -5.4671029705653, "b_star": 83.9095189008187},
{ "name": "7.5Y6/10","L_star": 61.0454854659928, "a_star": -6.412386356069, "b_star": 71.6659072127213},
{ "name": "7.5Y6/8","L_star": 61.0454854659928, "a_star": -6.198353444687, "b_star": 58.3984724171053},
{ "name": "7.5Y6/6","L_star": 61.0454854659928, "a_star": -5.59104085018169, "b_star": 43.985509230196},
{ "name": "7.5Y6/4","L_star": 61.0542495981883, "a_star": -4.407443727309, "b_star": 28.9774392523687},
{ "name": "7.5Y6/3","L_star": 61.0542495981883, "a_star": -3.66391735589844, "b_star": 21.9821543268176},
{ "name": "7.5Y6/2","L_star": 61.0542495981883, "a_star": -2.73713194095448, "b_star": 14.9602613960156},
{ "name": "7.5Y6/1","L_star": 61.0542495981883, "a_star": -1.56413833124336, "b_star": 7.74488775333815},
{ "name": "7.5Y5/12","L_star": 50.3107001044409, "a_star": -3.22568577976734, "b_star": 95.5725706127555},
{ "name": "7.5Y5/10","L_star": 50.8273057000631, "a_star": -4.56309603238747, "b_star": 71.2387820566314},
{ "name": "7.5Y5/8","L_star": 51.0016082780202, "a_star": -5.10516088193269, "b_star": 57.5836753543396},
{ "name": "7.5Y5/6","L_star": 51.0016082780202, "a_star": -4.75711243031823, "b_star": 44.1359865153223},
{ "name": "7.5Y5/4","L_star": 51.0016082780202, "a_star": -3.81870544787638, "b_star": 29.0352846790466},
{ "name": "7.5Y5/3","L_star": 51.0016082780202, "a_star": -3.13486375886823, "b_star": 21.6051433911668},
{ "name": "7.5Y5/2","L_star": 51.0016082780202, "a_star": -2.27682111660277, "b_star": 14.2899955442231},
{ "name": "7.5Y5/1","L_star": 51.0016082780202, "a_star": -1.22758033574816, "b_star": 7.07726885598385},
{ "name": "7.5Y4/8","L_star": 40.7352896595015, "a_star": -3.94241911130733, "b_star": 56.570478962446},
{ "name": "7.5Y4/6","L_star": 40.7352896595015, "a_star": -3.83888226812834, "b_star": 42.4884736750794},
{ "name": "7.5Y4/4","L_star": 40.7352896595015, "a_star": -3.34482881268269, "b_star": 28.9529534329369},
{ "name": "7.5Y4/3","L_star": 40.7352896595015, "a_star": -2.83156001871257, "b_star": 21.6840508147488},
{ "name": "7.5Y4/2","L_star": 40.7352896595015, "a_star": -2.13144295766604, "b_star": 14.335581422392},
{ "name": "7.5Y4/1","L_star": 40.7352896595015, "a_star": -1.1952112591864, "b_star": 7.08139187666927},
{ "name": "7.5Y3/6","L_star": 30.3709818560962, "a_star": -2.9001830812875, "b_star": 42.1136172151227},
{ "name": "7.5Y3/4","L_star": 30.3758207355903, "a_star": -2.5934639650102, "b_star": 26.4048261527762},
{ "name": "7.5Y3/3","L_star": 30.3782397967149, "a_star": -2.23177146409317, "b_star": 19.488632980457},
{ "name": "7.5Y3/2","L_star": 30.3806586055123, "a_star": -1.71381663501438, "b_star": 12.8299216128494},
{ "name": "7.5Y3/1","L_star": 30.3830771620482, "a_star": -0.965580331534349, "b_star": 6.3541955436307},
{ "name": "7.5Y2/4","L_star": 20.2311518015721, "a_star": -2.90602821869676, "b_star": 29.8816227045416},
{ "name": "7.5Y2/3","L_star": 20.2351149553776, "a_star": -1.71878034113862, "b_star": 18.6365741821872},
{ "name": "7.5Y2/2","L_star": 20.2390772424465, "a_star": -1.34563266514959, "b_star": 11.375114342265},
{ "name": "7.5Y2/1","L_star": 20.2430386632522, "a_star": -0.781089927573114, "b_star": 5.30472587023833},
{ "name": "7.5Y1/2","L_star": 10.4164601972582, "a_star": -1.03087030577463, "b_star": 13.892779022764},
{ "name": "7.5Y1/1","L_star": 10.4164601972582, "a_star": -0.327592688558503, "b_star": 4.33946243456461},
{ "name": "10Y9/18","L_star": 90.2525299808178, "a_star": -18.2006452571442, "b_star": 130.090329239942},
{ "name": "10Y9/16","L_star": 90.2479211263841, "a_star": -17.8673128118059, "b_star": 115.131317823596},
{ "name": "10Y9/14","L_star": 90.2387022177802, "a_star": -17.489364713442, "b_star": 101.732960648019},
{ "name": "10Y9/12","L_star": 90.2340921634365, "a_star": -16.5526782890114, "b_star": 87.2933659272782},
{ "name": "10Y9/10","L_star": 90.2294817089492, "a_star": -15.091851597419, "b_star": 73.0036991116786},
{ "name": "10Y9/8","L_star": 90.2248708542314, "a_star": -13.4072127547686, "b_star": 59.7291525666559},
{ "name": "10Y9/6","L_star": 90.2202595991964, "a_star": -10.9721397106769, "b_star": 45.2558919363071},
{ "name": "10Y9/4","L_star": 90.2110358878267, "a_star": -8.26556461333294, "b_star": 31.2265637569954},
{ "name": "10Y9/3","L_star": 90.2064234313182, "a_star": -6.69263523363151, "b_star": 24.3511652198013},
{ "name": "10Y9/2","L_star": 90.2018105741446, "a_star": -4.90679494558688, "b_star": 17.0677911852583},
{ "name": "10Y9/1","L_star": 90.1971973162189, "a_star": -2.70021372015977, "b_star": 9.0834564400835},
{ "name": "10Y8/18","L_star": 80.0669717866102, "a_star": -15.4974487862919, "b_star": 128.685052326475},
{ "name": "10Y8/16","L_star": 80.1739694988219, "a_star": -15.4687656186822, "b_star": 113.867276052492},
{ "name": "10Y8/14","L_star": 80.3480353301775, "a_star": -15.4982762086665, "b_star": 99.82632230347},
{ "name": "10Y8/12","L_star": 80.571711200428, "a_star": -15.485733767398, "b_star": 86.2014774990211},
{ "name": "10Y8/10","L_star": 80.5661319230916, "a_star": -14.1439023865666, "b_star": 71.7989264206633},
{ "name": "10Y8/8","L_star": 80.5661319230916, "a_star": -12.5038374950943, "b_star": 57.714480313164},
{ "name": "10Y8/6","L_star": 80.5605520009757, "a_star": -10.4099517862165, "b_star": 43.7772286826293},
{ "name": "10Y8/4","L_star": 80.5549714338939, "a_star": -7.70349284538241, "b_star": 29.7095126457441},
{ "name": "10Y8/3","L_star": 80.5493902216599, "a_star": -6.24456934680928, "b_star": 23.0414000433478},
{ "name": "10Y8/2","L_star": 80.5438083640871, "a_star": -4.61495644025289, "b_star": 16.1451239141441},
{ "name": "10Y8/1","L_star": 80.5382258609889, "a_star": -2.59848111925598, "b_star": 8.61983227907144},
{ "name": "10Y7/16","L_star": 70.1969359540931, "a_star": -13.3085430391573, "b_star": 119.505241617613},
{ "name": "10Y7/14","L_star": 70.4065087117747, "a_star": -13.534504815983, "b_star": 99.9854557836551},
{ "name": "10Y7/12","L_star": 70.7120554465534, "a_star": -13.7566777410573, "b_star": 85.6341919063465},
{ "name": "10Y7/10","L_star": 70.905378608356, "a_star": -13.4453172952536, "b_star": 72.6182995258652},
{ "name": "10Y7/8","L_star": 70.8984890111646, "a_star": -11.9003549348937, "b_star": 57.3674610949556},
{ "name": "10Y7/6","L_star": 70.8984890111646, "a_star": -9.89901273128863, "b_star": 42.9646129533983},
{ "name": "10Y7/4","L_star": 70.8915983213402, "a_star": -7.54820200277029, "b_star": 29.2430942671723},
{ "name": "10Y7/3","L_star": 70.8847065384495, "a_star": -6.04705665133276, "b_star": 22.2581102671185},
{ "name": "10Y7/2","L_star": 70.8847065384495, "a_star": -4.32458509192596, "b_star": 15.1239767467457},
{ "name": "10Y7/1","L_star": 70.877813662059, "a_star": -2.33845563733659, "b_star": 7.75726066913343},
{ "name": "10Y6/14","L_star": 60.0959626214371, "a_star": -11.5509077648699, "b_star": 113.259564844756},
{ "name": "10Y6/12","L_star": 60.7021174155531, "a_star": -12.0436550917482, "b_star": 84.4100835328129},
{ "name": "10Y6/10","L_star": 61.0542495981883, "a_star": -12.5150896575738, "b_star": 71.5728284693497},
{ "name": "10Y6/8","L_star": 61.0805300400617, "a_star": -11.4725366216152, "b_star": 58.197433050289},
{ "name": "10Y6/6","L_star": 61.0717718840904, "a_star": -9.66964111298663, "b_star": 43.4222145248171},
{ "name": "10Y6/4","L_star": 61.0717718840904, "a_star": -7.21283001526535, "b_star": 28.6901793548159},
{ "name": "10Y6/3","L_star": 61.0630117371766, "a_star": -5.77353475688752, "b_star": 21.7522913418894},
{ "name": "10Y6/2","L_star": 61.0630117371766, "a_star": -4.14381112060025, "b_star": 14.7979649197677},
{ "name": "10Y6/1","L_star": 61.0542495981883, "a_star": -2.29055025280156, "b_star": 7.64782694067481},
{ "name": "10Y5/12","L_star": 50.1921608811649, "a_star": -10.1103850777092, "b_star": 96.9473298009466},
{ "name": "10Y5/10","L_star": 50.8505985934447, "a_star": -10.7430621520831, "b_star": 71.2325699287115},
{ "name": "10Y5/8","L_star": 51.0363601557647, "a_star": -10.5232296879305, "b_star": 57.2892483161317},
{ "name": "10Y5/6","L_star": 51.0247802004202, "a_star": -9.19273024651696, "b_star": 43.4491385996986},
{ "name": "10Y5/4","L_star": 51.0247802004202, "a_star": -6.96462088473421, "b_star": 28.6354480871174},
{ "name": "10Y5/3","L_star": 51.0247802004202, "a_star": -5.50398707585281, "b_star": 21.2019068744234},
{ "name": "10Y5/2","L_star": 51.0131962423256, "a_star": -3.87803414685844, "b_star": 13.9068443621382},
{ "name": "10Y5/1","L_star": 51.0131962423256, "a_star": -2.04031469929411, "b_star": 6.7933400336988},
{ "name": "10Y4/8","L_star": 40.7352896595015, "a_star": -9.35877135732663, "b_star": 57.0064003872283},
{ "name": "10Y4/6","L_star": 40.7675990569515, "a_star": -8.36327501704373, "b_star": 42.2765373238841},
{ "name": "10Y4/4","L_star": 40.751448956775, "a_star": -6.61158500411743, "b_star": 28.4200663667124},
{ "name": "10Y4/3","L_star": 40.751448956775, "a_star": -5.31590445501262, "b_star": 21.0664950043475},
{ "name": "10Y4/2","L_star": 40.751448956775, "a_star": -3.80822371733608, "b_star": 13.7818142703958},
{ "name": "10Y4/1","L_star": 40.751448956775, "a_star": -2.02842709647352, "b_star": 6.69675968452775},
{ "name": "10Y3/6","L_star": 30.4024165408095, "a_star": -7.44670540648562, "b_star": 41.5754049046783},
{ "name": "10Y3/4","L_star": 30.3975832074544, "a_star": -5.77627030507522, "b_star": 26.1205807948666},
{ "name": "10Y3/3","L_star": 30.395166163107, "a_star": -4.6811025313247, "b_star": 19.0959140646499},
{ "name": "10Y3/2","L_star": 30.3927488668924, "a_star": -3.37231544869437, "b_star": 12.3564297956087},
{ "name": "10Y3/1","L_star": 30.3903313187449, "a_star": -1.83915126015011, "b_star": 5.97988835044149},
{ "name": "10Y2/4","L_star": 20.2549177328224, "a_star": -5.26185195864534, "b_star": 26.8299825565016},
{ "name": "10Y2/3","L_star": 20.2509589079675, "a_star": -4.06960673020462, "b_star": 17.5749356411115},
{ "name": "10Y2/2","L_star": 20.2509589079675, "a_star": -2.75801174280219, "b_star": 10.5827357986451},
{ "name": "10Y2/1","L_star": 20.2469992182683, "a_star": -1.39660599853159, "b_star": 4.82751546824408},
{ "name": "10Y1/2","L_star": 10.4239140521113, "a_star": -2.73697826785421, "b_star": 11.8686951767691},
{ "name": "10Y1/1","L_star": 10.4239140521113, "a_star": -0.964273519486789, "b_star": 3.84538829665429},
{ "name": "2.5GY9/18","L_star": 90.3400223523973, "a_star": -29.6658983960358, "b_star": 130.447581719672},
{ "name": "2.5GY9/16","L_star": 90.3308194079677, "a_star": -28.7781694647826, "b_star": 113.618952706269},
{ "name": "2.5GY9/14","L_star": 90.3170120036561, "a_star": -27.6640366959051, "b_star": 100.198377815335},
{ "name": "2.5GY9/12","L_star": 90.2985965506391, "a_star": -25.6203959419467, "b_star": 85.9007087669634},
{ "name": "2.5GY9/10","L_star": 90.2847807724031, "a_star": -22.7126005719114, "b_star": 71.1544401927038},
{ "name": "2.5GY9/8","L_star": 90.2709614014539, "a_star": -19.563123864661, "b_star": 58.0546518425528},
{ "name": "2.5GY9/6","L_star": 90.2525299808178, "a_star": -15.5816118407673, "b_star": 44.0692861088617},
{ "name": "2.5GY9/4","L_star": 90.2340921634365, "a_star": -11.3959001733014, "b_star": 30.5929971381129},
{ "name": "2.5GY9/3","L_star": 90.2202595991964, "a_star": -8.99893233499355, "b_star": 23.8320163047068},
{ "name": "2.5GY9/2","L_star": 90.2064234313182, "a_star": -6.28883829296134, "b_star": 16.6497173788968},
{ "name": "2.5GY9/1","L_star": 90.1971973162189, "a_star": -3.35354746349831, "b_star": 8.83463450636683},
{ "name": "2.5GY8/18","L_star": 80.1964650190854, "a_star": -26.4157109672084, "b_star": 127.68934537739},
{ "name": "2.5GY8/16","L_star": 80.2807296601208, "a_star": -26.1756098122424, "b_star": 112.821916455501},
{ "name": "2.5GY8/14","L_star": 80.420843848224, "a_star": -25.535533936467, "b_star": 99.0553840632181},
{ "name": "2.5GY8/12","L_star": 80.6386123033635, "a_star": -24.5007800521614, "b_star": 83.8999885107462},
{ "name": "2.5GY8/10","L_star": 80.6274685518459, "a_star": -21.9957077980995, "b_star": 69.6585660783635},
{ "name": "2.5GY8/8","L_star": 80.6107481035883, "a_star": -18.6631092110024, "b_star": 55.1514719363952},
{ "name": "2.5GY8/6","L_star": 80.5940218657014, "a_star": -15.0637425877428, "b_star": 41.964408390518},
{ "name": "2.5GY8/4","L_star": 80.5772898331711, "a_star": -10.9170290397022, "b_star": 28.5167417949878},
{ "name": "2.5GY8/3","L_star": 80.5661319230916, "a_star": -8.62488733769812, "b_star": 22.2495920358815},
{ "name": "2.5GY8/2","L_star": 80.5549714338939, "a_star": -6.06796535812498, "b_star": 15.7270682841893},
{ "name": "2.5GY8/1","L_star": 80.5382258609889, "a_star": -3.26696584797387, "b_star": 8.50305488845997},
{ "name": "2.5GY7/16","L_star": 70.2669067096778, "a_star": -23.6629679253572, "b_star": 117.377940219106},
{ "name": "2.5GY7/14","L_star": 70.5039623609524, "a_star": -23.4188947587609, "b_star": 98.8460904304123},
{ "name": "2.5GY7/12","L_star": 70.7881065627101, "a_star": -22.8882312744098, "b_star": 83.2688228197131},
{ "name": "2.5GY7/10","L_star": 70.9604561030066, "a_star": -21.5217015968568, "b_star": 69.8448092801251},
{ "name": "2.5GY7/8","L_star": 70.9466932704427, "a_star": -18.3133939439431, "b_star": 54.7431528877033},
{ "name": "2.5GY7/6","L_star": 70.9329260794485, "a_star": -14.7288137537934, "b_star": 40.7649555830655},
{ "name": "2.5GY7/4","L_star": 70.9122671133475, "a_star": -10.7864259257898, "b_star": 27.8631736810242},
{ "name": "2.5GY7/3","L_star": 70.905378608356, "a_star": -8.49751746669913, "b_star": 21.3045320496181},
{ "name": "2.5GY7/2","L_star": 70.8915983213402, "a_star": -5.92764309625465, "b_star": 14.5375877800927},
{ "name": "2.5GY7/1","L_star": 70.877813662059, "a_star": -3.09824242899792, "b_star": 7.50869630078383},
{ "name": "2.5GY6/14","L_star": 60.4091550846504, "a_star": -21.3446337022404, "b_star": 104.821489276341},
{ "name": "2.5GY6/12","L_star": 60.8080962963385, "a_star": -21.0785560283081, "b_star": 82.7958725820823},
{ "name": "2.5GY6/10","L_star": 61.1067925736237, "a_star": -20.4544274637454, "b_star": 69.0909941826882},
{ "name": "2.5GY6/8","L_star": 61.1330372293513, "a_star": -18.291322499161, "b_star": 55.7126391079436},
{ "name": "2.5GY6/6","L_star": 61.1155427771229, "a_star": -14.7026582044433, "b_star": 41.0031981671453},
{ "name": "2.5GY6/4","L_star": 61.098040383699, "a_star": -10.6815768035787, "b_star": 27.2417906072523},
{ "name": "2.5GY6/3","L_star": 61.0805300400617, "a_star": -8.38886915439357, "b_star": 20.6312854212056},
{ "name": "2.5GY6/2","L_star": 61.0717718840904, "a_star": -5.92092875778694, "b_star": 13.9696842289853},
{ "name": "2.5GY6/1","L_star": 61.0630117371766, "a_star": -3.20762671496849, "b_star": 7.1849264669152},
{ "name": "2.5GY5/12","L_star": 50.499486034432, "a_star": -19.2452365195898, "b_star": 89.4344782836368},
{ "name": "2.5GY5/10","L_star": 50.9784203169096, "a_star": -19.0236185939742, "b_star": 68.4095709220881},
{ "name": "2.5GY5/8","L_star": 51.0826400186909, "a_star": -17.3302253616714, "b_star": 54.8112682136929},
{ "name": "2.5GY5/6","L_star": 51.0710760398342, "a_star": -14.3535153034177, "b_star": 40.6843283628463},
{ "name": "2.5GY5/4","L_star": 51.0479361118152, "a_star": -10.3591107299073, "b_star": 26.4556074266938},
{ "name": "2.5GY5/3","L_star": 51.0363601557647, "a_star": -8.08237646139331, "b_star": 19.5838789674636},
{ "name": "2.5GY5/2","L_star": 51.0247802004202, "a_star": -5.60008446405735, "b_star": 12.8779298480099},
{ "name": "2.5GY5/1","L_star": 51.0131962423256, "a_star": -2.9542023769995, "b_star": 6.34468359886993},
{ "name": "2.5GY4/8","L_star": 40.8321077477464, "a_star": -16.433588430199, "b_star": 54.3180926331837},
{ "name": "2.5GY4/6","L_star": 40.7998717182846, "a_star": -13.7845396624939, "b_star": 39.8187104791814},
{ "name": "2.5GY4/4","L_star": 40.7837399731063, "a_star": -10.2932312820229, "b_star": 26.5949230085067},
{ "name": "2.5GY4/3","L_star": 40.7837399731063, "a_star": -8.05860585765591, "b_star": 19.4447812661044},
{ "name": "2.5GY4/2","L_star": 40.7675990569515, "a_star": -5.5346039622447, "b_star": 12.4171988087939},
{ "name": "2.5GY4/1","L_star": 40.751448956775, "a_star": -2.80772123354764, "b_star": 5.82324919858996},
{ "name": "2.5GY3/6","L_star": 30.4434592823515, "a_star": -12.8093259225703, "b_star": 38.3859615872807},
{ "name": "2.5GY3/4","L_star": 30.4289818938566, "a_star": -9.51896637378191, "b_star": 24.5868228657121},
{ "name": "2.5GY3/3","L_star": 30.4193252833673, "a_star": -7.39795983973932, "b_star": 17.785716379657},
{ "name": "2.5GY3/2","L_star": 30.4072488674824, "a_star": -5.00250419041676, "b_star": 11.254799815445},
{ "name": "2.5GY3/1","L_star": 30.395166163107, "a_star": -2.47740888981704, "b_star": 5.21949199628987},
{ "name": "2.5GY2/4","L_star": 20.2865572538282, "a_star": -8.82310007754336, "b_star": 23.2521304514007},
{ "name": "2.5GY2/3","L_star": 20.27469890094, "a_star": -6.55230667547699, "b_star": 15.4274008440795},
{ "name": "2.5GY2/2","L_star": 20.262832789887, "a_star": -4.23241671502325, "b_star": 9.23084863223281},
{ "name": "2.5GY2/1","L_star": 20.2549177328224, "a_star": -1.97786534031322, "b_star": 4.13125816706085},
{ "name": "2.5GY1/2","L_star": 10.4536875015812, "a_star": -4.44691101587322, "b_star": 9.55487183315339},
{ "name": "2.5GY1/1","L_star": 10.4313637040572, "a_star": -1.39583174058555, "b_star": 3.33156556919922},
{ "name": "5GY9/18","L_star": 90.4365572460867, "a_star": -43.0643516507352, "b_star": 125.156865023158},
{ "name": "5GY9/16","L_star": 90.3998027018785, "a_star": -41.0145531623734, "b_star": 109.612836299514},
{ "name": "5GY9/14","L_star": 90.3860131990411, "a_star": -37.9997745536502, "b_star": 94.9835064552485},
{ "name": "5GY9/12","L_star": 90.3676216327313, "a_star": -34.4153953294294, "b_star": 81.3188631224478},
{ "name": "5GY9/10","L_star": 90.3400223523973, "a_star": -29.8881539849641, "b_star": 67.7110404992384},
{ "name": "5GY9/8","L_star": 90.3170120036561, "a_star": -25.214028304964, "b_star": 55.0213214254946},
{ "name": "5GY9/6","L_star": 90.2847807724031, "a_star": -19.7712185879272, "b_star": 42.0361101673646},
{ "name": "5GY9/4","L_star": 90.2525299808178, "a_star": -14.0115902029823, "b_star": 29.2649709768571},
{ "name": "5GY9/3","L_star": 90.2340921634365, "a_star": -10.9935508307163, "b_star": 22.7599708714945},
{ "name": "5GY9/2","L_star": 90.2156479437571, "a_star": -7.79906093790317, "b_star": 15.83451663558},
{ "name": "5GY9/1","L_star": 90.1971973162189, "a_star": -4.20694466135774, "b_star": 8.35958794056142},
{ "name": "5GY8/20","L_star": 80.1570909499497, "a_star": -40.7509000851289, "b_star": 140.398643921114},
{ "name": "5GY8/18","L_star": 80.3592438034441, "a_star": -40.2431796911305, "b_star": 122.595453696533},
{ "name": "5GY8/16","L_star": 80.448817867389, "a_star": -38.7287704775001, "b_star": 106.055691413046},
{ "name": "5GY8/14","L_star": 80.5438083640871, "a_star": -36.757043465239, "b_star": 93.4014922216019},
{ "name": "5GY8/12","L_star": 80.7109841237154, "a_star": -33.7804132743364, "b_star": 78.5756118402018},
{ "name": "5GY8/10","L_star": 80.6831616296243, "a_star": -29.3421717793367, "b_star": 64.4910882063091},
{ "name": "5GY8/8","L_star": 80.6553231133634, "a_star": -24.3054405349707, "b_star": 51.3555301689331},
{ "name": "5GY8/6","L_star": 80.6274685518459, "a_star": -19.0526808360023, "b_star": 39.2663610527845},
{ "name": "5GY8/4","L_star": 80.5940218657014, "a_star": -13.2650192426423, "b_star": 26.8603434138212},
{ "name": "5GY8/3","L_star": 80.5772898331711, "a_star": -10.3973449214277, "b_star": 20.8794930300259},
{ "name": "5GY8/2","L_star": 80.5605520009757, "a_star": -7.42500565583315, "b_star": 14.6698495715529},
{ "name": "5GY8/1","L_star": 80.5382258609889, "a_star": -4.01813055870653, "b_star": 7.80279713205987},
{ "name": "5GY7/16","L_star": 70.5942588381602, "a_star": -36.6851766943418, "b_star": 107.086525622133},
{ "name": "5GY7/14","L_star": 70.6982136304596, "a_star": -35.0717404499885, "b_star": 91.9444946329739},
{ "name": "5GY7/12","L_star": 70.905378608356, "a_star": -32.9004536417508, "b_star": 77.4828884834519},
{ "name": "5GY7/10","L_star": 71.0223350063132, "a_star": -29.2600043341581, "b_star": 64.0299486695351},
{ "name": "5GY7/8","L_star": 70.9948441388999, "a_star": -24.2525360888384, "b_star": 50.1330723105398},
{ "name": "5GY7/6","L_star": 70.9673358859548, "a_star": -18.9930069278737, "b_star": 37.4778262586259},
{ "name": "5GY7/4","L_star": 70.9329260794485, "a_star": -13.5490186004894, "b_star": 25.7065602185689},
{ "name": "5GY7/3","L_star": 70.9191545265718, "a_star": -10.5257575988826, "b_star": 19.79030760947},
{ "name": "5GY7/2","L_star": 70.8984890111646, "a_star": -7.28570162504999, "b_star": 13.6529836422992},
{ "name": "5GY7/1","L_star": 70.877813662059, "a_star": -3.72227562982003, "b_star": 7.10627573438645},
{ "name": "5GY6/14","L_star": 60.887388950337, "a_star": -33.5315399070471, "b_star": 92.8741080716969},
{ "name": "5GY6/12","L_star": 61.0104089824442, "a_star": -31.6309360582685, "b_star": 77.3197252245025},
{ "name": "5GY6/10","L_star": 61.1942054050492, "a_star": -29.0639220937423, "b_star": 63.8053603498975},
{ "name": "5GY6/8","L_star": 61.1854730287976, "a_star": -24.6403037153818, "b_star": 50.6112079014759},
{ "name": "5GY6/6","L_star": 61.1505237493858, "a_star": -19.1816938124544, "b_star": 37.2913957533701},
{ "name": "5GY6/4","L_star": 61.1155427771229, "a_star": -13.3744077987065, "b_star": 24.9771887886802},
{ "name": "5GY6/3","L_star": 61.098040383699, "a_star": -10.4174191935157, "b_star": 18.9866786978184},
{ "name": "5GY6/2","L_star": 61.0805300400617, "a_star": -7.33092500716143, "b_star": 12.9504228090505},
{ "name": "5GY6/1","L_star": 61.0630117371766, "a_star": -3.9356728195612, "b_star": 6.68783198098524},
{ "name": "5GY5/12","L_star": 50.966820313159, "a_star": -30.2424223689164, "b_star": 79.8773478291549},
{ "name": "5GY5/10","L_star": 51.1057560232836, "a_star": -27.7897931136164, "b_star": 62.9737291920426},
{ "name": "5GY5/8","L_star": 51.1404001987838, "a_star": -24.2081661706443, "b_star": 49.6442181213694},
{ "name": "5GY5/6","L_star": 51.1057560232836, "a_star": -19.135054137596, "b_star": 36.6180035637296},
{ "name": "5GY5/4","L_star": 51.0710760398342, "a_star": -13.0795605076617, "b_star": 23.7511942617146},
{ "name": "5GY5/3","L_star": 51.0595080720228, "a_star": -10.0215534402558, "b_star": 17.6465689861389},
{ "name": "5GY5/2","L_star": 51.0363601557647, "a_star": -6.85207302815821, "b_star": 11.6946932058413},
{ "name": "5GY5/1","L_star": 51.0131962423256, "a_star": -3.56533532533049, "b_star": 5.82533628529571},
{ "name": "5GY4/10","L_star": 40.6381400008437, "a_star": -26.6127662399912, "b_star": 68.6933985072019},
{ "name": "5GY4/8","L_star": 40.8964703256786, "a_star": -23.602878764474, "b_star": 49.3544041753422},
{ "name": "5GY4/6","L_star": 40.848212057973, "a_star": -18.8425621220581, "b_star": 35.9685410497443},
{ "name": "5GY4/4","L_star": 40.8159943055026, "a_star": -13.2773466751044, "b_star": 23.8675088810224},
{ "name": "5GY4/3","L_star": 40.7998717182846, "a_star": -10.089433971677, "b_star": 17.5377143729431},
{ "name": "5GY4/2","L_star": 40.7675990569515, "a_star": -6.66943152448957, "b_star": 11.2639646469754},
{ "name": "5GY4/1","L_star": 40.751448956775, "a_star": -3.23095502157106, "b_star": 5.33773113535613},
{ "name": "5GY3/8","L_star": 30.4458713030924, "a_star": -22.8883125194161, "b_star": 53.1262299996292},
{ "name": "5GY3/6","L_star": 30.4868373828619, "a_star": -18.0192297530431, "b_star": 33.7357731319192},
{ "name": "5GY3/4","L_star": 30.4555168817515, "a_star": -12.6587472991732, "b_star": 21.66799856877},
{ "name": "5GY3/3","L_star": 30.4386344891231, "a_star": -9.51614666267789, "b_star": 15.7161757561554},
{ "name": "5GY3/2","L_star": 30.4193252833673, "a_star": -6.21895531705141, "b_star": 9.98603219924097},
{ "name": "5GY3/1","L_star": 30.4, "a_star": -2.95381216972992, "b_star": 4.66245422815778},
{ "name": "5GY2/6","L_star": 20.3772148994357, "a_star": -18.478289478862, "b_star": 37.4331318440067},
{ "name": "5GY2/4","L_star": 20.3141966443599, "a_star": -11.8496180188234, "b_star": 19.5315553331465},
{ "name": "5GY2/3","L_star": 20.2944585185861, "a_star": -8.47491347973861, "b_star": 13.2859122565095},
{ "name": "5GY2/2","L_star": 20.27469890094, "a_star": -5.20953895791662, "b_star": 7.88921015172772},
{ "name": "5GY2/1","L_star": 20.258875693305, "a_star": -2.26757922751275, "b_star": 3.384632572136},
{ "name": "5GY1/4","L_star": 10.4685491268993, "a_star": -15.1655868392071, "b_star": 29.4931853675179},
{ "name": "5GY1/3","L_star": 10.5204338421202, "a_star": -10.3263044038682, "b_star": 15.4193367016777},
{ "name": "5GY1/2","L_star": 10.4685491268993, "a_star": -5.4943796654926, "b_star": 7.86196058164774},
{ "name": "5GY1/1","L_star": 10.4462504228925, "a_star": -1.7917850640814, "b_star": 2.840779675165},
{ "name": "7.5GY9/18","L_star": 90.4824647597565, "a_star": -66.6808850273776, "b_star": 106.648067566781},
{ "name": "7.5GY9/16","L_star": 90.4365572460867, "a_star": -61.2090267892931, "b_star": 93.2194703518251},
{ "name": "7.5GY9/14","L_star": 90.4181831475392, "a_star": -54.7732596861409, "b_star": 79.8274632596897},
{ "name": "7.5GY9/12","L_star": 90.4227772670039, "a_star": -48.4386869118527, "b_star": 68.5929372357667},
{ "name": "7.5GY9/10","L_star": 90.4227772670039, "a_star": -41.8018767923747, "b_star": 57.4498750403533},
{ "name": "7.5GY9/8","L_star": 90.3906100971672, "a_star": -34.4309096618887, "b_star": 46.2078211002345},
{ "name": "7.5GY9/6","L_star": 90.3446232273371, "a_star": -27.0165847298058, "b_star": 35.5667041475822},
{ "name": "7.5GY9/4","L_star": 90.2893864308902, "a_star": -18.8306021532937, "b_star": 24.1306936256531},
{ "name": "7.5GY9/3","L_star": 90.2617464903823, "a_star": -14.8063294812384, "b_star": 18.7963631351823},
{ "name": "7.5GY9/2","L_star": 90.2340921634365, "a_star": -10.5152023196641, "b_star": 13.2685619392183},
{ "name": "7.5GY9/1","L_star": 90.2018105741446, "a_star": -5.67196174960111, "b_star": 7.15695662554201},
{ "name": "7.5GY8/20","L_star": 80.6163222293855, "a_star": -66.5770006812667, "b_star": 117.12447747582},
{ "name": "7.5GY8/18","L_star": 80.6831616296243, "a_star": -63.5918221287234, "b_star": 104.066944188429},
{ "name": "7.5GY8/16","L_star": 80.7221086403056, "a_star": -58.9689242485988, "b_star": 90.0874040695956},
{ "name": "7.5GY8/14","L_star": 80.7776928768033, "a_star": -53.3668118463401, "b_star": 77.4468870700195},
{ "name": "7.5GY8/12","L_star": 80.8110127943531, "a_star": -46.8574962044098, "b_star": 65.1516631706024},
{ "name": "7.5GY8/10","L_star": 80.7665811374695, "a_star": -39.7993897558688, "b_star": 53.6797347636626},
{ "name": "7.5GY8/8","L_star": 80.7221086403056, "a_star": -32.6099370351864, "b_star": 42.951903460667},
{ "name": "7.5GY8/6","L_star": 80.6775952088837, "a_star": -25.6217291704721, "b_star": 32.6161785418553},
{ "name": "7.5GY8/4","L_star": 80.6330407488799, "a_star": -18.0263686969646, "b_star": 22.2554379061126},
{ "name": "7.5GY8/3","L_star": 80.605173334499, "a_star": -14.0281914014428, "b_star": 17.2570738768911},
{ "name": "7.5GY8/2","L_star": 80.5772898331711, "a_star": -9.8026587268969, "b_star": 12.0581636345988},
{ "name": "7.5GY8/1","L_star": 80.5438083640871, "a_star": -5.163439378017, "b_star": 6.41176953293892},
{ "name": "7.5GY7/18","L_star": 70.8502312128167, "a_star": -60.6656019997313, "b_star": 103.231957996284},
{ "name": "7.5GY7/14","L_star": 71.056674183978, "a_star": -52.3363497114528, "b_star": 76.3353985458444},
{ "name": "7.5GY7/12","L_star": 71.1252713868811, "a_star": -46.2581961527373, "b_star": 63.6537906594534},
{ "name": "7.5GY7/10","L_star": 71.1047035694533, "a_star": -39.5532914410267, "b_star": 52.3796351618631},
{ "name": "7.5GY7/8","L_star": 71.056674183978, "a_star": -32.1158147569273, "b_star": 41.2714448459557},
{ "name": "7.5GY7/6","L_star": 71.0154639178515, "a_star": -24.7288275985608, "b_star": 30.9505653128457},
{ "name": "7.5GY7/4","L_star": 70.9673358859548, "a_star": -17.4000382665801, "b_star": 21.0622351044224},
{ "name": "7.5GY7/3","L_star": 70.939810219965, "a_star": -13.561380613291, "b_star": 16.1838333122774},
{ "name": "7.5GY7/2","L_star": 70.9122671133475, "a_star": -9.48018991906219, "b_star": 11.1617051223691},
{ "name": "7.5GY7/1","L_star": 70.8847065384495, "a_star": -4.9797302701447, "b_star": 5.85372771478661},
{ "name": "7.5GY6/16","L_star": 61.0892862062213, "a_star": -55.9614640401925, "b_star": 90.6743602382313},
{ "name": "7.5GY6/14","L_star": 61.333655554213, "a_star": -52.1005389676957, "b_star": 76.6405202178964},
{ "name": "7.5GY6/12","L_star": 61.3075469581861, "a_star": -46.1268171706281, "b_star": 63.5316482886275},
{ "name": "7.5GY6/10","L_star": 61.2901314255044, "a_star": -39.7735219460527, "b_star": 52.344755039036},
{ "name": "7.5GY6/8","L_star": 61.246558222321, "a_star": -32.6587395436737, "b_star": 41.5404951064282},
{ "name": "7.5GY6/6","L_star": 61.1942054050492, "a_star": -24.7693422466564, "b_star": 30.5114282408871},
{ "name": "7.5GY6/4","L_star": 61.1505237493858, "a_star": -17.1614094098704, "b_star": 20.4754001230749},
{ "name": "7.5GY6/3","L_star": 61.1242909953236, "a_star": -13.34336288301, "b_star": 15.6443036449113},
{ "name": "7.5GY6/2","L_star": 61.098040383699, "a_star": -9.34349387040545, "b_star": 10.75093901143},
{ "name": "7.5GY6/1","L_star": 61.0630117371766, "a_star": -4.94983929722764, "b_star": 5.61524360212302},
{ "name": "7.5GY5/14","L_star": 51.0826400186909, "a_star": -51.0561845722775, "b_star": 78.2011149108781},
{ "name": "7.5GY5/12","L_star": 51.381914509339, "a_star": -46.3170968307092, "b_star": 64.1939427587085},
{ "name": "7.5GY5/10","L_star": 51.2901134222191, "a_star": -39.6052937404613, "b_star": 51.9400395839621},
{ "name": "7.5GY5/8","L_star": 51.1980611685518, "a_star": -32.2102335259776, "b_star": 40.1809203322886},
{ "name": "7.5GY5/6","L_star": 51.1519403159239, "a_star": -24.7042863010657, "b_star": 29.6985179947808},
{ "name": "7.5GY5/4","L_star": 51.1057560232836, "a_star": -16.6175563024425, "b_star": 19.2287848358416},
{ "name": "7.5GY5/3","L_star": 51.0826400186909, "a_star": -12.6286188893099, "b_star": 14.3080201983605},
{ "name": "7.5GY5/2","L_star": 51.0479361118152, "a_star": -8.55066352026973, "b_star": 9.52346770437948},
{ "name": "7.5GY5/1","L_star": 51.0247802004202, "a_star": -4.37318409867221, "b_star": 4.76372566314545},
{ "name": "7.5GY4/12","L_star": 40.7352896595015, "a_star": -45.9479295648814, "b_star": 68.7967048673387},
{ "name": "7.5GY4/10","L_star": 41.0886895909649, "a_star": -40.083053067588, "b_star": 52.2116122512395},
{ "name": "7.5GY4/8","L_star": 40.9446468688704, "a_star": -32.6762508693136, "b_star": 40.3579109142766},
{ "name": "7.5GY4/6","L_star": 40.8964703256786, "a_star": -25.0093165968638, "b_star": 29.378119519756},
{ "name": "7.5GY4/4","L_star": 40.848212057973, "a_star": -16.9634699414711, "b_star": 19.0491530269272},
{ "name": "7.5GY4/3","L_star": 40.8159943055026, "a_star": -12.6640408562686, "b_star": 13.844200559421},
{ "name": "7.5GY4/2","L_star": 40.7837399731063, "a_star": -8.2767462341928, "b_star": 8.8057256705962},
{ "name": "7.5GY4/1","L_star": 40.751448956775, "a_star": -3.99067723767996, "b_star": 4.11492710263373},
{ "name": "7.5GY3/10","L_star": 30.5517531439122, "a_star": -39.2829323588189, "b_star": 52.7544692936229},
{ "name": "7.5GY3/8","L_star": 30.6212764134076, "a_star": -31.5492680819034, "b_star": 37.8524963865327},
{ "name": "7.5GY3/6","L_star": 30.5325377232702, "a_star": -23.6321258022881, "b_star": 26.8560360597997},
{ "name": "7.5GY3/4","L_star": 30.4868373828619, "a_star": -16.0227755724457, "b_star": 17.23029738509},
{ "name": "7.5GY3/3","L_star": 30.4603381694081, "a_star": -11.8496053472883, "b_star": 12.3759210471913},
{ "name": "7.5GY3/2","L_star": 30.4313954189373, "a_star": -7.61094138123022, "b_star": 7.73446059696948},
{ "name": "7.5GY3/1","L_star": 30.4048328299485, "a_star": -3.53136730628445, "b_star": 3.50606450150456},
{ "name": "7.5GY2/8","L_star": 20.4204136190585, "a_star": -32.564448728932, "b_star": 40.2679400596051},
{ "name": "7.5GY2/6","L_star": 20.3929353918417, "a_star": -23.0761824043371, "b_star": 24.8207337262715},
{ "name": "7.5GY2/4","L_star": 20.341794025138, "a_star": -14.69797949616, "b_star": 15.0534214470441},
{ "name": "7.5GY2/3","L_star": 20.3181416956288, "a_star": -10.4180576340914, "b_star": 10.3411346693735},
{ "name": "7.5GY2/2","L_star": 20.2905083162782, "a_star": -6.29309177270285, "b_star": 6.02424747223944},
{ "name": "7.5GY2/1","L_star": 20.262832789887, "a_star": -2.66150659562958, "b_star": 2.43480469476918},
{ "name": "7.5GY1/4","L_star": 10.572116332781, "a_star": -18.3688437986739, "b_star": 17.420528046888},
{ "name": "7.5GY1/3","L_star": 10.5204338421202, "a_star": -12.1525991712649, "b_star": 10.8593067734514},
{ "name": "7.5GY1/2","L_star": 10.4833940818493, "a_star": -6.58090919364318, "b_star": 5.85351038546951},
{ "name": "7.5GY1/1","L_star": 10.4536875015812, "a_star": -2.27008421437365, "b_star": 2.17666289789635},
{ "name": "10GY9/18","L_star": 90.4089937184406, "a_star": -85.1281216276436, "b_star": 78.7539347196971},
{ "name": "10GY9/16","L_star": 90.4273709898596, "a_star": -76.3216030407446, "b_star": 70.0951736540027},
{ "name": "10GY9/14","L_star": 90.4273709898596, "a_star": -67.2559382589353, "b_star": 61.3819625000358},
{ "name": "10GY9/12","L_star": 90.4273709898596, "a_star": -58.3424148877852, "b_star": 52.9333116033587},
{ "name": "10GY9/10","L_star": 90.4227772670039, "a_star": -49.9868860948542, "b_star": 45.1998252884199},
{ "name": "10GY9/8","L_star": 90.4365572460867, "a_star": -40.777296411078, "b_star": 36.39521224712},
{ "name": "10GY9/6","L_star": 90.3906100971672, "a_star": -32.4733722764141, "b_star": 28.7111981069993},
{ "name": "10GY9/4","L_star": 90.3124087384961, "a_star": -21.7867585256674, "b_star": 19.2366932202943},
{ "name": "10GY9/3","L_star": 90.2801747147258, "a_star": -17.0190395402994, "b_star": 14.9203412887992},
{ "name": "10GY9/2","L_star": 90.2433118720672, "a_star": -12.1645077023059, "b_star": 10.4758662403421},
{ "name": "10GY9/1","L_star": 90.2018105741446, "a_star": -6.67417867366049, "b_star": 5.63428173426943},
{ "name": "10GY8/24","L_star": 81.0988295163557, "a_star": -106.556546877413, "b_star": 104.675368376033},
{ "name": "10GY8/22","L_star": 81.0215083066333, "a_star": -99.4769337866094, "b_star": 94.7088649336317},
{ "name": "10GY8/20","L_star": 80.9551347553501, "a_star": -91.4015152819201, "b_star": 85.3346076937305},
{ "name": "10GY8/18","L_star": 80.9219138739572, "a_star": -82.5953440583057, "b_star": 76.0168771719973},
{ "name": "10GY8/16","L_star": 80.9163748458348, "a_star": -73.8074172906124, "b_star": 66.9916607583477},
{ "name": "10GY8/14","L_star": 80.9385271608568, "a_star": -65.0227445824009, "b_star": 58.5856741282673},
{ "name": "10GY8/12","L_star": 80.8886702034112, "a_star": -56.2822631766622, "b_star": 50.5213470604451},
{ "name": "10GY8/10","L_star": 80.8332133369634, "a_star": -47.4030237598946, "b_star": 42.0761399665188},
{ "name": "10GY8/8","L_star": 80.7721373261093, "a_star": -38.1553404777687, "b_star": 33.6336632806389},
{ "name": "10GY8/6","L_star": 80.7165467019011, "a_star": -29.8302922653018, "b_star": 26.2203881941345},
{ "name": "10GY8/4","L_star": 80.6553231133634, "a_star": -20.5662628614288, "b_star": 17.8779354251029},
{ "name": "10GY8/3","L_star": 80.6218957120763, "a_star": -16.1153630745974, "b_star": 13.84275707307},
{ "name": "10GY8/2","L_star": 80.5884451656218, "a_star": -11.3762176284242, "b_star": 9.62877839332885},
{ "name": "10GY8/1","L_star": 80.5438083640871, "a_star": -6.1108341641104, "b_star": 5.10616752121371},
{ "name": "10GY7/22","L_star": 71.4394359366807, "a_star": -97.9123230319351, "b_star": 95.0234229447278},
{ "name": "10GY7/20","L_star": 71.3781465923443, "a_star": -90.3323848494081, "b_star": 84.4521652176723},
{ "name": "10GY7/18","L_star": 71.3372392143654, "a_star": -81.6135028448697, "b_star": 74.3899663125418},
{ "name": "10GY7/16","L_star": 71.3031204362961, "a_star": -73.1985173262652, "b_star": 66.0082489394182},
{ "name": "10GY7/14","L_star": 71.2689749695638, "a_star": -64.1480720173868, "b_star": 57.312292592415},
{ "name": "10GY7/12","L_star": 71.2142865798885, "a_star": -55.3679309430221, "b_star": 49.0257613366899},
{ "name": "10GY7/10","L_star": 71.1595295185127, "a_star": -46.3640279823985, "b_star": 40.7482410626522},
{ "name": "10GY7/8","L_star": 71.1047035694533, "a_star": -37.3643375493289, "b_star": 32.6623129434872},
{ "name": "10GY7/6","L_star": 71.0498085156385, "a_star": -28.4033969205819, "b_star": 24.5906883364816},
{ "name": "10GY7/4","L_star": 70.9879687066308, "a_star": -19.9421453374483, "b_star": 16.9478918896838},
{ "name": "10GY7/3","L_star": 70.9535752313128, "a_star": -15.4698469983511, "b_star": 12.988256225979},
{ "name": "10GY7/2","L_star": 70.9191545265718, "a_star": -10.7297618510679, "b_star": 8.88204828669195},
{ "name": "10GY7/1","L_star": 70.8847065384495, "a_star": -5.61363239303148, "b_star": 4.59518611563847},
{ "name": "10GY6/20","L_star": 61.63696187951, "a_star": -90.4143800467989, "b_star": 86.1264736158351},
{ "name": "10GY6/18","L_star": 61.6110569854016, "a_star": -82.6220251776382, "b_star": 75.2758254214403},
{ "name": "10GY6/16","L_star": 61.5678436930684, "a_star": -73.851804847207, "b_star": 65.7060924469489},
{ "name": "10GY6/14","L_star": 61.533238360038, "a_star": -65.2706429905244, "b_star": 57.4459292089041},
{ "name": "10GY6/12","L_star": 61.4118757581295, "a_star": -55.5664893321926, "b_star": 48.3775053365556},
{ "name": "10GY6/10","L_star": 61.3423545031676, "a_star": -46.5657354117682, "b_star": 40.2442615640408},
{ "name": "10GY6/8","L_star": 61.2901314255044, "a_star": -37.8096391159342, "b_star": 32.4246369410629},
{ "name": "10GY6/6","L_star": 61.229115169177, "a_star": -28.5707572392955, "b_star": 24.1938181970467},
{ "name": "10GY6/4","L_star": 61.1680023462115, "a_star": -19.6307683227927, "b_star": 16.2624118699171},
{ "name": "10GY6/3","L_star": 61.1417814803308, "a_star": -15.2095975876483, "b_star": 12.4024227411459},
{ "name": "10GY6/2","L_star": 61.1067925736237, "a_star": -10.5566382714956, "b_star": 8.48280643032402},
{ "name": "10GY6/1","L_star": 61.0717718840904, "a_star": -5.52567470463639, "b_star": 4.38811571382574},
{ "name": "10GY5/18","L_star": 51.6103295257169, "a_star": -82.8415342850696, "b_star": 78.5769340337893},
{ "name": "10GY5/16","L_star": 51.5761656343387, "a_star": -74.6155371080705, "b_star": 66.4093226706638},
{ "name": "10GY5/14","L_star": 51.5419671641083, "a_star": -65.5744823683392, "b_star": 56.742114622046},
{ "name": "10GY5/12","L_star": 51.4506015237771, "a_star": -56.815496259467, "b_star": 48.5230031046967},
{ "name": "10GY5/10","L_star": 51.301602258608, "a_star": -46.7033181130912, "b_star": 39.5695078162105},
{ "name": "10GY5/8","L_star": 51.2441187985769, "a_star": -37.4332662913133, "b_star": 31.3938062886208},
{ "name": "10GY5/6","L_star": 51.186536890986, "a_star": -28.0140753213172, "b_star": 23.2072706473469},
{ "name": "10GY5/4","L_star": 51.1288561132401, "a_star": -18.8217885343804, "b_star": 15.1702347269512},
{ "name": "10GY5/3","L_star": 51.0942000120299, "a_star": -14.2353122528563, "b_star": 11.3530857078568},
{ "name": "10GY5/2","L_star": 51.0595080720228, "a_star": -9.57830289120637, "b_star": 7.59669085637444},
{ "name": "10GY5/1","L_star": 51.0247802004202, "a_star": -4.83270214896458, "b_star": 3.83776915250744},
{ "name": "10GY4/16","L_star": 41.3746143416849, "a_star": -74.2855516373995, "b_star": 69.1999658258792},
{ "name": "10GY4/14","L_star": 41.3113218310235, "a_star": -67.3135377921281, "b_star": 57.728294315404},
{ "name": "10GY4/12","L_star": 41.2320090783868, "a_star": -58.2169925029639, "b_star": 48.1738828488078},
{ "name": "10GY4/10","L_star": 41.1046495273824, "a_star": -47.947871633246, "b_star": 39.2423366558803},
{ "name": "10GY4/8","L_star": 40.9927420326694, "a_star": -38.1530985908319, "b_star": 31.0347016292503},
{ "name": "10GY4/6","L_star": 40.9285970802895, "a_star": -28.783233205769, "b_star": 23.1800547530613},
{ "name": "10GY4/4","L_star": 40.8643072491098, "a_star": -18.8814065842794, "b_star": 15.0001071145827},
{ "name": "10GY4/3","L_star": 40.8321077477464, "a_star": -13.9515862512145, "b_star": 11.0324591827922},
{ "name": "10GY4/2","L_star": 40.7998717182846, "a_star": -9.08801179880367, "b_star": 7.17883260046656},
{ "name": "10GY4/1","L_star": 40.7675990569515, "a_star": -4.39343708274484, "b_star": 3.48888427362553},
{ "name": "10GY3/14","L_star": 31.059984663291, "a_star": -64.2620598273027, "b_star": 54.3493667036437},
{ "name": "10GY3/12","L_star": 30.8428433732511, "a_star": -55.1531534603834, "b_star": 43.0566546557171},
{ "name": "10GY3/10","L_star": 30.7335136949528, "a_star": -46.0026884577869, "b_star": 34.9609610445893},
{ "name": "10GY3/8","L_star": 30.6188825096081, "a_star": -36.3112956921145, "b_star": 27.6105498818177},
{ "name": "10GY3/6","L_star": 30.5661543058963, "a_star": -27.1537653458889, "b_star": 20.4364125308645},
{ "name": "10GY3/4","L_star": 30.5084961081804, "a_star": -17.9826115408366, "b_star": 13.2976548088617},
{ "name": "10GY3/3","L_star": 30.4747960319158, "a_star": -13.1664195724622, "b_star": 9.66407277976055},
{ "name": "10GY3/2","L_star": 30.4410470110501, "a_star": -8.38604360552114, "b_star": 6.11747578051736},
{ "name": "10GY3/1","L_star": 30.4072488674824, "a_star": -3.91606715484577, "b_star": 2.84370121002399},
{ "name": "10GY2/12","L_star": 21.079173864162, "a_star": -58.3258344340661, "b_star": 43.0766303401815},
{ "name": "10GY2/10","L_star": 20.6813472072113, "a_star": -47.2771505858963, "b_star": 32.6083712573597},
{ "name": "10GY2/8","L_star": 20.4986952432828, "a_star": -36.0407071551003, "b_star": 24.4105896944305},
{ "name": "10GY2/6","L_star": 20.4204136190585, "a_star": -25.7500062875926, "b_star": 17.5186370550888},
{ "name": "10GY2/4","L_star": 20.3654156073658, "a_star": -16.486708639852, "b_star": 11.179492662545},
{ "name": "10GY2/3","L_star": 20.3339133366061, "a_star": -11.6189413307964, "b_star": 7.86021274591769},
{ "name": "10GY2/2","L_star": 20.29840786122, "a_star": -6.94716331780429, "b_star": 4.69955710697386},
{ "name": "10GY2/1","L_star": 20.2707443932339, "a_star": -2.89413982436595, "b_star": 1.97646339981469},
{ "name": "10GY1/6","L_star": 10.7259698972709, "a_star": -33.7021618290475, "b_star": 19.7617375772373},
{ "name": "10GY1/4","L_star": 10.5647454263466, "a_star": -19.9354003291562, "b_star": 11.6906338270598},
{ "name": "10GY1/3","L_star": 10.5352208169387, "a_star": -13.3222595118359, "b_star": 7.83688607399946},
{ "name": "10GY1/2","L_star": 10.4982224131009, "a_star": -7.29584449474874, "b_star": 4.31074483077555},
{ "name": "10GY1/1","L_star": 10.4685491268993, "a_star": -2.72577616112447, "b_star": 1.7654056907481},
{ "name": "2.5G9/16","L_star": 90.3722201205514, "a_star": -84.7341984134351, "b_star": 46.0837293396143},
{ "name": "2.5G9/14","L_star": 90.3906100971672, "a_star": -74.8384800012139, "b_star": 41.0230703217525},
{ "name": "2.5G9/12","L_star": 90.4089937184406, "a_star": -64.9189292771934, "b_star": 35.9985497702806},
{ "name": "2.5G9/10","L_star": 90.4227772670039, "a_star": -55.4451760032176, "b_star": 31.1425624994895},
{ "name": "2.5G9/8","L_star": 90.431964316192, "a_star": -45.4435881399819, "b_star": 25.81946495371},
{ "name": "2.5G9/6","L_star": 90.41358863138, "a_star": -35.6160420763969, "b_star": 20.4758091841285},
{ "name": "2.5G9/4","L_star": 90.3308194079677, "a_star": -23.8297305088734, "b_star": 13.7537842529921},
{ "name": "2.5G9/3","L_star": 90.2939916902734, "a_star": -18.5280987341762, "b_star": 10.7131365067126},
{ "name": "2.5G9/2","L_star": 90.2525299808178, "a_star": -13.0988170348744, "b_star": 7.58898381152457},
{ "name": "2.5G9/1","L_star": 90.2064234313182, "a_star": -7.13148349671838, "b_star": 4.11635168377373},
{ "name": "2.5G8/24","L_star": 81.1153823345278, "a_star": -122.426436657682, "b_star": 59.3698431175632},
{ "name": "2.5G8/22","L_star": 81.1098653554593, "a_star": -112.60542752673, "b_star": 56.0264891060454},
{ "name": "2.5G8/20","L_star": 81.0933106559644, "a_star": -103.032318820449, "b_star": 52.3274226395784},
{ "name": "2.5G8/18","L_star": 81.0767503092922, "a_star": -93.0824374200354, "b_star": 48.0969287979524},
{ "name": "2.5G8/16","L_star": 81.0436126551356, "a_star": -82.5914268456547, "b_star": 43.4905621877373},
{ "name": "2.5G8/14","L_star": 80.9883328866058, "a_star": -72.2027981489579, "b_star": 38.525782178348},
{ "name": "2.5G8/12","L_star": 80.9329900312874, "a_star": -63.0059650658084, "b_star": 34.0971131110208},
{ "name": "2.5G8/10","L_star": 80.872039809984, "a_star": -52.3975275106485, "b_star": 28.8873945419614},
{ "name": "2.5G8/8","L_star": 80.8054610674918, "a_star": -42.5126173502219, "b_star": 23.6218899247372},
{ "name": "2.5G8/6","L_star": 80.7443499997881, "a_star": -33.1772270569614, "b_star": 18.5598976595271},
{ "name": "2.5G8/4","L_star": 80.6664604440045, "a_star": -22.3359437428917, "b_star": 12.5974684857385},
{ "name": "2.5G8/3","L_star": 80.6330407488799, "a_star": -17.3689111097658, "b_star": 9.80977363512487},
{ "name": "2.5G8/2","L_star": 80.5940218657014, "a_star": -12.279669912155, "b_star": 6.94364232546307},
{ "name": "2.5G8/1","L_star": 80.5493902216599, "a_star": -6.69499563552367, "b_star": 3.75818512637349},
{ "name": "2.5G7/26","L_star": 71.987214671463, "a_star": -135.614475381361, "b_star": 58.6246884118935},
{ "name": "2.5G7/24","L_star": 71.8862886656199, "a_star": -123.584833280417, "b_star": 55.5851989087955},
{ "name": "2.5G7/22","L_star": 71.7716249201182, "a_star": -113.30737580074, "b_star": 52.6517919258909},
{ "name": "2.5G7/20","L_star": 71.6363416833176, "a_star": -102.590248440381, "b_star": 49.0330897648636},
{ "name": "2.5G7/18","L_star": 71.5006394815849, "a_star": -91.8010779094531, "b_star": 45.1200871667717},
{ "name": "2.5G7/16","L_star": 71.3917738768124, "a_star": -82.0351518729594, "b_star": 40.9816047125282},
{ "name": "2.5G7/14","L_star": 71.3167711472427, "a_star": -71.862748929638, "b_star": 36.6683895814378},
{ "name": "2.5G7/12","L_star": 71.2621426689824, "a_star": -62.0927313056079, "b_star": 32.2594659980024},
{ "name": "2.5G7/10","L_star": 71.2006037609348, "a_star": -51.6498357612309, "b_star": 27.4670118817729},
{ "name": "2.5G7/8","L_star": 71.1389778720261, "a_star": -41.3968904572826, "b_star": 22.3105389071508},
{ "name": "2.5G7/6","L_star": 71.0704022728571, "a_star": -31.270975897379, "b_star": 16.9773601302227},
{ "name": "2.5G7/4","L_star": 71.0017184845736, "a_star": -21.7792467749421, "b_star": 11.9038199466931},
{ "name": "2.5G7/3","L_star": 70.9673358859548, "a_star": -16.8620322253029, "b_star": 9.23923642078059},
{ "name": "2.5G7/2","L_star": 70.9329260794485, "a_star": -11.6696394226817, "b_star": 6.42660910365362},
{ "name": "2.5G7/1","L_star": 70.8847065384495, "a_star": -6.06447151062917, "b_star": 3.36796381541102},
{ "name": "2.5G6/28","L_star": 62.5835353678145, "a_star": -151.819901525017, "b_star": 54.9408785047695},
{ "name": "2.5G6/26","L_star": 62.5076336444413, "a_star": -140.038512717827, "b_star": 53.3189033278309},
{ "name": "2.5G6/24","L_star": 62.3977380299771, "a_star": -128.471939081696, "b_star": 51.2440752878035},
{ "name": "2.5G6/22","L_star": 62.2365649912483, "a_star": -116.45963270081, "b_star": 48.7393093599343},
{ "name": "2.5G6/20","L_star": 62.0576503011181, "a_star": -105.40718221702, "b_star": 45.8463310683979},
{ "name": "2.5G6/18","L_star": 61.8693319572042, "a_star": -94.247106593594, "b_star": 42.4733888878345},
{ "name": "2.5G6/16","L_star": 61.6800983268115, "a_star": -82.907989894027, "b_star": 38.7281195391375},
{ "name": "2.5G6/14","L_star": 61.5245821984879, "a_star": -72.4615963268702, "b_star": 34.7216822108474},
{ "name": "2.5G6/12","L_star": 61.437914063409, "a_star": -61.4088882949467, "b_star": 30.286546649602},
{ "name": "2.5G6/10","L_star": 61.3858199245362, "a_star": -51.6885935396398, "b_star": 26.0673070198259},
{ "name": "2.5G6/8","L_star": 61.3162517828243, "a_star": -41.5236178388457, "b_star": 21.3884609522729},
{ "name": "2.5G6/6","L_star": 61.2552767955579, "a_star": -31.5635360673913, "b_star": 16.5712975980119},
{ "name": "2.5G6/4","L_star": 61.1854730287976, "a_star": -21.6825578009645, "b_star": 11.4342008183576},
{ "name": "2.5G6/3","L_star": 61.1505237493858, "a_star": -16.6120585566889, "b_star": 8.86535449215044},
{ "name": "2.5G6/2","L_star": 61.1155427771229, "a_star": -11.3895763750254, "b_star": 6.1742960160289},
{ "name": "2.5G6/1","L_star": 61.0717718840904, "a_star": -5.86263779780893, "b_star": 3.28844602273612},
{ "name": "2.5G5/28","L_star": 52.6085360835917, "a_star": -152.591231636652, "b_star": 47.4074176912205},
{ "name": "2.5G5/26","L_star": 52.5642939071549, "a_star": -140.690502247148, "b_star": 46.5343157469354},
{ "name": "2.5G5/24","L_star": 52.4645397014335, "a_star": -129.59258204139, "b_star": 45.3258534444514},
{ "name": "2.5G5/22","L_star": 52.353359675186, "a_star": -119.32459299011, "b_star": 43.8401037070324},
{ "name": "2.5G5/20","L_star": 52.2082827205336, "a_star": -108.566919206, "b_star": 41.7503145299079},
{ "name": "2.5G5/18","L_star": 52.0063823305922, "a_star": -97.6572978585906, "b_star": 39.2639887860422},
{ "name": "2.5G5/16","L_star": 51.7919565354378, "a_star": -85.7463058682442, "b_star": 36.2204239122571},
{ "name": "2.5G5/14","L_star": 51.5875574359089, "a_star": -74.1607552936389, "b_star": 32.6290249989806},
{ "name": "2.5G5/12","L_star": 51.4277213993881, "a_star": -63.5028235173519, "b_star": 28.6910648625365},
{ "name": "2.5G5/10","L_star": 51.3360452545173, "a_star": -51.9175171987517, "b_star": 24.3573838744923},
{ "name": "2.5G5/8","L_star": 51.2786206613769, "a_star": -41.6048695660599, "b_star": 20.2135263926303},
{ "name": "2.5G5/6","L_star": 51.2095814947108, "a_star": -30.8032192171882, "b_star": 15.4282157835023},
{ "name": "2.5G5/4","L_star": 51.1404001987838, "a_star": -20.5750260526639, "b_star": 10.4972784698247},
{ "name": "2.5G5/3","L_star": 51.1057560232836, "a_star": -15.531693458875, "b_star": 7.99954011854787},
{ "name": "2.5G5/2","L_star": 51.0710760398342, "a_star": -10.4142435738079, "b_star": 5.46256288956006},
{ "name": "2.5G5/1","L_star": 51.0247802004202, "a_star": -5.27882109313516, "b_star": 2.82741837816483},
{ "name": "2.5G4/26","L_star": 42.0463629294143, "a_star": -145.215539546652, "b_star": 38.424617449427},
{ "name": "2.5G4/24","L_star": 42.0463629294143, "a_star": -130.86629063767, "b_star": 37.919512379293},
{ "name": "2.5G4/22","L_star": 42, "a_star": -117.687630382397, "b_star": 37.1070053373619},
{ "name": "2.5G4/20","L_star": 41.9070511227818, "a_star": -106.884707655049, "b_star": 35.9177818115643},
{ "name": "2.5G4/18","L_star": 41.7982322467941, "a_star": -96.9044829448083, "b_star": 34.4696680394824},
{ "name": "2.5G4/16","L_star": 41.6107265259976, "a_star": -86.0996942559385, "b_star": 32.3961369531947},
{ "name": "2.5G4/14","L_star": 41.4377675179741, "a_star": -75.7273652044422, "b_star": 29.97122329285},
{ "name": "2.5G4/12","L_star": 41.2637605462474, "a_star": -65.364671833971, "b_star": 27.0175759375683},
{ "name": "2.5G4/10","L_star": 41.1046495273824, "a_star": -53.7896218599415, "b_star": 23.1069981516462},
{ "name": "2.5G4/8","L_star": 41.0087557289414, "a_star": -42.5436908852728, "b_star": 19.010555782139},
{ "name": "2.5G4/6","L_star": 40.9446468688704, "a_star": -31.8944081990585, "b_star": 14.8006954700346},
{ "name": "2.5G4/4","L_star": 40.8803933340554, "a_star": -20.5607611581123, "b_star": 9.88285719532814},
{ "name": "2.5G4/3","L_star": 40.848212057973, "a_star": -15.1418681486118, "b_star": 7.39863579610036},
{ "name": "2.5G4/2","L_star": 40.7998717182846, "a_star": -9.87823616259387, "b_star": 4.91527439501853},
{ "name": "2.5G4/1","L_star": 40.7675990569515, "a_star": -4.84122501026543, "b_star": 2.45031955448467},
{ "name": "2.5G3/22","L_star": 31.377352210005, "a_star": -126.837994567816, "b_star": 29.6978092925841},
{ "name": "2.5G3/20","L_star": 31.375034108807, "a_star": -108.289470672256, "b_star": 29.2119962604226},
{ "name": "2.5G3/18","L_star": 31.3518406095019, "a_star": -93.7956076438207, "b_star": 28.3791014673804},
{ "name": "2.5G3/16","L_star": 31.2448567151527, "a_star": -82.222125248829, "b_star": 27.0995462739697},
{ "name": "2.5G3/14","L_star": 31.102235158524, "a_star": -71.4912311613508, "b_star": 25.376913371083},
{ "name": "2.5G3/12","L_star": 30.9115071622873, "a_star": -60.8673480969715, "b_star": 22.8652702232308},
{ "name": "2.5G3/10","L_star": 30.7573245278772, "a_star": -50.3667283447092, "b_star": 19.9308645752928},
{ "name": "2.5G3/8","L_star": 30.6380268606166, "a_star": -39.3781240373144, "b_star": 16.4577142708743},
{ "name": "2.5G3/6","L_star": 30.5853420100737, "a_star": -29.8339241178245, "b_star": 13.0241873238253},
{ "name": "2.5G3/4","L_star": 30.5205200218775, "a_star": -19.4998410870509, "b_star": 8.94576560166922},
{ "name": "2.5G3/3","L_star": 30.4844296117184, "a_star": -14.1684742045192, "b_star": 6.68294588592453},
{ "name": "2.5G3/2","L_star": 30.4458713030924, "a_star": -8.95922511950725, "b_star": 4.37086260516739},
{ "name": "2.5G3/1","L_star": 30.4096646534768, "a_star": -4.15191186251579, "b_star": 2.10844949335576},
{ "name": "2.5G2/16","L_star": 20.9614855134017, "a_star": -102.179722248352, "b_star": 21.8362468359743},
{ "name": "2.5G2/14","L_star": 20.9652936203227, "a_star": -80.3145348029614, "b_star": 21.334157303619},
{ "name": "2.5G2/12","L_star": 20.9004487867803, "a_star": -64.5083093637411, "b_star": 20.0310108616196},
{ "name": "2.5G2/10","L_star": 20.6929442194311, "a_star": -50.5206885401553, "b_star": 17.6015315924933},
{ "name": "2.5G2/8","L_star": 20.5221143676061, "a_star": -37.6435887683584, "b_star": 14.1899993266707},
{ "name": "2.5G2/6","L_star": 20.4360968651002, "a_star": -27.6997353966866, "b_star": 11.0544012527269},
{ "name": "2.5G2/4","L_star": 20.3732826526318, "a_star": -17.4974495271238, "b_star": 7.44196599955157},
{ "name": "2.5G2/3","L_star": 20.3378541081486, "a_star": -12.3478532196789, "b_star": 5.41409826711348},
{ "name": "2.5G2/2","L_star": 20.3023563446473, "a_star": -7.48990791367063, "b_star": 3.38857995292061},
{ "name": "2.5G2/1","L_star": 20.2707443932339, "a_star": -3.24075311444524, "b_star": 1.53058615461374},
{ "name": "2.5G1/8","L_star": 10.8996610101788, "a_star": -63.6427487212422, "b_star": 14.749945959293},
{ "name": "2.5G1/6","L_star": 10.7768627811419, "a_star": -37.4738113379635, "b_star": 12.159641746431},
{ "name": "2.5G1/4","L_star": 10.5794831522138, "a_star": -20.509823023416, "b_star": 7.72374654247195},
{ "name": "2.5G1/3","L_star": 10.5352208169387, "a_star": -13.5669677146945, "b_star": 5.41779959220271},
{ "name": "2.5G1/2","L_star": 10.4982224131009, "a_star": -7.65114834245685, "b_star": 3.23764868439836},
{ "name": "2.5G1/1","L_star": 10.4685491268993, "a_star": -3.02541802688625, "b_star": 1.3645471475964},
{ "name": "5G9/12","L_star": 90.3814159036203, "a_star": -68.0438569799495, "b_star": 22.1925499569075},
{ "name": "5G9/8","L_star": 90.4227772670039, "a_star": -47.4591700594738, "b_star": 16.8521954021161},
{ "name": "5G9/6","L_star": 90.4273709898596, "a_star": -36.8226834712758, "b_star": 13.6693784679093},
{ "name": "5G9/4","L_star": 90.3400223523973, "a_star": -24.6143070189327, "b_star": 9.55080709348286},
{ "name": "5G9/3","L_star": 90.2985965506391, "a_star": -19.0143257861879, "b_star": 7.57942648788288},
{ "name": "5G9/2","L_star": 90.2571384354549, "a_star": -13.3144759850845, "b_star": 5.46008060294023},
{ "name": "5G9/1","L_star": 90.2110358878267, "a_star": -7.085577997403, "b_star": 3.02908395285406},
{ "name": "5G9/0","L_star": 90.4043984086353, "a_star": -57.3426729258896, "b_star": 19.6063599673272},
{ "name": "5G8/22","L_star": 81.0988295163557, "a_star": -116.601119544105, "b_star": 29.227615335967},
{ "name": "5G8/20","L_star": 81.0933106559644, "a_star": -106.781592013493, "b_star": 28.1684437759211},
{ "name": "5G8/18","L_star": 81.0767503092922, "a_star": -95.9737456518767, "b_star": 26.7928601423194},
{ "name": "5G8/16","L_star": 81.0546610542355, "a_star": -85.5095642802885, "b_star": 25.1767837718782},
{ "name": "5G8/14","L_star": 81.0104523543632, "a_star": -75.3332338019598, "b_star": 23.2217040251538},
{ "name": "5G8/12","L_star": 80.9551347553501, "a_star": -65.193258430933, "b_star": 20.9339456358631},
{ "name": "5G8/10","L_star": 80.8886702034112, "a_star": -54.4552645250183, "b_star": 18.3667172744084},
{ "name": "5G8/8","L_star": 80.8221143382603, "a_star": -44.2365608456893, "b_star": 15.5519465295664},
{ "name": "5G8/6","L_star": 80.7554668456199, "a_star": -34.1380376497489, "b_star": 12.4010554358355},
{ "name": "5G8/4","L_star": 80.6775952088837, "a_star": -22.9230464627166, "b_star": 8.66299837942868},
{ "name": "5G8/3","L_star": 80.6386123033635, "a_star": -17.7984514623931, "b_star": 6.84225234664444},
{ "name": "5G8/2","L_star": 80.5995979219319, "a_star": -12.5752691655184, "b_star": 4.93257552660837},
{ "name": "5G8/1","L_star": 80.5549714338939, "a_star": -6.74762825591141, "b_star": 2.69228044175578},
{ "name": "5G7/26","L_star": 71.9266868734077, "a_star": -134.346573447235, "b_star": 29.4869525877902},
{ "name": "5G7/24","L_star": 71.8997588610436, "a_star": -125.816601620716, "b_star": 28.7060036918328},
{ "name": "5G7/22","L_star": 71.8391104341108, "a_star": -116.524945846714, "b_star": 27.84528700853},
{ "name": "5G7/20","L_star": 71.7581153578115, "a_star": -106.670403736711, "b_star": 26.7837057320394},
{ "name": "5G7/18","L_star": 71.6363416833176, "a_star": -95.6805752221681, "b_star": 25.3756096919783},
{ "name": "5G7/16","L_star": 71.5074345922344, "a_star": -85.6309736927803, "b_star": 23.7583420051092},
{ "name": "5G7/14","L_star": 71.3645150559933, "a_star": -74.972493552045, "b_star": 21.8868377253751},
{ "name": "5G7/12","L_star": 71.2758062005091, "a_star": -63.7705677648622, "b_star": 19.6633677637757},
{ "name": "5G7/10","L_star": 71.2142865798885, "a_star": -53.1964381246949, "b_star": 17.2263280380391},
{ "name": "5G7/8","L_star": 71.1458294979289, "a_star": -42.3901421575991, "b_star": 14.4388648988139},
{ "name": "5G7/6","L_star": 71.07726469425, "a_star": -32.2121871604889, "b_star": 11.3727838463368},
{ "name": "5G7/4","L_star": 71.0085917440811, "a_star": -22.3900865105341, "b_star": 8.17324716063401},
{ "name": "5G7/3","L_star": 70.9742145805881, "a_star": -17.2596137172527, "b_star": 6.42610368825061},
{ "name": "5G7/2","L_star": 70.9329260794485, "a_star": -11.879216377385, "b_star": 4.52204287559241},
{ "name": "5G7/1","L_star": 70.8915983213402, "a_star": -6.15218939313855, "b_star": 2.39496283112168},
{ "name": "5G6/28","L_star": 62.3469129019359, "a_star": -152.508936755974, "b_star": 27.9443709806503},
{ "name": "5G6/26","L_star": 62.3384356336483, "a_star": -140.719704847387, "b_star": 27.4314615426116},
{ "name": "5G6/24","L_star": 62.3045081995074, "a_star": -129.255613541918, "b_star": 26.6501236386036},
{ "name": "5G6/22","L_star": 62.2365649912483, "a_star": -117.87449543959, "b_star": 25.8472363616782},
{ "name": "5G6/20","L_star": 62.1259049213583, "a_star": -106.856457847864, "b_star": 24.8189024383961},
{ "name": "5G6/18","L_star": 61.9892761062234, "a_star": -95.9703410499127, "b_star": 23.6318278813504},
{ "name": "5G6/16","L_star": 61.8178140248937, "a_star": -85.1428314493244, "b_star": 22.1904758435689},
{ "name": "5G6/14","L_star": 61.63696187951, "a_star": -74.4715951209784, "b_star": 20.456862156098},
{ "name": "5G6/12","L_star": 61.4552632101895, "a_star": -64.0132713225854, "b_star": 18.5443278863133},
{ "name": "5G6/10","L_star": 61.3945071518122, "a_star": -52.718286454609, "b_star": 16.2380122580724},
{ "name": "5G6/8","L_star": 61.333655554213, "a_star": -42.627398151826, "b_star": 13.8736170413261},
{ "name": "5G6/6","L_star": 61.2639934013856, "a_star": -32.5387165008256, "b_star": 11.1453754343454},
{ "name": "5G6/4","L_star": 61.1942054050492, "a_star": -22.5252743992936, "b_star": 7.96369004476023},
{ "name": "5G6/3","L_star": 61.1592640376388, "a_star": -17.1668484049242, "b_star": 6.186582152841},
{ "name": "5G6/2","L_star": 61.1155427771229, "a_star": -11.5541152340687, "b_star": 4.25642954737122},
{ "name": "5G6/1","L_star": 61.0717718840904, "a_star": -5.82610101664272, "b_star": 2.21350281784281},
{ "name": "5G5/28","L_star": 52.0625859779343, "a_star": -154.322063242005, "b_star": 24.0651270166671},
{ "name": "5G5/26","L_star": 52.1186969563382, "a_star": -141.641378250285, "b_star": 23.8373433813939},
{ "name": "5G5/24","L_star": 52.1411154858345, "a_star": -130.596580453431, "b_star": 23.5698702890377},
{ "name": "5G5/22","L_star": 52.1299080653222, "a_star": -118.942982946638, "b_star": 22.9896086076447},
{ "name": "5G5/20","L_star": 52.0738155749007, "a_star": -108.925523106563, "b_star": 22.4713007355944},
{ "name": "5G5/18","L_star": 52.0063823305922, "a_star": -99.4051532592668, "b_star": 21.765375842512},
{ "name": "5G5/16","L_star": 51.8824062178063, "a_star": -88.1208606149559, "b_star": 20.6285775683598},
{ "name": "5G5/14","L_star": 51.7012648465092, "a_star": -76.3555522425693, "b_star": 19.2004948728726},
{ "name": "5G5/12","L_star": 51.5305599747064, "a_star": -65.8853652125807, "b_star": 17.668208805143},
{ "name": "5G5/10","L_star": 51.3475184265326, "a_star": -53.3536121810105, "b_star": 15.4383012442602},
{ "name": "5G5/8","L_star": 51.2901134222191, "a_star": -42.8578895759976, "b_star": 13.2259375942425},
{ "name": "5G5/6","L_star": 51.2210978728482, "a_star": -31.6975045058524, "b_star": 10.3855628564748},
{ "name": "5G5/4","L_star": 51.1519403159239, "a_star": -21.2832225251109, "b_star": 7.31992477259136},
{ "name": "5G5/3","L_star": 51.1057560232836, "a_star": -16.0163744058859, "b_star": 5.62455650480722},
{ "name": "5G5/2","L_star": 51.0710760398342, "a_star": -10.6534302129107, "b_star": 3.79715550192361},
{ "name": "5G5/1","L_star": 51.0247802004202, "a_star": -5.30410420771033, "b_star": 1.93683161873286},
{ "name": "5G4/26","L_star": 41.3904156683835, "a_star": -145.872309304642, "b_star": 19.5096515573177},
{ "name": "5G4/24","L_star": 41.4850414220451, "a_star": -130.521342290092, "b_star": 19.50745468054},
{ "name": "5G4/22","L_star": 41.5479525982931, "a_star": -116.602575039262, "b_star": 19.4102114824272},
{ "name": "5G4/20","L_star": 41.5950458702078, "a_star": -106.978088785664, "b_star": 19.2170395492642},
{ "name": "5G4/18","L_star": 41.5793566714282, "a_star": -98.2275733352971, "b_star": 18.8400386032449},
{ "name": "5G4/16","L_star": 41.5479525982931, "a_star": -87.8368565398951, "b_star": 18.19810175493},
{ "name": "5G4/14","L_star": 41.4535341307665, "a_star": -77.3142062899782, "b_star": 17.3315966639459},
{ "name": "5G4/12","L_star": 41.3271580633385, "a_star": -67.3061341705465, "b_star": 16.1420914390704},
{ "name": "5G4/10","L_star": 41.1524758893756, "a_star": -54.7352104153631, "b_star": 14.3373155575287},
{ "name": "5G4/8","L_star": 41.0247604338049, "a_star": -43.3048834176445, "b_star": 12.3532462310919},
{ "name": "5G4/6","L_star": 40.9606876153157, "a_star": -32.3441273645756, "b_star": 9.99474540053736},
{ "name": "5G4/4","L_star": 40.8803933340554, "a_star": -21.2657219429115, "b_star": 6.97714285111392},
{ "name": "5G4/3","L_star": 40.848212057973, "a_star": -15.6439372264884, "b_star": 5.28615414161411},
{ "name": "5G4/2","L_star": 40.7998717182846, "a_star": -10.1450141104016, "b_star": 3.51951799300042},
{ "name": "5G4/1","L_star": 40.7675990569515, "a_star": -4.88635054128045, "b_star": 1.74316457006373},
{ "name": "5G3/22","L_star": 30.7882424015286, "a_star": -123.891861510235, "b_star": 15.170716638563},
{ "name": "5G3/20","L_star": 30.8902192269051, "a_star": -105.896059402381, "b_star": 15.2674643617515},
{ "name": "5G3/18","L_star": 30.9728973665266, "a_star": -92.9065259522876, "b_star": 15.2472113479627},
{ "name": "5G3/16","L_star": 30.9893981096696, "a_star": -82.5538554578962, "b_star": 14.9465461318644},
{ "name": "5G3/14","L_star": 30.9823277818305, "a_star": -72.0543814805796, "b_star": 14.4777358757175},
{ "name": "5G3/12","L_star": 30.9020482433572, "a_star": -61.4255764599972, "b_star": 13.5751659682723},
{ "name": "5G3/10","L_star": 30.7858655540568, "a_star": -51.0741321496331, "b_star": 12.3240616044148},
{ "name": "5G3/8","L_star": 30.6619352191697, "a_star": -40.0619738722736, "b_star": 10.6391858259103},
{ "name": "5G3/6","L_star": 30.592533325719, "a_star": -30.546538754767, "b_star": 8.79344079597071},
{ "name": "5G3/4","L_star": 30.5277313876761, "a_star": -20.2257609195243, "b_star": 6.27495344486662},
{ "name": "5G3/3","L_star": 30.4892449046116, "a_star": -14.6881840805249, "b_star": 4.69017919018722},
{ "name": "5G3/2","L_star": 30.4482830733378, "a_star": -9.263747317726, "b_star": 3.04365710722282},
{ "name": "5G3/1","L_star": 30.412080187997, "a_star": -4.24020682200726, "b_star": 1.44862539028681},
{ "name": "5G2/16","L_star": 20.4635101039511, "a_star": -100.408083031935, "b_star": 11.1215249411863},
{ "name": "5G2/14","L_star": 20.6038457145266, "a_star": -79.538055164965, "b_star": 11.4167646476784},
{ "name": "5G2/12","L_star": 20.6929442194311, "a_star": -64.1098486757365, "b_star": 11.2423617330026},
{ "name": "5G2/10","L_star": 20.6774799064755, "a_star": -50.5588399565177, "b_star": 10.4244894121849},
{ "name": "5G2/8","L_star": 20.5416073870883, "a_star": -38.1644511993118, "b_star": 8.95910581966942},
{ "name": "5G2/6","L_star": 20.4400155677402, "a_star": -28.1272718210085, "b_star": 7.35799671365966},
{ "name": "5G2/4","L_star": 20.3772148994357, "a_star": -17.9018505870993, "b_star": 5.20130440442955},
{ "name": "5G2/3","L_star": 20.341794025138, "a_star": -12.6734872717339, "b_star": 3.83035368013387},
{ "name": "5G2/2","L_star": 20.3063039693353, "a_star": -7.72499305599547, "b_star": 2.42293433726011},
{ "name": "5G2/1","L_star": 20.27469890094, "a_star": -3.36608482047104, "b_star": 1.09723176434317},
{ "name": "5G1/8","L_star": 10.6309369261204, "a_star": -61.6819915637799, "b_star": 7.85556386258245},
{ "name": "5G1/6","L_star": 10.7478048497097, "a_star": -38.528867566961, "b_star": 7.36362941238813},
{ "name": "5G1/4","L_star": 10.6015591450467, "a_star": -20.8792984582197, "b_star": 5.22121455849192},
{ "name": "5G1/3","L_star": 10.5426081282382, "a_star": -13.8794357787042, "b_star": 3.81035241564533},
{ "name": "5G1/2","L_star": 10.5056303593669, "a_star": -7.95511272416, "b_star": 2.37675408036193},
{ "name": "5G1/1","L_star": 10.4685491268993, "a_star": -3.20671575966806, "b_star": 1.02033530070164},
{ "name": "7.5G9/12","L_star": 90.3676216327313, "a_star": -68.4316786341994, "b_star": 15.4682359338182},
{ "name": "7.5G9/10","L_star": 90.3906100971672, "a_star": -57.4522461769455, "b_star": 13.7934741595364},
{ "name": "7.5G9/8","L_star": 90.4089937184406, "a_star": -47.6387028937291, "b_star": 11.9797507384437},
{ "name": "7.5G9/6","L_star": 90.431964316192, "a_star": -36.972656348384, "b_star": 9.71105146576863},
{ "name": "7.5G9/4","L_star": 90.3446232273371, "a_star": -24.8193888652284, "b_star": 6.82218102994521},
{ "name": "7.5G9/3","L_star": 90.3032010120739, "a_star": -19.1309167451675, "b_star": 5.3912883073979},
{ "name": "7.5G9/2","L_star": 90.2571384354549, "a_star": -13.3288450517587, "b_star": 3.84138086065144},
{ "name": "7.5G9/1","L_star": 90.2110358878267, "a_star": -7.07415238439729, "b_star": 2.07914346442206},
{ "name": "7.5G8/20","L_star": 81.0933106559644, "a_star": -106.010035046634, "b_star": 19.102688264721},
{ "name": "7.5G8/18","L_star": 81.0767503092922, "a_star": -96.0820056028329, "b_star": 18.3264767489502},
{ "name": "7.5G8/16","L_star": 81.0546610542355, "a_star": -85.6712722232221, "b_star": 17.3645669956962},
{ "name": "7.5G8/14","L_star": 81.0159806454827, "a_star": -75.2559121086902, "b_star": 15.9461993786351},
{ "name": "7.5G8/12","L_star": 80.9551347553501, "a_star": -65.2638611000936, "b_star": 14.4435479012874},
{ "name": "7.5G8/10","L_star": 80.8886702034112, "a_star": -54.4430134382626, "b_star": 12.6817939604066},
{ "name": "7.5G8/8","L_star": 80.8221143382603, "a_star": -44.0044840697926, "b_star": 10.6900289567224},
{ "name": "7.5G8/6","L_star": 80.7554668456199, "a_star": -34.0604860542866, "b_star": 8.72096938188438},
{ "name": "7.5G8/4","L_star": 80.6775952088837, "a_star": -23.0356470540192, "b_star": 6.17021333749088},
{ "name": "7.5G8/3","L_star": 80.6386123033635, "a_star": -17.8201568684483, "b_star": 4.88562538004818},
{ "name": "7.5G8/2","L_star": 80.5995979219319, "a_star": -12.5479871618982, "b_star": 3.47789170817794},
{ "name": "7.5G8/1","L_star": 80.5549714338939, "a_star": -6.68589872536801, "b_star": 1.93879940885573},
{ "name": "7.5G7/26","L_star": 71.9266868734077, "a_star": -132.898896590777, "b_star": 19.2468147353381},
{ "name": "7.5G7/24","L_star": 71.8930242794311, "a_star": -124.961617654708, "b_star": 18.839222698325},
{ "name": "7.5G7/22","L_star": 71.8593358804701, "a_star": -116.388994402562, "b_star": 18.3547408253505},
{ "name": "7.5G7/20","L_star": 71.778378142036, "a_star": -106.166646719356, "b_star": 17.7688413064073},
{ "name": "7.5G7/18","L_star": 71.6702016440638, "a_star": -95.7237473945382, "b_star": 16.974928734371},
{ "name": "7.5G7/16","L_star": 71.5481831124855, "a_star": -85.9676447101428, "b_star": 15.9975225729082},
{ "name": "7.5G7/14","L_star": 71.3917738768124, "a_star": -75.1041682315561, "b_star": 14.7794539189763},
{ "name": "7.5G7/12","L_star": 71.2826363622368, "a_star": -64.1643687088677, "b_star": 13.3014968927534},
{ "name": "7.5G7/10","L_star": 71.221126379664, "a_star": -53.419015698245, "b_star": 11.6829772623106},
{ "name": "7.5G7/8","L_star": 71.1526800466166, "a_star": -42.4677346970708, "b_star": 9.85682639265524},
{ "name": "7.5G7/6","L_star": 71.0841260341803, "a_star": -32.2686252768877, "b_star": 7.90127029668044},
{ "name": "7.5G7/4","L_star": 71.0154639178515, "a_star": -22.4094224564727, "b_star": 5.75121914699033},
{ "name": "7.5G7/3","L_star": 70.9742145805881, "a_star": -17.2420452188931, "b_star": 4.53369812702171},
{ "name": "7.5G7/2","L_star": 70.9329260794485, "a_star": -11.9352778005595, "b_star": 3.22894529119004},
{ "name": "7.5G7/1","L_star": 70.8915983213402, "a_star": -6.17345649816881, "b_star": 1.73160080321471},
{ "name": "7.5G6/28","L_star": 62.287533451913, "a_star": -149.865396783899, "b_star": 17.8753460369092},
{ "name": "7.5G6/26","L_star": 62.2790433169964, "a_star": -138.503995318526, "b_star": 17.3956814259382},
{ "name": "7.5G6/24","L_star": 62.2705513400023, "a_star": -128.106987849019, "b_star": 16.9223795627631},
{ "name": "7.5G6/22","L_star": 62.2195607383782, "a_star": -117.172516900633, "b_star": 16.5127418523965},
{ "name": "7.5G6/20","L_star": 62.1429499540626, "a_star": -106.845534932975, "b_star": 15.9740684356922},
{ "name": "7.5G6/18","L_star": 62.0149304752139, "a_star": -96.1906456041305, "b_star": 15.3024652851755},
{ "name": "7.5G6/16","L_star": 61.8607503688652, "a_star": -85.1097447890247, "b_star": 14.4450334025085},
{ "name": "7.5G6/14","L_star": 61.6800983268115, "a_star": -74.5732969801716, "b_star": 13.3736058047167},
{ "name": "7.5G6/12","L_star": 61.4899382067453, "a_star": -63.9608226308553, "b_star": 12.1635627965601},
{ "name": "7.5G6/10","L_star": 61.3945071518122, "a_star": -52.812891084408, "b_star": 10.7265362819636},
{ "name": "7.5G6/8","L_star": 61.333655554213, "a_star": -42.6909464263027, "b_star": 9.23376897954149},
{ "name": "7.5G6/6","L_star": 61.2727080409137, "a_star": -32.5766946670855, "b_star": 7.41727144372601},
{ "name": "7.5G6/4","L_star": 61.2029358060968, "a_star": -22.7225915416385, "b_star": 5.55694222668959},
{ "name": "7.5G6/3","L_star": 61.1592640376388, "a_star": -17.1748208001037, "b_star": 4.34750573553604},
{ "name": "7.5G6/2","L_star": 61.1155427771229, "a_star": -11.495693902569, "b_star": 2.97725335596588},
{ "name": "7.5G6/1","L_star": 61.0717718840904, "a_star": -5.68606176510528, "b_star": 1.49964993492595},
{ "name": "7.5G5/28","L_star": 51.8485158446159, "a_star": -150.687932204198, "b_star": 14.1772530408355},
{ "name": "7.5G5/26","L_star": 51.9275408097063, "a_star": -139.452282251864, "b_star": 14.0888838629418},
{ "name": "7.5G5/24","L_star": 51.9951304471422, "a_star": -129.110899211121, "b_star": 14.0607824493864},
{ "name": "7.5G5/22","L_star": 52.0063823305922, "a_star": -118.000587085659, "b_star": 13.8812781148707},
{ "name": "7.5G5/20","L_star": 52.0063823305922, "a_star": -108.137211168743, "b_star": 13.6190872246512},
{ "name": "7.5G5/18","L_star": 51.9613524334132, "a_star": -98.8304086513173, "b_star": 13.2540238829333},
{ "name": "7.5G5/16","L_star": 51.8598163970974, "a_star": -87.578277339408, "b_star": 12.666995826557},
{ "name": "7.5G5/14","L_star": 51.7239605454424, "a_star": -76.4068960970123, "b_star": 11.9932014901673},
{ "name": "7.5G5/12","L_star": 51.5647699906767, "a_star": -66.2103006316163, "b_star": 11.1528504580688},
{ "name": "7.5G5/10","L_star": 51.3589876907916, "a_star": -53.4519547216627, "b_star": 9.81546657400434},
{ "name": "7.5G5/8","L_star": 51.2901134222191, "a_star": -43.0352681598358, "b_star": 8.56900505508777},
{ "name": "7.5G5/6","L_star": 51.2210978728482, "a_star": -31.742916160131, "b_star": 6.85840780513396},
{ "name": "7.5G5/4","L_star": 51.1519403159239, "a_star": -21.404315726081, "b_star": 4.98004735556423},
{ "name": "7.5G5/3","L_star": 51.1173080558794, "a_star": -16.099322475857, "b_star": 3.89514230930919},
{ "name": "7.5G5/2","L_star": 51.0710760398342, "a_star": -10.6981504429416, "b_star": 2.68703781645272},
{ "name": "7.5G5/1","L_star": 51.0247802004202, "a_star": -5.354175529654, "b_star": 1.4154230522615},
{ "name": "7.5G4/26","L_star": 41.1046495273824, "a_star": -142.307646588818, "b_star": 10.296132577696},
{ "name": "7.5G4/24","L_star": 41.2478892149061, "a_star": -127.474322108369, "b_star": 10.5352257674404},
{ "name": "7.5G4/22","L_star": 41.3429855511839, "a_star": -115.041538702699, "b_star": 10.6857296365922},
{ "name": "7.5G4/20","L_star": 41.4062082986986, "a_star": -105.851600153508, "b_star": 10.7106800181285},
{ "name": "7.5G4/18","L_star": 41.4377675179741, "a_star": -97.2539680048843, "b_star": 10.6648761416456},
{ "name": "7.5G4/16","L_star": 41.4535341307665, "a_star": -86.7282011145379, "b_star": 10.4437720841209},
{ "name": "7.5G4/14","L_star": 41.4062082986986, "a_star": -76.7417537717655, "b_star": 10.0730624604429},
{ "name": "7.5G4/12","L_star": 41.3271580633385, "a_star": -67.1371696861185, "b_star": 9.59961930875182},
{ "name": "7.5G4/10","L_star": 41.168400235669, "a_star": -54.3448107507108, "b_star": 8.78145338479838},
{ "name": "7.5G4/8","L_star": 41.0247604338049, "a_star": -43.2782091625895, "b_star": 7.63802050535431},
{ "name": "7.5G4/6","L_star": 40.9606876153157, "a_star": -32.5838307170799, "b_star": 6.34693993484328},
{ "name": "7.5G4/4","L_star": 40.8803933340554, "a_star": -21.4121921025763, "b_star": 4.6062160669351},
{ "name": "7.5G4/3","L_star": 40.848212057973, "a_star": -15.7632219263061, "b_star": 3.5328393724361},
{ "name": "7.5G4/2","L_star": 40.7998717182846, "a_star": -10.1804045392171, "b_star": 2.35857553393657},
{ "name": "7.5G4/1","L_star": 40.7675990569515, "a_star": -4.86839885193777, "b_star": 1.16399283872733},
{ "name": "7.5G3/22","L_star": 30.4506945931528, "a_star": -120.018567218418, "b_star": 7.10142148762279},
{ "name": "7.5G3/20","L_star": 30.6236700713886, "a_star": -103.998143578604, "b_star": 7.55216072458229},
{ "name": "7.5G3/18","L_star": 30.7454221435609, "a_star": -91.7699039646367, "b_star": 7.8275961861705},
{ "name": "7.5G3/16","L_star": 30.8357287249987, "a_star": -81.3873996893788, "b_star": 7.92906478836594},
{ "name": "7.5G3/14","L_star": 30.8665432727528, "a_star": -71.3201934929879, "b_star": 7.84910545223909},
{ "name": "7.5G3/12","L_star": 30.8712803772752, "a_star": -61.432240016065, "b_star": 7.64738307716714},
{ "name": "7.5G3/10","L_star": 30.7811111344769, "a_star": -50.740725442852, "b_star": 7.10554570801216},
{ "name": "7.5G3/8","L_star": 30.6714917060928, "a_star": -40.204262334659, "b_star": 6.33010042444704},
{ "name": "7.5G3/6","L_star": 30.5949299376941, "a_star": -30.8077540177389, "b_star": 5.43368714024062},
{ "name": "7.5G3/4","L_star": 30.5301346795909, "a_star": -20.5918323399324, "b_star": 4.1023681232622},
{ "name": "7.5G3/3","L_star": 30.4916521770323, "a_star": -14.9624432642529, "b_star": 3.12600197859508},
{ "name": "7.5G3/2","L_star": 30.4506945931528, "a_star": -9.37897801704482, "b_star": 2.03295316314044},
{ "name": "7.5G3/1","L_star": 30.412080187997, "a_star": -4.2509862175665, "b_star": 0.945773607296052},
{ "name": "7.5G2/16","L_star": 20.1397589488837, "a_star": -96.7432496044495, "b_star": 4.62766483174341},
{ "name": "7.5G2/14","L_star": 20.4007905475771, "a_star": -77.9189840820268, "b_star": 5.33698151291112},
{ "name": "7.5G2/12","L_star": 20.5532932233141, "a_star": -63.1336855938128, "b_star": 5.66597073930192},
{ "name": "7.5G2/10","L_star": 20.6154918641239, "a_star": -49.8187494107126, "b_star": 5.6372142109491},
{ "name": "7.5G2/8","L_star": 20.549398774827, "a_star": -38.1286550710014, "b_star": 5.15800824230617},
{ "name": "7.5G2/6","L_star": 20.4439334277389, "a_star": -27.9103555149584, "b_star": 4.34891031113005},
{ "name": "7.5G2/4","L_star": 20.3772148994357, "a_star": -18.0618202619309, "b_star": 3.22382231814681},
{ "name": "7.5G2/3","L_star": 20.3457330880374, "a_star": -12.8695382279392, "b_star": 2.4336964580147},
{ "name": "7.5G2/2","L_star": 20.3063039693353, "a_star": -7.87924691971156, "b_star": 1.58379707976288},
{ "name": "7.5G2/1","L_star": 20.27469890094, "a_star": -3.46668297133004, "b_star": 0.757656776019699},
{ "name": "7.5G1/8","L_star": 10.4611204009613, "a_star": -59.720740289892, "b_star": 3.29661318106722},
{ "name": "7.5G1/6","L_star": 10.689498823172, "a_star": -38.1411389723885, "b_star": 3.7448726246845},
{ "name": "7.5G1/4","L_star": 10.6015591450467, "a_star": -20.8727515794116, "b_star": 2.92733615837498},
{ "name": "7.5G1/3","L_star": 10.5426081282382, "a_star": -13.995359311064, "b_star": 2.22100106858204},
{ "name": "7.5G1/2","L_star": 10.5056303593669, "a_star": -8.14722323027299, "b_star": 1.44476031414814},
{ "name": "7.5G1/1","L_star": 10.4685491268993, "a_star": -3.35130778242473, "b_star": 0.617167428328186},
{ "name": "10G9/12","L_star": 90.3584234640904, "a_star": -67.5449005261622, "b_star": 8.47032882957079},
{ "name": "10G9/10","L_star": 90.3814159036203, "a_star": -57.0235018312662, "b_star": 7.80248868397937},
{ "name": "10G9/8","L_star": 90.3998027018785, "a_star": -47.4737796406439, "b_star": 7.0400067362272},
{ "name": "10G9/6","L_star": 90.4181831475392, "a_star": -36.5028247094585, "b_star": 5.76323200939419},
{ "name": "10G9/4","L_star": 90.3446232273371, "a_star": -24.5332096600818, "b_star": 4.20691409358913},
{ "name": "10G9/3","L_star": 90.3032010120739, "a_star": -18.9322600576501, "b_star": 3.40215034299565},
{ "name": "10G9/2","L_star": 90.2571384354549, "a_star": -13.1253325827788, "b_star": 2.44169492929374},
{ "name": "10G9/1","L_star": 90.2064234313182, "a_star": -6.94556872407487, "b_star": 1.36957338272505},
{ "name": "10G8/20","L_star": 81.0933106559644, "a_star": -104.711275900918, "b_star": 9.42698763931569},
{ "name": "10G8/18","L_star": 81.0767503092922, "a_star": -95.5138487926013, "b_star": 9.6947791118539},
{ "name": "10G8/16","L_star": 81.0546610542355, "a_star": -84.8694600183361, "b_star": 9.30784590846168},
{ "name": "10G8/14","L_star": 81.0159806454827, "a_star": -74.7917958524142, "b_star": 8.73853297945129},
{ "name": "10G8/12","L_star": 80.9551347553501, "a_star": -64.7357399901138, "b_star": 7.99393667243553},
{ "name": "10G8/10","L_star": 80.8886702034112, "a_star": -53.7806025725951, "b_star": 7.16814690837417},
{ "name": "10G8/8","L_star": 80.8221143382603, "a_star": -43.5850055862191, "b_star": 6.17327249286532},
{ "name": "10G8/6","L_star": 80.7554668456199, "a_star": -33.6255208358909, "b_star": 5.06189111855813},
{ "name": "10G8/4","L_star": 80.6775952088837, "a_star": -22.9303039150813, "b_star": 3.70448865667314},
{ "name": "10G8/3","L_star": 80.6386123033635, "a_star": -17.7129343157499, "b_star": 2.92604185097745},
{ "name": "10G8/2","L_star": 80.5995979219319, "a_star": -12.3331924653093, "b_star": 2.12890246969846},
{ "name": "10G8/1","L_star": 80.5493902216599, "a_star": -6.49304053919197, "b_star": 1.18356516068181},
{ "name": "10G7/24","L_star": 71.9064924108524, "a_star": -124.251100581276, "b_star": 9.12452165251576},
{ "name": "10G7/22","L_star": 71.8660756270354, "a_star": -115.443850388028, "b_star": 9.09868589235183},
{ "name": "10G7/20","L_star": 71.7986315756441, "a_star": -104.926375964168, "b_star": 9.13161116240202},
{ "name": "10G7/18","L_star": 71.7040354702244, "a_star": -94.9891107460716, "b_star": 8.87172794812274},
{ "name": "10G7/16","L_star": 71.5753277334366, "a_star": -84.9722424772754, "b_star": 8.50827315151954},
{ "name": "10G7/14","L_star": 71.4258235077324, "a_star": -74.6049579003687, "b_star": 7.91795462106795},
{ "name": "10G7/12","L_star": 71.2826363622368, "a_star": -63.6918007399507, "b_star": 7.22406305598253},
{ "name": "10G7/10","L_star": 71.221126379664, "a_star": -53.1617609569097, "b_star": 6.49274485277906},
{ "name": "10G7/8","L_star": 71.1526800466166, "a_star": -42.3105985937184, "b_star": 5.57325600886083},
{ "name": "10G7/6","L_star": 71.0841260341803, "a_star": -32.1304630721276, "b_star": 4.55441162002792},
{ "name": "10G7/4","L_star": 71.0154639178515, "a_star": -22.1527302531048, "b_star": 3.33742198537881},
{ "name": "10G7/3","L_star": 70.9742145805881, "a_star": -17.031988453905, "b_star": 2.62382372292502},
{ "name": "10G7/2","L_star": 70.9329260794485, "a_star": -11.7699564485136, "b_star": 1.84995937509806},
{ "name": "10G7/1","L_star": 70.8915983213402, "a_star": -6.12196836111806, "b_star": 1.01081526222784},
{ "name": "10G6/26","L_star": 62.245064346556, "a_star": -136.57662748054, "b_star": 7.74149725424056},
{ "name": "10G6/24","L_star": 62.2280637888571, "a_star": -127.108873330601, "b_star": 7.72786999732209},
{ "name": "10G6/22","L_star": 62.2025490891383, "a_star": -116.228858436498, "b_star": 7.77427647011202},
{ "name": "10G6/20","L_star": 62.1429499540626, "a_star": -106.283895114589, "b_star": 7.82649335773333},
{ "name": "10G6/18","L_star": 62.0405679829665, "a_star": -95.5183443177392, "b_star": 7.70467400071184},
{ "name": "10G6/16","L_star": 61.8950653800811, "a_star": -84.8239736335142, "b_star": 7.43309388124367},
{ "name": "10G6/14","L_star": 61.7145730227062, "a_star": -74.2155492595746, "b_star": 7.03429996954577},
{ "name": "10G6/12","L_star": 61.533238360038, "a_star": -64.0404060056131, "b_star": 6.40362549141889},
{ "name": "10G6/10","L_star": 61.4031924293121, "a_star": -52.8238350570931, "b_star": 5.70125216939745},
{ "name": "10G6/8","L_star": 61.333655554213, "a_star": -42.5314226810032, "b_star": 4.94157993822737},
{ "name": "10G6/6","L_star": 61.2639934013856, "a_star": -32.2574642410894, "b_star": 4.00290204243694},
{ "name": "10G6/4","L_star": 61.2029358060968, "a_star": -22.6116652450702, "b_star": 3.01499333193789},
{ "name": "10G6/3","L_star": 61.1592640376388, "a_star": -17.0704903843819, "b_star": 2.37161157081562},
{ "name": "10G6/2","L_star": 61.1155427771229, "a_star": -11.2390663509882, "b_star": 1.62096963454432},
{ "name": "10G6/1","L_star": 61.0717718840904, "a_star": -5.4369414051576, "b_star": 0.821423848817471},
{ "name": "10G5/28","L_star": 51.6899112885491, "a_star": -145.78907872262, "b_star": 4.30187327709179},
{ "name": "10G5/26","L_star": 51.746641042967, "a_star": -136.121410198004, "b_star": 4.55957061149219},
{ "name": "10G5/24","L_star": 51.8259034397268, "a_star": -127.149337121755, "b_star": 4.75244149329341},
{ "name": "10G5/22","L_star": 51.8824062178063, "a_star": -117.12036297473, "b_star": 5.00301262847145},
{ "name": "10G5/20","L_star": 51.9162627851242, "a_star": -106.973325029249, "b_star": 5.3217801773747},
{ "name": "10G5/18","L_star": 51.9049810136899, "a_star": -97.7409355320218, "b_star": 5.49293705710556},
{ "name": "10G5/16","L_star": 51.8485158446159, "a_star": -86.7451976276548, "b_star": 5.56389752236464},
{ "name": "10G5/14","L_star": 51.7239605454424, "a_star": -75.6121708293855, "b_star": 5.48432878251894},
{ "name": "10G5/12","L_star": 51.5761656343387, "a_star": -65.861929864507, "b_star": 5.24893426849438},
{ "name": "10G5/10","L_star": 51.3704530506199, "a_star": -53.0214123225273, "b_star": 4.69431481983664},
{ "name": "10G5/8","L_star": 51.2901134222191, "a_star": -42.6666860690391, "b_star": 4.14724535821682},
{ "name": "10G5/6","L_star": 51.2210978728482, "a_star": -31.278503349368, "b_star": 3.39003953282198},
{ "name": "10G5/4","L_star": 51.1519403159239, "a_star": -21.2918986988407, "b_star": 2.51172904619761},
{ "name": "10G5/3","L_star": 51.1173080558794, "a_star": -15.950018005724, "b_star": 1.97229592957202},
{ "name": "10G5/2","L_star": 51.0710760398342, "a_star": -10.6237474405889, "b_star": 1.37195033971595},
{ "name": "10G5/1","L_star": 51.0247802004202, "a_star": -5.27092040467869, "b_star": 0.718743900818541},
{ "name": "10G4/26","L_star": 40.8159943055026, "a_star": -136.08537007773, "b_star": 1.05351892860968},
{ "name": "10G4/24","L_star": 40.97671933235, "a_star": -123.468882359022, "b_star": 1.52130105519604},
{ "name": "10G4/22","L_star": 41.120600547633, "a_star": -113.230469101874, "b_star": 2.01118458302},
{ "name": "10G4/20","L_star": 41.2161201244669, "a_star": -104.13642169659, "b_star": 2.47240224582024},
{ "name": "10G4/18","L_star": 41.295476842151, "a_star": -95.3844854825483, "b_star": 2.91166579299732},
{ "name": "10G4/16","L_star": 41.3429855511839, "a_star": -84.6742586800623, "b_star": 3.28009097400578},
{ "name": "10G4/14","L_star": 41.3588043066211, "a_star": -75.7129266278332, "b_star": 3.60962725431629},
{ "name": "10G4/12","L_star": 41.295476842151, "a_star": -66.1323443662017, "b_star": 3.67129585673226},
{ "name": "10G4/10","L_star": 41.168400235669, "a_star": -53.8272196307564, "b_star": 3.60035874459915},
{ "name": "10G4/8","L_star": 41.0247604338049, "a_star": -42.9625338938932, "b_star": 3.30490361110968},
{ "name": "10G4/6","L_star": 40.9606876153157, "a_star": -32.312407071615, "b_star": 2.87860524941705},
{ "name": "10G4/4","L_star": 40.8964703256786, "a_star": -21.3984461521591, "b_star": 2.213527666577},
{ "name": "10G4/3","L_star": 40.848212057973, "a_star": -15.7892254651966, "b_star": 1.7590079288041},
{ "name": "10G4/2","L_star": 40.8159943055026, "a_star": -10.2647598201211, "b_star": 1.23277183581887},
{ "name": "10G4/1","L_star": 40.7675990569515, "a_star": -4.94494772503554, "b_star": 0.658636043868199},
{ "name": "10G3/22","L_star": 30.0665257470739, "a_star": -114.584044705038, "b_star": -1.23105979615596},
{ "name": "10G3/18","L_star": 30.5253278474617, "a_star": -90.3858324911528, "b_star": 0.380640508356445},
{ "name": "10G3/16","L_star": 30.6404187996331, "a_star": -80.3313727132, "b_star": 1.03503925942481},
{ "name": "10G3/14","L_star": 30.7501838245701, "a_star": -70.1915200135542, "b_star": 1.63178926791152},
{ "name": "10G3/12","L_star": 30.8001230181586, "a_star": -60.3696470983855, "b_star": 2.0735895981571},
{ "name": "10G3/10","L_star": 30.7620837853306, "a_star": -50.0121245108022, "b_star": 2.27930699859088},
{ "name": "10G3/8","L_star": 30.6738802164197, "a_star": -39.9696136296604, "b_star": 2.29513860996979},
{ "name": "10G3/6","L_star": 30.5949299376941, "a_star": -30.5638415357339, "b_star": 2.11137102561015},
{ "name": "10G3/4","L_star": 30.5325377232702, "a_star": -20.7061313708167, "b_star": 1.6904641696219},
{ "name": "10G3/3","L_star": 30.4940592001884, "a_star": -15.0542617976582, "b_star": 1.34594137552262},
{ "name": "10G3/2","L_star": 30.4531058626024, "a_star": -9.43232418778805, "b_star": 0.918274375474526},
{ "name": "10G3/1","L_star": 30.412080187997, "a_star": -4.30056158036415, "b_star": 0.472893932451746},
{ "name": "10G3/0","L_star": 30.3322344122274, "a_star": -100.9368759469, "b_star": -0.382395322983453},
{ "name": "10G2/16","L_star": 19.7532078246955, "a_star": -91.2993365688707, "b_star": -2.47469376910249},
{ "name": "10G2/14","L_star": 20.1596662785396, "a_star": -74.611451653423, "b_star": -1.09437094539754},
{ "name": "10G2/12","L_star": 20.4047168544296, "a_star": -61.6548727779072, "b_star": 0.00422689750559924},
{ "name": "10G2/10","L_star": 20.5338126753784, "a_star": -49.18498164029, "b_star": 0.828582998577121},
{ "name": "10G2/8","L_star": 20.5338126753784, "a_star": -38.0557763782354, "b_star": 1.25746537790397},
{ "name": "10G2/6","L_star": 20.4439334277389, "a_star": -27.4573626835914, "b_star": 1.35363370308699},
{ "name": "10G2/4","L_star": 20.3811462962997, "a_star": -18.1176156196566, "b_star": 1.15835353453779},
{ "name": "10G2/3","L_star": 20.3457330880374, "a_star": -12.9234310495078, "b_star": 0.928653629714227},
{ "name": "10G2/2","L_star": 20.3102507357507, "a_star": -7.89386735694531, "b_star": 0.646466105267418},
{ "name": "10G2/1","L_star": 20.27469890094, "a_star": -3.47999984146605, "b_star": 0.329488573385983},
{ "name": "10G1/8","L_star": 10.2211650355277, "a_star": -56.4996417678698, "b_star": -1.93756741206392},
{ "name": "10G1/6","L_star": 10.6235985578993, "a_star": -37.0628387761869, "b_star": 0.0749787017397585},
{ "name": "10G1/4","L_star": 10.6015591450467, "a_star": -20.7515400724793, "b_star": 0.680884483447075},
{ "name": "10G1/3","L_star": 10.5499913297674, "a_star": -14.0397829437326, "b_star": 0.678862578063438},
{ "name": "10G1/2","L_star": 10.5056303593669, "a_star": -8.2896742003241, "b_star": 0.553183190096945},
{ "name": "10G1/1","L_star": 10.4685491268993, "a_star": -3.44870694811136, "b_star": 0.230986022934054},
{ "name": "2.5BG9/10","L_star": 90.3722201205514, "a_star": -56.4789363037729, "b_star": 2.53175480849428},
{ "name": "2.5BG9/8","L_star": 90.3860131990411, "a_star": -46.8865670621664, "b_star": 2.53440398181708},
{ "name": "2.5BG9/6","L_star": 90.4089937184406, "a_star": -36.0078046035929, "b_star": 2.33410993200158},
{ "name": "2.5BG9/4","L_star": 90.3446232273371, "a_star": -24.1273418557755, "b_star": 1.89974894129969},
{ "name": "2.5BG9/3","L_star": 90.3032010120739, "a_star": -18.5948786813196, "b_star": 1.6438000349855},
{ "name": "2.5BG9/2","L_star": 90.2617464903823, "a_star": -12.9370419783398, "b_star": 1.27878807723951},
{ "name": "2.5BG9/1","L_star": 90.2110358878267, "a_star": -6.8523441411581, "b_star": 0.798087441844153},
{ "name": "2.5BG8/18","L_star": 81.082271052611, "a_star": -93.9333046951453, "b_star": 0.398771919554752},
{ "name": "2.5BG8/16","L_star": 81.0491371691322, "a_star": -83.2065390219914, "b_star": 1.13965563690697},
{ "name": "2.5BG8/14","L_star": 81.0049234330954, "a_star": -73.2797550639135, "b_star": 1.49227400511587},
{ "name": "2.5BG8/12","L_star": 80.9495995227076, "a_star": -63.7256644484937, "b_star": 1.8289908833949},
{ "name": "2.5BG8/10","L_star": 80.8831273732959, "a_star": -53.0932212105293, "b_star": 2.04162719779104},
{ "name": "2.5BG8/8","L_star": 80.8221143382603, "a_star": -42.9822227689325, "b_star": 2.0600871677134},
{ "name": "2.5BG8/6","L_star": 80.7499087420435, "a_star": -32.872555463517, "b_star": 1.95929678034197},
{ "name": "2.5BG8/4","L_star": 80.6775952088837, "a_star": -22.7475962322873, "b_star": 1.59550258297871},
{ "name": "2.5BG8/3","L_star": 80.6441832154818, "a_star": -17.5353876735262, "b_star": 1.37595315943897},
{ "name": "2.5BG8/2","L_star": 80.5995979219319, "a_star": -12.1186448405097, "b_star": 1.08761861065911},
{ "name": "2.5BG8/1","L_star": 80.5549714338939, "a_star": -6.31914512484566, "b_star": 0.694867215283868},
{ "name": "2.5BG7/22","L_star": 71.8862886656199, "a_star": -112.70878485031, "b_star": -2.00992679071124},
{ "name": "2.5BG7/20","L_star": 71.8256216270139, "a_star": -102.413767513667, "b_star": -0.908193630133214},
{ "name": "2.5BG7/18","L_star": 71.7446016348757, "a_star": -93.3592751099524, "b_star": -0.312560652233529},
{ "name": "2.5BG7/16","L_star": 71.6024555374588, "a_star": -82.610581361906, "b_star": 0.162116652595556},
{ "name": "2.5BG7/14","L_star": 71.4530441286267, "a_star": -73.2303379644922, "b_star": 0.473715467414126},
{ "name": "2.5BG7/12","L_star": 71.2758062005091, "a_star": -62.3530295707975, "b_star": 0.719509510972949},
{ "name": "2.5BG7/10","L_star": 71.2142865798885, "a_star": -52.400844370055, "b_star": 1.04349903952579},
{ "name": "2.5BG7/8","L_star": 71.1526800466166, "a_star": -41.8240661572497, "b_star": 1.34873951614316},
{ "name": "2.5BG7/6","L_star": 71.07726469425, "a_star": -31.4246472211034, "b_star": 1.32015327201629},
{ "name": "2.5BG7/4","L_star": 71.0085917440811, "a_star": -21.7477655469226, "b_star": 1.18529197903874},
{ "name": "2.5BG7/3","L_star": 70.9742145805881, "a_star": -16.7586369534946, "b_star": 1.03841570377265},
{ "name": "2.5BG7/2","L_star": 70.9329260794485, "a_star": -11.5102550618543, "b_star": 0.857910475658707},
{ "name": "2.5BG7/1","L_star": 70.8915983213402, "a_star": -5.96975871148181, "b_star": 0.528824577913722},
{ "name": "2.5BG6/22","L_star": 62.1770177294756, "a_star": -114.167941497968, "b_star": -3.66985701817575},
{ "name": "2.5BG6/20","L_star": 62.1429499540626, "a_star": -104.429411147341, "b_star": -2.76798960459841},
{ "name": "2.5BG6/18","L_star": 62.0747251458716, "a_star": -94.3978061563015, "b_star": -1.95777459320126},
{ "name": "2.5BG6/16","L_star": 61.9379166734529, "a_star": -83.8190775704989, "b_star": -1.27297551501124},
{ "name": "2.5BG6/14","L_star": 61.7576234264935, "a_star": -73.042397314678, "b_star": -0.778881954283528},
{ "name": "2.5BG6/12","L_star": 61.559195254632, "a_star": -62.731408880529, "b_star": -0.477135573314458},
{ "name": "2.5BG6/10","L_star": 61.3945071518122, "a_star": -51.8532364770333, "b_star": -0.0607533262488191},
{ "name": "2.5BG6/8","L_star": 61.333655554213, "a_star": -42.0549673304502, "b_star": 0.390726620865189},
{ "name": "2.5BG6/6","L_star": 61.2639934013856, "a_star": -31.7687195883084, "b_star": 0.703923363151548},
{ "name": "2.5BG6/4","L_star": 61.2029358060968, "a_star": -22.2232277835102, "b_star": 0.71210445339116},
{ "name": "2.5BG6/3","L_star": 61.1592640376388, "a_star": -16.7780870290065, "b_star": 0.619080667266392},
{ "name": "2.5BG6/2","L_star": 61.1155427771229, "a_star": -11.1603658322776, "b_star": 0.447533849110715},
{ "name": "2.5BG6/1","L_star": 61.0717718840904, "a_star": -5.48844524901748, "b_star": 0.257618720346975},
{ "name": "2.5BG5/24","L_star": 51.6217098204152, "a_star": -123.792156712184, "b_star": -7.02335715021836},
{ "name": "2.5BG5/22","L_star": 51.6899112885491, "a_star": -114.79486132043, "b_star": -6.37086300435163},
{ "name": "2.5BG5/20","L_star": 51.7693063645118, "a_star": -105.432396251714, "b_star": -5.41400840030688},
{ "name": "2.5BG5/18","L_star": 51.8145915810381, "a_star": -95.752647691122, "b_star": -4.41994782246964},
{ "name": "2.5BG5/16","L_star": 51.8032759473254, "a_star": -85.3433926495112, "b_star": -3.41538133246766},
{ "name": "2.5BG5/14","L_star": 51.7126145977439, "a_star": -73.9563796739525, "b_star": -2.41566615589865},
{ "name": "2.5BG5/12","L_star": 51.5989453986242, "a_star": -64.8071686119217, "b_star": -1.75939026459331},
{ "name": "2.5BG5/10","L_star": 51.4048257367113, "a_star": -52.5604832804657, "b_star": -1.03097075216767},
{ "name": "2.5BG5/8","L_star": 51.2901134222191, "a_star": -41.7934154626075, "b_star": -0.496650627420503},
{ "name": "2.5BG5/6","L_star": 51.2095814947108, "a_star": -30.5534024299816, "b_star": -0.0245717269315993},
{ "name": "2.5BG5/4","L_star": 51.1519403159239, "a_star": -20.847533444741, "b_star": 0.215007829453029},
{ "name": "2.5BG5/3","L_star": 51.1057560232836, "a_star": -15.6882196611494, "b_star": 0.251810605320202},
{ "name": "2.5BG5/2","L_star": 51.0710760398342, "a_star": -10.4914373799625, "b_star": 0.224679349879509},
{ "name": "2.5BG5/1","L_star": 51.0363601557647, "a_star": -5.26590815472266, "b_star": 0.152563793799221},
{ "name": "2.5BG4/24","L_star": 40.6381400008437, "a_star": -119.307829967454, "b_star": -9.13017133098977},
{ "name": "2.5BG4/22","L_star": 40.7837399731063, "a_star": -109.831662316501, "b_star": -8.35751748757255},
{ "name": "2.5BG4/20","L_star": 40.9285970802895, "a_star": -101.152224960435, "b_star": -7.54405949045731},
{ "name": "2.5BG4/18","L_star": 41.0727207259106, "a_star": -92.5770365348587, "b_star": -6.66458877573322},
{ "name": "2.5BG4/16","L_star": 41.1843157153987, "a_star": -82.5426110116115, "b_star": -5.68162858221036},
{ "name": "2.5BG4/14","L_star": 41.2478892149061, "a_star": -73.6462721503216, "b_star": -4.74541202507582},
{ "name": "2.5BG4/12","L_star": 41.2478892149061, "a_star": -64.0482215547467, "b_star": -3.74129508339817},
{ "name": "2.5BG4/10","L_star": 41.168400235669, "a_star": -53.2816321135166, "b_star": -2.75615624291544},
{ "name": "2.5BG4/8","L_star": 41.0407561598708, "a_star": -42.2708084024261, "b_star": -1.82037818920378},
{ "name": "2.5BG4/6","L_star": 40.9606876153157, "a_star": -31.5321878140321, "b_star": -1.12948906000763},
{ "name": "2.5BG4/4","L_star": 40.8803933340554, "a_star": -21.0076554196113, "b_star": -0.565116267578247},
{ "name": "2.5BG4/3","L_star": 40.848212057973, "a_star": -15.4928541505129, "b_star": -0.31772568992513},
{ "name": "2.5BG4/2","L_star": 40.7998717182846, "a_star": -10.0478860244624, "b_star": -0.133598709358562},
{ "name": "2.5BG4/1","L_star": 40.7675990569515, "a_star": -4.83458904871481, "b_star": -0.023750114908494},
{ "name": "2.5BG3/20","L_star": 29.9115426722927, "a_star": -96.9282970532093, "b_star": -9.52167424350636},
{ "name": "2.5BG3/18","L_star": 30.1619474401831, "a_star": -87.0176980751948, "b_star": -8.49232545846124},
{ "name": "2.5BG3/16","L_star": 30.4, "a_star": -77.3161293791108, "b_star": -7.28057119630886},
{ "name": "2.5BG3/14","L_star": 30.5805465659089, "a_star": -68.1872109991411, "b_star": -6.0836866838263},
{ "name": "2.5BG3/12","L_star": 30.681044281032, "a_star": -58.707313407465, "b_star": -4.89697304295217},
{ "name": "2.5BG3/10","L_star": 30.7215991743225, "a_star": -49.0576281030857, "b_star": -3.73194210722215},
{ "name": "2.5BG3/8","L_star": 30.664324707868, "a_star": -39.1187635306254, "b_star": -2.6130342668126},
{ "name": "2.5BG3/6","L_star": 30.5877393619723, "a_star": -29.8293711328563, "b_star": -1.73838486416729},
{ "name": "2.5BG3/4","L_star": 30.5277313876761, "a_star": -20.2899666800208, "b_star": -1.02860683651159},
{ "name": "2.5BG3/3","L_star": 30.4892449046116, "a_star": -14.7178213093692, "b_star": -0.692481510752163},
{ "name": "2.5BG3/2","L_star": 30.4482830733378, "a_star": -9.19880533868336, "b_star": -0.379166884004556},
{ "name": "2.5BG3/1","L_star": 30.4096646534768, "a_star": -4.17300670788426, "b_star": -0.11498156586105},
{ "name": "2.5BG2/14","L_star": 19.8222688454216, "a_star": -70.7945071273683, "b_star": -8.27490555661921},
{ "name": "2.5BG2/12","L_star": 20.187499769181, "a_star": -59.0942151706459, "b_star": -6.62623631095209},
{ "name": "2.5BG2/10","L_star": 20.4204136190585, "a_star": -47.9830483022561, "b_star": -5.05185724095116},
{ "name": "2.5BG2/8","L_star": 20.4986952432828, "a_star": -37.5012367099334, "b_star": -3.58023261360102},
{ "name": "2.5BG2/6","L_star": 20.4360968651002, "a_star": -26.6675596896379, "b_star": -2.31612503861142},
{ "name": "2.5BG2/4","L_star": 20.3693495554285, "a_star": -17.6757675891448, "b_star": -1.35174040319124},
{ "name": "2.5BG2/3","L_star": 20.3378541081486, "a_star": -12.6717582238155, "b_star": -0.880676136957514},
{ "name": "2.5BG2/2","L_star": 20.3023563446473, "a_star": -7.8393432677889, "b_star": -0.466984932190351},
{ "name": "2.5BG2/1","L_star": 20.2707443932339, "a_star": -3.53121257811903, "b_star": -0.145699582536829},
{ "name": "2.5BG1/8","L_star": 9.93038989303505, "a_star": -53.3578000316629, "b_star": -7.4211311624622},
{ "name": "2.5BG1/6","L_star": 10.5130341671149, "a_star": -34.6153856566086, "b_star": -4.56360241975402},
{ "name": "2.5BG1/4","L_star": 10.5942045527048, "a_star": -20.5008506178394, "b_star": -2.3925172697348},
{ "name": "2.5BG1/3","L_star": 10.5352208169387, "a_star": -13.9664932079241, "b_star": -1.517755751693},
{ "name": "2.5BG1/2","L_star": 10.4908103225287, "a_star": -8.22087482248719, "b_star": -0.801236425480678},
{ "name": "2.5BG1/1","L_star": 10.4611204009613, "a_star": -3.50249857009495, "b_star": -0.274910280948015},
{ "name": "5BG9/10","L_star": 90.3584234640904, "a_star": -55.1548905252047, "b_star": -3.83427128250244},
{ "name": "5BG9/8","L_star": 90.3768182108191, "a_star": -45.9618730472375, "b_star": -2.71417863866583},
{ "name": "5BG9/6","L_star": 90.3906100971672, "a_star": -35.0313218061646, "b_star": -1.61732698790382},
{ "name": "5BG9/4","L_star": 90.3446232273371, "a_star": -23.4863589790532, "b_star": -0.772745412630971},
{ "name": "5BG9/3","L_star": 90.3032010120739, "a_star": -17.9568882518639, "b_star": -0.411737027849801},
{ "name": "5BG9/2","L_star": 90.2617464903823, "a_star": -12.2486994946174, "b_star": -0.165705084806933},
{ "name": "5BG9/1","L_star": 90.2156479437571, "a_star": -6.34382425910052, "b_star": 0.0139506278234558},
{ "name": "5BG8/16","L_star": 81.0546610542355, "a_star": -81.003185790849, "b_star": -7.95723737340555},
{ "name": "5BG8/14","L_star": 81.0049234330954, "a_star": -71.3103990801979, "b_star": -6.81575021213379},
{ "name": "5BG8/12","L_star": 80.9440636579367, "a_star": -61.9670727819485, "b_star": -5.67233255227231},
{ "name": "5BG8/10","L_star": 80.8831273732959, "a_star": -51.5800441084675, "b_star": -4.49222779148628},
{ "name": "5BG8/8","L_star": 80.8165638845484, "a_star": -41.938182330799, "b_star": -3.414494767114},
{ "name": "5BG8/6","L_star": 80.7499087420435, "a_star": -31.6974386808904, "b_star": -2.20861379141155},
{ "name": "5BG8/4","L_star": 80.6775952088837, "a_star": -21.7643255714696, "b_star": -1.2172623981719},
{ "name": "5BG8/3","L_star": 80.6386123033635, "a_star": -16.7881365968412, "b_star": -0.785581359386733},
{ "name": "5BG8/2","L_star": 80.5995979219319, "a_star": -11.4786202244989, "b_star": -0.443005689058351},
{ "name": "5BG8/1","L_star": 80.5605520009757, "a_star": -6.04162222273485, "b_star": -0.126218509063847},
{ "name": "5BG7/20","L_star": 71.8593358804701, "a_star": -99.4457120232822, "b_star": -12.461994870194},
{ "name": "5BG7/18","L_star": 71.7851303249967, "a_star": -90.7772933121118, "b_star": -10.8814849154011},
{ "name": "5BG7/16","L_star": 71.6431157686568, "a_star": -80.3252956347836, "b_star": -9.56394600917094},
{ "name": "5BG7/14","L_star": 71.4598466367537, "a_star": -70.0195136099394, "b_star": -8.34960545572749},
{ "name": "5BG7/12","L_star": 71.2758062005091, "a_star": -60.4064278684969, "b_star": -7.1396454227477},
{ "name": "5BG7/10","L_star": 71.2142865798885, "a_star": -50.8925923652038, "b_star": -5.77115989885377},
{ "name": "5BG7/8","L_star": 71.1458294979289, "a_star": -40.3605012250021, "b_star": -4.3063585243543},
{ "name": "5BG7/6","L_star": 71.07726469425, "a_star": -30.1200841268353, "b_star": -2.9642839812883},
{ "name": "5BG7/4","L_star": 71.0154639178515, "a_star": -21.1108710307625, "b_star": -1.83137680092997},
{ "name": "5BG7/3","L_star": 70.9742145805881, "a_star": -16.2409692780276, "b_star": -1.26172660832067},
{ "name": "5BG7/2","L_star": 70.939810219965, "a_star": -11.1696592217369, "b_star": -0.753345659448224},
{ "name": "5BG7/1","L_star": 70.8984890111646, "a_star": -5.80658186392402, "b_star": -0.279616239519576},
{ "name": "5BG6/20","L_star": 62.1259049213583, "a_star": -100.861333724795, "b_star": -14.5781598079168},
{ "name": "5BG6/18","L_star": 62.0832597682774, "a_star": -91.0809797099751, "b_star": -12.8469578490479},
{ "name": "5BG6/16","L_star": 61.9721638125076, "a_star": -80.8803759081328, "b_star": -11.3244125165575},
{ "name": "5BG6/14","L_star": 61.7920294590482, "a_star": -70.7644871486063, "b_star": -9.96005676950418},
{ "name": "5BG6/12","L_star": 61.5764902034168, "a_star": -60.4264703228994, "b_star": -8.61661198260735},
{ "name": "5BG6/10","L_star": 61.3858199245362, "a_star": -50.1972518665886, "b_star": -7.14797450016338},
{ "name": "5BG6/8","L_star": 61.3249546477991, "a_star": -40.3343923878277, "b_star": -5.6034223086292},
{ "name": "5BG6/6","L_star": 61.2639934013856, "a_star": -30.6355342367754, "b_star": -4.00881970203009},
{ "name": "5BG6/4","L_star": 61.1942054050492, "a_star": -21.0899685376661, "b_star": -2.50099171649993},
{ "name": "5BG6/3","L_star": 61.1592640376388, "a_star": -15.952783114747, "b_star": -1.77346045631022},
{ "name": "5BG6/2","L_star": 61.1155427771229, "a_star": -10.7083139160711, "b_star": -1.10504579806827},
{ "name": "5BG6/1","L_star": 61.0805300400617, "a_star": -5.38096335412969, "b_star": -0.471538594223397},
{ "name": "5BG5/22","L_star": 51.4848926336617, "a_star": -110.122459341261, "b_star": -19.5130188335027},
{ "name": "5BG5/20","L_star": 51.5647699906767, "a_star": -101.410219931744, "b_star": -18.1250656120433},
{ "name": "5BG5/18","L_star": 51.6558277423594, "a_star": -92.6667332346149, "b_star": -16.4284231151178},
{ "name": "5BG5/16","L_star": 51.7353026927899, "a_star": -81.7625440988907, "b_star": -14.0502491460095},
{ "name": "5BG5/14","L_star": 51.6899112885491, "a_star": -70.8301065341486, "b_star": -11.9558961000899},
{ "name": "5BG5/12","L_star": 51.5875574359089, "a_star": -62.2800953039878, "b_star": -10.5511398336569},
{ "name": "5BG5/10","L_star": 51.4048257367113, "a_star": -50.8090914611914, "b_star": -8.74096044590307},
{ "name": "5BG5/8","L_star": 51.2786206613769, "a_star": -39.8466813584755, "b_star": -6.71524259657752},
{ "name": "5BG5/6","L_star": 51.2095814947108, "a_star": -29.1854123079298, "b_star": -4.6383133332248},
{ "name": "5BG5/4","L_star": 51.1404001987838, "a_star": -19.9218026789704, "b_star": -3.04458032201793},
{ "name": "5BG5/3","L_star": 51.1057560232836, "a_star": -15.0789045145369, "b_star": -2.26970310296668},
{ "name": "5BG5/2","L_star": 51.0710760398342, "a_star": -10.094733935902, "b_star": -1.45710479998402},
{ "name": "5BG5/1","L_star": 51.0363601557647, "a_star": -5.10198557888791, "b_star": -0.677999182440003},
{ "name": "5BG4/20","L_star": 40.5243758552341, "a_star": -96.5510983836155, "b_star": -20.6922513718142},
{ "name": "5BG4/18","L_star": 40.7191211520259, "a_star": -87.1103646558032, "b_star": -18.8750044443563},
{ "name": "5BG4/16","L_star": 40.9285970802895, "a_star": -78.5358403642584, "b_star": -16.8120409090168},
{ "name": "5BG4/14","L_star": 41.0886895909649, "a_star": -70.1625562048004, "b_star": -14.6492434797085},
{ "name": "5BG4/12","L_star": 41.168400235669, "a_star": -61.0011072760285, "b_star": -12.3723345363764},
{ "name": "5BG4/10","L_star": 41.1365426641592, "a_star": -50.9764587987431, "b_star": -10.224970591112},
{ "name": "5BG4/8","L_star": 41.0247604338049, "a_star": -40.4073687310886, "b_star": -8.02280367788784},
{ "name": "5BG4/6","L_star": 40.9446468688704, "a_star": -30.0252917790973, "b_star": -5.81556236181845},
{ "name": "5BG4/4","L_star": 40.8803933340554, "a_star": -20.0048737585656, "b_star": -3.70107643728068},
{ "name": "5BG4/3","L_star": 40.848212057973, "a_star": -14.8578065771085, "b_star": -2.6421323984593},
{ "name": "5BG4/2","L_star": 40.7998717182846, "a_star": -9.71473238444309, "b_star": -1.62634888506541},
{ "name": "5BG4/1","L_star": 40.7675990569515, "a_star": -4.74939472946129, "b_star": -0.716499931671555},
{ "name": "5BG3/18","L_star": 29.6284068409682, "a_star": -80.9782922755888, "b_star": -20.2577387223787},
{ "name": "5BG3/16","L_star": 29.9214139997679, "a_star": -72.8879267251917, "b_star": -18.4776821918754},
{ "name": "5BG3/14","L_star": 30.2667007541636, "a_star": -63.9904788254574, "b_star": -16.1438299534028},
{ "name": "5BG3/12","L_star": 30.5036848014572, "a_star": -55.6671418297702, "b_star": -13.8432722036323},
{ "name": "5BG3/10","L_star": 30.6356346762228, "a_star": -46.9446106410687, "b_star": -11.3428705714425},
{ "name": "5BG3/8","L_star": 30.6452019417853, "a_star": -37.6239429454683, "b_star": -8.83597552911803},
{ "name": "5BG3/6","L_star": 30.5805465659089, "a_star": -28.2808234618523, "b_star": -6.49490289002601},
{ "name": "5BG3/4","L_star": 30.5205200218775, "a_star": -19.3744930632807, "b_star": -4.37682284424896},
{ "name": "5BG3/3","L_star": 30.4844296117184, "a_star": -14.1976972192593, "b_star": -3.15237076360132},
{ "name": "5BG3/2","L_star": 30.4458713030924, "a_star": -9.00471958632446, "b_star": -1.95105563438528},
{ "name": "5BG3/1","L_star": 30.4072488674824, "a_star": -4.18746893581917, "b_star": -0.860768155953384},
{ "name": "5BG2/12","L_star": 19.8263229727587, "a_star": -54.1433207192211, "b_star": -15.5698091851549},
{ "name": "5BG2/10","L_star": 20.2033854164229, "a_star": -45.3386307832452, "b_star": -12.8756729387785},
{ "name": "5BG2/8","L_star": 20.4204136190585, "a_star": -35.6950013839751, "b_star": -9.91314161303024},
{ "name": "5BG2/6","L_star": 20.4243356967993, "a_star": -25.1991312834301, "b_star": -6.84706699826468},
{ "name": "5BG2/4","L_star": 20.3614808079835, "a_star": -16.7289377019562, "b_star": -4.53318582051275},
{ "name": "5BG2/3","L_star": 20.3299717100471, "a_star": -12.0924186923401, "b_star": -3.24140186486805},
{ "name": "5BG2/2","L_star": 20.2944585185861, "a_star": -7.61768185303033, "b_star": -1.97746722002499},
{ "name": "5BG2/1","L_star": 20.262832789887, "a_star": -3.52372777148319, "b_star": -0.839801348710745},
{ "name": "5BG1/6","L_star": 10.3940733558595, "a_star": -31.0916698593104, "b_star": -10.1364399458613},
{ "name": "5BG1/4","L_star": 10.5647454263466, "a_star": -19.034913075973, "b_star": -6.15567351280302},
{ "name": "5BG1/3","L_star": 10.5204338421202, "a_star": -13.1850166673569, "b_star": -4.22642706991499},
{ "name": "5BG1/2","L_star": 10.4759736852486, "a_star": -7.98084096236601, "b_star": -2.47234868121985},
{ "name": "5BG1/1","L_star": 10.4462504228925, "a_star": -3.56410652011802, "b_star": -1.01594757665396},
{ "name": "7.5BG9/10","L_star": 90.3446232273371, "a_star": -53.3450148222587, "b_star": -10.9830635546555},
{ "name": "7.5BG9/8","L_star": 90.3584234640904, "a_star": -44.2264103568159, "b_star": -8.91626251649926},
{ "name": "7.5BG9/6","L_star": 90.3676216327313, "a_star": -33.2078173812675, "b_star": -6.4445322041883},
{ "name": "7.5BG9/4","L_star": 90.3446232273371, "a_star": -22.1951644901715, "b_star": -4.08249554250371},
{ "name": "7.5BG9/3","L_star": 90.3032010120739, "a_star": -16.7475397589379, "b_star": -3.03236684697705},
{ "name": "7.5BG9/2","L_star": 90.2617464903823, "a_star": -11.2984128172537, "b_star": -1.98134679143609},
{ "name": "7.5BG9/1","L_star": 90.2202595991964, "a_star": -5.69831795943893, "b_star": -0.930383672011348},
{ "name": "7.5BG8/16","L_star": 81.0712289379739, "a_star": -77.8839676157571, "b_star": -17.4513139639378},
{ "name": "7.5BG8/14","L_star": 81.0104523543632, "a_star": -68.3833326696415, "b_star": -15.4651682209631},
{ "name": "7.5BG8/12","L_star": 80.9495995227076, "a_star": -59.5607247871593, "b_star": -13.5554950856761},
{ "name": "7.5BG8/10","L_star": 80.8831273732959, "a_star": -49.3230869331595, "b_star": -11.2124794390215},
{ "name": "7.5BG8/8","L_star": 80.8165638845484, "a_star": -39.766376024913, "b_star": -9.04798454339682},
{ "name": "7.5BG8/6","L_star": 80.7499087420435, "a_star": -30.3102923352713, "b_star": -6.77253289434654},
{ "name": "7.5BG8/4","L_star": 80.6775952088837, "a_star": -20.3440096839532, "b_star": -4.10900727573884},
{ "name": "7.5BG8/3","L_star": 80.6386123033635, "a_star": -15.4964616778381, "b_star": -3.06273931025529},
{ "name": "7.5BG8/2","L_star": 80.5995979219319, "a_star": -10.5759494824473, "b_star": -2.12542275704115},
{ "name": "7.5BG8/1","L_star": 80.5605520009757, "a_star": -5.42193414470921, "b_star": -1.11119113691236},
{ "name": "7.5BG7/18","L_star": 71.8256216270139, "a_star": -87.3417820060915, "b_star": -20.9281087781027},
{ "name": "7.5BG7/16","L_star": 71.683738307854, "a_star": -77.3390072418911, "b_star": -18.6817041664658},
{ "name": "7.5BG7/14","L_star": 71.4734484793738, "a_star": -67.2082369344368, "b_star": -16.4101180411669},
{ "name": "7.5BG7/12","L_star": 71.2689749695638, "a_star": -57.6957350499769, "b_star": -14.4327733239363},
{ "name": "7.5BG7/10","L_star": 71.2142865798885, "a_star": -48.162000596917, "b_star": -12.0038803387919},
{ "name": "7.5BG7/8","L_star": 71.1458294979289, "a_star": -38.1407912729718, "b_star": -9.39831484301099},
{ "name": "7.5BG7/6","L_star": 71.07726469425, "a_star": -28.5066751429852, "b_star": -6.84799896060642},
{ "name": "7.5BG7/4","L_star": 71.0154639178515, "a_star": -20.0907555285373, "b_star": -4.57883140420681},
{ "name": "7.5BG7/3","L_star": 70.9742145805881, "a_star": -15.3474356354646, "b_star": -3.38844159767333},
{ "name": "7.5BG7/2","L_star": 70.939810219965, "a_star": -10.4372802924794, "b_star": -2.20862577442642},
{ "name": "7.5BG7/1","L_star": 70.8984890111646, "a_star": -5.31767027683305, "b_star": -1.0227957628391},
{ "name": "7.5BG6/18","L_star": 62.0832597682774, "a_star": -87.1460788449009, "b_star": -22.6150676275882},
{ "name": "7.5BG6/16","L_star": 61.9892761062234, "a_star": -77.5986786606936, "b_star": -20.1839624233706},
{ "name": "7.5BG6/14","L_star": 61.7920294590482, "a_star": -67.3699135051274, "b_star": -17.8532466185008},
{ "name": "7.5BG6/12","L_star": 61.5678436930684, "a_star": -57.6231175674096, "b_star": -15.7564179852565},
{ "name": "7.5BG6/10","L_star": 61.3858199245362, "a_star": -47.6417906040218, "b_star": -13.3328446735605},
{ "name": "7.5BG6/8","L_star": 61.3249546477991, "a_star": -38.2116544858316, "b_star": -10.5593689002184},
{ "name": "7.5BG6/6","L_star": 61.2639934013856, "a_star": -28.9312128900155, "b_star": -7.97475044988945},
{ "name": "7.5BG6/4","L_star": 61.1942054050492, "a_star": -19.9377632941592, "b_star": -5.2932095313406},
{ "name": "7.5BG6/3","L_star": 61.1592640376388, "a_star": -15.0082889496947, "b_star": -3.91940653483633},
{ "name": "7.5BG6/2","L_star": 61.1155427771229, "a_star": -10.0471387007902, "b_star": -2.53226867960692},
{ "name": "7.5BG6/1","L_star": 61.0805300400617, "a_star": -5.06006737570014, "b_star": -1.20482603500793},
{ "name": "7.5BG5/18","L_star": 51.4963152629647, "a_star": -87.498231093225, "b_star": -26.731257168796},
{ "name": "7.5BG5/16","L_star": 51.6444589255192, "a_star": -77.6505882027722, "b_star": -23.3537315227206},
{ "name": "7.5BG5/14","L_star": 51.6558277423594, "a_star": -67.4598986598843, "b_star": -20.2701901963557},
{ "name": "7.5BG5/12","L_star": 51.5647699906767, "a_star": -58.9073923613333, "b_star": -17.8704310076346},
{ "name": "7.5BG5/10","L_star": 51.3933720702654, "a_star": -48.0364622109685, "b_star": -14.8070397145217},
{ "name": "7.5BG5/8","L_star": 51.2786206613769, "a_star": -37.6621781713387, "b_star": -11.6359414581685},
{ "name": "7.5BG5/6","L_star": 51.2095814947108, "a_star": -27.9585789172976, "b_star": -8.41897623888441},
{ "name": "7.5BG5/4","L_star": 51.1404001987838, "a_star": -18.7679743370667, "b_star": -5.54032223971836},
{ "name": "7.5BG5/3","L_star": 51.1057560232836, "a_star": -14.2512324169641, "b_star": -4.18922731744236},
{ "name": "7.5BG5/2","L_star": 51.0710760398342, "a_star": -9.69942623000586, "b_star": -2.81368349389137},
{ "name": "7.5BG5/1","L_star": 51.0363601557647, "a_star": -4.95798179564017, "b_star": -1.42285253720724},
{ "name": "7.5BG4/18","L_star": 40.4428345189212, "a_star": -82.2718297767278, "b_star": -28.6075157373171},
{ "name": "7.5BG4/16","L_star": 40.6705602368878, "a_star": -73.777343069914, "b_star": -26.1880569113773},
{ "name": "7.5BG4/14","L_star": 40.9125382368196, "a_star": -66.0717853889227, "b_star": -23.2702809960847},
{ "name": "7.5BG4/12","L_star": 41.0727207259106, "a_star": -57.4868892718036, "b_star": -19.9672332880189},
{ "name": "7.5BG4/10","L_star": 41.1046495273824, "a_star": -47.8497404613565, "b_star": -16.6520918363436},
{ "name": "7.5BG4/8","L_star": 41.0247604338049, "a_star": -37.9287359908653, "b_star": -13.2342929643058},
{ "name": "7.5BG4/6","L_star": 40.9446468688704, "a_star": -28.183848626444, "b_star": -9.8357444149528},
{ "name": "7.5BG4/4","L_star": 40.8803933340554, "a_star": -18.7525139380191, "b_star": -6.41305929087967},
{ "name": "7.5BG4/3","L_star": 40.848212057973, "a_star": -14.0087456537478, "b_star": -4.69364001840125},
{ "name": "7.5BG4/2","L_star": 40.7998717182846, "a_star": -9.30051076158939, "b_star": -3.03023631890607},
{ "name": "7.5BG4/1","L_star": 40.7675990569515, "a_star": -4.63858728913272, "b_star": -1.42963900780082},
{ "name": "7.5BG3/16","L_star": 29.5908896337533, "a_star": -68.0139083489512, "b_star": -26.4081124869197},
{ "name": "7.5BG3/14","L_star": 29.9731689780403, "a_star": -59.8843572630058, "b_star": -23.7268405876328},
{ "name": "7.5BG3/12","L_star": 30.3273864228217, "a_star": -52.1616301235481, "b_star": -20.5643949724082},
{ "name": "7.5BG3/10","L_star": 30.5517531439122, "a_star": -44.2584528952706, "b_star": -17.3573044583767},
{ "name": "7.5BG3/8","L_star": 30.6236700713886, "a_star": -35.3221364896442, "b_star": -13.8601022667855},
{ "name": "7.5BG3/6","L_star": 30.5805465659089, "a_star": -26.8517485709554, "b_star": -10.6078433878713},
{ "name": "7.5BG3/4","L_star": 30.5181157363793, "a_star": -18.2647892916869, "b_star": -7.25870311356163},
{ "name": "7.5BG3/3","L_star": 30.4820215911168, "a_star": -13.4091079891346, "b_star": -5.30662061029623},
{ "name": "7.5BG3/2","L_star": 30.4434592823515, "a_star": -8.61326283669042, "b_star": -3.35334071442344},
{ "name": "7.5BG3/1","L_star": 30.4072488674824, "a_star": -4.10611282299186, "b_star": -1.51146006112475},
{ "name": "7.5BG2/12","L_star": 19.5279267337455, "a_star": -50.0427066476564, "b_star": -21.7519436761655},
{ "name": "7.5BG2/10","L_star": 20.0398917465438, "a_star": -41.9370573467852, "b_star": -18.4638032853099},
{ "name": "7.5BG2/8","L_star": 20.3457330880374, "a_star": -33.3500469658858, "b_star": -14.6489532938965},
{ "name": "7.5BG2/6","L_star": 20.4164906964065, "a_star": -23.878780441422, "b_star": -10.5158773993015},
{ "name": "7.5BG2/4","L_star": 20.3575451568209, "a_star": -15.7502873350697, "b_star": -6.95085902693008},
{ "name": "7.5BG2/3","L_star": 20.3220858900229, "a_star": -11.434479563432, "b_star": -5.04159173302587},
{ "name": "7.5BG2/2","L_star": 20.2905083162782, "a_star": -7.28082382990294, "b_star": -3.14459347133244},
{ "name": "7.5BG2/1","L_star": 20.258875693305, "a_star": -3.4365839295214, "b_star": -1.42499550622112},
{ "name": "7.5BG1/6","L_star": 10.2740304307288, "a_star": -27.6398932662509, "b_star": -14.277593024965},
{ "name": "7.5BG1/4","L_star": 10.5499913297674, "a_star": -17.2068468216567, "b_star": -8.72652798270938},
{ "name": "7.5BG1/3","L_star": 10.5056303593669, "a_star": -12.1716730299872, "b_star": -6.16797278045756},
{ "name": "7.5BG1/2","L_star": 10.4685491268993, "a_star": -7.62502144195326, "b_star": -3.8098719793602},
{ "name": "7.5BG1/1","L_star": 10.4313637040572, "a_star": -3.56058285574987, "b_star": -1.72624107345708},
{ "name": "10BG9/6","L_star": 90.3492237042086, "a_star": -31.1768449513283, "b_star": -10.5811578883109},
{ "name": "10BG9/4","L_star": 90.3492237042086, "a_star": -20.6510186626318, "b_star": -6.88696099174917},
{ "name": "10BG9/3","L_star": 90.3078050746641, "a_star": -15.3881542826221, "b_star": -4.90569841155311},
{ "name": "10BG9/2","L_star": 90.2617464903823, "a_star": -10.1452590810785, "b_star": -3.05080992431266},
{ "name": "10BG9/1","L_star": 90.2202595991964, "a_star": -5.09676596528219, "b_star": -1.28565945306436},
{ "name": "10BG8/14","L_star": 81.0270353379946, "a_star": -64.7989096896701, "b_star": -24.1693997819124},
{ "name": "10BG8/12","L_star": 80.9606693560446, "a_star": -56.3319280020184, "b_star": -21.1126357980126},
{ "name": "10BG8/10","L_star": 80.8886702034112, "a_star": -46.5443728453087, "b_star": -17.4829231714249},
{ "name": "10BG8/8","L_star": 80.8221143382603, "a_star": -37.2311426997577, "b_star": -13.9932426651},
{ "name": "10BG8/6","L_star": 80.7499087420435, "a_star": -28.2271576034372, "b_star": -10.4015602431742},
{ "name": "10BG8/4","L_star": 80.6775952088837, "a_star": -19.093683120907, "b_star": -6.75035700600226},
{ "name": "10BG8/3","L_star": 80.6441832154818, "a_star": -14.3363541601506, "b_star": -4.9680820527791},
{ "name": "10BG8/2","L_star": 80.605173334499, "a_star": -9.57946260008463, "b_star": -3.16035134156225},
{ "name": "10BG8/1","L_star": 80.5661319230916, "a_star": -4.83324549111913, "b_star": -1.46668738553557},
{ "name": "10BG7/16","L_star": 71.7310837481058, "a_star": -73.0565140307549, "b_star": -29.6162814058506},
{ "name": "10BG7/14","L_star": 71.432630252038, "a_star": -62.1465185101271, "b_star": -26.190307223499},
{ "name": "10BG7/12","L_star": 71.2758062005091, "a_star": -53.2099716904896, "b_star": -22.7760547637141},
{ "name": "10BG7/10","L_star": 71.2074457071191, "a_star": -44.0825185605306, "b_star": -18.7472895928624},
{ "name": "10BG7/8","L_star": 71.1458294979289, "a_star": -35.2325845377542, "b_star": -14.7873864358732},
{ "name": "10BG7/6","L_star": 71.07726469425, "a_star": -26.3317975885669, "b_star": -10.7256924726678},
{ "name": "10BG7/4","L_star": 71.0154639178515, "a_star": -18.453898563533, "b_star": -7.23345299710927},
{ "name": "10BG7/3","L_star": 70.9742145805881, "a_star": -13.9841249050185, "b_star": -5.36121599370656},
{ "name": "10BG7/2","L_star": 70.939810219965, "a_star": -9.3365764544715, "b_star": -3.43920728426588},
{ "name": "10BG7/1","L_star": 70.8984890111646, "a_star": -4.69010411691539, "b_star": -1.59972495952097},
{ "name": "10BG6/18","L_star": 62.0747251458716, "a_star": -80.0760336145402, "b_star": -34.8122466022161},
{ "name": "10BG6/16","L_star": 61.9978294377359, "a_star": -71.6016017999723, "b_star": -30.9761533123515},
{ "name": "10BG6/14","L_star": 61.76622778901, "a_star": -61.6812275598753, "b_star": -27.2994043873214},
{ "name": "10BG6/12","L_star": 61.4986021086743, "a_star": -52.4341793978776, "b_star": -23.9597405570996},
{ "name": "10BG6/10","L_star": 61.3858199245362, "a_star": -43.1565547234385, "b_star": -19.8638092644531},
{ "name": "10BG6/8","L_star": 61.3249546477991, "a_star": -34.8646898075883, "b_star": -15.997863367516},
{ "name": "10BG6/6","L_star": 61.2639934013856, "a_star": -26.5942180616148, "b_star": -12.0838117612645},
{ "name": "10BG6/4","L_star": 61.1942054050492, "a_star": -17.9996003092167, "b_star": -7.92492856969811},
{ "name": "10BG6/3","L_star": 61.1592640376388, "a_star": -13.6062632278738, "b_star": -5.787455453484},
{ "name": "10BG6/2","L_star": 61.1155427771229, "a_star": -9.19045677148289, "b_star": -3.69713861998937},
{ "name": "10BG6/1","L_star": 61.0805300400617, "a_star": -4.67922807408466, "b_star": -1.70155100519489},
{ "name": "10BG5/16","L_star": 51.4963152629647, "a_star": -70.8482981697023, "b_star": -34.1446896296498},
{ "name": "10BG5/14","L_star": 51.6103295257169, "a_star": -61.4666456966604, "b_star": -29.4786481937175},
{ "name": "10BG5/12","L_star": 51.5305599747064, "a_star": -53.2233989085438, "b_star": -25.9747847596838},
{ "name": "10BG5/10","L_star": 51.3475184265326, "a_star": -43.3990535848613, "b_star": -21.7985288934339},
{ "name": "10BG5/8","L_star": 51.2786206613769, "a_star": -34.148529569848, "b_star": -17.2839747209601},
{ "name": "10BG5/6","L_star": 51.2095814947108, "a_star": -25.3838737282482, "b_star": -12.8657093317196},
{ "name": "10BG5/4","L_star": 51.1404001987838, "a_star": -17.0148107612129, "b_star": -8.35569416198796},
{ "name": "10BG5/3","L_star": 51.1057560232836, "a_star": -12.9447954525087, "b_star": -6.18652202834393},
{ "name": "10BG5/2","L_star": 51.0710760398342, "a_star": -8.77127206347755, "b_star": -4.08139171806583},
{ "name": "10BG5/1","L_star": 51.0363601557647, "a_star": -4.54701955726422, "b_star": -1.98762140912234},
{ "name": "10BG4/16","L_star": 40.4264979523013, "a_star": -65.3936854992866, "b_star": -36.7749721556484},
{ "name": "10BG4/14","L_star": 40.6705602368878, "a_star": -59.4136480132562, "b_star": -33.1926187762073},
{ "name": "10BG4/12","L_star": 40.97671933235, "a_star": -51.5105169299564, "b_star": -28.273486788745},
{ "name": "10BG4/10","L_star": 41.0727207259106, "a_star": -43.2736219572069, "b_star": -23.6249928561711},
{ "name": "10BG4/8","L_star": 41.0087557289414, "a_star": -34.0879073670351, "b_star": -18.7983499122029},
{ "name": "10BG4/6","L_star": 40.9446468688704, "a_star": -25.3550221901477, "b_star": -13.907091658723},
{ "name": "10BG4/4","L_star": 40.8803933340554, "a_star": -17.1702673157735, "b_star": -9.15302415112704},
{ "name": "10BG4/3","L_star": 40.848212057973, "a_star": -12.9317607544661, "b_star": -6.69985623652425},
{ "name": "10BG4/2","L_star": 40.7998717182846, "a_star": -8.57859683956905, "b_star": -4.30525465610039},
{ "name": "10BG4/1","L_star": 40.7675990569515, "a_star": -4.30267409213908, "b_star": -2.01878632548241},
{ "name": "10BG3/14","L_star": 29.5332430378252, "a_star": -53.3100374060428, "b_star": -34.4160669335573},
{ "name": "10BG3/12","L_star": 30.0640738349048, "a_star": -46.0172061037753, "b_star": -29.6261030011389},
{ "name": "10BG3/10","L_star": 30.4434592823515, "a_star": -39.737818298496, "b_star": -24.8137566782948},
{ "name": "10BG3/8","L_star": 30.6093044349619, "a_star": -31.8794160988558, "b_star": -19.5082630832863},
{ "name": "10BG3/6","L_star": 30.5781484735156, "a_star": -24.268119389145, "b_star": -14.9377808424767},
{ "name": "10BG3/4","L_star": 30.5157112023247, "a_star": -16.5440683739327, "b_star": -10.0757901456537},
{ "name": "10BG3/3","L_star": 30.4796133209922, "a_star": -12.3478394528071, "b_star": -7.38572795728679},
{ "name": "10BG3/2","L_star": 30.4434592823515, "a_star": -8.12690519803258, "b_star": -4.70468649119615},
{ "name": "10BG3/1","L_star": 30.4072488674824, "a_star": -4.00318660844159, "b_star": -2.17959712567908},
{ "name": "10BG2/10","L_star": 19.7409928569444, "a_star": -36.0366157238154, "b_star": -26.3855430256514},
{ "name": "10BG2/8","L_star": 20.258875693305, "a_star": -29.1518944764837, "b_star": -20.732732543384},
{ "name": "10BG2/6","L_star": 20.4047168544296, "a_star": -21.479914673383, "b_star": -14.7391538048058},
{ "name": "10BG2/4","L_star": 20.3496712973095, "a_star": -14.1920246685699, "b_star": -9.73636777935168},
{ "name": "10BG2/3","L_star": 20.3181416956288, "a_star": -10.3789571543022, "b_star": -7.07984844619198},
{ "name": "10BG2/2","L_star": 20.2826053307678, "a_star": -6.7201159633363, "b_star": -4.46786078661138},
{ "name": "10BG2/1","L_star": 20.2509589079675, "a_star": -3.23312455366223, "b_star": -2.06186393367063},
{ "name": "10BG1/6","L_star": 10.2060214196641, "a_star": -22.4969863534488, "b_star": -18.613195021873},
{ "name": "10BG1/4","L_star": 10.5204338421202, "a_star": -14.7318446112144, "b_star": -11.6954247349701},
{ "name": "10BG1/3","L_star": 10.4908103225287, "a_star": -10.7251969249913, "b_star": -8.43185115743069},
{ "name": "10BG1/2","L_star": 10.4536875015812, "a_star": -6.94142373198656, "b_star": -5.38166861817889},
{ "name": "10BG1/1","L_star": 10.4164601972582, "a_star": -3.37003014025888, "b_star": -2.570440171011},
{ "name": "2.5B9/4","L_star": 90.3400223523973, "a_star": -18.6612772710951, "b_star": -9.56188958827486},
{ "name": "2.5B9/3","L_star": 90.2939916902734, "a_star": -13.2866114823464, "b_star": -6.77575117185225},
{ "name": "2.5B9/2","L_star": 90.2525299808178, "a_star": -8.44094847210891, "b_star": -4.20564095904357},
{ "name": "2.5B9/1","L_star": 90.2156479437571, "a_star": -4.02466882054364, "b_star": -1.89246457591565},
{ "name": "2.5B8/12","L_star": 80.9938636993964, "a_star": -51.2805636024173, "b_star": -29.471686337323},
{ "name": "2.5B8/8","L_star": 80.8110127943531, "a_star": -33.5816340080939, "b_star": -19.1190206968552},
{ "name": "2.5B8/6","L_star": 80.7443499997881, "a_star": -25.5613944071256, "b_star": -13.977457370955},
{ "name": "2.5B8/4","L_star": 80.672028147072, "a_star": -17.3588976081623, "b_star": -9.07352700993112},
{ "name": "2.5B8/3","L_star": 80.6330407488799, "a_star": -12.7204960669142, "b_star": -6.49080119225116},
{ "name": "2.5B8/2","L_star": 80.5940218657014, "a_star": -8.19006963402896, "b_star": -4.03461985080509},
{ "name": "2.5B8/1","L_star": 80.5605520009757, "a_star": -3.92731861074685, "b_star": -1.76419659321567},
{ "name": "2.5B8/0","L_star": 80.8942123994079, "a_star": -42.2951371430505, "b_star": -24.2982606226994},
{ "name": "2.5B7/16","L_star": 71.7918814693996, "a_star": -66.1279406230947, "b_star": -40.0185940950729},
{ "name": "2.5B7/14","L_star": 71.4122068384821, "a_star": -55.7898145620554, "b_star": -34.9349035049318},
{ "name": "2.5B7/12","L_star": 71.2689749695638, "a_star": -47.6247060657548, "b_star": -30.2451033039619},
{ "name": "2.5B7/10","L_star": 71.2006037609348, "a_star": -39.5570726617827, "b_star": -24.7966155697406},
{ "name": "2.5B7/8","L_star": 71.1321251684848, "a_star": -31.2203642629639, "b_star": -19.4372505256255},
{ "name": "2.5B7/6","L_star": 71.0704022728571, "a_star": -23.6947943106485, "b_star": -14.2965878296178},
{ "name": "2.5B7/4","L_star": 71.0017184845736, "a_star": -16.3243807594026, "b_star": -9.48738955352368},
{ "name": "2.5B7/3","L_star": 70.9673358859548, "a_star": -12.3053268645242, "b_star": -6.9097150311288},
{ "name": "2.5B7/2","L_star": 70.9329260794485, "a_star": -8.18749751910902, "b_star": -4.42397224873352},
{ "name": "2.5B7/1","L_star": 70.8984890111646, "a_star": -4.12627344453664, "b_star": -2.03527450698289},
{ "name": "2.5B6/16","L_star": 62.0149304752139, "a_star": -64.4240818222424, "b_star": -40.8903933154697},
{ "name": "2.5B6/14","L_star": 61.7490171593035, "a_star": -54.8731822967254, "b_star": -35.8603394000782},
{ "name": "2.5B6/12","L_star": 61.4465896084137, "a_star": -46.3653626864864, "b_star": -31.4337048737043},
{ "name": "2.5B6/10","L_star": 61.3771307463894, "a_star": -38.0040144243919, "b_star": -25.7742920946675},
{ "name": "2.5B6/8","L_star": 61.3162517828243, "a_star": -31.1409968579895, "b_star": -20.7744484783638},
{ "name": "2.5B6/6","L_star": 61.2552767955579, "a_star": -23.6993900274638, "b_star": -15.4201905825436},
{ "name": "2.5B6/4","L_star": 61.1854730287976, "a_star": -15.9324000680567, "b_star": -9.81476564412491},
{ "name": "2.5B6/3","L_star": 61.1505237493858, "a_star": -11.9643060137796, "b_star": -7.15648143701464},
{ "name": "2.5B6/2","L_star": 61.1155427771229, "a_star": -8.06757308567862, "b_star": -4.61195594158241},
{ "name": "2.5B6/1","L_star": 61.0805300400617, "a_star": -4.08218317375153, "b_star": -2.20321428021175},
{ "name": "2.5B5/16","L_star": 51.381914509339, "a_star": -61.5278315960349, "b_star": -44.2455868420093},
{ "name": "2.5B5/14","L_star": 51.5533705016818, "a_star": -53.9828408891003, "b_star": -37.9489235347242},
{ "name": "2.5B5/12","L_star": 51.4848926336617, "a_star": -46.683692930722, "b_star": -33.1568607201651},
{ "name": "2.5B5/10","L_star": 51.3360452545173, "a_star": -37.9548702181738, "b_star": -27.3969004194051},
{ "name": "2.5B5/8","L_star": 51.2671239727284, "a_star": -29.9544814620245, "b_star": -21.741196164379},
{ "name": "2.5B5/6","L_star": 51.1980611685518, "a_star": -22.5098261513464, "b_star": -16.1268397890459},
{ "name": "2.5B5/4","L_star": 51.1404001987838, "a_star": -15.1830449092108, "b_star": -10.530189092952},
{ "name": "2.5B5/3","L_star": 51.1057560232836, "a_star": -11.4742750197028, "b_star": -7.74666344516683},
{ "name": "2.5B5/2","L_star": 51.0710760398342, "a_star": -7.74977840340396, "b_star": -5.05606262939435},
{ "name": "2.5B5/1","L_star": 51.0363601557647, "a_star": -3.96307311531285, "b_star": -2.40371413803764},
{ "name": "2.5B4/16","L_star": 40.2626094371923, "a_star": -54.7065648579213, "b_star": -46.1711400515173},
{ "name": "2.5B4/14","L_star": 40.4917875655919, "a_star": -50.0490336415988, "b_star": -42.1789870049518},
{ "name": "2.5B4/12","L_star": 40.9125382368196, "a_star": -44.0213322577479, "b_star": -35.1744244772504},
{ "name": "2.5B4/10","L_star": 41.0407561598708, "a_star": -37.4785273866458, "b_star": -29.5609275571156},
{ "name": "2.5B4/8","L_star": 41.0087557289414, "a_star": -29.6668770717619, "b_star": -23.4909780217478},
{ "name": "2.5B4/6","L_star": 40.9285970802895, "a_star": -21.981687703363, "b_star": -17.2380375552005},
{ "name": "2.5B4/4","L_star": 40.8643072491098, "a_star": -15.100952584836, "b_star": -11.5600384731953},
{ "name": "2.5B4/3","L_star": 40.8321077477464, "a_star": -11.3452884633274, "b_star": -8.48625062873752},
{ "name": "2.5B4/2","L_star": 40.7998717182846, "a_star": -7.61798620884824, "b_star": -5.42313021993168},
{ "name": "2.5B4/1","L_star": 40.7675990569515, "a_star": -3.85046484486778, "b_star": -2.54741572231518},
{ "name": "2.5B3/12","L_star": 29.8026767820801, "a_star": -38.7713847775694, "b_star": -37.5598281975787},
{ "name": "2.5B3/10","L_star": 30.3395044946333, "a_star": -34.2270801054614, "b_star": -31.0671930263593},
{ "name": "2.5B3/8","L_star": 30.5949299376941, "a_star": -28.006827600238, "b_star": -24.4043982772606},
{ "name": "2.5B3/6","L_star": 30.5733515477892, "a_star": -21.3529375510849, "b_star": -18.4943612588457},
{ "name": "2.5B3/4","L_star": 30.5109013882896, "a_star": -14.7181355105102, "b_star": -12.6238720648151},
{ "name": "2.5B3/3","L_star": 30.4747960319158, "a_star": -11.015894243854, "b_star": -9.26968193291424},
{ "name": "2.5B3/2","L_star": 30.4410470110501, "a_star": -7.31897507037066, "b_star": -5.92199449828497},
{ "name": "2.5B3/1","L_star": 30.4048328299485, "a_star": -3.66950263991767, "b_star": -2.75611196043419},
{ "name": "2.5B2/10","L_star": 19.5155562847184, "a_star": -29.7972489840488, "b_star": -32.7318030193794},
{ "name": "2.5B2/8","L_star": 20.1954443357961, "a_star": -25.0319253940983, "b_star": -25.3557866378446},
{ "name": "2.5B2/6","L_star": 20.4047168544296, "a_star": -19.1027351036107, "b_star": -18.1193293387073},
{ "name": "2.5B2/4","L_star": 20.3457330880374, "a_star": -12.6378684776664, "b_star": -11.8233252403046},
{ "name": "2.5B2/3","L_star": 20.3141966443599, "a_star": -9.33366944675185, "b_star": -8.59221005858862},
{ "name": "2.5B2/2","L_star": 20.2826053307678, "a_star": -6.13332941563677, "b_star": -5.48998306637135},
{ "name": "2.5B2/1","L_star": 20.2509589079675, "a_star": -2.9968642570472, "b_star": -2.59017959108641},
{ "name": "2.5B1/6","L_star": 10.1908602815276, "a_star": -17.8252181563941, "b_star": -21.5691506178269},
{ "name": "2.5B1/4","L_star": 10.5130341671149, "a_star": -12.6117391558689, "b_star": -13.6441179437808},
{ "name": "2.5B1/3","L_star": 10.4908103225287, "a_star": -9.48214922679633, "b_star": -9.95988883044868},
{ "name": "2.5B1/2","L_star": 10.4536875015812, "a_star": -6.33160123754142, "b_star": -6.48449117810991},
{ "name": "2.5B1/1","L_star": 10.4164601972582, "a_star": -3.16602433367465, "b_star": -3.17182821641323},
{ "name": "5B9/4","L_star": 90.3170120036561, "a_star": -15.6883420024531, "b_star": -12.109517857013},
{ "name": "5B9/3","L_star": 90.2801747147258, "a_star": -10.9314463068673, "b_star": -8.20245666368138},
{ "name": "5B9/2","L_star": 90.2433118720672, "a_star": -6.77008832732895, "b_star": -4.88069632435069},
{ "name": "5B9/1","L_star": 90.2110358878267, "a_star": -3.21009509211528, "b_star": -2.07866774106871},
{ "name": "5B8/8","L_star": 80.8054610674918, "a_star": -28.0555893301483, "b_star": -25.0559221662909},
{ "name": "5B8/6","L_star": 80.7221086403056, "a_star": -20.9336940578727, "b_star": -17.8394957891506},
{ "name": "5B8/4","L_star": 80.6553231133634, "a_star": -14.1633015272051, "b_star": -11.2976394162965},
{ "name": "5B8/3","L_star": 80.6218957120763, "a_star": -10.271114469977, "b_star": -7.92330998683759},
{ "name": "5B8/2","L_star": 80.582867821507, "a_star": -6.4127146934263, "b_star": -4.79618437883507},
{ "name": "5B8/1","L_star": 80.5493902216599, "a_star": -3.01495110980238, "b_star": -2.0152467564803},
{ "name": "5B7/14","L_star": 71.5753277334366, "a_star": -45.239072378342, "b_star": -44.6656749130538},
{ "name": "5B7/12","L_star": 71.2484748572357, "a_star": -38.6121279450197, "b_star": -38.9015119641406},
{ "name": "5B7/10","L_star": 71.1800714776797, "a_star": -31.9155214543564, "b_star": -31.398501091055},
{ "name": "5B7/8","L_star": 71.1115605877898, "a_star": -25.6426130313706, "b_star": -23.9889664736986},
{ "name": "5B7/6","L_star": 71.0498085156385, "a_star": -19.8199968508827, "b_star": -17.6464337760308},
{ "name": "5B7/4","L_star": 70.9879687066308, "a_star": -13.5432496074341, "b_star": -11.4127724018347},
{ "name": "5B7/3","L_star": 70.9535752313128, "a_star": -10.056132528895, "b_star": -8.27720400165146},
{ "name": "5B7/2","L_star": 70.9260408484614, "a_star": -6.55621062552664, "b_star": -5.26971029353271},
{ "name": "5B7/1","L_star": 70.8915983213402, "a_star": -3.23347894199461, "b_star": -2.40596358226559},
{ "name": "5B6/16","L_star": 61.9892761062234, "a_star": -51.2632170115656, "b_star": -51.6048787390502},
{ "name": "5B6/14","L_star": 61.619693871405, "a_star": -42.9593117587315, "b_star": -44.9709502570977},
{ "name": "5B6/12","L_star": 61.4118757581295, "a_star": -36.159972835099, "b_star": -38.3813289075361},
{ "name": "5B6/10","L_star": 61.3510514957635, "a_star": -30.4276368460915, "b_star": -31.6928846964799},
{ "name": "5B6/8","L_star": 61.2901314255044, "a_star": -25.0475606453601, "b_star": -25.1645218545279},
{ "name": "5B6/6","L_star": 61.229115169177, "a_star": -19.3777492459588, "b_star": -18.4144218891106},
{ "name": "5B6/4","L_star": 61.1680023462115, "a_star": -13.0728384974149, "b_star": -11.6116289849356},
{ "name": "5B6/3","L_star": 61.1330372293513, "a_star": -9.85137761767174, "b_star": -8.41619967508636},
{ "name": "5B6/2","L_star": 61.1067925736237, "a_star": -6.58597161214874, "b_star": -5.42825058878984},
{ "name": "5B6/1","L_star": 61.0717718840904, "a_star": -3.29127939242646, "b_star": -2.57058825474827},
{ "name": "5B5/16","L_star": 51.3130871738921, "a_star": -46.9023908126501, "b_star": -54.5988181302185},
{ "name": "5B5/14","L_star": 51.4848926336617, "a_star": -41.6570660705529, "b_star": -46.4445186053297},
{ "name": "5B5/12","L_star": 51.3933720702654, "a_star": -35.6725808523015, "b_star": -39.9159163593861},
{ "name": "5B5/10","L_star": 51.301602258608, "a_star": -29.3633076490754, "b_star": -32.8115081572702},
{ "name": "5B5/8","L_star": 51.2441187985769, "a_star": -23.8744894673171, "b_star": -26.0658173751456},
{ "name": "5B5/6","L_star": 51.186536890986, "a_star": -18.4198610764146, "b_star": -19.1201866132229},
{ "name": "5B5/4","L_star": 51.1288561132401, "a_star": -12.7946910542026, "b_star": -12.278547486622},
{ "name": "5B5/3","L_star": 51.0942000120299, "a_star": -9.64701761462222, "b_star": -9.05215421667405},
{ "name": "5B5/2","L_star": 51.0595080720228, "a_star": -6.46644633598503, "b_star": -5.90224772874715},
{ "name": "5B5/1","L_star": 51.0247802004202, "a_star": -3.22180692816687, "b_star": -2.86823074894655},
{ "name": "5B4/14","L_star": 40.4428345189212, "a_star": -36.4835127855086, "b_star": -50.1153335482786},
{ "name": "5B4/12","L_star": 40.9125382368196, "a_star": -33.3137375019659, "b_star": -41.7277170690616},
{ "name": "5B4/10","L_star": 41.0247604338049, "a_star": -28.8200985562185, "b_star": -34.495786239001},
{ "name": "5B4/8","L_star": 40.97671933235, "a_star": -23.4593480998694, "b_star": -27.6236339012254},
{ "name": "5B4/6","L_star": 40.9125382368196, "a_star": -17.7133003900392, "b_star": -20.1661658382412},
{ "name": "5B4/4","L_star": 40.8643072491098, "a_star": -12.5570430425068, "b_star": -13.3882288782227},
{ "name": "5B4/3","L_star": 40.8321077477464, "a_star": -9.61876559790206, "b_star": -9.86114929347069},
{ "name": "5B4/2","L_star": 40.7998717182846, "a_star": -6.55337820787971, "b_star": -6.3693340682158},
{ "name": "5B4/1","L_star": 40.7675990569515, "a_star": -3.36071936730481, "b_star": -3.01397417288412},
{ "name": "5B3/12","L_star": 29.7206864611895, "a_star": -27.1414385000584, "b_star": -45.1674571427822},
{ "name": "5B3/10","L_star": 30.3346580265234, "a_star": -25.5911316278651, "b_star": -36.9318993866749},
{ "name": "5B3/8","L_star": 30.5997224221648, "a_star": -22.0949066749588, "b_star": -28.7902567567637},
{ "name": "5B3/6","L_star": 30.558954838511, "a_star": -17.1525844396457, "b_star": -21.5705290012053},
{ "name": "5B3/4","L_star": 30.5012787747144, "a_star": -12.2775042733654, "b_star": -14.7181571636665},
{ "name": "5B3/3","L_star": 30.4699777439717, "a_star": -9.42017353278155, "b_star": -10.8902681901166},
{ "name": "5B3/2","L_star": 30.4362217165054, "a_star": -6.49004296872632, "b_star": -7.04817366579362},
{ "name": "5B3/1","L_star": 30.4048328299485, "a_star": -3.33289699489039, "b_star": -3.35865062169396},
{ "name": "5B2/10","L_star": 19.4783931362464, "a_star": -20.4310386171125, "b_star": -38.8958607441362},
{ "name": "5B2/8","L_star": 20.1477245120574, "a_star": -18.7819904808011, "b_star": -30.234378404104},
{ "name": "5B2/6","L_star": 20.4007905475771, "a_star": -15.359770586921, "b_star": -21.3346321594159},
{ "name": "5B2/4","L_star": 20.341794025138, "a_star": -10.5254472208949, "b_star": -13.7552782943338},
{ "name": "5B2/3","L_star": 20.3102507357507, "a_star": -7.95403190539035, "b_star": -10.0134593217628},
{ "name": "5B2/2","L_star": 20.2786525466281, "a_star": -5.37824775564755, "b_star": -6.43629501641055},
{ "name": "5B2/1","L_star": 20.2509589079675, "a_star": -2.71670170860958, "b_star": -3.07471428463953},
{ "name": "5B1/6","L_star": 10.2664912537208, "a_star": -12.5864454520166, "b_star": -23.6914978595325},
{ "name": "5B1/4","L_star": 10.5056303593669, "a_star": -9.94952103198952, "b_star": -15.4537911478057},
{ "name": "5B1/3","L_star": 10.4833940818493, "a_star": -7.84992774755207, "b_star": -11.464759850627},
{ "name": "5B1/2","L_star": 10.4462504228925, "a_star": -5.4759313446221, "b_star": -7.61790203578282},
{ "name": "5B1/1","L_star": 10.4164601972582, "a_star": -2.84211658723302, "b_star": -3.82261855466544},
{ "name": "7.5B9/4","L_star": 90.2939916902734, "a_star": -12.8434486632561, "b_star": -13.5491833939264},
{ "name": "7.5B9/3","L_star": 90.2663541456864, "a_star": -8.52132634114289, "b_star": -8.98058315002179},
{ "name": "7.5B9/2","L_star": 90.2294817089492, "a_star": -5.10490015205317, "b_star": -5.15780009353044},
{ "name": "7.5B9/1","L_star": 90.2018105741446, "a_star": -2.2849769529924, "b_star": -2.07444935978782},
{ "name": "7.5B8/8","L_star": 80.8165638845484, "a_star": -22.9014078695287, "b_star": -28.1687647552752},
{ "name": "7.5B8/6","L_star": 80.7054209055644, "a_star": -16.9748876009396, "b_star": -19.9708532120768},
{ "name": "7.5B8/4","L_star": 80.6386123033635, "a_star": -11.4093309828637, "b_star": -12.471598366558},
{ "name": "7.5B8/3","L_star": 80.605173334499, "a_star": -8.12672245492246, "b_star": -8.69489214657464},
{ "name": "7.5B8/2","L_star": 80.571711200428, "a_star": -4.95359465249579, "b_star": -5.2070977533166},
{ "name": "7.5B8/1","L_star": 80.5438083640871, "a_star": -2.22729289705226, "b_star": -2.17002151417471},
{ "name": "7.5B7/12","L_star": 71.3099463253338, "a_star": -30.3035141013771, "b_star": -43.0560020244651},
{ "name": "7.5B7/10","L_star": 71.1458294979289, "a_star": -24.9622907501293, "b_star": -35.0707225225224},
{ "name": "7.5B7/8","L_star": 71.0841260341803, "a_star": -20.2343068025076, "b_star": -26.9931353654287},
{ "name": "7.5B7/6","L_star": 71.0292050098948, "a_star": -15.640833638231, "b_star": -19.4813760803666},
{ "name": "7.5B7/4","L_star": 70.9742145805881, "a_star": -10.8356021539815, "b_star": -12.5442534022422},
{ "name": "7.5B7/3","L_star": 70.9466932704427, "a_star": -8.03492729521171, "b_star": -9.03797616399051},
{ "name": "7.5B7/2","L_star": 70.9122671133475, "a_star": -5.15768879260325, "b_star": -5.70129558126742},
{ "name": "7.5B7/1","L_star": 70.8847065384495, "a_star": -2.52461002905624, "b_star": -2.58904890751024},
{ "name": "7.5B6/16","L_star": 61.8264050841404, "a_star": -37.6934046229341, "b_star": -58.178636572998},
{ "name": "7.5B6/14","L_star": 61.437914063409, "a_star": -31.4348686423871, "b_star": -50.3722832558681},
{ "name": "7.5B6/12","L_star": 61.3771307463894, "a_star": -27.2205034816968, "b_star": -42.68771369468},
{ "name": "7.5B6/10","L_star": 61.3162517828243, "a_star": -22.8302924751553, "b_star": -34.8893493199203},
{ "name": "7.5B6/8","L_star": 61.2639934013856, "a_star": -18.8939019235987, "b_star": -27.7469930624888},
{ "name": "7.5B6/6","L_star": 61.2116642330571, "a_star": -14.8024874602347, "b_star": -20.418964653648},
{ "name": "7.5B6/4","L_star": 61.1505237493858, "a_star": -10.0195049716622, "b_star": -12.9355472839442},
{ "name": "7.5B6/3","L_star": 61.1242909953236, "a_star": -7.62272352928894, "b_star": -9.37858903941096},
{ "name": "7.5B6/2","L_star": 61.098040383699, "a_star": -5.17133860735353, "b_star": -5.98319128276239},
{ "name": "7.5B6/1","L_star": 61.0717718840904, "a_star": -2.70002014156778, "b_star": -2.78394808224232},
{ "name": "7.5B5/16","L_star": 51.3589876907916, "a_star": -33.8462156877414, "b_star": -59.3288600581307},
{ "name": "7.5B5/14","L_star": 51.4048257367113, "a_star": -29.8534470712943, "b_star": -50.9643904714236},
{ "name": "7.5B5/12","L_star": 51.3245681714155, "a_star": -25.6220341636456, "b_star": -43.6116550814895},
{ "name": "7.5B5/10","L_star": 51.2671239727284, "a_star": -21.5759350516649, "b_star": -35.8375721423091},
{ "name": "7.5B5/8","L_star": 51.2095814947108, "a_star": -17.9127323109707, "b_star": -28.4950365441821},
{ "name": "7.5B5/6","L_star": 51.163476468069, "a_star": -13.9766743122277, "b_star": -21.1454277539024},
{ "name": "7.5B5/4","L_star": 51.1057560232836, "a_star": -9.83054141819128, "b_star": -13.8384336703093},
{ "name": "7.5B5/3","L_star": 51.0826400186909, "a_star": -7.54336773107483, "b_star": -10.1528978188112},
{ "name": "7.5B5/2","L_star": 51.0479361118152, "a_star": -5.14360247499857, "b_star": -6.55812773496167},
{ "name": "7.5B5/1","L_star": 51.0247802004202, "a_star": -2.64579769698831, "b_star": -3.10021900912962},
{ "name": "7.5B4/14","L_star": 40.6056826069156, "a_star": -25.8515028648341, "b_star": -53.3416460772445},
{ "name": "7.5B4/12","L_star": 40.9606876153157, "a_star": -23.7712389017374, "b_star": -44.8044787844247},
{ "name": "7.5B4/10","L_star": 41.0087557289414, "a_star": -20.3470200494305, "b_star": -37.2521547497347},
{ "name": "7.5B4/8","L_star": 40.9446468688704, "a_star": -17.0389682238048, "b_star": -30.0916270658951},
{ "name": "7.5B4/6","L_star": 40.8964703256786, "a_star": -13.0225999499549, "b_star": -22.2001500853433},
{ "name": "7.5B4/4","L_star": 40.848212057973, "a_star": -9.48828224909667, "b_star": -14.8596788189746},
{ "name": "7.5B4/3","L_star": 40.8159943055026, "a_star": -7.38408363355328, "b_star": -11.0035386891418},
{ "name": "7.5B4/2","L_star": 40.7837399731063, "a_star": -5.0258397797539, "b_star": -7.17020985748907},
{ "name": "7.5B4/1","L_star": 40.751448956775, "a_star": -2.59514678104053, "b_star": -3.45058542996728},
{ "name": "7.5B3/12","L_star": 29.8447999059177, "a_star": -18.1296448402836, "b_star": -48.7372306616967},
{ "name": "7.5B3/10","L_star": 30.4531058626024, "a_star": -18.4582481904557, "b_star": -39.4306916435871},
{ "name": "7.5B3/8","L_star": 30.5973263031549, "a_star": -16.1020256831916, "b_star": -31.5505731767023},
{ "name": "7.5B3/6","L_star": 30.5397453655359, "a_star": -12.6650136310453, "b_star": -23.7603947796138},
{ "name": "7.5B3/4","L_star": 30.4868373828619, "a_star": -9.21514730485753, "b_star": -16.3462796971698},
{ "name": "7.5B3/3","L_star": 30.4579276506651, "a_star": -7.26197514451471, "b_star": -12.2518418994867},
{ "name": "7.5B3/2","L_star": 30.4313954189373, "a_star": -5.07865279440303, "b_star": -8.12879020745827},
{ "name": "7.5B3/1","L_star": 30.4024165408095, "a_star": -2.70325712707398, "b_star": -3.98758644118058},
{ "name": "7.5B2/10","L_star": 19.5773225939979, "a_star": -12.7028419431159, "b_star": -42.1524143128666},
{ "name": "7.5B2/8","L_star": 20.2351149553776, "a_star": -13.3105128562888, "b_star": -32.6427587089834},
{ "name": "7.5B2/6","L_star": 20.385076843683, "a_star": -11.0740544067257, "b_star": -23.6040545983065},
{ "name": "7.5B2/4","L_star": 20.3299717100471, "a_star": -7.81881856925168, "b_star": -15.4632134057927},
{ "name": "7.5B2/3","L_star": 20.3023563446473, "a_star": -6.13839747891598, "b_star": -11.4313676101723},
{ "name": "7.5B2/2","L_star": 20.27469890094, "a_star": -4.34350704129072, "b_star": -7.49051502627269},
{ "name": "7.5B2/1","L_star": 20.2509589079675, "a_star": -2.30213775834418, "b_star": -3.68025906741289},
{ "name": "7.5B1/8","L_star": 9.82940173128609, "a_star": -8.15876136768089, "b_star": -34.3380742623543},
{ "name": "7.5B1/6","L_star": 10.3491853893754, "a_star": -8.41973174990158, "b_star": -24.9708434787492},
{ "name": "7.5B1/4","L_star": 10.4982224131009, "a_star": -7.00404626638219, "b_star": -16.784051717337},
{ "name": "7.5B1/3","L_star": 10.4685491268993, "a_star": -5.70068567866443, "b_star": -12.7303670917248},
{ "name": "7.5B1/2","L_star": 10.4462504228925, "a_star": -4.1442789732485, "b_star": -8.62530135813486},
{ "name": "7.5B1/1","L_star": 10.4164601972582, "a_star": -2.21894170259394, "b_star": -4.42685670180055},
{ "name": "10B9/4","L_star": 90.2801747147258, "a_star": -9.53522802011236, "b_star": -14.5828907443408},
{ "name": "10B9/3","L_star": 90.2525299808178, "a_star": -6.42747735090621, "b_star": -9.51104645524019},
{ "name": "10B9/2","L_star": 90.2248708542314, "a_star": -3.9262509119532, "b_star": -5.33934093528834},
{ "name": "10B9/1","L_star": 90.1971973162189, "a_star": -1.80082155605088, "b_star": -2.11645658061443},
{ "name": "10B8/8","L_star": 80.8443097919206, "a_star": -16.5581102582352, "b_star": -30.5861824478533},
{ "name": "10B8/6","L_star": 80.672028147072, "a_star": -11.804607853795, "b_star": -21.6278157977293},
{ "name": "10B8/4","L_star": 80.6163222293855, "a_star": -7.74195572453368, "b_star": -13.5975902841628},
{ "name": "10B8/3","L_star": 80.5884451656218, "a_star": -5.67556160696392, "b_star": -9.38701705118106},
{ "name": "10B8/2","L_star": 80.5661319230916, "a_star": -3.61722499925199, "b_star": -5.52252782158822},
{ "name": "10B8/1","L_star": 80.5438083640871, "a_star": -1.7439996891217, "b_star": -2.24108115713373},
{ "name": "10B7/12","L_star": 71.3917738768124, "a_star": -21.3794532607569, "b_star": -46.1386161115672},
{ "name": "10B7/10","L_star": 71.1047035694533, "a_star": -17.1374626450996, "b_star": -37.4799731230804},
{ "name": "10B7/8","L_star": 71.0498085156385, "a_star": -14.078600417065, "b_star": -28.8872464225425},
{ "name": "10B7/6","L_star": 71.0017184845736, "a_star": -10.9196616865581, "b_star": -20.9075492053437},
{ "name": "10B7/4","L_star": 70.9535752313128, "a_star": -7.52909930257867, "b_star": -13.4088992064213},
{ "name": "10B7/3","L_star": 70.9260408484614, "a_star": -5.5084187009628, "b_star": -9.61919846109613},
{ "name": "10B7/2","L_star": 70.905378608356, "a_star": -3.49569623278762, "b_star": -6.01339849860956},
{ "name": "10B7/1","L_star": 70.877813662059, "a_star": -1.68885450172895, "b_star": -2.68884824366213},
{ "name": "10B6/16","L_star": 61.6714748699993, "a_star": -25.933224718579, "b_star": -62.222741238525},
{ "name": "10B6/14","L_star": 61.3858199245362, "a_star": -21.6708249805825, "b_star": -53.5816640669099},
{ "name": "10B6/12","L_star": 61.333655554213, "a_star": -18.7010718947647, "b_star": -45.3605101782427},
{ "name": "10B6/10","L_star": 61.2727080409137, "a_star": -15.759756473771, "b_star": -36.8272165266137},
{ "name": "10B6/8","L_star": 61.229115169177, "a_star": -13.1121686242575, "b_star": -29.3638416126768},
{ "name": "10B6/6","L_star": 61.1767386762246, "a_star": -10.2640648304364, "b_star": -21.6364448973565},
{ "name": "10B6/4","L_star": 61.1330372293513, "a_star": -6.95840098117401, "b_star": -13.6980043573956},
{ "name": "10B6/3","L_star": 61.1067925736237, "a_star": -5.28861302613004, "b_star": -9.99596682816422},
{ "name": "10B6/2","L_star": 61.0892862062213, "a_star": -3.59077645123851, "b_star": -6.45778914231769},
{ "name": "10B6/1","L_star": 61.0630117371766, "a_star": -1.82645092549472, "b_star": -3.09204267318384},
{ "name": "10B5/18","L_star": 51.2326103063449, "a_star": -24.1890160862984, "b_star": -69.3834266176265},
{ "name": "10B5/16","L_star": 51.381914509339, "a_star": -23.116290584116, "b_star": -61.7441624223578},
{ "name": "10B5/14","L_star": 51.3475184265326, "a_star": -20.2540380415708, "b_star": -53.3738266852407},
{ "name": "10B5/12","L_star": 51.2901134222191, "a_star": -17.3865025537919, "b_star": -45.704402703449},
{ "name": "10B5/10","L_star": 51.2326103063449, "a_star": -14.6863074074642, "b_star": -37.8526499386415},
{ "name": "10B5/8","L_star": 51.186536890986, "a_star": -12.2723413637697, "b_star": -30.1101777208679},
{ "name": "10B5/6","L_star": 51.1288561132401, "a_star": -9.48040772422454, "b_star": -22.4059374861717},
{ "name": "10B5/4","L_star": 51.0826400186909, "a_star": -6.70454178152013, "b_star": -14.7862069003809},
{ "name": "10B5/3","L_star": 51.0595080720228, "a_star": -5.09263117281822, "b_star": -10.9556999152767},
{ "name": "10B5/2","L_star": 51.0363601557647, "a_star": -3.44321395020536, "b_star": -7.17621104765738},
{ "name": "10B5/1","L_star": 51.0247802004202, "a_star": -1.70053373606194, "b_star": -3.47380779081616},
{ "name": "10B4/16","L_star": 40.2954635231621, "a_star": -15.8487430258198, "b_star": -64.3452078147984},
{ "name": "10B4/14","L_star": 40.8159943055026, "a_star": -16.9849928985101, "b_star": -54.9314784610055},
{ "name": "10B4/12","L_star": 41.0247604338049, "a_star": -15.8156998310276, "b_star": -46.5214734534715},
{ "name": "10B4/10","L_star": 40.9606876153157, "a_star": -13.5772987675151, "b_star": -38.8905008072689},
{ "name": "10B4/8","L_star": 40.9125382368196, "a_star": -11.3659700741786, "b_star": -31.5264900840242},
{ "name": "10B4/6","L_star": 40.8643072491098, "a_star": -8.88797613555339, "b_star": -23.3301367021714},
{ "name": "10B4/4","L_star": 40.8159943055026, "a_star": -6.43287619378768, "b_star": -15.7679717105155},
{ "name": "10B4/3","L_star": 40.7998717182846, "a_star": -4.99209409557858, "b_star": -11.7783738540103},
{ "name": "10B4/2","L_star": 40.7837399731063, "a_star": -3.4243657778732, "b_star": -7.74858495591876},
{ "name": "10B4/1","L_star": 40.751448956775, "a_star": -1.7882159225682, "b_star": -3.76972346154414},
{ "name": "10B3/14","L_star": 29.4477580072064, "a_star": -8.61586290271049, "b_star": -59.6162265270691},
{ "name": "10B3/12","L_star": 30.0738799179759, "a_star": -11.1580643833929, "b_star": -50.310475351869},
{ "name": "10B3/10","L_star": 30.5733515477892, "a_star": -12.1554109988944, "b_star": -40.9137781384696},
{ "name": "10B3/8","L_star": 30.5661543058963, "a_star": -10.4212214449013, "b_star": -33.1931907109154},
{ "name": "10B3/6","L_star": 30.5157112023247, "a_star": -8.22464544997722, "b_star": -25.2543138360994},
{ "name": "10B3/4","L_star": 30.4675682252624, "a_star": -6.06382890317489, "b_star": -17.3889532086907},
{ "name": "10B3/3","L_star": 30.4458713030924, "a_star": -4.81452630895513, "b_star": -13.2257018983173},
{ "name": "10B3/2","L_star": 30.4217398126451, "a_star": -3.40507062514245, "b_star": -8.96331093544137},
{ "name": "10B3/1","L_star": 30.395166163107, "a_star": -1.78890551521571, "b_star": -4.56004671466461},
{ "name": "10B2/10","L_star": 19.7694814763849, "a_star": -6.24228603925403, "b_star": -43.6041174934318},
{ "name": "10B2/8","L_star": 20.3496712973095, "a_star": -8.46822941547662, "b_star": -33.9915007022034},
{ "name": "10B2/6","L_star": 20.3654156073658, "a_star": -7.06543788018157, "b_star": -25.0661420418891},
{ "name": "10B2/4","L_star": 20.3141966443599, "a_star": -4.99590986888015, "b_star": -16.6688202164436},
{ "name": "10B2/3","L_star": 20.2905083162782, "a_star": -3.9057036835346, "b_star": -12.5424760368617},
{ "name": "10B2/2","L_star": 20.2667890230397, "a_star": -2.71144661260839, "b_star": -8.42117651146271},
{ "name": "10B2/1","L_star": 20.2469992182683, "a_star": -1.39748412737864, "b_star": -4.27037164687198},
{ "name": "10B1/8","L_star": 9.90715476420058, "a_star": -2.94862305803882, "b_star": -35.3446599508175},
{ "name": "10B1/6","L_star": 10.4462504228925, "a_star": -4.99379140658857, "b_star": -25.7045253032911},
{ "name": "10B1/4","L_star": 10.4833940818493, "a_star": -4.22654660335207, "b_star": -17.5758159265094},
{ "name": "10B1/3","L_star": 10.4611204009613, "a_star": -3.51676860540823, "b_star": -13.6356766019612},
{ "name": "10B1/2","L_star": 10.4388091590155, "a_star": -2.56886816632237, "b_star": -9.50180098156169},
{ "name": "10B1/1","L_star": 10.4090021335651, "a_star": -1.34548845025549, "b_star": -5.04137957907887},
{ "name": "2.5PB9/2","L_star": 90.2064234313182, "a_star": -1.84582468028943, "b_star": -5.43447598373028},
{ "name": "2.5PB9/1","L_star": 90.192583657454, "a_star": -1.00016952252119, "b_star": -2.61961514263405},
{ "name": "2.5PB8/6","L_star": 80.6218957120763, "a_star": -6.4389275186893, "b_star": -22.8112658336963},
{ "name": "2.5PB8/4","L_star": 80.5884451656218, "a_star": -4.12772687061536, "b_star": -14.1528558528779},
{ "name": "2.5PB8/3","L_star": 80.5661319230916, "a_star": -2.90231276718472, "b_star": -9.7575510257305},
{ "name": "2.5PB8/2","L_star": 80.5493902216599, "a_star": -1.85909421473218, "b_star": -5.70463679096034},
{ "name": "2.5PB8/1","L_star": 80.5382258609889, "a_star": -0.904207786188826, "b_star": -2.28576736933923},
{ "name": "2.5PB7/10","L_star": 71.0292050098948, "a_star": -8.36852170747698, "b_star": -39.3458748130936},
{ "name": "2.5PB7/8","L_star": 70.9879687066308, "a_star": -7.00901581864755, "b_star": -30.1998996149645},
{ "name": "2.5PB7/6","L_star": 70.9535752313128, "a_star": -5.60107836295792, "b_star": -21.880683616464},
{ "name": "2.5PB7/4","L_star": 70.9191545265718, "a_star": -3.73261878861647, "b_star": -14.0702325865209},
{ "name": "2.5PB7/3","L_star": 70.905378608356, "a_star": -2.8230340605393, "b_star": -9.99451694950237},
{ "name": "2.5PB7/2","L_star": 70.8915983213402, "a_star": -1.85645795874451, "b_star": -6.12479740302434},
{ "name": "2.5PB7/1","L_star": 70.8709196917347, "a_star": -0.965570813984518, "b_star": -2.65777411672983},
{ "name": "2.5PB6/14","L_star": 61.2814207152509, "a_star": -8.89770070931572, "b_star": -55.8194265956019},
{ "name": "2.5PB6/12","L_star": 61.2378376805646, "a_star": -8.13621543430021, "b_star": -47.2645436778013},
{ "name": "2.5PB6/10","L_star": 61.2029358060968, "a_star": -7.03716778101682, "b_star": -38.490369253942},
{ "name": "2.5PB6/8","L_star": 61.1680023462115, "a_star": -6.00266942661798, "b_star": -30.5508522525453},
{ "name": "2.5PB6/6","L_star": 61.1330372293513, "a_star": -5.10615037898515, "b_star": -22.5104086560583},
{ "name": "2.5PB6/4","L_star": 61.098040383699, "a_star": -3.52989077107457, "b_star": -14.2238804115196},
{ "name": "2.5PB6/3","L_star": 61.0892862062213, "a_star": -2.72602731336674, "b_star": -10.3704320413245},
{ "name": "2.5PB6/2","L_star": 61.0717718840904, "a_star": -1.76574924378775, "b_star": -6.7501047196475},
{ "name": "2.5PB6/1","L_star": 61.0542495981883, "a_star": -0.940618274632377, "b_star": -3.24871426144644},
{ "name": "2.5PB5/18","L_star": 51.2556233529158, "a_star": -7.11268481503768, "b_star": -72.3276507459089},
{ "name": "2.5PB5/16","L_star": 51.2671239727284, "a_star": -7.13846043723265, "b_star": -63.6948639561942},
{ "name": "2.5PB5/14","L_star": 51.2326103063449, "a_star": -7.15458088108706, "b_star": -55.009539626524},
{ "name": "2.5PB5/12","L_star": 51.186536890986, "a_star": -6.77708956464001, "b_star": -47.2355598750027},
{ "name": "2.5PB5/10","L_star": 51.1519403159239, "a_star": -6.03399396415777, "b_star": -39.284660118033},
{ "name": "2.5PB5/8","L_star": 51.1173080558794, "a_star": -5.3910428851125, "b_star": -31.3501912160669},
{ "name": "2.5PB5/6","L_star": 51.0942000120299, "a_star": -4.45296446943788, "b_star": -23.4588853933712},
{ "name": "2.5PB5/4","L_star": 51.0595080720228, "a_star": -3.34572127255978, "b_star": -15.2318396682682},
{ "name": "2.5PB5/3","L_star": 51.0479361118152, "a_star": -2.57740285040731, "b_star": -11.3319343059468},
{ "name": "2.5PB5/2","L_star": 51.0247802004202, "a_star": -1.70540364601962, "b_star": -7.49932201810235},
{ "name": "2.5PB5/1","L_star": 51.0131962423256, "a_star": -0.867796628376383, "b_star": -3.72429013319009},
{ "name": "2.5PB4/18","L_star": 40.2132562456911, "a_star": 0.5026866416597, "b_star": -74.0078824788313},
{ "name": "2.5PB4/16","L_star": 40.6705602368878, "a_star": -2.79784758983367, "b_star": -64.8658377100238},
{ "name": "2.5PB4/14","L_star": 40.9125382368196, "a_star": -4.67711404707352, "b_star": -56.0778483596326},
{ "name": "2.5PB4/12","L_star": 40.9285970802895, "a_star": -5.06220868610444, "b_star": -47.7161061367592},
{ "name": "2.5PB4/10","L_star": 40.8964703256786, "a_star": -4.88683920250954, "b_star": -40.1153165468295},
{ "name": "2.5PB4/8","L_star": 40.8643072491098, "a_star": -4.52414334259971, "b_star": -32.6471935637713},
{ "name": "2.5PB4/6","L_star": 40.8159943055026, "a_star": -3.75459365844513, "b_star": -24.3223187933791},
{ "name": "2.5PB4/4","L_star": 40.7998717182846, "a_star": -2.90151306753725, "b_star": -16.479909725871},
{ "name": "2.5PB4/3","L_star": 40.7837399731063, "a_star": -2.28174852535926, "b_star": -12.3780991819012},
{ "name": "2.5PB4/2","L_star": 40.7675990569515, "a_star": -1.56863834764509, "b_star": -8.20771632234751},
{ "name": "2.5PB4/1","L_star": 40.751448956775, "a_star": -0.838957374827975, "b_star": -4.04731770635181},
{ "name": "2.5PB3/14","L_star": 29.8497504809689, "a_star": 2.41644115724157, "b_star": -60.0368629970305},
{ "name": "2.5PB3/12","L_star": 30.3854954663885, "a_star": -1.46143887993552, "b_star": -50.7770478206676},
{ "name": "2.5PB3/10","L_star": 30.5397453655359, "a_star": -2.86953161987463, "b_star": -42.313240879444},
{ "name": "2.5PB3/8","L_star": 30.5036848014572, "a_star": -3.03488349235428, "b_star": -34.4036749240582},
{ "name": "2.5PB3/6","L_star": 30.4699777439717, "a_star": -2.90197638453468, "b_star": -26.3895620815667},
{ "name": "2.5PB3/4","L_star": 30.4362217165054, "a_star": -2.37230195221294, "b_star": -18.2060491933883},
{ "name": "2.5PB3/3","L_star": 30.4217398126451, "a_star": -2.00094549766039, "b_star": -13.9537113378501},
{ "name": "2.5PB3/2","L_star": 30.4048328299485, "a_star": -1.48204788339601, "b_star": -9.54409064534042},
{ "name": "2.5PB3/1","L_star": 30.3879135185988, "a_star": -0.823852870518949, "b_star": -4.89884675437318},
{ "name": "2.5PB2/12","L_star": 19.5814327283453, "a_star": 6.54924831481796, "b_star": -53.5117586858434},
{ "name": "2.5PB2/10","L_star": 20.0958854537657, "a_star": 1.71367377467013, "b_star": -44.0896537574036},
{ "name": "2.5PB2/8","L_star": 20.3575451568209, "a_star": -1.17835915868292, "b_star": -35.0215829370378},
{ "name": "2.5PB2/6","L_star": 20.3181416956288, "a_star": -1.57631858634211, "b_star": -26.2028874811119},
{ "name": "2.5PB2/4","L_star": 20.2865572538282, "a_star": -1.63041734559186, "b_star": -17.4586383268098},
{ "name": "2.5PB2/3","L_star": 20.2707443932339, "a_star": -1.38144746867783, "b_star": -13.2133352237206},
{ "name": "2.5PB2/2","L_star": 20.2549177328224, "a_star": -1.03629025950835, "b_star": -8.95024580707153},
{ "name": "2.5PB2/1","L_star": 20.2390772424465, "a_star": -0.547336756888478, "b_star": -4.5846285808545},
{ "name": "2.5PB1/8","L_star": 10.1680856089276, "a_star": 4.2111133199231, "b_star": -35.8156034048059},
{ "name": "2.5PB1/6","L_star": 10.4833940818493, "a_star": 0.238008748332116, "b_star": -26.3412132445621},
{ "name": "2.5PB1/4","L_star": 10.4536875015812, "a_star": -0.573741669529379, "b_star": -18.1342265506465},
{ "name": "2.5PB1/3","L_star": 10.4388091590155, "a_star": -0.695604827751287, "b_star": -14.2615275318362},
{ "name": "2.5PB1/2","L_star": 10.4239140521113, "a_star": -0.594065950696554, "b_star": -10.1152979433292},
{ "name": "2.5PB1/1","L_star": 10.4090021335651, "a_star": -0.283852231141371, "b_star": -5.47752642119739},
{ "name": "5PB9/2","L_star": 90.1971973162189, "a_star": -0.652857282537833, "b_star": -5.40620365345903},
{ "name": "5PB9/1","L_star": 90.1879695977629, "a_star": -0.365606142927299, "b_star": -2.61187193659178},
{ "name": "5PB8/6","L_star": 80.582867821507, "a_star": -1.50707812119022, "b_star": -23.6105290823383},
{ "name": "5PB8/4","L_star": 80.5605520009757, "a_star": -1.1946824414249, "b_star": -14.2471815208397},
{ "name": "5PB8/3","L_star": 80.5493902216599, "a_star": -0.850246084999839, "b_star": -9.78658111025565},
{ "name": "5PB8/2","L_star": 80.5438083640871, "a_star": -0.591892162515095, "b_star": -5.7650814278019},
{ "name": "5PB8/1","L_star": 80.5326427121788, "a_star": -0.332362764842997, "b_star": -2.3743829119254},
{ "name": "5PB7/10","L_star": 70.9604561030066, "a_star": -0.596769726255886, "b_star": -39.9967560756201},
{ "name": "5PB7/8","L_star": 70.939810219965, "a_star": -0.882134663172551, "b_star": -30.8173718761386},
{ "name": "5PB7/6","L_star": 70.9191545265718, "a_star": -1.00517591652433, "b_star": -22.3639291854341},
{ "name": "5PB7/4","L_star": 70.8984890111646, "a_star": -0.803712736530127, "b_star": -14.1748876534399},
{ "name": "5PB7/3","L_star": 70.8847065384495, "a_star": -0.582803621041461, "b_star": -10.1253592758997},
{ "name": "5PB7/2","L_star": 70.877813662059, "a_star": -0.383900798162473, "b_star": -6.29079143219446},
{ "name": "5PB7/1","L_star": 70.8709196917347, "a_star": -0.212691942243104, "b_star": -2.82886136674971},
{ "name": "5PB6/14","L_star": 61.1854730287976, "a_star": 1.64607609329082, "b_star": -56.7507174870379},
{ "name": "5PB6/12","L_star": 61.1592640376388, "a_star": 0.743908575799257, "b_star": -47.7621240035751},
{ "name": "5PB6/10","L_star": 61.1330372293513, "a_star": 0.357865470439922, "b_star": -38.8480657487444},
{ "name": "5PB6/8","L_star": 61.1155427771229, "a_star": 0.230902978871295, "b_star": -31.0569783003011},
{ "name": "5PB6/6","L_star": 61.098040383699, "a_star": -0.191582523320855, "b_star": -23.0758379328869},
{ "name": "5PB6/4","L_star": 61.0717718840904, "a_star": -0.383299418780125, "b_star": -14.4165069909475},
{ "name": "5PB6/3","L_star": 61.0630117371766, "a_star": -0.294863640712073, "b_star": -10.5220341278053},
{ "name": "5PB6/2","L_star": 61.0630117371766, "a_star": -0.255452418818358, "b_star": -6.80482557337934},
{ "name": "5PB6/1","L_star": 61.0542495981883, "a_star": -0.154972871939829, "b_star": -3.28942233854774},
{ "name": "5PB5/18","L_star": 51.1404001987838, "a_star": 7.32054850469049, "b_star": -73.5054068298796},
{ "name": "5PB5/16","L_star": 51.163476468069, "a_star": 4.70564625811926, "b_star": -64.3433905080224},
{ "name": "5PB5/14","L_star": 51.1404001987838, "a_star": 2.98336762918938, "b_star": -55.5972362912842},
{ "name": "5PB5/12","L_star": 51.1173080558794, "a_star": 1.9513696677585, "b_star": -47.5783686953723},
{ "name": "5PB5/10","L_star": 51.0942000120299, "a_star": 1.37943003136365, "b_star": -39.632355715147},
{ "name": "5PB5/8","L_star": 51.0710760398342, "a_star": 0.794722820829263, "b_star": -31.6909315786052},
{ "name": "5PB5/6","L_star": 51.0479361118152, "a_star": 0.513346372257861, "b_star": -23.8423917293914},
{ "name": "5PB5/4","L_star": 51.0363601557647, "a_star": 0.176172838651012, "b_star": -15.5079630918865},
{ "name": "5PB5/3","L_star": 51.0247802004202, "a_star": 0.162141823852324, "b_star": -11.5153818317595},
{ "name": "5PB5/2","L_star": 51.0131962423256, "a_star": 0.147594261774686, "b_star": -7.62660939087474},
{ "name": "5PB5/1","L_star": 51.0016082780202, "a_star": 0.103764811081264, "b_star": -3.76728819065208},
{ "name": "5PB4/20","L_star": 40.1638162410762, "a_star": 18.5323994039294, "b_star": -84.1289985389187},
{ "name": "5PB4/18","L_star": 40.5406559183046, "a_star": 13.0024549380462, "b_star": -73.933365829697},
{ "name": "5PB4/16","L_star": 40.7352896595015, "a_star": 8.8888033995147, "b_star": -64.8056479634853},
{ "name": "5PB4/14","L_star": 40.8643072491098, "a_star": 5.74016030232899, "b_star": -56.323903950261},
{ "name": "5PB4/12","L_star": 40.848212057973, "a_star": 3.99910333485196, "b_star": -48.1804091001377},
{ "name": "5PB4/10","L_star": 40.8321077477464, "a_star": 2.70095169826173, "b_star": -40.7521039439717},
{ "name": "5PB4/8","L_star": 40.7998717182846, "a_star": 1.80098529406114, "b_star": -33.1457232413912},
{ "name": "5PB4/6","L_star": 40.7837399731063, "a_star": 1.14525182718075, "b_star": -24.7789676537829},
{ "name": "5PB4/4","L_star": 40.7675990569515, "a_star": 0.842848035904259, "b_star": -16.7401917778623},
{ "name": "5PB4/3","L_star": 40.751448956775, "a_star": 0.645078052431858, "b_star": -12.7303262731481},
{ "name": "5PB4/2","L_star": 40.751448956775, "a_star": 0.444375699894362, "b_star": -8.66007237963768},
{ "name": "5PB4/1","L_star": 40.7352896595015, "a_star": 0.219722745613243, "b_star": -4.40443355534732},
{ "name": "5PB3/18","L_star": 29.3973221007118, "a_star": 25.3089745914848, "b_star": -78.6703197267274},
{ "name": "5PB3/16","L_star": 29.8101159046378, "a_star": 17.900176228718, "b_star": -69.0176543121341},
{ "name": "5PB3/14","L_star": 30.247247679603, "a_star": 11.7273978357894, "b_star": -59.4246857943931},
{ "name": "5PB3/12","L_star": 30.4531058626024, "a_star": 7.56085377064519, "b_star": -51.0205477721227},
{ "name": "5PB3/10","L_star": 30.4699777439717, "a_star": 5.06553999989967, "b_star": -42.7506477310963},
{ "name": "5PB3/8","L_star": 30.4482830733378, "a_star": 3.20819065316655, "b_star": -34.5948986937842},
{ "name": "5PB3/6","L_star": 30.4289818938566, "a_star": 2.0732354732223, "b_star": -26.9456306866252},
{ "name": "5PB3/4","L_star": 30.4096646534768, "a_star": 1.27443090509879, "b_star": -18.6087837272192},
{ "name": "5PB3/3","L_star": 30.4, "a_star": 0.919784547398161, "b_star": -14.3392035701952},
{ "name": "5PB3/2","L_star": 30.3903313187449, "a_star": 0.596725358582295, "b_star": -9.91002855968102},
{ "name": "5PB3/1","L_star": 30.3830771620482, "a_star": 0.311254140180628, "b_star": -5.14509038355216},
{ "name": "5PB2/14","L_star": 19.5814327283453, "a_star": 22.21314031728, "b_star": -62.3947083713113},
{ "name": "5PB2/12","L_star": 19.9515484820954, "a_star": 14.2687596192143, "b_star": -52.5896533847919},
{ "name": "5PB2/10","L_star": 20.262832789887, "a_star": 8.86554592132024, "b_star": -43.9624754236289},
{ "name": "5PB2/8","L_star": 20.3023563446473, "a_star": 5.3383600588657, "b_star": -35.4029118657701},
{ "name": "5PB2/6","L_star": 20.2786525466281, "a_star": 3.21726625619742, "b_star": -26.6426734908257},
{ "name": "5PB2/4","L_star": 20.258875693305, "a_star": 1.84395895449113, "b_star": -17.9481364884969},
{ "name": "5PB2/3","L_star": 20.2509589079675, "a_star": 1.36020797962291, "b_star": -13.7440879309852},
{ "name": "5PB2/2","L_star": 20.2430386632522, "a_star": 0.958489112111965, "b_star": -9.45099742632757},
{ "name": "5PB2/1","L_star": 20.2351149553776, "a_star": 0.507285603083246, "b_star": -4.91643121358483},
{ "name": "5PB1/10","L_star": 9.99984662631419, "a_star": 17.1814126880452, "b_star": -45.3977397773465},
{ "name": "5PB1/8","L_star": 10.3791276712618, "a_star": 9.18277869377415, "b_star": -35.4074210224242},
{ "name": "5PB1/6","L_star": 10.4462504228925, "a_star": 5.02134238203725, "b_star": -26.7447326324221},
{ "name": "5PB1/4","L_star": 10.4239140521113, "a_star": 2.79483801463347, "b_star": -18.3987553220894},
{ "name": "5PB1/3","L_star": 10.4164601972582, "a_star": 2.04946231155514, "b_star": -14.6156127708007},
{ "name": "5PB1/2","L_star": 10.4090021335651, "a_star": 1.36862991751201, "b_star": -10.4826385444143},
{ "name": "5PB1/1","L_star": 10.4015398550854, "a_star": 0.7897209494208, "b_star": -5.74586208432908},
{ "name": "7.5PB9/2","L_star": 90.1879695977629, "a_star": 0.842816519544209, "b_star": -5.2250760259714},
{ "name": "7.5PB9/1","L_star": 90.1879695977629, "a_star": 0.370068140418456, "b_star": -2.51779798010683},
{ "name": "7.5PB8/6","L_star": 80.5438083640871, "a_star": 4.45580266586837, "b_star": -23.2426039877948},
{ "name": "7.5PB8/4","L_star": 80.5382258609889, "a_star": 2.56210798046946, "b_star": -13.9980851755159},
{ "name": "7.5PB8/3","L_star": 80.5326427121788, "a_star": 1.76839529023132, "b_star": -9.54104624204921},
{ "name": "7.5PB8/2","L_star": 80.5326427121788, "a_star": 1.00497405710637, "b_star": -5.52715466150346},
{ "name": "7.5PB8/1","L_star": 80.5270589174699, "a_star": 0.428430334306673, "b_star": -2.2101435318765},
{ "name": "7.5PB7/10","L_star": 70.8847065384495, "a_star": 10.7697118889724, "b_star": -39.946019952287},
{ "name": "7.5PB7/8","L_star": 70.877813662059, "a_star": 7.53801399603721, "b_star": -30.88035909512},
{ "name": "7.5PB7/6","L_star": 70.877813662059, "a_star": 5.00470454970359, "b_star": -22.3388179186484},
{ "name": "7.5PB7/4","L_star": 70.8709196917347, "a_star": 2.86978066584254, "b_star": -14.0588114765993},
{ "name": "7.5PB7/3","L_star": 70.8709196917347, "a_star": 2.04706127502119, "b_star": -10.007306870648},
{ "name": "7.5PB7/2","L_star": 70.8640246270425, "a_star": 1.30730434290044, "b_star": -6.17282773461363},
{ "name": "7.5PB7/1","L_star": 70.8640246270425, "a_star": 0.591670045248849, "b_star": -2.77046738411841},
{ "name": "7.5PB6/14","L_star": 61.0630117371766, "a_star": 18.1266519992412, "b_star": -55.4157256887528},
{ "name": "7.5PB6/12","L_star": 61.0630117371766, "a_star": 14.3042582911225, "b_star": -46.6389768119409},
{ "name": "7.5PB6/10","L_star": 61.0630117371766, "a_star": 10.8405610227868, "b_star": -37.7718791822771},
{ "name": "7.5PB6/8","L_star": 61.0542495981883, "a_star": 8.15443603284, "b_star": -30.1984321644095},
{ "name": "7.5PB6/6","L_star": 61.0542495981883, "a_star": 5.80023587379253, "b_star": -22.8333516473029},
{ "name": "7.5PB6/4","L_star": 61.0542495981883, "a_star": 3.39519874831457, "b_star": -14.4707407208565},
{ "name": "7.5PB6/3","L_star": 61.0454854659928, "a_star": 2.44352201552323, "b_star": -10.5835860461925},
{ "name": "7.5PB6/2","L_star": 61.0454854659928, "a_star": 1.58567046629193, "b_star": -6.86871066504116},
{ "name": "7.5PB6/1","L_star": 61.0454854659928, "a_star": 0.774401204182873, "b_star": -3.31991493176946},
{ "name": "7.5PB5/20","L_star": 50.8622389556151, "a_star": 36.3197514532774, "b_star": -78.7181649784542},
{ "name": "7.5PB5/18","L_star": 51.0131962423256, "a_star": 29.9496102160266, "b_star": -70.5227548150614},
{ "name": "7.5PB5/16","L_star": 51.0131962423256, "a_star": 24.2187834453705, "b_star": -62.2842789058301},
{ "name": "7.5PB5/14","L_star": 51.0131962423256, "a_star": 19.8341798152641, "b_star": -54.5389005115505},
{ "name": "7.5PB5/12","L_star": 51.0131962423256, "a_star": 15.9648567885781, "b_star": -46.6282628814516},
{ "name": "7.5PB5/10","L_star": 51.0131962423256, "a_star": 12.2383293568347, "b_star": -38.594400152173},
{ "name": "7.5PB5/8","L_star": 51.0131962423256, "a_star": 9.52140379813832, "b_star": -31.3697039698601},
{ "name": "7.5PB5/6","L_star": 51.0131962423256, "a_star": 6.57694092491251, "b_star": -23.7033329849639},
{ "name": "7.5PB5/4","L_star": 51.0016082780202, "a_star": 3.88011315785297, "b_star": -15.3497182839105},
{ "name": "7.5PB5/3","L_star": 51.0016082780202, "a_star": 2.83948276516799, "b_star": -11.4453870182773},
{ "name": "7.5PB5/2","L_star": 51.0016082780202, "a_star": 1.93980401658989, "b_star": -7.63989984603679},
{ "name": "7.5PB5/1","L_star": 51.0016082780202, "a_star": 0.981257619252884, "b_star": -3.80910236570529},
{ "name": "7.5PB4/26","L_star": 39.8986541151014, "a_star": 64.4731328390925, "b_star": -96.5267763303005},
{ "name": "7.5PB4/24","L_star": 40.0977605303268, "a_star": 56.7942412957068, "b_star": -89.9228644828539},
{ "name": "7.5PB4/22","L_star": 40.2626094371923, "a_star": 49.683903097279, "b_star": -83.4305137020288},
{ "name": "7.5PB4/20","L_star": 40.4101519206937, "a_star": 43.6491879720535, "b_star": -77.8214046017358},
{ "name": "7.5PB4/18","L_star": 40.6056826069156, "a_star": 35.3427654634107, "b_star": -69.5067821645425},
{ "name": "7.5PB4/16","L_star": 40.7352896595015, "a_star": 28.5786759462773, "b_star": -62.0373496690418},
{ "name": "7.5PB4/14","L_star": 40.7352896595015, "a_star": 22.8713945624153, "b_star": -54.5313943922205},
{ "name": "7.5PB4/12","L_star": 40.7352896595015, "a_star": 18.2693863710026, "b_star": -47.41041704269},
{ "name": "7.5PB4/10","L_star": 40.7352896595015, "a_star": 14.4890313338196, "b_star": -40.1295116961949},
{ "name": "7.5PB4/8","L_star": 40.7352896595015, "a_star": 11.0157897548367, "b_star": -32.4618664227343},
{ "name": "7.5PB4/6","L_star": 40.7352896595015, "a_star": 7.72672208493541, "b_star": -24.5482200095026},
{ "name": "7.5PB4/4","L_star": 40.7352896595015, "a_star": 5.01612448888397, "b_star": -16.7385425964749},
{ "name": "7.5PB4/3","L_star": 40.7352896595015, "a_star": 3.76216528191373, "b_star": -12.8343552560707},
{ "name": "7.5PB4/2","L_star": 40.7352896595015, "a_star": 2.48013910682465, "b_star": -8.76978465350085},
{ "name": "7.5PB4/1","L_star": 40.7352896595015, "a_star": 1.21352888015758, "b_star": -4.50683890716348},
{ "name": "7.5PB3/34","L_star": 29.3189242518151, "a_star": 100.04130516492, "b_star": -113.884301368809},
{ "name": "7.5PB3/32","L_star": 29.4074182474031, "a_star": 93.9788930713299, "b_star": -109.533808290852},
{ "name": "7.5PB3/30","L_star": 29.5056212545388, "a_star": 86.1183503522673, "b_star": -103.655038654689},
{ "name": "7.5PB3/28","L_star": 29.6309057952884, "a_star": 78.2449501128158, "b_star": -97.6113779889464},
{ "name": "7.5PB3/26","L_star": 29.7157078923625, "a_star": 71.8909882026925, "b_star": -92.8220440617617},
{ "name": "7.5PB3/24","L_star": 29.8125950753269, "a_star": 65.1183217141362, "b_star": -87.4441023220694},
{ "name": "7.5PB3/22","L_star": 29.9214139997679, "a_star": 57.5416886476373, "b_star": -81.4429980599176},
{ "name": "7.5PB3/20","L_star": 30.0297196226254, "a_star": 49.7949537071438, "b_star": -75.0442343376788},
{ "name": "7.5PB3/18","L_star": 30.1057208955728, "a_star": 43.7690763888483, "b_star": -69.7269807538721},
{ "name": "7.5PB3/16","L_star": 30.2180373464382, "a_star": 37.042778552287, "b_star": -63.7476743028872},
{ "name": "7.5PB3/14","L_star": 30.3104104643861, "a_star": 29.6859358926918, "b_star": -56.4328434626176},
{ "name": "7.5PB3/12","L_star": 30.3661419665029, "a_star": 23.1494395250107, "b_star": -49.0129818408605},
{ "name": "7.5PB3/10","L_star": 30.3734014220727, "a_star": 18.0225474011765, "b_star": -41.9367717319043},
{ "name": "7.5PB3/8","L_star": 30.3758207355903, "a_star": 13.5570715875679, "b_star": -34.2456040074151},
{ "name": "7.5PB3/6","L_star": 30.3758207355903, "a_star": 9.73262759633986, "b_star": -26.828789635147},
{ "name": "7.5PB3/4","L_star": 30.3758207355903, "a_star": 6.20118041109335, "b_star": -18.6865396238962},
{ "name": "7.5PB3/3","L_star": 30.3758207355903, "a_star": 4.64731848170247, "b_star": -14.4762979454867},
{ "name": "7.5PB3/2","L_star": 30.3758207355903, "a_star": 3.19788810200122, "b_star": -10.0419382399942},
{ "name": "7.5PB3/1","L_star": 30.3758207355903, "a_star": 1.63684572290526, "b_star": -5.24292898834493},
{ "name": "7.5PB2/38","L_star": 19.4701240779965, "a_star": 123.717093201169, "b_star": -119.543017020237},
{ "name": "7.5PB2/36","L_star": 19.5485250261232, "a_star": 114.808335262738, "b_star": -113.276065642896},
{ "name": "7.5PB2/34","L_star": 19.6388751779558, "a_star": 106.610802827275, "b_star": -107.678228498549},
{ "name": "7.5PB2/32","L_star": 19.7042977666094, "a_star": 98.5954737915805, "b_star": -102.045117127949},
{ "name": "7.5PB2/30","L_star": 19.7654144517363, "a_star": 90.9637563950929, "b_star": -96.6661804519519},
{ "name": "7.5PB2/28","L_star": 19.8222688454216, "a_star": 83.4334816273124, "b_star": -91.3087695045222},
{ "name": "7.5PB2/26","L_star": 19.8465798599047, "a_star": 76.9932910843791, "b_star": -86.6999982140108},
{ "name": "7.5PB2/24","L_star": 19.9273793665876, "a_star": 70.0696864815614, "b_star": -81.7708794436358},
{ "name": "7.5PB2/22","L_star": 19.9756851449229, "a_star": 63.1500474616451, "b_star": -76.6621039278833},
{ "name": "7.5PB2/20","L_star": 20.0158410942582, "a_star": 55.5756933182656, "b_star": -70.9186386189399},
{ "name": "7.5PB2/18","L_star": 20.0519050446824, "a_star": 48.7615205804366, "b_star": -65.6661843746234},
{ "name": "7.5PB2/16","L_star": 20.091891662468, "a_star": 41.6850854269106, "b_star": -59.8527964147021},
{ "name": "7.5PB2/14","L_star": 20.1317898727821, "a_star": 34.5191873112684, "b_star": -53.8163315410899},
{ "name": "7.5PB2/12","L_star": 20.1716001626663, "a_star": 27.8551944309953, "b_star": -47.6359103744898},
{ "name": "7.5PB2/10","L_star": 20.211323014883, "a_star": 21.7084188322228, "b_star": -41.3893064459748},
{ "name": "7.5PB2/8","L_star": 20.2232228918536, "a_star": 15.6542049158075, "b_star": -33.8405013916368},
{ "name": "7.5PB2/6","L_star": 20.2271877805556, "a_star": 10.8320489132699, "b_star": -25.9626804987743},
{ "name": "7.5PB2/4","L_star": 20.2271877805556, "a_star": 6.71286461446707, "b_star": -17.9651064402031},
{ "name": "7.5PB2/3","L_star": 20.2271877805556, "a_star": 4.98440459003466, "b_star": -13.9356984796031},
{ "name": "7.5PB2/2","L_star": 20.2271877805556, "a_star": 3.34844710118962, "b_star": -9.68799786862193},
{ "name": "7.5PB2/1","L_star": 20.2311518015721, "a_star": 1.71912129355831, "b_star": -5.09765402956123},
{ "name": "7.5PB1/38","L_star": 9.57739149763381, "a_star": 145.256748710297, "b_star": -121.426282182046},
{ "name": "7.5PB1/36","L_star": 9.71188690427136, "a_star": 134.076979739799, "b_star": -114.109077136403},
{ "name": "7.5PB1/34","L_star": 9.80598430980695, "a_star": 124.560628897926, "b_star": -107.842497269215},
{ "name": "7.5PB1/32","L_star": 9.88387788305692, "a_star": 115.200676723306, "b_star": -101.727737509952},
{ "name": "7.5PB1/30","L_star": 9.96902283140891, "a_star": 104.069159449523, "b_star": -94.3270833840579},
{ "name": "7.5PB1/28","L_star": 10.0229165999318, "a_star": 94.0432723569452, "b_star": -87.6496217081782},
{ "name": "7.5PB1/26","L_star": 10.0842373959684, "a_star": 84.9530206165785, "b_star": -81.5544173488074},
{ "name": "7.5PB1/24","L_star": 10.1376575749622, "a_star": 76.6281039178779, "b_star": -75.9480352059791},
{ "name": "7.5PB1/22","L_star": 10.1832731257492, "a_star": 68.5042657075015, "b_star": -70.3948385162069},
{ "name": "7.5PB1/20","L_star": 10.2211650355277, "a_star": 61.5613028419284, "b_star": -65.5874378424838},
{ "name": "7.5PB1/18","L_star": 10.2513999023643, "a_star": 54.2410835447664, "b_star": -60.4524064014341},
{ "name": "7.5PB1/16","L_star": 10.2740304307288, "a_star": 48.1804472115278, "b_star": -56.0497526798453},
{ "name": "7.5PB1/14","L_star": 10.3041439590119, "a_star": 40.1393874438424, "b_star": -50.1016394723703},
{ "name": "7.5PB1/12","L_star": 10.3266839391574, "a_star": 33.1612438272818, "b_star": -44.7284791944508},
{ "name": "7.5PB1/10","L_star": 10.3491853893754, "a_star": 26.4499946258627, "b_star": -39.167657111018},
{ "name": "7.5PB1/8","L_star": 10.3716484739028, "a_star": 19.9960531621153, "b_star": -33.2184946934289},
{ "name": "7.5PB1/6","L_star": 10.3866026299139, "a_star": 13.5499901256873, "b_star": -26.1860183231945},
{ "name": "7.5PB1/4","L_star": 10.3940733558595, "a_star": 8.19452084356465, "b_star": -18.3586848520386},
{ "name": "7.5PB1/3","L_star": 10.3940733558595, "a_star": 6.26490072770883, "b_star": -14.8017930886585},
{ "name": "7.5PB1/2","L_star": 10.3940733558595, "a_star": 4.34076897952763, "b_star": -10.743908975545},
{ "name": "7.5PB1/1","L_star": 10.4015398550854, "a_star": 2.31597289352629, "b_star": -5.93801580468131},
{ "name": "10PB9/4","L_star": 90.1787402752534, "a_star": 5.62219119189367, "b_star": -14.4356465280517},
{ "name": "10PB9/3","L_star": 90.6748454214823, "a_star": 18.7946235273629, "b_star": -19.9992358014608},
{ "name": "10PB9/2","L_star": 90.1833551370584, "a_star": 1.99697261628157, "b_star": -4.78836890034244},
{ "name": "10PB9/1","L_star": 90.1833551370584, "a_star": 0.838978004266355, "b_star": -1.78533517087232},
{ "name": "10PB8/8","L_star": 80.6664604440045, "a_star": 13.9626082003831, "b_star": -31.4138457950675},
{ "name": "10PB8/6","L_star": 80.5158893896084, "a_star": 9.27685282884511, "b_star": -22.0410962869319},
{ "name": "10PB8/4","L_star": 80.5214744766754, "a_star": 5.25278771766124, "b_star": -13.1775162586053},
{ "name": "10PB8/3","L_star": 80.5214744766754, "a_star": 3.5950848897997, "b_star": -8.91174103160457},
{ "name": "10PB8/2","L_star": 80.5270589174699, "a_star": 2.14845593583673, "b_star": -5.149214272744},
{ "name": "10PB8/1","L_star": 80.5270589174699, "a_star": 0.965134183908201, "b_star": -2.07421688796694},
{ "name": "10PB7/12","L_star": 70.905378608356, "a_star": 23.6943461791955, "b_star": -46.1841412522048},
{ "name": "10PB7/10","L_star": 70.8364334159028, "a_star": 18.2490872980115, "b_star": -37.5207094302107},
{ "name": "10PB7/8","L_star": 70.8433328624133, "a_star": 13.610607698277, "b_star": -29.1656287709086},
{ "name": "10PB7/6","L_star": 70.8502312128167, "a_star": 9.37880485017784, "b_star": -21.2616755872786},
{ "name": "10PB7/4","L_star": 70.8571284675481, "a_star": 5.70159525784214, "b_star": -13.6540293532829},
{ "name": "10PB7/3","L_star": 70.8571284675481, "a_star": 3.95322251295654, "b_star": -9.71655636755107},
{ "name": "10PB7/2","L_star": 70.8640246270425, "a_star": 2.42188597908838, "b_star": -5.9824183793036},
{ "name": "10PB7/1","L_star": 70.8640246270425, "a_star": 1.07717375065597, "b_star": -2.64888886539896},
{ "name": "10PB6/16","L_star": 61.0016348671828, "a_star": 35.3654645159701, "b_star": -59.9975159972266},
{ "name": "10PB6/14","L_star": 60.9928587518952, "a_star": 28.7163458556242, "b_star": -51.2730834651486},
{ "name": "10PB6/12","L_star": 61.0016348671828, "a_star": 23.4766868278278, "b_star": -43.690880264682},
{ "name": "10PB6/10","L_star": 61.0104089824442, "a_star": 18.5371358070413, "b_star": -35.853770121303},
{ "name": "10PB6/8","L_star": 61.0191810988183, "a_star": 14.0857085567919, "b_star": -28.6573067761911},
{ "name": "10PB6/6","L_star": 61.0279512174433, "a_star": 10.2602648180854, "b_star": -21.7112789339974},
{ "name": "10PB6/4","L_star": 61.0367193394561, "a_star": 6.29688823312974, "b_star": -13.935764133093},
{ "name": "10PB6/3","L_star": 61.0367193394561, "a_star": 4.57529702436571, "b_star": -10.1633130407796},
{ "name": "10PB6/2","L_star": 61.0367193394561, "a_star": 2.95616238755886, "b_star": -6.54386242319587},
{ "name": "10PB6/1","L_star": 61.0454854659928, "a_star": 1.46103439619383, "b_star": -3.12787637117786},
{ "name": "10PB5/22","L_star": 50.6755045527692, "a_star": 55.7544215529172, "b_star": -77.1986789285374},
{ "name": "10PB5/20","L_star": 50.8855075285586, "a_star": 48.2236021339253, "b_star": -70.2176207345831},
{ "name": "10PB5/18","L_star": 50.9087599231063, "a_star": 41.9082325551386, "b_star": -63.5958277320536},
{ "name": "10PB5/16","L_star": 50.9203800622637, "a_star": 36.536201540268, "b_star": -57.7492588860086},
{ "name": "10PB5/14","L_star": 50.9319961673508, "a_star": 30.7968983072025, "b_star": -50.7198812180414},
{ "name": "10PB5/12","L_star": 50.9436082418668, "a_star": 25.4483040368663, "b_star": -43.6122341013602},
{ "name": "10PB5/10","L_star": 50.9552162893063, "a_star": 20.5230867010322, "b_star": -36.5452448858671},
{ "name": "10PB5/8","L_star": 50.966820313159, "a_star": 15.7974021496505, "b_star": -29.7727054254928},
{ "name": "10PB5/6","L_star": 50.9784203169096, "a_star": 11.7127056913022, "b_star": -22.7459053230021},
{ "name": "10PB5/4","L_star": 50.9900163040383, "a_star": 7.16189293822411, "b_star": -14.7700550552482},
{ "name": "10PB5/3","L_star": 50.9900163040383, "a_star": 5.26836004208903, "b_star": -11.0300049377332},
{ "name": "10PB5/2","L_star": 50.9900163040383, "a_star": 3.44116023137286, "b_star": -7.33654159140034},
{ "name": "10PB5/1","L_star": 51.0016082780202, "a_star": 1.67698840599723, "b_star": -3.64540404980824},
{ "name": "10PB4/30","L_star": 39.6477591126523, "a_star": 89.4056141441983, "b_star": -98.5533496844661},
{ "name": "10PB4/28","L_star": 39.8486552922544, "a_star": 81.6136702341242, "b_star": -92.6341781954792},
{ "name": "10PB4/26","L_star": 39.9817872746467, "a_star": 74.6051324798993, "b_star": -86.9759856980261},
{ "name": "10PB4/24","L_star": 40.1638162410762, "a_star": 66.6749411033874, "b_star": -80.6757495150975},
{ "name": "10PB4/22","L_star": 40.2790412749949, "a_star": 60.0138110441145, "b_star": -75.0629218564131},
{ "name": "10PB4/20","L_star": 40.3937964103769, "a_star": 54.2624294021025, "b_star": -70.1129062524542},
{ "name": "10PB4/18","L_star": 40.5406559183046, "a_star": 46.5382138265914, "b_star": -63.2253032628254},
{ "name": "10PB4/16","L_star": 40.6381400008437, "a_star": 40.1028183602168, "b_star": -57.0562628108466},
{ "name": "10PB4/14","L_star": 40.6543547569545, "a_star": 33.8815676703117, "b_star": -51.0265940785507},
{ "name": "10PB4/12","L_star": 40.6705602368878, "a_star": 28.2041724310035, "b_star": -44.3889531210949},
{ "name": "10PB4/10","L_star": 40.686756453899, "a_star": 22.730039694447, "b_star": -37.5712072636221},
{ "name": "10PB4/8","L_star": 40.7029434212134, "a_star": 17.7551195135854, "b_star": -30.6009651768317},
{ "name": "10PB4/6","L_star": 40.7029434212134, "a_star": 13.0209016267057, "b_star": -23.5087363955971},
{ "name": "10PB4/4","L_star": 40.7191211520259, "a_star": 8.54054920183059, "b_star": -16.0918783360759},
{ "name": "10PB4/3","L_star": 40.7191211520259, "a_star": 6.50463376862595, "b_star": -12.469081982684},
{ "name": "10PB4/2","L_star": 40.7191211520259, "a_star": 4.45848533026846, "b_star": -8.69299301194865},
{ "name": "10PB4/1","L_star": 40.7352896595015, "a_star": 2.29099909714894, "b_star": -4.55494303503785},
{ "name": "10PB3/34","L_star": 29.4074182474031, "a_star": 113.107919736435, "b_star": -108.045156380664},
{ "name": "10PB3/32","L_star": 29.4955686510738, "a_star": 105.687131207712, "b_star": -102.904858513781},
{ "name": "10PB3/30","L_star": 29.6034022289676, "a_star": 96.0442070995944, "b_star": -96.179704350813},
{ "name": "10PB3/28","L_star": 29.6708518950624, "a_star": 88.6941517920579, "b_star": -90.9109227749329},
{ "name": "10PB3/26","L_star": 29.7480492289331, "a_star": 81.4460370369075, "b_star": -85.5658695407903},
{ "name": "10PB3/24","L_star": 29.8447999059177, "a_star": 74.2434890684896, "b_star": -80.087134965902},
{ "name": "10PB3/22","L_star": 29.9386786180762, "a_star": 66.8619690271125, "b_star": -74.4313639804098},
{ "name": "10PB3/20","L_star": 30.017437824822, "a_star": 59.649212169599, "b_star": -68.8180864386386},
{ "name": "10PB3/18","L_star": 30.0885812207981, "a_star": 52.9296686940856, "b_star": -63.3999939294493},
{ "name": "10PB3/16","L_star": 30.1595056550719, "a_star": 46.4816470544496, "b_star": -58.1090015537967},
{ "name": "10PB3/14","L_star": 30.2423818530976, "a_star": 39.2051497467359, "b_star": -51.752635437641},
{ "name": "10PB3/12","L_star": 30.2934220510993, "a_star": 32.2777418254754, "b_star": -45.159866908854},
{ "name": "10PB3/10","L_star": 30.3104104643861, "a_star": 26.7570143703046, "b_star": -39.3141418686583},
{ "name": "10PB3/8","L_star": 30.3273864228217, "a_star": 20.6300044845964, "b_star": -32.1013076694052},
{ "name": "10PB3/6","L_star": 30.3419273485797, "a_star": 15.1997884655488, "b_star": -25.0933659618623},
{ "name": "10PB3/4","L_star": 30.3540378198749, "a_star": 10.1756811864449, "b_star": -17.8161487333662},
{ "name": "10PB3/3","L_star": 30.3588802370237, "a_star": 7.83792297378255, "b_star": -14.0296820144358},
{ "name": "10PB3/2","L_star": 30.3637216427544, "a_star": 5.40471107921914, "b_star": -9.8945172484847},
{ "name": "10PB3/1","L_star": 30.3709818560962, "a_star": 2.80704095738832, "b_star": -5.250099706922},
{ "name": "10PB2/34","L_star": 19.4701240779965, "a_star": 119.922599555686, "b_star": -104.815022635197},
{ "name": "10PB2/32","L_star": 19.5485250261232, "a_star": 111.442146698867, "b_star": -99.1831421225126},
{ "name": "10PB2/30","L_star": 19.6019691710487, "a_star": 102.990872708977, "b_star": -93.5138424597085},
{ "name": "10PB2/28","L_star": 19.7083787242135, "a_star": 93.8660730040528, "b_star": -87.1825627509952},
{ "name": "10PB2/26","L_star": 19.8019844306396, "a_star": 86.1436606016419, "b_star": -81.7827061284186},
{ "name": "10PB2/24","L_star": 19.870857943764, "a_star": 78.301162612844, "b_star": -76.325842540536},
{ "name": "10PB2/22","L_star": 19.9233480178816, "a_star": 70.6511501288467, "b_star": -70.9208222965141},
{ "name": "10PB2/20","L_star": 19.9636208632855, "a_star": 63.1888231072736, "b_star": -65.5120281860928},
{ "name": "10PB2/18","L_star": 20.0078170679051, "a_star": 55.5405631029478, "b_star": -59.9042270597659},
{ "name": "10PB2/16","L_star": 20.0479015016129, "a_star": 48.4185298241437, "b_star": -54.4210313147395},
{ "name": "10PB2/14","L_star": 20.0878969870992, "a_star": 41.503744118367, "b_star": -48.9798950412593},
{ "name": "10PB2/12","L_star": 20.1238172798771, "a_star": 34.7120933903265, "b_star": -43.1790350136599},
{ "name": "10PB2/10","L_star": 20.1517059775056, "a_star": 28.5389404561561, "b_star": -37.6007101565423},
{ "name": "10PB2/8","L_star": 20.1755763746542, "a_star": 21.9418938517583, "b_star": -31.0293806182123},
{ "name": "10PB2/6","L_star": 20.1914724884762, "a_star": 15.8239117109129, "b_star": -23.7249656397399},
{ "name": "10PB2/4","L_star": 20.2073546506852, "a_star": 10.5556244981629, "b_star": -16.8602935119429},
{ "name": "10PB2/3","L_star": 20.211323014883, "a_star": 8.09808111212087, "b_star": -13.3448426174502},
{ "name": "10PB2/2","L_star": 20.2192571349912, "a_star": 5.52701139570896, "b_star": -9.45557488735753},
{ "name": "10PB2/1","L_star": 20.2271877805556, "a_star": 2.88187216154864, "b_star": -5.07390962441978},
{ "name": "10PB1/30","L_star": 9.72761764353012, "a_star": 112.299287870248, "b_star": -91.213763607983},
{ "name": "10PB1/28","L_star": 9.79816905060406, "a_star": 101.503030258539, "b_star": -84.3763961927424},
{ "name": "10PB1/26","L_star": 9.94585687546375, "a_star": 91.3230273147167, "b_star": -77.7815150274033},
{ "name": "10PB1/24","L_star": 10.0612759075204, "a_star": 81.5078962795482, "b_star": -71.3362006601838},
{ "name": "10PB1/22","L_star": 10.1224169548449, "a_star": 73.1374267101398, "b_star": -65.7117632118291},
{ "name": "10PB1/20","L_star": 10.1604852351022, "a_star": 65.986841292836, "b_star": -60.9401425234546},
{ "name": "10PB1/18","L_star": 10.1908602815276, "a_star": 58.5794907949217, "b_star": -55.8783327237597},
{ "name": "10PB1/16","L_star": 10.2211650355277, "a_star": 52.1747816056359, "b_star": -51.4096164007743},
{ "name": "10PB1/14","L_star": 10.2589477463405, "a_star": 44.7952401834681, "b_star": -46.025447301863},
{ "name": "10PB1/12","L_star": 10.281565283574, "a_star": 37.773345391375, "b_star": -40.850072679104},
{ "name": "10PB1/10","L_star": 10.3116615770145, "a_star": 30.8534089316676, "b_star": -35.3883359810231},
{ "name": "10PB1/8","L_star": 10.3341886955329, "a_star": 24.3135312870371, "b_star": -29.8347197091502},
{ "name": "10PB1/6","L_star": 10.3491853893754, "a_star": 18.410983954194, "b_star": -24.2334870185273},
{ "name": "10PB1/4","L_star": 10.3716484739028, "a_star": 12.328189996111, "b_star": -17.4779916951093},
{ "name": "10PB1/3","L_star": 10.3791276712618, "a_star": 9.80486416623699, "b_star": -14.3818314558606},
{ "name": "10PB1/2","L_star": 10.3866026299139, "a_star": 6.96305707524104, "b_star": -10.6455970347485},
{ "name": "10PB1/1","L_star": 10.3940733558595, "a_star": 3.73230544442728, "b_star": -5.96700808449073},
{ "name": "2.5P9/4","L_star": 90.1648932823636, "a_star": 7.76508184016467, "b_star": -13.0617383490179},
{ "name": "2.5P9/3","L_star": 90.1741250122608, "a_star": 4.9451888647204, "b_star": -8.41665383296444},
{ "name": "2.5P9/2","L_star": 90.1787402752534, "a_star": 2.78844280175078, "b_star": -4.73305995206863},
{ "name": "2.5P9/1","L_star": 90.1833551370584, "a_star": 1.12568218187564, "b_star": -1.89506518520515},
{ "name": "2.5P8/8","L_star": 80.5772898331711, "a_star": 18.3975978247577, "b_star": -28.0587468457049},
{ "name": "2.5P8/6","L_star": 80.4935425748736, "a_star": 12.7696474919766, "b_star": -19.9912355641879},
{ "name": "2.5P8/4","L_star": 80.5103036560818, "a_star": 7.76755914352262, "b_star": -12.3948443521064},
{ "name": "2.5P8/3","L_star": 80.5158893896084, "a_star": 5.17370790054261, "b_star": -8.3539627335073},
{ "name": "2.5P8/2","L_star": 80.5214744766754, "a_star": 2.92766920724408, "b_star": -4.69761524236518},
{ "name": "2.5P8/1","L_star": 80.5270589174699, "a_star": 1.11088518301111, "b_star": -1.73999909916518},
{ "name": "2.5P7/12","L_star": 70.905378608356, "a_star": 30.1711735123135, "b_star": -40.8680907748454},
{ "name": "2.5P7/10","L_star": 70.7950136941808, "a_star": 24.1699261956889, "b_star": -33.6354359212873},
{ "name": "2.5P7/8","L_star": 70.8157284953717, "a_star": 18.2905553660875, "b_star": -26.4102566108709},
{ "name": "2.5P7/6","L_star": 70.8295328728496, "a_star": 12.9879826468634, "b_star": -19.3755734226289},
{ "name": "2.5P7/4","L_star": 70.8433328624133, "a_star": 8.24200935180192, "b_star": -12.6041413486104},
{ "name": "2.5P7/3","L_star": 70.8502312128167, "a_star": 5.808599835877, "b_star": -9.03889303998098},
{ "name": "2.5P7/2","L_star": 70.8571284675481, "a_star": 3.57638692262296, "b_star": -5.64574611436395},
{ "name": "2.5P7/1","L_star": 70.8640246270425, "a_star": 1.56020370703736, "b_star": -2.57036835356736},
{ "name": "2.5P6/18","L_star": 60.9137835246555, "a_star": 49.6393101754841, "b_star": -58.2524842839074},
{ "name": "2.5P6/16","L_star": 60.9225776916113, "a_star": 43.914798126322, "b_star": -52.4925854306919},
{ "name": "2.5P6/14","L_star": 60.9313698482419, "a_star": 36.7823377311068, "b_star": -45.7608599309683},
{ "name": "2.5P6/12","L_star": 60.957734267663, "a_star": 30.6077300614329, "b_star": -39.4067615327805},
{ "name": "2.5P6/10","L_star": 60.9753005166793, "a_star": 24.7594386730621, "b_star": -33.1322815749087},
{ "name": "2.5P6/8","L_star": 60.9928587518952, "a_star": 19.1480138556297, "b_star": -26.5908669491648},
{ "name": "2.5P6/6","L_star": 61.0104089824442, "a_star": 13.9199339164892, "b_star": -20.0849407867654},
{ "name": "2.5P6/4","L_star": 61.0191810988183, "a_star": 8.64543180105393, "b_star": -12.8211069499467},
{ "name": "2.5P6/3","L_star": 61.0279512174433, "a_star": 6.31053755533068, "b_star": -9.50105749190426},
{ "name": "2.5P6/2","L_star": 61.0367193394561, "a_star": 4.06150068309524, "b_star": -6.26266468676735},
{ "name": "2.5P6/1","L_star": 61.0367193394561, "a_star": 1.96256327633015, "b_star": -3.07281759853877},
{ "name": "2.5P5/26","L_star": 50.2988652478439, "a_star": 80.5822067293444, "b_star": -80.707878083459},
{ "name": "2.5P5/24","L_star": 50.4759463913911, "a_star": 72.7537789913832, "b_star": -74.7838436625998},
{ "name": "2.5P5/22","L_star": 50.6169350351417, "a_star": 65.2659835260258, "b_star": -68.8040806627108},
{ "name": "2.5P5/20","L_star": 50.7690023152908, "a_star": 57.9084783515856, "b_star": -62.5860951287013},
{ "name": "2.5P5/18","L_star": 50.8273057000631, "a_star": 50.9041217577906, "b_star": -56.5598305144568},
{ "name": "2.5P5/16","L_star": 50.8505985934447, "a_star": 44.7895341644661, "b_star": -51.0959601862085},
{ "name": "2.5P5/14","L_star": 50.8738752661451, "a_star": 38.4597710997611, "b_star": -45.2604563974615},
{ "name": "2.5P5/12","L_star": 50.8971357463743, "a_star": 32.4335938806315, "b_star": -39.5413767092486},
{ "name": "2.5P5/10","L_star": 50.9203800622637, "a_star": 26.2479025437102, "b_star": -33.3347743779462},
{ "name": "2.5P5/8","L_star": 50.9436082418668, "a_star": 20.8389500969888, "b_star": -27.4534885972693},
{ "name": "2.5P5/6","L_star": 50.9552162893063, "a_star": 15.0188206164762, "b_star": -20.7291059921261},
{ "name": "2.5P5/4","L_star": 50.9784203169096, "a_star": 9.68843165970318, "b_star": -13.8479890229241},
{ "name": "2.5P5/3","L_star": 50.9784203169096, "a_star": 7.06354401143205, "b_star": -10.270452993682},
{ "name": "2.5P5/2","L_star": 50.9900163040383, "a_star": 4.61018592918666, "b_star": -6.7597132885963},
{ "name": "2.5P5/1","L_star": 50.9900163040383, "a_star": 2.21821534914041, "b_star": -3.29783264918198},
{ "name": "2.5P4/32","L_star": 39.6981191031338, "a_star": 107.299202758436, "b_star": -95.9349969213742},
{ "name": "2.5P4/30","L_star": 39.8486552922544, "a_star": 97.6301546148914, "b_star": -89.1297948456751},
{ "name": "2.5P4/28","L_star": 39.9651803952164, "a_star": 90.2752656818892, "b_star": -83.895174195996},
{ "name": "2.5P4/26","L_star": 40.081222274495, "a_star": 82.9159297991513, "b_star": -78.5418855365989},
{ "name": "2.5P4/24","L_star": 40.1803059120886, "a_star": 75.1947473892114, "b_star": -72.7320352279106},
{ "name": "2.5P4/22","L_star": 40.2790412749949, "a_star": 68.4023122750753, "b_star": -67.5772821326134},
{ "name": "2.5P4/20","L_star": 40.3774314075978, "a_star": 62.478350547287, "b_star": -63.0021074277055},
{ "name": "2.5P4/18","L_star": 40.4917875655919, "a_star": 55.007556371693, "b_star": -57.0338408182802},
{ "name": "2.5P4/16","L_star": 40.5731879484504, "a_star": 47.932718280947, "b_star": -51.2978762942669},
{ "name": "2.5P4/14","L_star": 40.5894399424351, "a_star": 41.1728503289932, "b_star": -45.5653116512204},
{ "name": "2.5P4/12","L_star": 40.6219159552695, "a_star": 34.87491286727, "b_star": -39.8713119177598},
{ "name": "2.5P4/10","L_star": 40.6381400008437, "a_star": 28.3647311834321, "b_star": -33.7079588333581},
{ "name": "2.5P4/8","L_star": 40.6705602368878, "a_star": 22.3568006494342, "b_star": -27.7095580378785},
{ "name": "2.5P4/6","L_star": 40.686756453899, "a_star": 16.5543877481958, "b_star": -21.4544878335702},
{ "name": "2.5P4/4","L_star": 40.7029434212134, "a_star": 11.300320852486, "b_star": -15.1211166081885},
{ "name": "2.5P4/3","L_star": 40.7029434212134, "a_star": 8.64313244552337, "b_star": -11.8065448583942},
{ "name": "2.5P4/2","L_star": 40.7191211520259, "a_star": 5.85920768707251, "b_star": -8.20209013532206},
{ "name": "2.5P4/1","L_star": 40.7191211520259, "a_star": 2.96756343141666, "b_star": -4.30410026753895},
{ "name": "2.5P3/34","L_star": 29.467901060484, "a_star": 120.952531356726, "b_star": -100.165137728848},
{ "name": "2.5P3/32","L_star": 29.5508031024421, "a_star": 113.213093056536, "b_star": -95.0010808467878},
{ "name": "2.5P3/30","L_star": 29.6184082854074, "a_star": 105.033931641945, "b_star": -89.6667927533862},
{ "name": "2.5P3/28","L_star": 29.6808275087461, "a_star": 96.7555400241875, "b_star": -84.0566453386766},
{ "name": "2.5P3/26","L_star": 29.7629605889294, "a_star": 88.7119150709444, "b_star": -78.4346447646678},
{ "name": "2.5P3/24","L_star": 29.8423242173987, "a_star": 81.5880108694456, "b_star": -73.31677274709},
{ "name": "2.5P3/22","L_star": 29.9337471942159, "a_star": 74.2245152207787, "b_star": -67.9063933493124},
{ "name": "2.5P3/20","L_star": 30.0198947088536, "a_star": 66.5142393344587, "b_star": -62.314578819493},
{ "name": "2.5P3/18","L_star": 30.0812317419267, "a_star": 59.8795642836465, "b_star": -57.3779740492118},
{ "name": "2.5P3/16","L_star": 30.1375179545452, "a_star": 53.1226053626078, "b_star": -52.3003120174967},
{ "name": "2.5P3/14","L_star": 30.1961053418733, "a_star": 46.0789371310441, "b_star": -46.7525345255929},
{ "name": "2.5P3/12","L_star": 30.239948555795, "a_star": 38.7340620248186, "b_star": -40.7524962521867},
{ "name": "2.5P3/10","L_star": 30.2667007541636, "a_star": 32.4016667619267, "b_star": -35.3576163119649},
{ "name": "2.5P3/8","L_star": 30.2934220510993, "a_star": 25.4364292679009, "b_star": -29.1171459218173},
{ "name": "2.5P3/6","L_star": 30.3176873988969, "a_star": 19.0563938170458, "b_star": -23.0023656814402},
{ "name": "2.5P3/4","L_star": 30.3370813873033, "a_star": 13.0668898499201, "b_star": -16.4913488100774},
{ "name": "2.5P3/3","L_star": 30.3467722965866, "a_star": 10.1167232806766, "b_star": -13.0498956319384},
{ "name": "2.5P3/2","L_star": 30.3564591549096, "a_star": 7.01969099707336, "b_star": -9.21548600261995},
{ "name": "2.5P3/1","L_star": 30.3661419665029, "a_star": 3.66001185491618, "b_star": -4.90662100998006},
{ "name": "2.5P2/30","L_star": 19.4038323704814, "a_star": 113.274876709579, "b_star": -89.3724479674285},
{ "name": "2.5P2/28","L_star": 19.507304528779, "a_star": 102.912775269309, "b_star": -82.8723963028395},
{ "name": "2.5P2/26","L_star": 19.6224819487902, "a_star": 93.279241728371, "b_star": -76.5788294171652},
{ "name": "2.5P2/24","L_star": 20.0118295280528, "a_star": 85.4914629730663, "b_star": -71.4707122476263},
{ "name": "2.5P2/22","L_star": 19.8425303140207, "a_star": 76.6107303920869, "b_star": -65.2407913383323},
{ "name": "2.5P2/20","L_star": 19.9233480178816, "a_star": 68.7469427592248, "b_star": -59.8242578024525},
{ "name": "2.5P2/18","L_star": 19.9756851449229, "a_star": 61.2005811321608, "b_star": -54.5623130878802},
{ "name": "2.5P2/16","L_star": 20.0158410942582, "a_star": 53.85359585934, "b_star": -49.3878325400849},
{ "name": "2.5P2/14","L_star": 20.0479015016129, "a_star": 46.9508967258536, "b_star": -44.3365622907909},
{ "name": "2.5P2/12","L_star": 20.07990498219, "a_star": 39.9965028751555, "b_star": -39.0216673070487},
{ "name": "2.5P2/10","L_star": 20.111851788018, "a_star": 32.9075637397185, "b_star": -33.4385535583626},
{ "name": "2.5P2/8","L_star": 20.1397589488837, "a_star": 26.0541490645494, "b_star": -27.6324804895419},
{ "name": "2.5P2/6","L_star": 20.1716001626663, "a_star": 18.6370690137469, "b_star": -21.0099605433313},
{ "name": "2.5P2/4","L_star": 20.1914724884762, "a_star": 12.8642225737903, "b_star": -15.3041585615052},
{ "name": "2.5P2/3","L_star": 20.199415311619, "a_star": 10.018787003371, "b_star": -12.2454338385002},
{ "name": "2.5P2/2","L_star": 20.211323014883, "a_star": 6.97273775943491, "b_star": -8.7773355832791},
{ "name": "2.5P2/1","L_star": 20.2192571349912, "a_star": 3.67888335699595, "b_star": -4.74830417611865},
{ "name": "2.5P1/26","L_star": 9.71975467921293, "a_star": 95.6972568165798, "b_star": -72.1287807562568},
{ "name": "2.5P1/24","L_star": 9.78252431113926, "a_star": 85.365888773751, "b_star": -65.8294767843004},
{ "name": "2.5P1/22","L_star": 9.89940045222461, "a_star": 77.8052971935499, "b_star": -61.0254531440643},
{ "name": "2.5P1/20","L_star": 10.0382738868248, "a_star": 69.8842087193043, "b_star": -55.9140433894585},
{ "name": "2.5P1/18","L_star": 10.1300394872165, "a_star": 62.468028988969, "b_star": -51.0786937676028},
{ "name": "2.5P1/16","L_star": 10.1756815703438, "a_star": 56.1860044738929, "b_star": -46.9364794586412},
{ "name": "2.5P1/14","L_star": 10.2135954147157, "a_star": 48.6101611520824, "b_star": -41.7992112197895},
{ "name": "2.5P1/12","L_star": 10.2438477155545, "a_star": 41.6372671161219, "b_star": -36.9816570712043},
{ "name": "2.5P1/10","L_star": 10.2740304307288, "a_star": 34.3261616208257, "b_star": -31.7072396187458},
{ "name": "2.5P1/8","L_star": 10.2966220415417, "a_star": 27.9713554029289, "b_star": -26.8961949703452},
{ "name": "2.5P1/6","L_star": 10.3266839391574, "a_star": 21.2669306442614, "b_star": -21.5392268622336},
{ "name": "2.5P1/4","L_star": 10.3491853893754, "a_star": 14.9219235853435, "b_star": -16.0344915378827},
{ "name": "2.5P1/3","L_star": 10.3641650318228, "a_star": 12.0686933529007, "b_star": -13.3507862619602},
{ "name": "2.5P1/2","L_star": 10.3716484739028, "a_star": 8.70681319018821, "b_star": -9.98324732564469},
{ "name": "2.5P1/1","L_star": 10.3866026299139, "a_star": 4.72245892833192, "b_star": -5.62938867652328},
{ "name": "5P9/4","L_star": 90.1556599466689, "a_star": 9.670961630521, "b_star": -12.2376403520185},
{ "name": "5P9/3","L_star": 90.1648932823636, "a_star": 5.96138965353532, "b_star": -7.65500223382558},
{ "name": "5P9/2","L_star": 90.1741250122608, "a_star": 3.19994792042644, "b_star": -4.12031162037949},
{ "name": "5P9/1","L_star": 90.1833551370584, "a_star": 1.23641816383535, "b_star": -1.54432924150758},
{ "name": "5P8/10","L_star": 80.6887274094783, "a_star": 28.3316506711029, "b_star": -31.351224332069},
{ "name": "5P8/8","L_star": 80.5382258609889, "a_star": 21.9060567471225, "b_star": -24.6990389968394},
{ "name": "5P8/6","L_star": 80.4767756687889, "a_star": 15.261786783936, "b_star": -17.6297073589481},
{ "name": "5P8/4","L_star": 80.4991302489017, "a_star": 9.30521112601201, "b_star": -11.0593944368246},
{ "name": "5P8/3","L_star": 80.5103036560818, "a_star": 6.19121245835325, "b_star": -7.48426007398435},
{ "name": "5P8/2","L_star": 80.5158893896084, "a_star": 3.38913565404586, "b_star": -4.20634262898076},
{ "name": "5P8/1","L_star": 80.5270589174699, "a_star": 1.20585496283293, "b_star": -1.56515998749247},
{ "name": "5P7/14","L_star": 70.8915983213402, "a_star": 41.2999821387285, "b_star": -41.7799001221361},
{ "name": "5P7/12","L_star": 70.8295328728496, "a_star": 34.8042051134308, "b_star": -35.9948231721729},
{ "name": "5P7/10","L_star": 70.7673785689525, "a_star": 28.2800509458471, "b_star": -29.826519585887},
{ "name": "5P7/8","L_star": 70.7950136941808, "a_star": 21.6548175228442, "b_star": -23.3107343110706},
{ "name": "5P7/6","L_star": 70.8157284953717, "a_star": 15.5844289820797, "b_star": -17.2741351727901},
{ "name": "5P7/4","L_star": 70.8295328728496, "a_star": 9.9196208913368, "b_star": -11.1699290202711},
{ "name": "5P7/3","L_star": 70.8433328624133, "a_star": 7.05882463228746, "b_star": -8.0135576545608},
{ "name": "5P7/2","L_star": 70.8502312128167, "a_star": 4.36231981504903, "b_star": -4.99235942056717},
{ "name": "5P7/1","L_star": 70.8571284675481, "a_star": 1.97136531955117, "b_star": -2.21554809148705},
{ "name": "5P6/20","L_star": 60.7728025078901, "a_star": 63.542911661036, "b_star": -58.1726368239215},
{ "name": "5P6/18","L_star": 60.8169146771618, "a_star": 55.7657563034824, "b_star": -52.0837605969972},
{ "name": "5P6/16","L_star": 60.8609762416457, "a_star": 48.9882776683689, "b_star": -46.514176321923},
{ "name": "5P6/14","L_star": 60.8961891551695, "a_star": 42.0924163066531, "b_star": -40.7457950978304},
{ "name": "5P6/12","L_star": 60.9225776916113, "a_star": 35.1733425728102, "b_star": -34.8293790354397},
{ "name": "5P6/10","L_star": 60.9489481351208, "a_star": 28.9610195974209, "b_star": -29.4432965913429},
{ "name": "5P6/8","L_star": 60.9665183944677, "a_star": 22.6373170753207, "b_star": -23.5207636410361},
{ "name": "5P6/6","L_star": 60.9928587518952, "a_star": 16.8411275167642, "b_star": -17.8080161652768},
{ "name": "5P6/4","L_star": 61.0104089824442, "a_star": 10.6330326973674, "b_star": -11.5117267661872},
{ "name": "5P6/3","L_star": 61.0191810988183, "a_star": 7.87002814726945, "b_star": -8.56476819586585},
{ "name": "5P6/2","L_star": 61.0279512174433, "a_star": 5.12529803048589, "b_star": -5.67874974410876},
{ "name": "5P6/1","L_star": 61.0367193394561, "a_star": 2.53963360532511, "b_star": -2.83922645673591},
{ "name": "5P5/28","L_star": 50.27518285067, "a_star": 95.820895166807, "b_star": -78.2838960862684},
{ "name": "5P5/26","L_star": 50.4052272461001, "a_star": 86.5758990964304, "b_star": -72.1574635952013},
{ "name": "5P5/24","L_star": 50.499486034432, "a_star": 79.3222385850765, "b_star": -67.1581024414637},
{ "name": "5P5/22","L_star": 50.6052087657505, "a_star": 71.5537555242609, "b_star": -61.668193904154},
{ "name": "5P5/20","L_star": 50.6989035736593, "a_star": 63.9380625014314, "b_star": -56.1981743997583},
{ "name": "5P5/18","L_star": 50.7806711380232, "a_star": 56.9654713403072, "b_star": -51.0169824239358},
{ "name": "5P5/16","L_star": 50.8039965577119, "a_star": 50.2134844083376, "b_star": -45.8539340780527},
{ "name": "5P5/14","L_star": 50.8389541761052, "a_star": 43.4154444512628, "b_star": -40.44655916999},
{ "name": "5P5/12","L_star": 50.8622389556151, "a_star": 37.2511756933578, "b_star": -35.4022363429978},
{ "name": "5P5/10","L_star": 50.8971357463743, "a_star": 30.4207757544208, "b_star": -29.6265588007371},
{ "name": "5P5/8","L_star": 50.9203800622637, "a_star": 24.4403805904527, "b_star": -24.3142500818077},
{ "name": "5P5/6","L_star": 50.9436082418668, "a_star": 17.9238151883878, "b_star": -18.3528021284347},
{ "name": "5P5/4","L_star": 50.966820313159, "a_star": 11.6045404034276, "b_star": -12.1789944827844},
{ "name": "5P5/3","L_star": 50.9784203169096, "a_star": 8.55014330826537, "b_star": -9.03172811192448},
{ "name": "5P5/2","L_star": 50.9784203169096, "a_star": 5.60265096327595, "b_star": -5.9231864147467},
{ "name": "5P5/1","L_star": 50.9900163040383, "a_star": 2.75740823249843, "b_star": -2.88267946089829},
{ "name": "5P4/32","L_star": 39.9485636542174, "a_star": 113.544835546706, "b_star": -87.359572445549},
{ "name": "5P4/30","L_star": 40.0315488895355, "a_star": 104.050163956339, "b_star": -81.267785937205},
{ "name": "5P4/28","L_star": 40.0646742586997, "a_star": 96.8618987826244, "b_star": -76.5729331842576},
{ "name": "5P4/26","L_star": 40.1473168816541, "a_star": 88.5999157998547, "b_star": -71.0360070471087},
{ "name": "5P4/24","L_star": 40.2132562456911, "a_star": 81.3007552811611, "b_star": -66.0183414394287},
{ "name": "5P4/22","L_star": 40.2954635231621, "a_star": 74.1041612888836, "b_star": -61.0819804160662},
{ "name": "5P4/20","L_star": 40.3774314075978, "a_star": 67.8599809669689, "b_star": -56.7335010245295},
{ "name": "5P4/18","L_star": 40.4754793119243, "a_star": 60.3525293571021, "b_star": -51.4325910130982},
{ "name": "5P4/16","L_star": 40.5243758552341, "a_star": 53.5893506835317, "b_star": -46.4817121081084},
{ "name": "5P4/14","L_star": 40.5569266115533, "a_star": 45.9894476627166, "b_star": -40.903273131254},
{ "name": "5P4/12","L_star": 40.5894399424351, "a_star": 38.9380884867414, "b_star": -35.5261192328356},
{ "name": "5P4/10","L_star": 40.6219159552695, "a_star": 32.2908111323978, "b_star": -30.230828566054},
{ "name": "5P4/8","L_star": 40.6381400008437, "a_star": 25.5878868152692, "b_star": -24.6806306327713},
{ "name": "5P4/6","L_star": 40.6705602368878, "a_star": 19.3581571661006, "b_star": -19.1612110020977},
{ "name": "5P4/4","L_star": 40.686756453899, "a_star": 13.3565317584268, "b_star": -13.5248234367526},
{ "name": "5P4/3","L_star": 40.7029434212134, "a_star": 10.2713122353017, "b_star": -10.5439482517353},
{ "name": "5P4/2","L_star": 40.7191211520259, "a_star": 7.09105500187032, "b_star": -7.33870998929928},
{ "name": "5P4/1","L_star": 40.7191211520259, "a_star": 3.66864106805997, "b_star": -3.84347897040492},
{ "name": "5P3/32","L_star": 29.5031085200635, "a_star": 119.556404474214, "b_star": -87.6033713608561},
{ "name": "5P3/30","L_star": 29.5808746101715, "a_star": 110.60731078043, "b_star": -82.1477816069396},
{ "name": "5P3/28","L_star": 29.6483908167372, "a_star": 102.678739662347, "b_star": -77.23332151429},
{ "name": "5P3/26","L_star": 29.7430766148473, "a_star": 94.1400564533929, "b_star": -71.8675294396077},
{ "name": "5P3/24","L_star": 29.8274644678268, "a_star": 86.9421080907217, "b_star": -67.2086227214494},
{ "name": "5P3/22","L_star": 29.9263480724393, "a_star": 78.9325099116525, "b_star": -61.8632028072108},
{ "name": "5P3/20","L_star": 30.010065598194, "a_star": 71.0531915597821, "b_star": -56.5723377283745},
{ "name": "5P3/18","L_star": 30.0665257470739, "a_star": 64.5909358408218, "b_star": -52.207077138679},
{ "name": "5P3/16","L_star": 30.1179557228363, "a_star": 57.5834437693356, "b_star": -47.3564614499434},
{ "name": "5P3/14","L_star": 30.1619474401831, "a_star": 50.4865918292167, "b_star": -42.3412083740768},
{ "name": "5P3/12","L_star": 30.2009809203097, "a_star": 43.4606816055433, "b_star": -37.253279414804},
{ "name": "5P3/10","L_star": 30.2350811927532, "a_star": 36.1960581238339, "b_star": -31.9133421337546},
{ "name": "5P3/8","L_star": 30.2667007541636, "a_star": 28.8318209037131, "b_star": -26.1213008891403},
{ "name": "5P3/6","L_star": 30.2958497306911, "a_star": 21.8768086548655, "b_star": -20.4326531793054},
{ "name": "5P3/4","L_star": 30.3225374185553, "a_star": 15.332008727947, "b_star": -14.7396364071596},
{ "name": "5P3/3","L_star": 30.3346580265234, "a_star": 11.9637682282299, "b_star": -11.7154556280236},
{ "name": "5P3/2","L_star": 30.3491943907795, "a_star": 8.33111673420683, "b_star": -8.36174007402086},
{ "name": "5P3/1","L_star": 30.3613010662833, "a_star": 4.36956665169355, "b_star": -4.48573302928199},
{ "name": "5P2/28","L_star": 19.3372914738314, "a_star": 107.858301089684, "b_star": -76.1834772867331},
{ "name": "5P2/26","L_star": 19.4494345468919, "a_star": 98.2450431695462, "b_star": -70.8620766396966},
{ "name": "5P2/24","L_star": 19.5732115097697, "a_star": 89.1732483675247, "b_star": -65.3986384342796},
{ "name": "5P2/22","L_star": 19.7042977666094, "a_star": 80.5961086286694, "b_star": -60.0348650926993},
{ "name": "5P2/20","L_star": 19.8344284759703, "a_star": 72.5709412394289, "b_star": -54.8816288312739},
{ "name": "5P2/18","L_star": 19.9273793665876, "a_star": 65.0865812048688, "b_star": -49.9484828786661},
{ "name": "5P2/16","L_star": 19.9837235062997, "a_star": 57.5942522568935, "b_star": -45.0867682425462},
{ "name": "5P2/14","L_star": 20.019851767019, "a_star": 50.8357115007481, "b_star": -40.5622794941655},
{ "name": "5P2/12","L_star": 20.0519050446824, "a_star": 43.2092573770152, "b_star": -35.3157656174067},
{ "name": "5P2/10","L_star": 20.0839014271698, "a_star": 35.8982153586802, "b_star": -30.0920230506717},
{ "name": "5P2/8","L_star": 20.1198296634102, "a_star": 28.7792463658818, "b_star": -24.8663214340814},
{ "name": "5P2/6","L_star": 20.1517059775056, "a_star": 20.7723488336272, "b_star": -18.6925280107977},
{ "name": "5P2/4","L_star": 20.1795517127494, "a_star": 14.6852584940859, "b_star": -13.6593699680784},
{ "name": "5P2/3","L_star": 20.1914724884762, "a_star": 11.5532496603494, "b_star": -10.968585231324},
{ "name": "5P2/2","L_star": 20.2033854164229, "a_star": 8.13378195376593, "b_star": -7.88406864153067},
{ "name": "5P2/1","L_star": 20.2192571349912, "a_star": 4.33927575409565, "b_star": -4.29221790979669},
{ "name": "5P1/22","L_star": 9.75117774611292, "a_star": 81.7302307500453, "b_star": -56.8831474942527},
{ "name": "5P1/20","L_star": 9.82160064300392, "a_star": 73.0922196625509, "b_star": -51.7679862218369},
{ "name": "5P1/18","L_star": 9.96902283140891, "a_star": 65.5759433304655, "b_star": -47.2472319141121},
{ "name": "5P1/16","L_star": 10.0765880590501, "a_star": 59.3951915487321, "b_star": -43.391540178967},
{ "name": "5P1/14","L_star": 10.1680856089276, "a_star": 51.6745619169897, "b_star": -38.5555262102115},
{ "name": "5P1/12","L_star": 10.2211650355277, "a_star": 44.074044993315, "b_star": -33.7109548377959},
{ "name": "5P1/10","L_star": 10.2513999023643, "a_star": 36.7477836278439, "b_star": -28.9302913694043},
{ "name": "5P1/8","L_star": 10.2740304307288, "a_star": 30.1187243902635, "b_star": -24.3797728628037},
{ "name": "5P1/6","L_star": 10.3041439590119, "a_star": 23.3217941467292, "b_star": -19.5071738034495},
{ "name": "5P1/4","L_star": 10.3341886955329, "a_star": 16.7881544646273, "b_star": -14.5935292614923},
{ "name": "5P1/3","L_star": 10.3491853893754, "a_star": 13.7393945804457, "b_star": -12.1826941567009},
{ "name": "5P1/2","L_star": 10.3641650318228, "a_star": 10.0213388550315, "b_star": -9.10885040421861},
{ "name": "5P1/1","L_star": 10.3866026299139, "a_star": 5.4944345364842, "b_star": -5.14780649910093},
{ "name": "7.5P9/6","L_star": 90.0632381169517, "a_star": 20.7738490595285, "b_star": -13.765438242605},
{ "name": "7.5P9/4","L_star": 90.1279502978124, "a_star": 12.6903262087092, "b_star": -8.28015546908181},
{ "name": "7.5P9/3","L_star": 90.1556599466689, "a_star": 8.1622182098805, "b_star": -5.30584333327073},
{ "name": "7.5P9/2","L_star": 90.1695093479933, "a_star": 4.21276238993046, "b_star": -2.71930822079447},
{ "name": "7.5P9/1","L_star": 90.1787402752534, "a_star": 1.31385090565711, "b_star": -0.788144445767158},
{ "name": "7.5P8/12","L_star": 80.3536398927712, "a_star": 42.8970316826433, "b_star": -28.790916860185},
{ "name": "7.5P8/10","L_star": 80.3984529402843, "a_star": 34.8071117926899, "b_star": -23.5258839314174},
{ "name": "7.5P8/8","L_star": 80.4320354035271, "a_star": 27.2262057178982, "b_star": -18.4469420250478},
{ "name": "7.5P8/6","L_star": 80.4544107243379, "a_star": 18.5205641251512, "b_star": -12.588661769077},
{ "name": "7.5P8/4","L_star": 80.4823652850047, "a_star": 11.9624993034385, "b_star": -7.98497677043135},
{ "name": "7.5P8/3","L_star": 80.4991302489017, "a_star": 8.00665175482596, "b_star": -5.30879905755703},
{ "name": "7.5P8/2","L_star": 80.5103036560818, "a_star": 4.35480551126544, "b_star": -2.81932810995462},
{ "name": "7.5P8/1","L_star": 80.5214744766754, "a_star": 1.49286494297635, "b_star": -0.944280230670214},
{ "name": "7.5P7/18","L_star": 70.5317659666793, "a_star": 64.8175893585059, "b_star": -43.351112313496},
{ "name": "7.5P7/16","L_star": 70.6011969247885, "a_star": 56.7758591359439, "b_star": -38.1369047944802},
{ "name": "7.5P7/14","L_star": 70.6497324336259, "a_star": 48.8212129167756, "b_star": -33.1007990585631},
{ "name": "7.5P7/12","L_star": 70.7051350909417, "a_star": 40.8300852447506, "b_star": -27.899329969055},
{ "name": "7.5P7/10","L_star": 70.732809888573, "a_star": 33.9009810127994, "b_star": -23.2186354099361},
{ "name": "7.5P7/8","L_star": 70.7604670364579, "a_star": 26.3441893771519, "b_star": -18.1367976035265},
{ "name": "7.5P7/6","L_star": 70.7881065627101, "a_star": 19.163831286778, "b_star": -13.1508933979606},
{ "name": "7.5P7/4","L_star": 70.8157284953717, "a_star": 12.2132916105404, "b_star": -8.3304849052261},
{ "name": "7.5P7/3","L_star": 70.8295328728496, "a_star": 8.68124084252042, "b_star": -5.82745519010992},
{ "name": "7.5P7/2","L_star": 70.8433328624133, "a_star": 5.34886314258054, "b_star": -3.52380388130487},
{ "name": "7.5P7/1","L_star": 70.8571284675481, "a_star": 2.43058737660751, "b_star": -1.49954604395806},
{ "name": "7.5P6/24","L_star": 60.5248323331319, "a_star": 88.6404584723705, "b_star": -58.25754853643},
{ "name": "7.5P6/22","L_star": 60.5958448618355, "a_star": 80.6065953779177, "b_star": -53.4243302839798},
{ "name": "7.5P6/20","L_star": 60.6667259617269, "a_star": 71.9858180315499, "b_star": -48.0936630454546},
{ "name": "7.5P6/18","L_star": 60.7374762392385, "a_star": 62.8006346581294, "b_star": -42.3688401826853},
{ "name": "7.5P6/16","L_star": 60.7904534574736, "a_star": 55.6499662873824, "b_star": -37.8596089289273},
{ "name": "7.5P6/14","L_star": 60.8345453674015, "a_star": 48.1366532987506, "b_star": -33.0824456253971},
{ "name": "7.5P6/12","L_star": 60.8785867305749, "a_star": 40.9161954637354, "b_star": -28.2931200070069},
{ "name": "7.5P6/10","L_star": 60.9049873462251, "a_star": 34.060769945511, "b_star": -23.8031542512828},
{ "name": "7.5P6/8","L_star": 60.940159995696, "a_star": 26.7772267463557, "b_star": -18.7730672354222},
{ "name": "7.5P6/6","L_star": 60.9665183944677, "a_star": 20.0671965534224, "b_star": -14.1707845128568},
{ "name": "7.5P6/4","L_star": 61.0016348671828, "a_star": 12.6452994303145, "b_star": -8.81859200886357},
{ "name": "7.5P6/3","L_star": 61.0104089824442, "a_star": 9.40592698512171, "b_star": -6.5012720160317},
{ "name": "7.5P6/2","L_star": 61.0191810988183, "a_star": 6.31898055523006, "b_star": -4.33973569807262},
{ "name": "7.5P6/1","L_star": 61.0367193394561, "a_star": 3.26508143493187, "b_star": -2.16963022492362},
{ "name": "7.5P5/30","L_star": 50.5112496090687, "a_star": 110.128242163109, "b_star": -71.3566654911893},
{ "name": "7.5P5/28","L_star": 50.5112496090687, "a_star": 101.400106318393, "b_star": -66.1511551371397},
{ "name": "7.5P5/26","L_star": 50.5230090240254, "a_star": 93.8202597124307, "b_star": -61.7039795296459},
{ "name": "7.5P5/24","L_star": 50.5582623475418, "a_star": 85.6994756887726, "b_star": -56.7962433662825},
{ "name": "7.5P5/22","L_star": 50.5934783659366, "a_star": 79.374563870618, "b_star": -52.865964632262},
{ "name": "7.5P5/20","L_star": 50.6520890970938, "a_star": 71.0841892365262, "b_star": -47.8905393869198},
{ "name": "7.5P5/18","L_star": 50.7105969300653, "a_star": 63.7515331957669, "b_star": -43.2575932219498},
{ "name": "7.5P5/16","L_star": 50.7573294125503, "a_star": 56.1565425274674, "b_star": -38.3971774522498},
{ "name": "7.5P5/14","L_star": 50.7923358843108, "a_star": 49.1561447598778, "b_star": -33.8642913500043},
{ "name": "7.5P5/12","L_star": 50.8273057000631, "a_star": 42.1137488991854, "b_star": -29.3586380192197},
{ "name": "7.5P5/10","L_star": 50.8622389556151, "a_star": 34.7911443908526, "b_star": -24.4476916334799},
{ "name": "7.5P5/8","L_star": 50.8855075285586, "a_star": 28.3522538476624, "b_star": -20.0492628622727},
{ "name": "7.5P5/6","L_star": 50.9203800622637, "a_star": 20.9069551789525, "b_star": -14.9233309108442},
{ "name": "7.5P5/4","L_star": 50.9436082418668, "a_star": 13.6912808219567, "b_star": -9.76814420505461},
{ "name": "7.5P5/3","L_star": 50.966820313159, "a_star": 10.065331745022, "b_star": -7.16246268640668},
{ "name": "7.5P5/2","L_star": 50.9784203169096, "a_star": 6.50944703555206, "b_star": -4.62710175950396},
{ "name": "7.5P5/1","L_star": 50.9900163040383, "a_star": 3.14222688076826, "b_star": -2.18323068212884},
{ "name": "7.5P4/32","L_star": 40.1142890405732, "a_star": 120.291586493684, "b_star": -77.3560151747095},
{ "name": "7.5P4/30","L_star": 40.1638162410762, "a_star": 111.364929124212, "b_star": -72.1581538424947},
{ "name": "7.5P4/28","L_star": 40.1803059120886, "a_star": 102.61234944938, "b_star": -67.0160976704663},
{ "name": "7.5P4/26","L_star": 40.2297169366013, "a_star": 95.3256854191999, "b_star": -62.6741116341333},
{ "name": "7.5P4/24","L_star": 40.2790412749949, "a_star": 87.1623758499209, "b_star": -57.7358466644786},
{ "name": "7.5P4/22","L_star": 40.3282793064743, "a_star": 80.4486345936099, "b_star": -53.623635046629},
{ "name": "7.5P4/20","L_star": 40.3774314075978, "a_star": 73.5565861566185, "b_star": -49.3711967552292},
{ "name": "7.5P4/18","L_star": 40.4264979523013, "a_star": 66.2235150082932, "b_star": -44.8996511411424},
{ "name": "7.5P4/16","L_star": 40.4754793119243, "a_star": 58.6353317779344, "b_star": -40.0725828716008},
{ "name": "7.5P4/14","L_star": 40.5080864088407, "a_star": 50.8787369743227, "b_star": -35.2390620838033},
{ "name": "7.5P4/12","L_star": 40.5569266115533, "a_star": 43.2123626500629, "b_star": -30.2882399813636},
{ "name": "7.5P4/10","L_star": 40.5894399424351, "a_star": 36.1149031878528, "b_star": -25.5693565784379},
{ "name": "7.5P4/8","L_star": 40.6219159552695, "a_star": 29.1709236856645, "b_star": -20.9018120847371},
{ "name": "7.5P4/6","L_star": 40.6543547569545, "a_star": 22.1460909982912, "b_star": -16.0595599907714},
{ "name": "7.5P4/4","L_star": 40.6705602368878, "a_star": 15.2435858533997, "b_star": -11.2105664998555},
{ "name": "7.5P4/3","L_star": 40.686756453899, "a_star": 11.7705419451266, "b_star": -8.70805804742031},
{ "name": "7.5P4/2","L_star": 40.7029434212134, "a_star": 8.11234235224709, "b_star": -6.05114051282526},
{ "name": "7.5P4/1","L_star": 40.7191211520259, "a_star": 4.22922923490682, "b_star": -3.16719999926619},
{ "name": "7.5P3/30","L_star": 29.4980822184077, "a_star": 115.188537297089, "b_star": -73.5433867325614},
{ "name": "7.5P3/28","L_star": 29.5758654475826, "a_star": 107.730165713563, "b_star": -69.5265045391748},
{ "name": "7.5P3/26","L_star": 29.6733462070223, "a_star": 98.8358777684152, "b_star": -64.4962051436207},
{ "name": "7.5P3/24","L_star": 29.7728960995735, "a_star": 91.2513091683612, "b_star": -60.0571620024553},
{ "name": "7.5P3/22","L_star": 29.8794315059155, "a_star": 83.1453703245256, "b_star": -55.1640708150389},
{ "name": "7.5P3/20","L_star": 29.9903947682641, "a_star": 75.6032977529946, "b_star": -50.52970098243},
{ "name": "7.5P3/18","L_star": 30.04444913391, "a_star": 69.2571734491725, "b_star": -46.6609158585306},
{ "name": "7.5P3/16","L_star": 30.0959283564591, "a_star": 61.6599185772325, "b_star": -41.951324824905},
{ "name": "7.5P3/14","L_star": 30.1277389127189, "a_star": 54.8286440220433, "b_star": -37.6829308679142},
{ "name": "7.5P3/12","L_star": 30.1643889669999, "a_star": 47.3801145843714, "b_star": -33.0305935461152},
{ "name": "7.5P3/10","L_star": 30.205855469926, "a_star": 39.4038319211245, "b_star": -27.8509990395933},
{ "name": "7.5P3/8","L_star": 30.2423818530976, "a_star": 32.0700682150265, "b_star": -22.9367431269983},
{ "name": "7.5P3/6","L_star": 30.2788506236177, "a_star": 24.2998247229158, "b_star": -17.5889102694194},
{ "name": "7.5P3/4","L_star": 30.3104104643861, "a_star": 17.1047395004718, "b_star": -12.5658971818913},
{ "name": "7.5P3/3","L_star": 30.3249620475793, "a_star": 13.4608872466716, "b_star": -9.92868121057551},
{ "name": "7.5P3/2","L_star": 30.3395044946333, "a_star": 9.50205304650795, "b_star": -7.05630627466801},
{ "name": "7.5P3/1","L_star": 30.3588802370237, "a_star": 5.0418129327417, "b_star": -3.76154548951754},
{ "name": "7.5P2/24","L_star": 19.4370092270703, "a_star": 92.2331087854296, "b_star": -59.8484993743942},
{ "name": "7.5P2/22","L_star": 19.5608725523075, "a_star": 84.190004015464, "b_star": -55.3340795054117},
{ "name": "7.5P2/20","L_star": 19.7165378422226, "a_star": 75.3732637973212, "b_star": -50.1037930528647},
{ "name": "7.5P2/18","L_star": 19.8425303140207, "a_star": 68.2254002694595, "b_star": -45.7488027110226},
{ "name": "7.5P2/16","L_star": 19.9394679877443, "a_star": 60.8346366743087, "b_star": -41.2347134776958},
{ "name": "7.5P2/14","L_star": 19.9957743188236, "a_star": 53.4502338203936, "b_star": -36.6512810935559},
{ "name": "7.5P2/12","L_star": 20.0318784296045, "a_star": 45.7738846714731, "b_star": -31.8346276275101},
{ "name": "7.5P2/10","L_star": 20.0599094643603, "a_star": 38.5058805471234, "b_star": -27.145432000508},
{ "name": "7.5P2/8","L_star": 20.0998783614812, "a_star": 30.8772154592144, "b_star": -22.178248739667},
{ "name": "7.5P2/6","L_star": 20.1397589488837, "a_star": 22.5582941052388, "b_star": -16.5770434563806},
{ "name": "7.5P2/4","L_star": 20.1676230763052, "a_star": 15.9825302341335, "b_star": -11.9544708205057},
{ "name": "7.5P2/3","L_star": 20.1835261774318, "a_star": 12.7123868247368, "b_star": -9.60705141328304},
{ "name": "7.5P2/2","L_star": 20.199415311619, "a_star": 9.08626657843037, "b_star": -6.93940669527039},
{ "name": "7.5P2/1","L_star": 20.2152905094929, "a_star": 4.89903635168251, "b_star": -3.78686675725637},
{ "name": "7.5P1/20","L_star": 9.74332916966028, "a_star": 74.4137299910978, "b_star": -48.3391719696154},
{ "name": "7.5P1/18","L_star": 9.83719811018735, "a_star": 67.2722709435066, "b_star": -44.2558974532959},
{ "name": "7.5P1/16","L_star": 9.96902283140891, "a_star": 61.037562917564, "b_star": -40.5658801686162},
{ "name": "7.5P1/14","L_star": 10.1071585302547, "a_star": 53.8726755671262, "b_star": -36.2072095743766},
{ "name": "7.5P1/12","L_star": 10.1908602815276, "a_star": 45.6549837434422, "b_star": -31.2701188449257},
{ "name": "7.5P1/10","L_star": 10.2362911796586, "a_star": 38.4503393827481, "b_star": -26.8401431813945},
{ "name": "7.5P1/8","L_star": 10.2664912537208, "a_star": 31.6711403388975, "b_star": -22.5284068349674},
{ "name": "7.5P1/6","L_star": 10.2966220415417, "a_star": 24.6281799654283, "b_star": -17.9641554833521},
{ "name": "7.5P1/4","L_star": 10.3266839391574, "a_star": 17.9112205652899, "b_star": -13.4276531276544},
{ "name": "7.5P1/3","L_star": 10.3416891769113, "a_star": 14.751822110363, "b_star": -11.2123219470411},
{ "name": "7.5P1/2","L_star": 10.3566773389942, "a_star": 10.8561253840365, "b_star": -8.40431758116372},
{ "name": "7.5P1/1","L_star": 10.3791276712618, "a_star": 5.99019616155498, "b_star": -4.75749956841803},
{ "name": "10P9/6","L_star": 89.8732681693111, "a_star": 23.2054649303991, "b_star": -10.271534953785},
{ "name": "10P9/4","L_star": 90.0956040750537, "a_star": 13.8072872221383, "b_star": -5.91313577589998},
{ "name": "10P9/3","L_star": 90.1464250044782, "a_star": 8.89294188268785, "b_star": -3.76128579467778},
{ "name": "10P9/2","L_star": 90.1648932823636, "a_star": 4.66580975515574, "b_star": -1.92807302486513},
{ "name": "10P9/1","L_star": 90.1787402752534, "a_star": 1.51509366178781, "b_star": -0.561685462810124},
{ "name": "10P8/14","L_star": 79.778582532371, "a_star": 57.2757938869932, "b_star": -27.2460475518135},
{ "name": "10P8/12","L_star": 80.0218487307532, "a_star": 47.2989649099013, "b_star": -22.3679986386885},
{ "name": "10P8/10","L_star": 80.1852185742504, "a_star": 38.5678196748924, "b_star": -18.1015998948245},
{ "name": "10P8/8","L_star": 80.3143942536544, "a_star": 30.0167593592256, "b_star": -13.9331388298013},
{ "name": "10P8/6","L_star": 80.4376302071718, "a_star": 20.9721020723436, "b_star": -9.57001255601568},
{ "name": "10P8/4","L_star": 80.4767756687889, "a_star": 12.9742430286331, "b_star": -5.73287817336681},
{ "name": "10P8/3","L_star": 80.4935425748736, "a_star": 8.75567965478957, "b_star": -3.77401218325244},
{ "name": "10P8/2","L_star": 80.5103036560818, "a_star": 4.85475128189688, "b_star": -2.01997555632796},
{ "name": "10P8/1","L_star": 80.5214744766754, "a_star": 1.80873362903605, "b_star": -0.691883853911324},
{ "name": "10P7/22","L_star": 70.1478887518532, "a_star": 88.1472862477522, "b_star": -43.0061550870154},
{ "name": "10P7/20","L_star": 70.2389320318477, "a_star": 78.7120787591116, "b_star": -38.2880067479406},
{ "name": "10P7/18","L_star": 70.3297834860997, "a_star": 70.2660372975766, "b_star": -34.0310582997188},
{ "name": "10P7/16","L_star": 70.4204441218562, "a_star": 61.8453365048497, "b_star": -29.8965226297616},
{ "name": "10P7/14","L_star": 70.5178663975755, "a_star": 52.922515512528, "b_star": -25.5398892838726},
{ "name": "10P7/12","L_star": 70.6011969247885, "a_star": 44.6136521559112, "b_star": -21.4254204036504},
{ "name": "10P7/10","L_star": 70.6774426153785, "a_star": 37.1660359676988, "b_star": -17.735154593658},
{ "name": "10P7/8","L_star": 70.7397258291086, "a_star": 29.2463643431685, "b_star": -13.8303941331027},
{ "name": "10P7/6","L_star": 70.7742890005338, "a_star": 21.2936073932008, "b_star": -9.91063953473938},
{ "name": "10P7/4","L_star": 70.8088246600745, "a_star": 13.4711793095859, "b_star": -6.17029143405532},
{ "name": "10P7/3","L_star": 70.8226312328179, "a_star": 9.50448356297756, "b_star": -4.3409896893758},
{ "name": "10P7/2","L_star": 70.8433328624133, "a_star": 5.91490282582841, "b_star": -2.62904688347325},
{ "name": "10P7/1","L_star": 70.8571284675481, "a_star": 2.67778955086617, "b_star": -1.15588595032752},
{ "name": "10P6/26","L_star": 60.5337161075712, "a_star": 101.119063792246, "b_star": -50.2729294206412},
{ "name": "10P6/24","L_star": 60.5514774718913, "a_star": 92.3068072276466, "b_star": -45.9227347627451},
{ "name": "10P6/22","L_star": 60.5869754957463, "a_star": 84.74129484209, "b_star": -42.1122467165205},
{ "name": "10P6/20","L_star": 60.6401609124894, "a_star": 76.5803735552042, "b_star": -38.019199300299},
{ "name": "10P6/18","L_star": 60.6932726141122, "a_star": 67.7146042357786, "b_star": -33.589349222789},
{ "name": "10P6/16","L_star": 60.7374762392385, "a_star": 60.3306672339937, "b_star": -29.8636363888423},
{ "name": "10P6/14","L_star": 60.7904534574736, "a_star": 52.8196305887791, "b_star": -26.0810227461342},
{ "name": "10P6/12","L_star": 60.8433576791392, "a_star": 44.5019839832488, "b_star": -21.9359940635943},
{ "name": "10P6/10","L_star": 60.8785867305749, "a_star": 37.0740829036388, "b_star": -18.2230031517151},
{ "name": "10P6/8","L_star": 60.9225776916113, "a_star": 29.2353900395807, "b_star": -14.3809178634901},
{ "name": "10P6/6","L_star": 60.9489481351208, "a_star": 22.044281731937, "b_star": -10.7905083612294},
{ "name": "10P6/4","L_star": 60.9928587518952, "a_star": 13.8877260144629, "b_star": -6.84219362551164},
{ "name": "10P6/3","L_star": 61.0016348671828, "a_star": 10.33460719242, "b_star": -4.98340395244761},
{ "name": "10P6/2","L_star": 61.0191810988183, "a_star": 6.88368100156084, "b_star": -3.20562180688235},
{ "name": "10P6/1","L_star": 61.0279512174433, "a_star": 3.45331903051227, "b_star": -1.52861297272009},
{ "name": "10P5/30","L_star": 50.6637988810833, "a_star": 114.814840208294, "b_star": -57.1701047520219},
{ "name": "10P5/28","L_star": 50.6286571777442, "a_star": 106.83192314004, "b_star": -53.397065597561},
{ "name": "10P5/26","L_star": 50.6169350351417, "a_star": 100.0096090843, "b_star": -50.0532616972752},
{ "name": "10P5/24","L_star": 50.6052087657505, "a_star": 91.3377371147101, "b_star": -45.8668466949314},
{ "name": "10P5/22","L_star": 50.6169350351417, "a_star": 83.8640235036982, "b_star": -42.1291525771477},
{ "name": "10P5/20","L_star": 50.640375197187, "a_star": 76.625514294392, "b_star": -38.5232080949036},
{ "name": "10P5/18","L_star": 50.6637988810833, "a_star": 68.6430578985086, "b_star": -34.655549620821},
{ "name": "10P5/16","L_star": 50.7105969300653, "a_star": 60.9624683946101, "b_star": -30.7526281063595},
{ "name": "10P5/14","L_star": 50.7456524262333, "a_star": 53.4894937648027, "b_star": -26.9799478320291},
{ "name": "10P5/12","L_star": 50.7923358843108, "a_star": 46.1800187075629, "b_star": -23.3073476699044},
{ "name": "10P5/10","L_star": 50.8273057000631, "a_star": 38.2936001173594, "b_star": -19.2551227130972},
{ "name": "10P5/8","L_star": 50.8622389556151, "a_star": 31.1726513784088, "b_star": -15.5828126912833},
{ "name": "10P5/6","L_star": 50.9087599231063, "a_star": 23.0533303610164, "b_star": -11.4298546266475},
{ "name": "10P5/4","L_star": 50.9436082418668, "a_star": 14.9129558279368, "b_star": -7.35797381908994},
{ "name": "10P5/3","L_star": 50.9552162893063, "a_star": 10.8986081100398, "b_star": -5.39542292001312},
{ "name": "10P5/2","L_star": 50.9784203169096, "a_star": 7.08660018460561, "b_star": -3.53693169696589},
{ "name": "10P5/1","L_star": 50.9900163040383, "a_star": 3.45756383764029, "b_star": -1.70622975043151},
{ "name": "10P4/30","L_star": 40.1967859089013, "a_star": 117.122613788088, "b_star": -60.2474925407952},
{ "name": "10P4/28","L_star": 40.2297169366013, "a_star": 107.674859276159, "b_star": -55.9654949040477},
{ "name": "10P4/26","L_star": 40.2790412749949, "a_star": 100.603778207176, "b_star": -52.3240707822798},
{ "name": "10P4/24","L_star": 40.3282793064743, "a_star": 92.6927336800647, "b_star": -48.2292900758167},
{ "name": "10P4/22","L_star": 40.3610568985711, "a_star": 85.4148357219744, "b_star": -44.4815854169737},
{ "name": "10P4/20","L_star": 40.3774314075978, "a_star": 79.4235876755854, "b_star": -41.4590627853076},
{ "name": "10P4/18","L_star": 40.3937964103769, "a_star": 70.4486007537597, "b_star": -36.8091921281317},
{ "name": "10P4/16","L_star": 40.4264979523013, "a_star": 61.9110726607075, "b_star": -32.3848260800727},
{ "name": "10P4/14","L_star": 40.4754793119243, "a_star": 54.6552404404668, "b_star": -28.6702979836406},
{ "name": "10P4/12","L_star": 40.5080864088407, "a_star": 46.9100520777461, "b_star": -24.6162889349813},
{ "name": "10P4/10","L_star": 40.5569266115533, "a_star": 39.3315504191228, "b_star": -20.6544598221273},
{ "name": "10P4/8","L_star": 40.5894399424351, "a_star": 31.9982127133504, "b_star": -16.7792478256065},
{ "name": "10P4/6","L_star": 40.6219159552695, "a_star": 24.4618887423544, "b_star": -12.7957570460568},
{ "name": "10P4/4","L_star": 40.6705602368878, "a_star": 16.7912289778457, "b_star": -8.75027756075452},
{ "name": "10P4/3","L_star": 40.686756453899, "a_star": 12.8648737150886, "b_star": -6.70641519253708},
{ "name": "10P4/2","L_star": 40.7029434212134, "a_star": 8.82780816120463, "b_star": -4.58736367995961},
{ "name": "10P4/1","L_star": 40.7191211520259, "a_star": 4.51769864982365, "b_star": -2.34994255042068},
{ "name": "10P3/26","L_star": 29.5783701665065, "a_star": 102.354260283385, "b_star": -55.5155172683447},
{ "name": "10P3/24","L_star": 29.7728960995735, "a_star": 95.3816189977865, "b_star": -52.1434335080445},
{ "name": "10P3/22","L_star": 29.8225090758691, "a_star": 87.3981844507824, "b_star": -47.8504443973712},
{ "name": "10P3/20","L_star": 29.9288147112578, "a_star": 79.9670225850009, "b_star": -43.9190793554341},
{ "name": "10P3/18","L_star": 30.010065598194, "a_star": 73.4197992863131, "b_star": -40.2875274570058},
{ "name": "10P3/16","L_star": 30.0738799179759, "a_star": 65.2890870090336, "b_star": -35.9322831764836},
{ "name": "10P3/14","L_star": 30.1057208955728, "a_star": 58.3878473121811, "b_star": -32.2867571010772},
{ "name": "10P3/12","L_star": 30.1277389127189, "a_star": 50.8293195618332, "b_star": -28.1334250212916},
{ "name": "10P3/10","L_star": 30.1765927290591, "a_star": 42.4288529271872, "b_star": -23.5391767003219},
{ "name": "10P3/8","L_star": 30.2204729512776, "a_star": 34.4602506635401, "b_star": -19.1274470724793},
{ "name": "10P3/6","L_star": 30.2618390191671, "a_star": 26.130435262702, "b_star": -14.5156576234228},
{ "name": "10P3/4","L_star": 30.2982771557024, "a_star": 18.5034028829337, "b_star": -10.2559246374016},
{ "name": "10P3/3","L_star": 30.3152620081296, "a_star": 14.543299097054, "b_star": -8.03794497873437},
{ "name": "10P3/2","L_star": 30.3346580265234, "a_star": 10.1694979005003, "b_star": -5.62378055038961},
{ "name": "10P3/1","L_star": 30.3564591549096, "a_star": 5.35026045554379, "b_star": -2.9630190094622},
{ "name": "10P2/22","L_star": 19.4328655168797, "a_star": 86.9419914561099, "b_star": -49.4419782359614},
{ "name": "10P2/20","L_star": 19.5773225939979, "a_star": 78.6472946933268, "b_star": -45.2167054215906},
{ "name": "10P2/18","L_star": 19.7246932341814, "a_star": 71.2577927162159, "b_star": -41.1212506334762},
{ "name": "10P2/16","L_star": 19.870857943764, "a_star": 63.6762787411636, "b_star": -36.774485511526},
{ "name": "10P2/14","L_star": 19.9676431897773, "a_star": 55.9663634442725, "b_star": -32.4147873877875},
{ "name": "10P2/12","L_star": 20.0158410942582, "a_star": 48.1350212139469, "b_star": -28.0061617183123},
{ "name": "10P2/10","L_star": 20.0398917465438, "a_star": 40.4132627544617, "b_star": -23.641591356935},
{ "name": "10P2/8","L_star": 20.07990498219, "a_star": 32.7416435995456, "b_star": -19.3331259319961},
{ "name": "10P2/6","L_star": 20.1238172798771, "a_star": 24.1193790711559, "b_star": -14.2952811232517},
{ "name": "10P2/4","L_star": 20.1596662785396, "a_star": 17.0346548756775, "b_star": -10.1267171373447},
{ "name": "10P2/3","L_star": 20.1755763746542, "a_star": 13.6561861884319, "b_star": -8.10148064550375},
{ "name": "10P2/2","L_star": 20.1914724884762, "a_star": 9.84205388954004, "b_star": -5.84074532750939},
{ "name": "10P2/1","L_star": 20.211323014883, "a_star": 5.34044230903211, "b_star": -3.18291102359238},
{ "name": "10P1/18","L_star": 9.75117774611292, "a_star": 68.6124570068433, "b_star": -41.4008544537786},
{ "name": "10P1/16","L_star": 9.87610961191719, "a_star": 62.104586746076, "b_star": -37.7089986461154},
{ "name": "10P1/14","L_star": 10.0229165999318, "a_star": 55.1027708241297, "b_star": -33.7011007239587},
{ "name": "10P1/12","L_star": 10.1604852351022, "a_star": 47.0609330808443, "b_star": -29.0188895320741},
{ "name": "10P1/10","L_star": 10.2211650355277, "a_star": 39.8264264339551, "b_star": -24.8393066209113},
{ "name": "10P1/8","L_star": 10.2513999023643, "a_star": 32.9870577886349, "b_star": -20.7809540572927},
{ "name": "10P1/6","L_star": 10.281565283574, "a_star": 25.7995701394011, "b_star": -16.441997219818},
{ "name": "10P1/4","L_star": 10.3191749016881, "a_star": 18.7598634801472, "b_star": -12.095989633003},
{ "name": "10P1/3","L_star": 10.3341886955329, "a_star": 15.5362364079782, "b_star": -10.0631978255136},
{ "name": "10P1/2","L_star": 10.3566773389942, "a_star": 11.5388054657386, "b_star": -7.52616689558084},
{ "name": "10P1/1","L_star": 10.3791276712618, "a_star": 6.45277572759434, "b_star": -4.25804978348775},
{ "name": "2.5RP9/6","L_star": 89.7989482970964, "a_star": 25.238839411002, "b_star": -6.08866983949226},
{ "name": "2.5RP9/4","L_star": 90.0771116603561, "a_star": 14.5468164693487, "b_star": -3.39283902555523},
{ "name": "2.5RP9/3","L_star": 90.1418069307284, "a_star": 9.41574857429966, "b_star": -2.12252091317759},
{ "name": "2.5RP9/2","L_star": 90.1602768152846, "a_star": 5.11541734379006, "b_star": -1.08282660747425},
{ "name": "2.5RP9/1","L_star": 90.1741250122608, "a_star": 1.76439899731512, "b_star": -0.31800570952647},
{ "name": "2.5RP8/14","L_star": 79.7331871431278, "a_star": 60.7813571217198, "b_star": -16.8884845831248},
{ "name": "2.5RP8/12","L_star": 79.8975411496159, "a_star": 51.2236543871221, "b_star": -13.9023777459957},
{ "name": "2.5RP8/10","L_star": 80.0838820090659, "a_star": 41.3708043152609, "b_star": -10.8328268223706},
{ "name": "2.5RP8/8","L_star": 80.2526578381145, "a_star": 32.3808247879796, "b_star": -8.25680877047219},
{ "name": "2.5RP8/6","L_star": 80.4264399506073, "a_star": 22.8140716145193, "b_star": -5.4606789551007},
{ "name": "2.5RP8/4","L_star": 80.4655944928565, "a_star": 13.7679093222793, "b_star": -3.24326552077474},
{ "name": "2.5RP8/3","L_star": 80.4879542536371, "a_star": 9.26798908065946, "b_star": -2.16834929649445},
{ "name": "2.5RP8/2","L_star": 80.5047172759086, "a_star": 5.21803850729874, "b_star": -1.16923519973153},
{ "name": "2.5RP8/1","L_star": 80.5158893896084, "a_star": 1.99385076346587, "b_star": -0.389593668966604},
{ "name": "2.5RP7/20","L_star": 70.1969359540931, "a_star": 84.4168676666696, "b_star": -25.7208044243017},
{ "name": "2.5RP7/18","L_star": 70.2948632559789, "a_star": 75.2209841050074, "b_star": -22.6583443558931},
{ "name": "2.5RP7/16","L_star": 70.3855971648033, "a_star": 65.8832442822282, "b_star": -19.5231278532597},
{ "name": "2.5RP7/14","L_star": 70.4761408707772, "a_star": 56.4445644562913, "b_star": -16.4490773933689},
{ "name": "2.5RP7/12","L_star": 70.5664953675691, "a_star": 47.9919012395393, "b_star": -13.7497452093304},
{ "name": "2.5RP7/10","L_star": 70.6635897395448, "a_star": 39.3516028922374, "b_star": -10.9909821745582},
{ "name": "2.5RP7/8","L_star": 70.7258928449287, "a_star": 31.4461742403648, "b_star": -8.64631130374571},
{ "name": "2.5RP7/6","L_star": 70.7673785689525, "a_star": 22.9814655037466, "b_star": -6.16045462998542},
{ "name": "2.5RP7/4","L_star": 70.8019197264899, "a_star": 14.2652246431715, "b_star": -3.608020349927},
{ "name": "2.5RP7/3","L_star": 70.8226312328179, "a_star": 10.1825005878691, "b_star": -2.54009385258509},
{ "name": "2.5RP7/2","L_star": 70.8364334159028, "a_star": 6.35410487951682, "b_star": -1.55569554938548},
{ "name": "2.5RP7/1","L_star": 70.8502312128167, "a_star": 2.92675780273483, "b_star": -0.681727932485554},
{ "name": "2.5RP6/24","L_star": 60.5958448618355, "a_star": 95.9450836425794, "b_star": -30.7495132357245},
{ "name": "2.5RP6/22","L_star": 60.613577434494, "a_star": 87.9679998620713, "b_star": -28.1154551537726},
{ "name": "2.5RP6/20","L_star": 60.6401609124894, "a_star": 81.0238736229988, "b_star": -25.6705876634769},
{ "name": "2.5RP6/18","L_star": 60.6932726141122, "a_star": 72.0653291279955, "b_star": -22.5566612474043},
{ "name": "2.5RP6/16","L_star": 60.7286395882807, "a_star": 64.2795848320161, "b_star": -19.8897239956686},
{ "name": "2.5RP6/14","L_star": 60.7816289971043, "a_star": 55.7942859082584, "b_star": -17.1087240947573},
{ "name": "2.5RP6/12","L_star": 60.8345453674015, "a_star": 47.4356583730329, "b_star": -14.4121958587064},
{ "name": "2.5RP6/10","L_star": 60.8785867305749, "a_star": 38.945657815798, "b_star": -11.5977745539935},
{ "name": "2.5RP6/8","L_star": 60.9137835246555, "a_star": 31.162538085631, "b_star": -9.11866909933359},
{ "name": "2.5RP6/6","L_star": 60.940159995696, "a_star": 23.5659724576566, "b_star": -6.74971045949593},
{ "name": "2.5RP6/4","L_star": 60.984080635441, "a_star": 14.9661185580621, "b_star": -4.16454945328331},
{ "name": "2.5RP6/3","L_star": 60.9928587518952, "a_star": 11.1140054791142, "b_star": -3.02504555597825},
{ "name": "2.5RP6/2","L_star": 61.0104089824442, "a_star": 7.32904411179919, "b_star": -1.91171775570049},
{ "name": "2.5RP6/1","L_star": 61.0279512174433, "a_star": 3.63996918129883, "b_star": -0.887702785531364},
{ "name": "2.5RP5/26","L_star": 50.6989035736593, "a_star": 103.00124353109, "b_star": -34.6930124595537},
{ "name": "2.5RP5/24","L_star": 50.68720611576, "a_star": 95.3190642457932, "b_star": -32.123839290986},
{ "name": "2.5RP5/22","L_star": 50.6637988810833, "a_star": 88.6140702445185, "b_star": -29.7697372302472},
{ "name": "2.5RP5/20","L_star": 50.640375197187, "a_star": 80.5637859030487, "b_star": -26.9672346121608},
{ "name": "2.5RP5/18","L_star": 50.6520890970938, "a_star": 72.6675802657272, "b_star": -24.1242558017836},
{ "name": "2.5RP5/16","L_star": 50.68720611576, "a_star": 64.8790517428478, "b_star": -21.3656462040608},
{ "name": "2.5RP5/14","L_star": 50.7339713527665, "a_star": 57.3399504049507, "b_star": -18.7231513957462},
{ "name": "2.5RP5/12","L_star": 50.7806711380232, "a_star": 49.3656987662258, "b_star": -15.9576533154033},
{ "name": "2.5RP5/10","L_star": 50.8156531617798, "a_star": 40.9708546383983, "b_star": -13.0295140248292},
{ "name": "2.5RP5/8","L_star": 50.8505985934447, "a_star": 33.6116779307357, "b_star": -10.5095677535796},
{ "name": "2.5RP5/6","L_star": 50.8971357463743, "a_star": 24.6420194090107, "b_star": -7.57467864857848},
{ "name": "2.5RP5/4","L_star": 50.9319961673508, "a_star": 15.9379308519219, "b_star": -4.81028870225368},
{ "name": "2.5RP5/3","L_star": 50.9552162893063, "a_star": 11.760781978855, "b_star": -3.50142458411598},
{ "name": "2.5RP5/2","L_star": 50.966820313159, "a_star": 7.70752645829509, "b_star": -2.25228547917733},
{ "name": "2.5RP5/1","L_star": 50.9784203169096, "a_star": 3.77122977422245, "b_star": -1.0601830516936},
{ "name": "2.5RP4/26","L_star": 40.2790412749949, "a_star": 104.027844304293, "b_star": -36.8199168733022},
{ "name": "2.5RP4/24","L_star": 40.3446728694798, "a_star": 97.5576261175725, "b_star": -34.9118952150605},
{ "name": "2.5RP4/22","L_star": 40.3937964103769, "a_star": 90.039361556272, "b_star": -32.1683414367236},
{ "name": "2.5RP4/20","L_star": 40.4264979523013, "a_star": 83.2963835585619, "b_star": -29.6302153396079},
{ "name": "2.5RP4/18","L_star": 40.4264979523013, "a_star": 74.0069674307909, "b_star": -26.2031133907435},
{ "name": "2.5RP4/16","L_star": 40.4264979523013, "a_star": 65.6420609979985, "b_star": -23.1005701915448},
{ "name": "2.5RP4/14","L_star": 40.4591616342428, "a_star": 57.9293690808098, "b_star": -20.2079213289562},
{ "name": "2.5RP4/12","L_star": 40.5080864088407, "a_star": 50.2080278972818, "b_star": -17.3511664989908},
{ "name": "2.5RP4/10","L_star": 40.5406559183046, "a_star": 42.0475254939627, "b_star": -14.3653581675166},
{ "name": "2.5RP4/8","L_star": 40.5731879484504, "a_star": 34.5415675471044, "b_star": -11.6416159591227},
{ "name": "2.5RP4/6","L_star": 40.6219159552695, "a_star": 26.2731491390294, "b_star": -8.76012452989767},
{ "name": "2.5RP4/4","L_star": 40.6543547569545, "a_star": 17.7154338551706, "b_star": -5.79807650980929},
{ "name": "2.5RP4/3","L_star": 40.6705602368878, "a_star": 13.556566845813, "b_star": -4.42154063970778},
{ "name": "2.5RP4/2","L_star": 40.686756453899, "a_star": 9.29870671766245, "b_star": -2.98884449681466},
{ "name": "2.5RP4/1","L_star": 40.7191211520259, "a_star": 4.85289796612817, "b_star": -1.51403626906005},
{ "name": "2.5RP3/22","L_star": 29.615907961572, "a_star": 92.9309255607129, "b_star": -35.7334793033434},
{ "name": "2.5RP3/20","L_star": 29.7902728718213, "a_star": 84.4313216302481, "b_star": -32.5066238831604},
{ "name": "2.5RP3/18","L_star": 29.9238811686331, "a_star": 77.7123434909872, "b_star": -29.6917711655405},
{ "name": "2.5RP3/16","L_star": 30.0419948697865, "a_star": 69.5040377642008, "b_star": -26.2633389646094},
{ "name": "2.5RP3/14","L_star": 30.1057208955728, "a_star": 61.5451524983305, "b_star": -23.0585790764406},
{ "name": "2.5RP3/12","L_star": 30.1228478365595, "a_star": 53.5384248598959, "b_star": -19.9244061824613},
{ "name": "2.5RP3/10","L_star": 30.1619474401831, "a_star": 45.2202098446491, "b_star": -16.6738571144583},
{ "name": "2.5RP3/8","L_star": 30.2082923590962, "a_star": 36.5474836453333, "b_star": -13.2754871684309},
{ "name": "2.5RP3/6","L_star": 30.2521124824243, "a_star": 27.9288505460017, "b_star": -10.0231625916203},
{ "name": "2.5RP3/4","L_star": 30.2885659279072, "a_star": 19.7851885358672, "b_star": -6.95856326830483},
{ "name": "2.5RP3/3","L_star": 30.3079843112769, "a_star": 15.3996437579386, "b_star": -5.35865491491786},
{ "name": "2.5RP3/2","L_star": 30.3298105443489, "a_star": 10.7183005513222, "b_star": -3.69368402145024},
{ "name": "2.5RP3/1","L_star": 30.3491943907795, "a_star": 5.62240961685739, "b_star": -1.91093071746005},
{ "name": "2.5RP2/20","L_star": 19.4370092270703, "a_star": 83.0469619970225, "b_star": -37.7450212888435},
{ "name": "2.5RP2/18","L_star": 19.5526418211833, "a_star": 74.7481736963302, "b_star": -33.22694301071},
{ "name": "2.5RP2/16","L_star": 19.7246932341814, "a_star": 67.0635084716186, "b_star": -29.7047120478375},
{ "name": "2.5RP2/14","L_star": 19.9031776890943, "a_star": 59.1879015314453, "b_star": -25.932768786344},
{ "name": "2.5RP2/12","L_star": 19.9997894637893, "a_star": 50.7997936366411, "b_star": -21.9870245943898},
{ "name": "2.5RP2/10","L_star": 20.0358855335555, "a_star": 42.737324104564, "b_star": -18.3327920274948},
{ "name": "2.5RP2/8","L_star": 20.0679103320434, "a_star": 34.6627796631476, "b_star": -14.7205708026089},
{ "name": "2.5RP2/6","L_star": 20.115841166286, "a_star": 25.6147571023898, "b_star": -10.6996203367298},
{ "name": "2.5RP2/4","L_star": 20.1517059775056, "a_star": 18.0736593890872, "b_star": -7.45000602331295},
{ "name": "2.5RP2/3","L_star": 20.1676230763052, "a_star": 14.6044860672773, "b_star": -5.97536906192964},
{ "name": "2.5RP2/2","L_star": 20.1835261774318, "a_star": 10.6886526273336, "b_star": -4.33211044558706},
{ "name": "2.5RP2/1","L_star": 20.2033854164229, "a_star": 5.91394496454273, "b_star": -2.3752786807132},
{ "name": "2.5RP1/16","L_star": 9.74332916966028, "a_star": 63.3164242820819, "b_star": -34.3613404549445},
{ "name": "2.5RP1/14","L_star": 9.93038989303505, "a_star": 56.3478887352516, "b_star": -30.48745170465},
{ "name": "2.5RP1/12","L_star": 10.1071585302547, "a_star": 48.2734934067164, "b_star": -26.0825663724085},
{ "name": "2.5RP1/10","L_star": 10.1984430440478, "a_star": 41.238912715392, "b_star": -22.2661457396606},
{ "name": "2.5RP1/8","L_star": 10.2513999023643, "a_star": 34.272365949258, "b_star": -18.5228685534383},
{ "name": "2.5RP1/6","L_star": 10.2740304307288, "a_star": 26.7853257367363, "b_star": -14.5000291085525},
{ "name": "2.5RP1/4","L_star": 10.3116615770145, "a_star": 19.6981916567031, "b_star": -10.6521274349314},
{ "name": "2.5RP1/3","L_star": 10.3266839391574, "a_star": 16.2980467451835, "b_star": -8.7693481508946},
{ "name": "2.5RP1/2","L_star": 10.3416891769113, "a_star": 12.0940063847948, "b_star": -6.49421125884589},
{ "name": "2.5RP1/1","L_star": 10.3641650318228, "a_star": 6.79654729842523, "b_star": -3.64312232371331},
{ "name": "5RP9/6","L_star": 89.8128912266592, "a_star": 26.3020544636068, "b_star": -1.19746874362774},
{ "name": "5RP9/4","L_star": 90.0678630346794, "a_star": 15.2947577956375, "b_star": -0.419594477948548},
{ "name": "5RP9/3","L_star": 90.132569577483, "a_star": 9.9726206387904, "b_star": -0.19680036698162},
{ "name": "5RP9/2","L_star": 90.1510426764292, "a_star": 5.32288714262647, "b_star": -0.0567255727094818},
{ "name": "5RP9/1","L_star": 90.1648932823636, "a_star": 1.87261221776863, "b_star": 0.0330142522724675},
{ "name": "5RP8/12","L_star": 79.7955947143509, "a_star": 54.4870326612367, "b_star": -4.11897418362486},
{ "name": "5RP8/10","L_star": 80.0669717866102, "a_star": 43.3153793484877, "b_star": -2.837376154244},
{ "name": "5RP8/8","L_star": 80.2358068822662, "a_star": 34.0931157486171, "b_star": -1.93592287905076},
{ "name": "5RP8/6","L_star": 80.420843848224, "a_star": 23.9148340899955, "b_star": -1.00635882396081},
{ "name": "5RP8/4","L_star": 80.4600029327644, "a_star": 14.4376830637167, "b_star": -0.486589510599189},
{ "name": "5RP8/3","L_star": 80.4767756687889, "a_star": 9.72089682472871, "b_star": -0.283978676850039},
{ "name": "5RP8/2","L_star": 80.4935425748736, "a_star": 5.53510788274703, "b_star": -0.139790779685423},
{ "name": "5RP8/1","L_star": 80.5103036560818, "a_star": 2.13716114381146, "b_star": -0.00660980839799485},
{ "name": "5RP7/18","L_star": 70.2179391069599, "a_star": 78.1330570837124, "b_star": -7.40121361504873},
{ "name": "5RP7/16","L_star": 70.3158187845481, "a_star": 69.2627810121856, "b_star": -6.49070038918249},
{ "name": "5RP7/14","L_star": 70.4552629991617, "a_star": 59.590679196612, "b_star": -5.13196189896072},
{ "name": "5RP7/12","L_star": 70.5526069515811, "a_star": 50.4398260900835, "b_star": -3.97586045168892},
{ "name": "5RP7/10","L_star": 70.6635897395448, "a_star": 40.73704888076, "b_star": -2.81268852314784},
{ "name": "5RP7/8","L_star": 70.7120554465534, "a_star": 32.9591886494087, "b_star": -2.04505517853366},
{ "name": "5RP7/6","L_star": 70.7535544026117, "a_star": 23.9795712688431, "b_star": -1.38047649784343},
{ "name": "5RP7/4","L_star": 70.7950136941808, "a_star": 14.8552866278057, "b_star": -0.743151148533738},
{ "name": "5RP7/3","L_star": 70.8157284953717, "a_star": 10.5982746257449, "b_star": -0.46826573989962},
{ "name": "5RP7/2","L_star": 70.8295328728496, "a_star": 6.70733840040394, "b_star": -0.246969456642154},
{ "name": "5RP7/1","L_star": 70.8433328624133, "a_star": 3.13739667672897, "b_star": -0.0911747563081544},
{ "name": "5RP6/22","L_star": 60.613577434494, "a_star": 91.4925256410369, "b_star": -10.7049349220891},
{ "name": "5RP6/20","L_star": 60.6401609124894, "a_star": 85.0708091049252, "b_star": -9.96218603467116},
{ "name": "5RP6/18","L_star": 60.6844257721126, "a_star": 75.1362417018364, "b_star": -8.44042102968372},
{ "name": "5RP6/16","L_star": 60.7198009014627, "a_star": 66.7435844185164, "b_star": -7.12617021794102},
{ "name": "5RP6/14","L_star": 60.7639739886647, "a_star": 58.3500494814674, "b_star": -5.93138846858148},
{ "name": "5RP6/12","L_star": 60.8169146771618, "a_star": 49.4500763431653, "b_star": -4.60187426900014},
{ "name": "5RP6/10","L_star": 60.8609762416457, "a_star": 40.6788382055475, "b_star": -3.52491857010031},
{ "name": "5RP6/8","L_star": 60.8961891551695, "a_star": 32.905279751453, "b_star": -2.67196975034003},
{ "name": "5RP6/6","L_star": 60.9313698482419, "a_star": 24.9048809149234, "b_star": -1.86539460937052},
{ "name": "5RP6/4","L_star": 60.9753005166793, "a_star": 15.7500742233049, "b_star": -0.998281289296821},
{ "name": "5RP6/3","L_star": 60.9928587518952, "a_star": 11.5728199793106, "b_star": -0.691309725831202},
{ "name": "5RP6/2","L_star": 61.0104089824442, "a_star": 7.62342103781766, "b_star": -0.433139405404748},
{ "name": "5RP6/1","L_star": 61.0191810988183, "a_star": 3.79038191993436, "b_star": -0.181865782220791},
{ "name": "5RP5/24","L_star": 50.7339713527665, "a_star": 100.194023362108, "b_star": -13.5210426057743},
{ "name": "5RP5/22","L_star": 50.7105969300653, "a_star": 92.0716198824842, "b_star": -12.6134080505934},
{ "name": "5RP5/20","L_star": 50.6755045527692, "a_star": 84.5838432923533, "b_star": -11.4423871469072},
{ "name": "5RP5/18","L_star": 50.6286571777442, "a_star": 76.2306356200929, "b_star": -10.1215948471494},
{ "name": "5RP5/16","L_star": 50.6637988810833, "a_star": 68.2457630199749, "b_star": -8.73661084817088},
{ "name": "5RP5/14","L_star": 50.7105969300653, "a_star": 59.9736115611597, "b_star": -7.28603193775719},
{ "name": "5RP5/12","L_star": 50.7573294125503, "a_star": 52.0084842602628, "b_star": -5.9967163145501},
{ "name": "5RP5/10","L_star": 50.8039965577119, "a_star": 43.1596043659428, "b_star": -4.68147094920175},
{ "name": "5RP5/8","L_star": 50.8389541761052, "a_star": 35.1225410358003, "b_star": -3.54372547338209},
{ "name": "5RP5/6","L_star": 50.8855075285586, "a_star": 25.8941541924033, "b_star": -2.47631188253865},
{ "name": "5RP5/4","L_star": 50.9203800622637, "a_star": 16.8623927999195, "b_star": -1.47335130127113},
{ "name": "5RP5/3","L_star": 50.9436082418668, "a_star": 12.3994948463469, "b_star": -1.03456024394575},
{ "name": "5RP5/2","L_star": 50.9552162893063, "a_star": 7.99847722495678, "b_star": -0.597501170446657},
{ "name": "5RP5/1","L_star": 50.9784203169096, "a_star": 3.90119861248117, "b_star": -0.24338113318092},
{ "name": "5RP4/22","L_star": 40.3446728694798, "a_star": 93.2193802850367, "b_star": -16.8194486301231},
{ "name": "5RP4/20","L_star": 40.4264979523013, "a_star": 86.4128027015687, "b_star": -15.5063558818051},
{ "name": "5RP4/18","L_star": 40.4591616342428, "a_star": 77.3358539309461, "b_star": -13.4875420253468},
{ "name": "5RP4/16","L_star": 40.4264979523013, "a_star": 68.6401186804124, "b_star": -11.6118630670454},
{ "name": "5RP4/14","L_star": 40.4428345189212, "a_star": 60.5957593366136, "b_star": -9.89700856797343},
{ "name": "5RP4/12","L_star": 40.4754793119243, "a_star": 52.6073506688532, "b_star": -8.22320550930452},
{ "name": "5RP4/10","L_star": 40.5243758552341, "a_star": 43.8132429514606, "b_star": -6.4794240621322},
{ "name": "5RP4/8","L_star": 40.5731879484504, "a_star": 36.4275473563299, "b_star": -5.09384491046412},
{ "name": "5RP4/6","L_star": 40.6056826069156, "a_star": 27.6658569895534, "b_star": -3.64018846119433},
{ "name": "5RP4/4","L_star": 40.6543547569545, "a_star": 18.5882801204139, "b_star": -2.33425877440024},
{ "name": "5RP4/3","L_star": 40.6705602368878, "a_star": 14.2272383135205, "b_star": -1.73192071474183},
{ "name": "5RP4/2","L_star": 40.686756453899, "a_star": 9.76549597729262, "b_star": -1.11851165712455},
{ "name": "5RP4/1","L_star": 40.7029434212134, "a_star": 5.10171723530209, "b_star": -0.53258939330173},
{ "name": "5RP3/20","L_star": 29.6658624536183, "a_star": 86.5887286521277, "b_star": -21.198397202016},
{ "name": "5RP3/18","L_star": 29.8497504809689, "a_star": 79.5814699425756, "b_star": -19.1533502787279},
{ "name": "5RP3/16","L_star": 30.010065598194, "a_star": 72.1398776189472, "b_star": -16.7841282283697},
{ "name": "5RP3/14","L_star": 30.1032731507766, "a_star": 63.68849057789, "b_star": -14.2174939366687},
{ "name": "5RP3/12","L_star": 30.1228478365595, "a_star": 55.2223261345055, "b_star": -11.8315071923772},
{ "name": "5RP3/10","L_star": 30.1497359303174, "a_star": 46.8958319750395, "b_star": -9.61587972336926},
{ "name": "5RP3/8","L_star": 30.1961053418733, "a_star": 38.2470874340428, "b_star": -7.41013851476184},
{ "name": "5RP3/6","L_star": 30.239948555795, "a_star": 29.3108637695388, "b_star": -5.2930166656345},
{ "name": "5RP3/4","L_star": 30.2812798320954, "a_star": 20.6359075839966, "b_star": -3.48252442929},
{ "name": "5RP3/3","L_star": 30.3007043262, "a_star": 15.9607091237354, "b_star": -2.55975198506659},
{ "name": "5RP3/2","L_star": 30.3225374185553, "a_star": 10.9721281411252, "b_star": -1.66221707112254},
{ "name": "5RP3/1","L_star": 30.3419273485797, "a_star": 5.71667101472942, "b_star": -0.825064907723772},
{ "name": "5RP2/18","L_star": 19.4452937415106, "a_star": 76.6989784742783, "b_star": -25.8460893403738},
{ "name": "5RP2/16","L_star": 19.5896501495909, "a_star": 69.0423010047294, "b_star": -22.3854803525736},
{ "name": "5RP2/14","L_star": 19.8060431524101, "a_star": 60.8604754186646, "b_star": -19.409623292248},
{ "name": "5RP2/12","L_star": 19.9756851449229, "a_star": 52.3745255851528, "b_star": -16.2529406893022},
{ "name": "5RP2/10","L_star": 20.0318784296045, "a_star": 44.2211517722018, "b_star": -13.4383854177512},
{ "name": "5RP2/8","L_star": 20.0599094643603, "a_star": 35.8766068076034, "b_star": -10.4876777079027},
{ "name": "5RP2/6","L_star": 20.1078615281194, "a_star": 26.6367963360348, "b_star": -7.39568169730403},
{ "name": "5RP2/4","L_star": 20.1397589488837, "a_star": 18.8593812322208, "b_star": -4.97709968136905},
{ "name": "5RP2/3","L_star": 20.1556865661722, "a_star": 15.1402676519303, "b_star": -3.86988678319602},
{ "name": "5RP2/2","L_star": 20.1755763746542, "a_star": 11.001289785498, "b_star": -2.73467607047089},
{ "name": "5RP2/1","L_star": 20.1954443357961, "a_star": 6.1135132738703, "b_star": -1.46424848425047},
{ "name": "5RP1/14","L_star": 9.774694816485, "a_star": 57.2423251113291, "b_star": -25.4327747131203},
{ "name": "5RP1/12","L_star": 9.99214753003156, "a_star": 50.1985028661522, "b_star": -22.0413019069338},
{ "name": "5RP1/10","L_star": 10.1528804424542, "a_star": 42.8945032532341, "b_star": -18.5439357476016},
{ "name": "5RP1/8","L_star": 10.2362911796586, "a_star": 35.5060586505428, "b_star": -15.280314325909},
{ "name": "5RP1/6","L_star": 10.2664912537208, "a_star": 27.7853399357726, "b_star": -11.7791049827974},
{ "name": "5RP1/4","L_star": 10.2966220415417, "a_star": 20.5228564229231, "b_star": -8.50377172367723},
{ "name": "5RP1/3","L_star": 10.3116615770145, "a_star": 16.9326345924375, "b_star": -6.90168045725681},
{ "name": "5RP1/2","L_star": 10.3341886955329, "a_star": 12.5612416001951, "b_star": -5.04935311239577},
{ "name": "5RP1/1","L_star": 10.3491853893754, "a_star": 7.09028336867222, "b_star": -2.80773345008826},
{ "name": "7.5RP9/6","L_star": 89.7942998371477, "a_star": 26.8024410938789, "b_star": 2.63805215013555},
{ "name": "7.5RP9/4","L_star": 90.0678630346794, "a_star": 15.7206884631939, "b_star": 1.83659201107376},
{ "name": "7.5RP9/3","L_star": 90.1279502978124, "a_star": 10.2389379459304, "b_star": 1.30355630112249},
{ "name": "7.5RP9/2","L_star": 90.1464250044782, "a_star": 5.57322007203009, "b_star": 0.773998212480032},
{ "name": "7.5RP9/1","L_star": 90.1602768152846, "a_star": 1.93017838005177, "b_star": 0.314124356213052},
{ "name": "7.5RP8/12","L_star": 79.7445400263819, "a_star": 55.4992521481416, "b_star": 3.41531205418542},
{ "name": "7.5RP8/10","L_star": 79.9879785872975, "a_star": 44.2763098840201, "b_star": 3.07155439302484},
{ "name": "7.5RP8/8","L_star": 80.2358068822662, "a_star": 34.6738614941738, "b_star": 2.7727881038262},
{ "name": "7.5RP8/6","L_star": 80.4096496943127, "a_star": 24.3783980548115, "b_star": 2.43795707245074},
{ "name": "7.5RP8/4","L_star": 80.448817867389, "a_star": 14.9438183917756, "b_star": 1.61604052791533},
{ "name": "7.5RP8/3","L_star": 80.4711854048021, "a_star": 10.1300230045945, "b_star": 1.09680446672771},
{ "name": "7.5RP8/2","L_star": 80.4879542536371, "a_star": 5.75971910529027, "b_star": 0.664970753871774},
{ "name": "7.5RP8/1","L_star": 80.5047172759086, "a_star": 2.23425870627431, "b_star": 0.312817392827136},
{ "name": "7.5RP7/16","L_star": 70.2669067096778, "a_star": 70.5424764362296, "b_star": 3.45933215408962},
{ "name": "7.5RP7/14","L_star": 70.3995393208176, "a_star": 61.1754734837257, "b_star": 3.16399831688858},
{ "name": "7.5RP7/12","L_star": 70.5456610718505, "a_star": 51.5152173545406, "b_star": 3.16498326494501},
{ "name": "7.5RP7/10","L_star": 70.6566616405674, "a_star": 41.4825789909707, "b_star": 2.97448881133184},
{ "name": "7.5RP7/8","L_star": 70.7051350909417, "a_star": 33.1567013626129, "b_star": 2.61404060836503},
{ "name": "7.5RP7/6","L_star": 70.746640666975, "a_star": 24.4144581515781, "b_star": 2.14966543021482},
{ "name": "7.5RP7/4","L_star": 70.7881065627101, "a_star": 15.1523102361446, "b_star": 1.47160681175675},
{ "name": "7.5RP7/3","L_star": 70.8088246600745, "a_star": 10.9023276793194, "b_star": 1.07512836846655},
{ "name": "7.5RP7/2","L_star": 70.8226312328179, "a_star": 6.93865095090818, "b_star": 0.713414877367113},
{ "name": "7.5RP7/1","L_star": 70.8364334159028, "a_star": 3.30397823915801, "b_star": 0.355392898590878},
{ "name": "7.5RP6/20","L_star": 60.6401609124894, "a_star": 85.7971931560142, "b_star": 2.13957244643823},
{ "name": "7.5RP6/18","L_star": 60.675576888377, "a_star": 76.6235296069255, "b_star": 1.94807711126321},
{ "name": "7.5RP6/16","L_star": 60.7109601776115, "a_star": 68.5777641850106, "b_star": 2.23171651962089},
{ "name": "7.5RP6/14","L_star": 60.7551434382602, "a_star": 59.2910170668685, "b_star": 2.40021325468891},
{ "name": "7.5RP6/12","L_star": 60.8080962963385, "a_star": 50.4369085898619, "b_star": 2.5433612982503},
{ "name": "7.5RP6/10","L_star": 60.852167970168, "a_star": 41.657009498245, "b_star": 2.52032938268356},
{ "name": "7.5RP6/8","L_star": 60.887388950337, "a_star": 33.1288420759691, "b_star": 2.25958227158669},
{ "name": "7.5RP6/6","L_star": 60.9225776916113, "a_star": 25.5351612550943, "b_star": 1.87078123229176},
{ "name": "7.5RP6/4","L_star": 60.9665183944677, "a_star": 16.088991382874, "b_star": 1.272569071852},
{ "name": "7.5RP6/3","L_star": 60.984080635441, "a_star": 11.7363380051103, "b_star": 0.975965286392277},
{ "name": "7.5RP6/2","L_star": 61.0016348671828, "a_star": 7.64072980176383, "b_star": 0.632337893070756},
{ "name": "7.5RP6/1","L_star": 61.0191810988183, "a_star": 3.72934095401878, "b_star": 0.37167417744135},
{ "name": "7.5RP5/22","L_star": 50.7339713527665, "a_star": 93.8815419349682, "b_star": -0.251104587123829},
{ "name": "7.5RP5/20","L_star": 50.7105969300653, "a_star": 86.3148904554145, "b_star": -0.0277212707923802},
{ "name": "7.5RP5/18","L_star": 50.6169350351417, "a_star": 77.5484466684201, "b_star": 0.256106254312871},
{ "name": "7.5RP5/16","L_star": 50.6520890970938, "a_star": 69.4922514831349, "b_star": 0.814286408091935},
{ "name": "7.5RP5/14","L_star": 50.6989035736593, "a_star": 60.8962705287396, "b_star": 1.24987236477003},
{ "name": "7.5RP5/12","L_star": 50.7456524262333, "a_star": 53.2966160646695, "b_star": 1.49411873203154},
{ "name": "7.5RP5/10","L_star": 50.7923358843108, "a_star": 43.8333206118084, "b_star": 1.71057995549411},
{ "name": "7.5RP5/8","L_star": 50.8273057000631, "a_star": 35.7585225084761, "b_star": 1.6442785291088},
{ "name": "7.5RP5/6","L_star": 50.8738752661451, "a_star": 26.5005171768596, "b_star": 1.51702881686264},
{ "name": "7.5RP5/4","L_star": 50.9203800622637, "a_star": 17.3650416337145, "b_star": 1.17503354315094},
{ "name": "7.5RP5/3","L_star": 50.9319961673508, "a_star": 12.7797936637863, "b_star": 0.885079548821222},
{ "name": "7.5RP5/2","L_star": 50.9552162893063, "a_star": 8.20630036543835, "b_star": 0.581127692190719},
{ "name": "7.5RP5/1","L_star": 50.966820313159, "a_star": 3.96580608409869, "b_star": 0.282581985908781},
{ "name": "7.5RP4/20","L_star": 40.3937964103769, "a_star": 88.5691834724021, "b_star": -3.21605248837104},
{ "name": "7.5RP4/18","L_star": 40.4591616342428, "a_star": 79.368996719683, "b_star": -2.33429163588537},
{ "name": "7.5RP4/16","L_star": 40.4591616342428, "a_star": 70.6074778970884, "b_star": -1.5341111808481},
{ "name": "7.5RP4/14","L_star": 40.4264979523013, "a_star": 62.0272029205629, "b_star": -0.899690963844746},
{ "name": "7.5RP4/12","L_star": 40.4754793119243, "a_star": 53.6479578557102, "b_star": -0.310294468409034},
{ "name": "7.5RP4/10","L_star": 40.5080864088407, "a_star": 44.974899765879, "b_star": 0.26602693126484},
{ "name": "7.5RP4/8","L_star": 40.5569266115533, "a_star": 37.0581690465554, "b_star": 0.557890555443363},
{ "name": "7.5RP4/6","L_star": 40.6056826069156, "a_star": 28.0611561430476, "b_star": 0.746617902313895},
{ "name": "7.5RP4/4","L_star": 40.6381400008437, "a_star": 18.9621823488038, "b_star": 0.640988842427881},
{ "name": "7.5RP4/3","L_star": 40.6543547569545, "a_star": 14.5447434338001, "b_star": 0.560352512716433},
{ "name": "7.5RP4/2","L_star": 40.686756453899, "a_star": 9.98016004923144, "b_star": 0.42567613000426},
{ "name": "7.5RP4/1","L_star": 40.7029434212134, "a_star": 5.19601610524695, "b_star": 0.2498301828187},
{ "name": "7.5RP3/18","L_star": 29.6982743563919, "a_star": 81.4580377029597, "b_star": -7.97316682144408},
{ "name": "7.5RP3/16","L_star": 29.9312810851596, "a_star": 73.7393656060209, "b_star": -6.61648169323509},
{ "name": "7.5RP3/14","L_star": 30.0763307866394, "a_star": 65.4834592464295, "b_star": -5.10671940428543},
{ "name": "7.5RP3/12","L_star": 30.1252935043003, "a_star": 57.0593918047947, "b_star": -3.84338314140233},
{ "name": "7.5RP3/10","L_star": 30.139962067383, "a_star": 47.8283309979317, "b_star": -2.7280388875539},
{ "name": "7.5RP3/8","L_star": 30.1839118911295, "a_star": 39.2593239843967, "b_star": -1.82172774749036},
{ "name": "7.5RP3/6","L_star": 30.2302128046796, "a_star": 29.8877210652997, "b_star": -0.981389948108968},
{ "name": "7.5RP3/4","L_star": 30.2715614676263, "a_star": 20.9799420258662, "b_star": -0.430411221610205},
{ "name": "7.5RP3/3","L_star": 30.2934220510993, "a_star": 16.2633861109064, "b_star": -0.218342886555678},
{ "name": "7.5RP3/2","L_star": 30.3152620081296, "a_star": 11.2787806790532, "b_star": -0.0851417130100174},
{ "name": "7.5RP3/1","L_star": 30.3370813873033, "a_star": 5.92149601806724, "b_star": -0.00912336102507627},
{ "name": "7.5RP2/16","L_star": 19.4866583417024, "a_star": 70.0681733217453, "b_star": -15.6289874834153},
{ "name": "7.5RP2/14","L_star": 19.7124587492403, "a_star": 62.2289980184605, "b_star": -13.4541274355994},
{ "name": "7.5RP2/12","L_star": 19.9314098107954, "a_star": 53.7742983797773, "b_star": -10.8899967302142},
{ "name": "7.5RP2/10","L_star": 20.0238615468325, "a_star": 45.2780919916766, "b_star": -8.57689020117719},
{ "name": "7.5RP2/8","L_star": 20.0519050446824, "a_star": 36.5836209479221, "b_star": -6.48591535936767},
{ "name": "7.5RP2/6","L_star": 20.0958854537657, "a_star": 27.3693639281361, "b_star": -4.44623417940843},
{ "name": "7.5RP2/4","L_star": 20.1317898727821, "a_star": 19.3676405792634, "b_star": -2.90695373238846},
{ "name": "7.5RP2/3","L_star": 20.1477245120574, "a_star": 15.4059418412394, "b_star": -2.18070544235621},
{ "name": "7.5RP2/2","L_star": 20.1676230763052, "a_star": 11.0762400304455, "b_star": -1.46766724253299},
{ "name": "7.5RP2/1","L_star": 20.187499769181, "a_star": 6.10464192949609, "b_star": -0.761722067891979},
{ "name": "7.5RP1/12","L_star": 9.86833667374694, "a_star": 51.8818584395723, "b_star": -17.5332961606398},
{ "name": "7.5RP1/10","L_star": 10.0842373959684, "a_star": 44.1936546037018, "b_star": -14.5744205893945},
{ "name": "7.5RP1/8","L_star": 10.2135954147157, "a_star": 36.3450266844172, "b_star": -11.7060241410627},
{ "name": "7.5RP1/6","L_star": 10.2589477463405, "a_star": 28.6528936548715, "b_star": -8.99470262840795},
{ "name": "7.5RP1/4","L_star": 10.2890958184513, "a_star": 21.1014949728499, "b_star": -6.42071945390537},
{ "name": "7.5RP1/3","L_star": 10.3041439590119, "a_star": 17.3337097683497, "b_star": -5.1696503456823},
{ "name": "7.5RP1/2","L_star": 10.3191749016881, "a_star": 12.8897941462597, "b_star": -3.76031901074996},
{ "name": "7.5RP1/1","L_star": 10.3416891769113, "a_star": 7.31514402304222, "b_star": -2.09760083469204},
{ "name": "10RP9/6","L_star": 89.8314760854284, "a_star": 26.8303739968481, "b_star": 6.52187305171448},
{ "name": "10RP9/4","L_star": 90.058612795847, "a_star": 16.1320154526521, "b_star": 4.1496814227145},
{ "name": "10RP9/3","L_star": 90.1187105319379, "a_star": 10.5479299129398, "b_star": 2.76838395015779},
{ "name": "10RP9/2","L_star": 90.1371884550925, "a_star": 5.67713287150073, "b_star": 1.49976390789377},
{ "name": "10RP9/1","L_star": 90.1556599466689, "a_star": 1.89076061990678, "b_star": 0.560927034934178},
{ "name": "10RP8/10","L_star": 79.9371285096615, "a_star": 44.4690651909369, "b_star": 9.8594345083536},
{ "name": "10RP8/8","L_star": 80.2414245232154, "a_star": 34.6252055933564, "b_star": 8.12568979050536},
{ "name": "10RP8/6","L_star": 80.3984529402843, "a_star": 24.4337431071491, "b_star": 5.99111996489565},
{ "name": "10RP8/4","L_star": 80.4432243617297, "a_star": 15.3050826882123, "b_star": 3.8215504968756},
{ "name": "10RP8/3","L_star": 80.4655944928565, "a_star": 10.4014372688398, "b_star": 2.5776327921464},
{ "name": "10RP8/2","L_star": 80.4823652850047, "a_star": 5.89545614217102, "b_star": 1.43842038659856},
{ "name": "10RP8/1","L_star": 80.4991302489017, "a_star": 2.28555714829165, "b_star": 0.568612383153533},
{ "name": "10RP7/16","L_star": 70.2179391069599, "a_star": 70.7801698761613, "b_star": 15.359353896326},
{ "name": "10RP7/14","L_star": 70.3437436712535, "a_star": 61.7445456282104, "b_star": 13.1728830740307},
{ "name": "10RP7/12","L_star": 70.5317659666793, "a_star": 52.160515464251, "b_star": 11.3858956780008},
{ "name": "10RP7/10","L_star": 70.6497324336259, "a_star": 41.8483332318937, "b_star": 9.46135261189247},
{ "name": "10RP7/8","L_star": 70.6982136304596, "a_star": 33.305710983412, "b_star": 7.73143580425377},
{ "name": "10RP7/6","L_star": 70.7397258291086, "a_star": 24.4167910879131, "b_star": 5.65857019870157},
{ "name": "10RP7/4","L_star": 70.7811983316404, "a_star": 15.4831729939649, "b_star": 3.61398315567478},
{ "name": "10RP7/3","L_star": 70.7950136941808, "a_star": 11.1989805509197, "b_star": 2.61950527442198},
{ "name": "10RP7/2","L_star": 70.8157284953717, "a_star": 7.05042212726864, "b_star": 1.71871193059887},
{ "name": "10RP7/1","L_star": 70.8295328728496, "a_star": 3.31510077574249, "b_star": 0.875711374944554},
{ "name": "10RP6/18","L_star": 60.675576888377, "a_star": 77.3507785989898, "b_star": 14.9429011597029},
{ "name": "10RP6/16","L_star": 60.7021174155531, "a_star": 69.2373183890272, "b_star": 13.2576879433929},
{ "name": "10RP6/14","L_star": 60.7463108555081, "a_star": 59.2167692333866, "b_star": 11.6301189381391},
{ "name": "10RP6/12","L_star": 60.7904534574736, "a_star": 50.9361546006699, "b_star": 10.3443871639046},
{ "name": "10RP6/10","L_star": 60.8345453674015, "a_star": 42.2482876110445, "b_star": 8.79221422923404},
{ "name": "10RP6/8","L_star": 60.8785867305749, "a_star": 33.2918736721993, "b_star": 7.10600164033128},
{ "name": "10RP6/6","L_star": 60.9137835246555, "a_star": 25.6349535778659, "b_star": 5.58288034523124},
{ "name": "10RP6/4","L_star": 60.957734267663, "a_star": 16.4097402585917, "b_star": 3.62688006061767},
{ "name": "10RP6/3","L_star": 60.9753005166793, "a_star": 11.9977467598869, "b_star": 2.64311354698288},
{ "name": "10RP6/2","L_star": 60.9928587518952, "a_star": 7.75996677596102, "b_star": 1.73531708458197},
{ "name": "10RP6/1","L_star": 61.0104089824442, "a_star": 3.77138376942271, "b_star": 0.885272781212842},
{ "name": "10RP5/20","L_star": 50.7339713527665, "a_star": 87.5301422292721, "b_star": 13.8970420024874},
{ "name": "10RP5/18","L_star": 50.6520890970938, "a_star": 78.6862487035181, "b_star": 12.2635588863635},
{ "name": "10RP5/16","L_star": 50.640375197187, "a_star": 70.3121508656985, "b_star": 11.2045839021024},
{ "name": "10RP5/14","L_star": 50.68720611576, "a_star": 61.2495045724255, "b_star": 10.2939000998085},
{ "name": "10RP5/12","L_star": 50.7339713527665, "a_star": 53.7639633997643, "b_star": 9.43618140319398},
{ "name": "10RP5/10","L_star": 50.7806711380232, "a_star": 44.278440365303, "b_star": 8.16814428718415},
{ "name": "10RP5/8","L_star": 50.8273057000631, "a_star": 35.8607486238219, "b_star": 6.86094063698401},
{ "name": "10RP5/6","L_star": 50.8622389556151, "a_star": 26.6762213977217, "b_star": 5.28015011489731},
{ "name": "10RP5/4","L_star": 50.9087599231063, "a_star": 17.5487677427777, "b_star": 3.56245000494244},
{ "name": "10RP5/3","L_star": 50.9319961673508, "a_star": 12.8672771928103, "b_star": 2.6245616445858},
{ "name": "10RP5/2","L_star": 50.9436082418668, "a_star": 8.32112694725651, "b_star": 1.69255125424708},
{ "name": "10RP5/1","L_star": 50.966820313159, "a_star": 3.9705401169074, "b_star": 0.820141175666578},
{ "name": "10RP4/20","L_star": 40.3610568985711, "a_star": 88.9754334020043, "b_star": 10.3160706162989},
{ "name": "10RP4/18","L_star": 40.4754793119243, "a_star": 80.3704349192722, "b_star": 9.56438398559699},
{ "name": "10RP4/16","L_star": 40.4754793119243, "a_star": 71.2437934060817, "b_star": 8.89794917795057},
{ "name": "10RP4/14","L_star": 40.4101519206937, "a_star": 63.0457072911588, "b_star": 8.20350572088201},
{ "name": "10RP4/12","L_star": 40.4591616342428, "a_star": 54.5113324309681, "b_star": 7.61551180404635},
{ "name": "10RP4/10","L_star": 40.5080864088407, "a_star": 45.4766638920038, "b_star": 6.71675227614735},
{ "name": "10RP4/8","L_star": 40.5406559183046, "a_star": 37.2818442744472, "b_star": 5.77230329737584},
{ "name": "10RP4/6","L_star": 40.5894399424351, "a_star": 28.1586752637873, "b_star": 4.55441807976011},
{ "name": "10RP4/4","L_star": 40.6381400008437, "a_star": 19.2335466838209, "b_star": 3.20275236494919},
{ "name": "10RP4/3","L_star": 40.6543547569545, "a_star": 14.6684056047332, "b_star": 2.47105078656288},
{ "name": "10RP4/2","L_star": 40.6705602368878, "a_star": 9.95066994299285, "b_star": 1.699276517809},
{ "name": "10RP4/1","L_star": 40.686756453899, "a_star": 5.13687106160213, "b_star": 0.889819101855527},
{ "name": "10RP3/16","L_star": 29.8249869058137, "a_star": 74.9048580668821, "b_star": 5.34313041597784},
{ "name": "10RP3/14","L_star": 30.0493568774064, "a_star": 66.2018750000615, "b_star": 5.39705810076984},
{ "name": "10RP3/12","L_star": 30.1252935043003, "a_star": 58.0265079167387, "b_star": 5.32154250037288},
{ "name": "10RP3/10","L_star": 30.1228478365595, "a_star": 48.7731984078517, "b_star": 4.90008110158894},
{ "name": "10RP3/8","L_star": 30.1717119983675, "a_star": 39.6694539515248, "b_star": 4.41226498306048},
{ "name": "10RP3/6","L_star": 30.2204729512776, "a_star": 30.0279513020769, "b_star": 3.6804046572236},
{ "name": "10RP3/4","L_star": 30.2642700143906, "a_star": 21.0090010629211, "b_star": 2.75436014221604},
{ "name": "10RP3/3","L_star": 30.2837087855916, "a_star": 16.3095984381755, "b_star": 2.21250125282446},
{ "name": "10RP3/2","L_star": 30.3055579039205, "a_star": 11.3187806147812, "b_star": 1.56843899560412},
{ "name": "10RP3/1","L_star": 30.3298105443489, "a_star": 5.99665960918611, "b_star": 0.831930756490451},
{ "name": "10RP2/14","L_star": 19.6142796731263, "a_star": 63.73345567585, "b_star": -6.72598169107452},
{ "name": "10RP2/12","L_star": 19.8789433397199, "a_star": 54.7427330070639, "b_star": -4.97321886288113},
{ "name": "10RP2/10","L_star": 20.0118295280528, "a_star": 45.9402855861904, "b_star": -3.48378776051413},
{ "name": "10RP2/8","L_star": 20.0438970690648, "a_star": 37.1980121804436, "b_star": -2.29963216107106},
{ "name": "10RP2/6","L_star": 20.0878969870992, "a_star": 28.0269524735322, "b_star": -1.30966597265113},
{ "name": "10RP2/4","L_star": 20.1238172798771, "a_star": 19.7602746580892, "b_star": -0.707024879347451},
{ "name": "10RP2/3","L_star": 20.1397589488837, "a_star": 15.5759830282499, "b_star": -0.417077503609753},
{ "name": "10RP2/2","L_star": 20.1596662785396, "a_star": 11.0831106079779, "b_star": -0.208104749487459},
{ "name": "10RP2/1","L_star": 20.1795517127494, "a_star": 6.03088401521676, "b_star": -0.0659678894836646},
{ "name": "10RP1/12","L_star": 9.74332916966028, "a_star": 52.9006142387319, "b_star": -13.500956536766},
{ "name": "10RP1/10","L_star": 10.0229165999318, "a_star": 44.8671656012019, "b_star": -10.6577771339351},
{ "name": "10RP1/8","L_star": 10.1832731257492, "a_star": 37.1974155928974, "b_star": -8.21831517109647},
{ "name": "10RP1/6","L_star": 10.2438477155545, "a_star": 29.1186153991081, "b_star": -5.98280617341013},
{ "name": "10RP1/4","L_star": 10.2740304307288, "a_star": 21.4144178976998, "b_star": -4.07302105686628},
{ "name": "10RP1/3","L_star": 10.2890958184513, "a_star": 17.5586009928317, "b_star": -3.18449791427551},
{ "name": "10RP1/2","L_star": 10.3041439590119, "a_star": 13.0562774005111, "b_star": -2.25537096580956},
{ "name": "10RP1/1","L_star": 10.3266839391574, "a_star": 7.47757864823678, "b_star": -1.2605469155619}
]
};